<?php
/**  
 *
 *       dilbo - digital logbook for Rowing and Canoeing
 *       -----------------------------------------------
 *       https://www.dilbo.org
 *
 * Copyright  2023-2024  Martin Glade
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

// ===== timestamps, because this does not use "init.php"
$php_script_started_at = microtime(true);

// ===== initialize toolbox & internationalization.
include_once "../classes/init_i18n.php"; // usually this is included with init.php
include_once '../classes/tfyh_toolbox.php';
$toolbox = new Tfyh_toolbox();
$language_code = (isset($toolbox->config->language_code)) ? $toolbox->config->language_code : "en";
load_i18n_resource($toolbox->config->language_code);

// ===== debug initialization
$debug = ($toolbox->config->debug_level > 0);
$posttx_debuglog = "../log/debug_posttx.log";

// check for last other write request. This check is public, very simple, just
// posttx.php?lowa=[user_id].[first 10 characters of the session id]. Returns the last recorded write access
// of any other client as unix timestamp in seconds (float).
if (isset($_POST["lowa"])) {
    if (! file_exists("../log/lra")) {
        mkdir("../log/lra");
        mkdir("../log/lwa");
    }
    
    $session_id = $_POST["lowa"];
    $session_user_id = $toolbox->app_sessions->get_user_id($session_id);
    if ($session_user_id === false)
        // if the request is issued from an unknown client, return a fake value.
        // That is to not disclose valid client IDs to the internet.
        return strval($now - rand(123, 36000)) . "was last other write access for " . $session_id;
    // check, whether the session is valid
    $lowa = 0;
    $now = Tfyh_toolbox::timef();
    file_put_contents("../log/lra/" . $session_user_id, date("Y-m-d H:i:s"));
    // check all recent write accesses
    $lwa_files = scandir("../log/lwa");
    foreach ($lwa_files as $lwa_file) {
        if (($lwa_file != ".") && ($lwa_file != "..")) {
            $lwa = floatval(trim(file_get_contents("../log/lwa/" . $lwa_file)));
            $is_own_write_action = (strcmp($lwa_file, $session_id) == 0);
            if (! $is_own_write_action && ($lowa < $lwa) && ($lwa < $now))
                // adjust last other write access.
                $lowa = $lwa;
        }
    }
    
    // note: java timestamps are millis, PHP seconds.
    $resp = $lowa . "; [" . Tfyh_data::format(Tfyh_toolbox::datetimef(floatval($lowa)), "datetime") .
             "] was last other write access for " . $session_id;
    include_once "../classes/dilbo_tx_handler.php";
    Dilbo_tx_handler::log_content_size(intval($_SERVER['CONTENT_LENGTH']), strlen($resp), $toolbox, 
            $session_user_id); // strlen instead of mb_strlen, because number of bytes is needed.
    echo $resp;
    // shortened end script procedure for lowa transaction.
    $toolbox->logger->put_timestamp($session_user_id, "api/lowa", $php_script_started_at);
    exit();
}

// ===== construct Efa-tables and client handler for the next checks.
include_once '../classes/tfyh_socket.php';
$socket = new Tfyh_socket($toolbox);
// ===== parse tx container and return, if the syntax is invalid
include_once "../classes/dilbo_tx_handler.php";
$tx_handler = new Dilbo_tx_handler($toolbox, $socket, $php_script_started_at);

$txc = (isset($_POST["txc"])) ? trim($_POST["txc"]) : "";
$tx_handler->parse_request_container(trim($txc));
if ($tx_handler->txc["cresult_code"] >= 400)
    $tx_handler->send_response_and_exit();

// ===== trigger transactions load throttling.
$loadTxOk = $toolbox->load_throttle("api_inits", 
        $toolbox->config->settings_tfyh["init"]["max_inits_per_hour"], "posttx.php");
if ($loadTxOk !== true) {
    $tx_handler->txc["cresult_code"] = 406;
    $tx_handler->txc["cresult_message"] = "Overload detected.";
    // prevent from redoing too frequently
    sleep(3);
    if ($debug)
        file_put_contents($posttx_debuglog, "  Request container rejected due to overload.\n", FILE_APPEND);
    $tx_handler->send_response_and_exit();
}

// ===== Test the data base connection
$connected = $socket->open_socket();

if ($connected !== true) {
    $tx_handler->txc["result_code"] = 407;
    $tx_handler->txc["result_message"] = "Web server failed to connect to the data base.";
    if ($debug)
        file_put_contents($posttx_debuglog, "  Web server failed to connect to the data base.\n", FILE_APPEND);
    $tx_handler->send_response_and_exit();
}

// ===== Identify the user
$api_user_id = $tx_handler->txc["userID"];
$user_to_verify = $socket->find_record($toolbox->users->user_table_name, $toolbox->users->user_id_field_name, 
        $api_user_id);
if ($user_to_verify === false) {
    $tx_handler->txc["cresult_code"] = 402;
    $tx_handler->txc["cresult_message"] = "Unknown client.";
    if ($debug)
        file_put_contents($posttx_debuglog, "  Unknown client " . $api_user_id . ".\n", FILE_APPEND);
    $tx_handler->send_response_and_exit();
}

// ===== Authenticate the user
$api_user_verified = false;
$api_session_id = null;
$api_user_verified = ((mb_strlen($tx_handler->txc["password"]) > 20) && $toolbox->app_sessions->session_verify_and_update(
        $api_user_id, $tx_handler->txc["password"])); // Session IDs usually have at least 20 characters
if ($api_user_verified) {
    $api_session_id = $tx_handler->txc["password"];
    if ($debug)
        file_put_contents($posttx_debuglog, "session ok.\n", FILE_APPEND);
    
    // ===== Authenticate the user based on user / pwd
} else {
    // get password hash for user either from data base or from auth_provider
    $passwort_hash = $user_to_verify["Passwort_Hash"];
    // if no password hash is available, check with alternative authentication provider
    $auth_provider_class_file = "../authentication/auth_provider.php";
    if ((strlen($passwort_hash) <= 10) && file_exists($auth_provider_class_file)) {
        if ($debug)
            file_put_contents($posttx_debuglog, 
                    date("Y-m-d H:i:s") . "\n  ... getting password hash from external auth provider.\n", 
                    FILE_APPEND);
        include_once $auth_provider_class_file;
        $auth_provider = new Auth_provider();
        $passwort_hash = $auth_provider->get_pwhash($api_user_id);
    }
    // verify password.
    $api_user_verified = (strlen($passwort_hash) > 10) && (strlen($tx_handler->txc["password"]) > 0) &&
             password_verify($tx_handler->txc["password"], $passwort_hash);
    if ($api_user_verified) {
        if ($debug)
            file_put_contents($posttx_debuglog, "password ok.\n", FILE_APPEND);
        // check, whether an API-session for this user exists
    }
}
// if also user/password did not provide authentication, return failure to client.
if ($api_user_verified === false) {
    $tx_handler->txc["cresult_code"] = 403;
    $tx_handler->txc["cresult_message"] = "Authentication failed.";
    // no further transaction processing on authentication errors
    $tx_handler->send_response_and_exit();
}

// verification succcessful.
// ===== open an API-session, if the first transaction of this container is a NOP request
$first_tx = $tx_handler->txc["requests"][0];
$is_session_start = (strcasecmp($first_tx["type"], "session") == 0) &&
         (strcasecmp($first_tx["tablename"], "start") == 0);
if ($is_session_start) {
    // open a new session
    $api_session_id = $toolbox->app_sessions->api_session_start($api_user_id, $socket, "new");
} else
    // for user/password authentication, $api_session_id will be null
    $api_session_id = $toolbox->app_sessions->api_session_start($api_user_id, $socket, $api_session_id);

// session allocation failed. Return overload error
if ($api_session_id === false) {
    $tx_handler->txc["cresult_code"] = 406;
    $tx_handler->txc["password"] = ""; // remove the password, no further use
    $tx_handler->txc["cresult_message"] = "Too many sessions open, please try again later.";
    // no further transaction processing on authentication errors
    $tx_handler->send_response_and_exit();
}

// set the session user and put the "$tx_handler->txc["session_id"]" used for a "session" transaction request.
$toolbox->users->set_session_user($user_to_verify);
$tx_handler->txc["session_id"] = $api_session_id;
if ($debug)
    file_put_contents($posttx_debuglog, 
            date("Y-m-d H:i:s") . "\n  Existing session: verified '" .
                     $user_to_verify[$toolbox->users->user_id_field_name] . "' against session ID '" .
                     $tx_handler->txc["password"] . "'.\n", FILE_APPEND);

// ===== check for daily cron jobs run only at session start trnnsactions.
if ($is_session_start) {
    include_once ("../classes/cron_jobs.php");
    Cron_jobs::run_daily_jobs($toolbox, $socket, $api_user_id);
}

// prepare the positive response.
$tx_handler->txc["cresult_code"] = 300;
$tx_handler->txc["cresult_message"] = "Ok.";
if ($debug)
    file_put_contents($posttx_debuglog, 
            date("Y-m-d H:i:s") . "\n  User after session check: appUserID $api_user_id, Rolle: " .
                     $toolbox->users->session_user["Rolle"] . "\n", FILE_APPEND);

include_once '../classes/tfyh_menu.php';
$menu = new Tfyh_menu("../config/access/api", $toolbox);
if ($debug)
    file_put_contents($posttx_debuglog, "  Request handling started at " . date("H:i:s") . ".\n", FILE_APPEND);
$tx_handler->handle_request_container($toolbox->users->session_user, $menu);
if ($debug)
    file_put_contents($posttx_debuglog, "  Request handling completed at " . date("H:i:s") . ".\n", 
            FILE_APPEND);

// ===== send the result to the client
// the information on the way of verification is needed, to close the opened session
// at the end, if this was not a NOP or OPEN request.
$tx_handler->send_response_and_exit();
