<?php
/**  
 *
 *       dilbo - digital logbook for Rowing and Canoeing
 *       -----------------------------------------------
 *       https://www.dilbo.org
 *
 * Copyright  2023-2024  Martin Glade
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */


/**
 * static class container file for a daily jobs routine. It may be triggered by whatever, checks whther it was
 * already run this day and if not, starts the sequence.
 */
include_once '../classes/tfyh_cron_jobs.php';

class Cron_jobs extends Tfyh_cron_jobs
{

    /**
     * run all daily jobs.
     * 
     * @param Tfyh_toolbox $toolbox
     *            application toolbox
     * @param Tfyh_socket $socket
     *            the socket to connect to the database
     * @param int $app_user_id
     *            the id of the invoking user.
     * @param bool $skip_configured_jobs
     *            set true to skip the application configured jobs
     */
    public static function run_daily_jobs (Tfyh_toolbox $toolbox, Tfyh_socket $socket, int $app_user_id, 
            bool $skip_configured_jobs = false)
    {
        $daily_run = Tfyh_cron_jobs::run_daily_jobs($toolbox, $socket, $app_user_id);
        
        // add application specific cron jobs here.
        // The sequence is an implicit priority, in case one of the jobs fails.
        if ($daily_run) {
            
            // OPEN LOG
            // --------
            $cronlog = "../log/sys_cronjobs.log";
            file_put_contents($cronlog, date("Y-m-d H:i:s") . " +0: specific dilbo cronjobs started.\n", 
                    FILE_APPEND);
            $cron_started = time();
            $last_step_ended = time();
            
            
            // CLOSE LOG
            // ---------
            file_put_contents($cronlog, 
                    date("Y-m-d H:i:s") . ": Cron jobs done. Total cron jobs duration = " .
                             (time() - $cron_started) . ".\n", FILE_APPEND);
        }
    }

    /**
     * Returns true, if a task is due today based on the $task_day specification: starts with a letter (D =
     * Daily, W = Weekly, M = Monthly), continues with a number (for D : 1 always, for W : 1 = Monday, 2 =
     * Tuesday asf., for M : day of month, 31 is the same as ultimo).
     * 
     * @param String $task_day            
     */
    private static function due_today (String $task_day)
    {
        $period = substr($task_day, 0, 1);
        $day = intval(substr(trim($task_day), 1));
        // daily run
        if (strcasecmp($period, "defaultValue") == 0) {
            file_put_contents("../log/due_today.log", date("Y-m-d") . ": due today, $task_day = daily run", 
                    FILE_APPEND);
            return true;
        }
        // weekly run
        if ((strcasecmp($period, "W") == 0) && ($day == intval(date("w")))) {
            file_put_contents("../log/due_today.log", date("Y-m-d") . ": due today, $task_day = weekly run", 
                    FILE_APPEND);
            return true;
        }
        // monthly run, any day
        if ((strcasecmp($period, "M") == 0) && ($day == intval(date("j")))) {
            file_put_contents("../log/due_today.log", date("Y-m-d") . ": due today, $task_day = monthly run", 
                    FILE_APPEND);
            return true;
        }
        // monthly run, ultimo. 86400 seconds are 1 day
        if ((strcasecmp($period, "M") == 0) && ($day == 31) && (intval(date("j", time() + 86400)) == 1)) {
            file_put_contents("../log/due_today.log", date("Y-m-d") . ": due today, $task_day = ultimo run", 
                    FILE_APPEND);
            return true;
        }
        return false;
    }

    /**
     * Jobs can be configured to be run together with the cron jobs trigger, which should be called on a daily
     * basis. A job consists of a scheduled day (see Tfyh_tasks->due_today for details) and a task type. 
     * 
     * @param Tfyh_toolbox $toolbox
     *            application toolbox
     * @param Tfyh_socket $socket
     *            the socket to connect to the database
     */
    private static function run_configured_jobs (Tfyh_toolbox $toolbox, Tfyh_socket $socket, int $app_user_id)
    {
        // get the job list
        $cfg = $toolbox->config->get_cfg();
        if (! isset($cfg["configured_jobs"]) || (mb_strlen($cfg["configured_jobs"]) < 4))
            return;
    }
}
