<?php
/**  
 *
 *       dilbo - digital logbook for Rowing and Canoeing
 *       -----------------------------------------------
 *       https://www.dilbo.org
 *
 * Copyright  2023-2024  Martin Glade
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */
include_once "../classes/dilbo_record.php";
include_once "../classes/dilbo_tx_handler.php";
include_once "../classes/tfyh_app_sessions.php";

/**
 * class file for the specific handling of eFa tables, e. g. GUID generation, autoincrementation etc.
 */
class Dilbo_api
{

    /**
     * The data base connection socket.
     */
    private $socket;

    /**
     * The generic toolbox.
     */
    private $toolbox;

    /**
     * Debug level to add mor information for support cases.
     */
    private $debug_on;

    /**
     * public Constructor.
     * 
     * @param Tfyh_toolbox $toolbox
     *            application toolbox
     * @param Tfyh_socket $socket
     *            the socket to connect to the database
     */
    public function __construct (Tfyh_toolbox $toolbox, Tfyh_socket $socket)
    {
        $this->toolbox = $toolbox;
        $this->socket = $socket;
        $this->debug_on = $toolbox->config->debug_level > 0;
    }

    /* --------------------------------------------------------------------------------------- */
    /* ------------------ WRITE DATA TO SERVER - PUBLIC FUNCTIONS ---------------------------- */
    /* --------------------------------------------------------------------------------------- */
    
    /**
     * Modify a record of a table using the API syntax and return the result as
     * "<result_code>;<result_message>". Set the system generated values except uid and uuid.
     * 
     * @param array $client_verified
     *            the verified client which requests the execution
     * @param String $tablename
     *            the table into which the record shall be imported.
     * @param array $record
     *            record which shall be inserted.
     * @param int $mode
     *            Set 1 for insert, 2 for update, 3 for delete
     * @param String $debug_log
     *            the api debug log file path
     * @return the api result-code and result
     */
    public function api_modify (array $client_verified, String $tablename, array $record, int $mode, 
            String $debug_log)
    {
        $mode_str = ($mode == 1) ? "insert" : (($mode == 2) ? "update" : "delete");
        if ($this->debug_on)
            file_put_contents($debug_log, 
                    date("Y-m-d H:i:s") . ": starting api_modify ($mode_str) for client " .
                             $client_verified[$this->toolbox->users->user_id_field_name] .
                             " at table $tablename, API version.\n", FILE_APPEND);
        $user_id = $client_verified[$this->toolbox->users->user_id_field_name];
        include_once "../classes/dilbo_record.php";
        $dilbo_record = new Dilbo_record($this->toolbox, $this->socket);
        if ($this->debug_on)
            file_put_contents($debug_log, date("Y-m-d H:i:s") . ": dilbo_record instance created.\n", 
                    FILE_APPEND);
        
        Tfyh_validate::clear_findings();
        $record_item = $this->toolbox->config->get_by_path(".appTables.$tablename");
        $findings = $dilbo_record->modify_record($record, $record_item, $mode, $user_id);
        $findings .= Tfyh_validate::get_findings(false);
        
        // Return response
        if (strlen($findings) == 0) {
            // return response, depending on whether a key was modified, or not.
            if ($this->debug_on)
                file_put_contents($debug_log, 
                        date("Y-m-d H:i:s") . ": api_$mode_str: " . i("vOXx8v|completed") . ". Uid '" .
                                 $record["uid"] . "' \n", FILE_APPEND);
            return "300;ok."; // 300 => "Transaction completed."
        } else {
            if ($this->debug_on)
                file_put_contents($debug_log, 
                        date("Y-m-d H:i:s") . ": api_$mode_str: " . i("fposcc|Failed. Reason:") . " '" .
                                 $findings . "' \n", FILE_APPEND);
            return "502;" . $findings; // 502 => "Transaction failed."
        }
    }

    /* --------------------------------------------------------------------------------------- */
    /* ----------------------- READ DATA FROM SERVER ----------------------------------------- */
    /* --------------------------------------------------------------------------------------- */
    
    /**
     * LProvide a statistic at the API. Note that this uses the list transaction with the tablename
     * "statistics"
     * 
     * @param array $client_verified
     *            the verified client which requests the execution
     * @param array $record
     *            the request record with the statistics type and field.
     */
    public function api_statistics (array $client_verified, array $record)
    {
        if (! isset($record["type"]))
            return "502;" . i("0yaw5j|Missing type definition ...");
        if (! isset($record["field"]))
            return "502;" . i("rgoqVn|Missing field definition...");
        include_once "../classes/dilbo_logbook.php";
        $dilbo_logbook = new Dilbo_logbook($this->toolbox, $this->socket);
        $statistic = $dilbo_logbook->get_statistics($record["type"], $record["field"]);
        return "300;" . $statistic;
    }

    /**
     * Load a list of a table using the API syntax and return the result.
     * 
     * @param array $client_verified
     *            the verified client which requests the execution
     * @param String $listname
     *            the name of the list to be used. '@definitions' returns the config file of the set.
     * @param array $record
     *            the request record with set name and list arguments.
     */
    public function api_list (array $client_verified, String $listname, array $record)
    {
        if (! isset($record["setname"]))
            return "502;" . i("dAwKF1|Missing name of set in r...");
        if (! file_exists("../config/lists/" . $record["setname"]))
            return "502;" . i("X6z2gz|Invalid name of set in r...");
        if (strcasecmp($listname, "@definitions") == 0)
            return "300;" . file_get_contents("../config/lists/" . $record["setname"]);
        $modified_after = (isset($record["modified_after"])) ? floatval($record["modified_after"]) : 0;
        $list_args = ["{modified_after}" => $modified_after
        ];
        foreach ($record as $key => $value)
            $list_args["{" . $key . "}"] = $value;
        
        include_once "../classes/tfyh_list.php";
        $list = new Tfyh_list("../config/lists/" . $record["setname"], 0, $listname, $this->socket, 
                $this->toolbox, $list_args);
        $list->entry_size_limit = 10000;
        $is_versionized_table = ($list->get_field_index("InvalidFrom") !== false);
        $csv = $list->get_csv($client_verified, ($is_versionized_table) ? "Id" : null);
        return "300;" . $csv;
    }

    /**
     * Load a configuration file to update the client.
     * 
     * @param array $client_verified
     *            the verified client which requests the execution
     * @param String $config_filename
     *            the name of the configuration file to load.
     */
    public function api_config (array $client_verified, String $config_filename)
    {
        if (strcasecmp($client_verified["Rolle"], $this->toolbox->users->anonymous_role) == 0)
            return "503;Transcation forbidden";
        $config_file = file_get_contents("../config/settings/" . $config_filename);
        if (($config_file == false) || (strlen($config_file) < 10))
            return "502;Could not open config file.";
        return "300;" . $config_file;
    }

    /**
     * Return just an ok message, like a ping.
     * 
     * @param array $record
     *            the record with the one and only parameter "sleep" in seconds.
     * @return string
     */
    public function api_nop (array $record)
    {
        // wait some time as this is also a nop function.
        $wait_for_secs = (isset($record["sleep"])) ? intval(trim($record["sleep"])) : 1;
        $wait_for_secs = ($wait_for_secs > 100) ? 100 : $wait_for_secs;
        if ($wait_for_secs > 0)
            sleep($wait_for_secs);
        return "300;ok.";
    }

    /**
     * Add the api configuration information to a transaction response
     * 
     * @param array $client_verified
     *            the verified client which requests the execution
     * @return string the api configuration parameters and server welcome message as ';' separated
     *         name=value-pairs, preceded by a ';'.
     */
    private function api_configuration (array $client_verified)
    {
        $api_configuration = ";max_session_keepalive=" . $this->toolbox->app_sessions->max_session_keepalive;
        $api_configuration .= ";max_session_duration=" . $this->toolbox->app_sessions->max_session_duration;
        
        // add the synchronisation period settings
        $synch_period = $this->toolbox->config->get_by_path(
                ".appSettings.communication.synchronization.synch_period");
        $synch_check_period = $this->toolbox->config->get_by_path(
                ".appSettings.communication.synchronization.synch_check_period");
        $api_configuration .= ";synch_check_period=" .
                 ((is_null($synch_check_period)) ? 0 : $synch_check_period->get_value());
        $api_configuration .= ";synch_check=" . ((is_null($synch_period)) ? 0 : $synch_period->get_value());
        
        // add the server welcome message
        $username = $client_verified[$this->toolbox->users->user_firstname_field_name] . " " .
                 $client_verified[$this->toolbox->users->user_lastname_field_name] . " (Rolle: " .
                 $client_verified["Rolle"] . ")";
        $version = (file_exists("../public/version")) ? file_get_contents("../public/version") : "";
        $acronym = $this->toolbox->config->get_by_path(".clubSettings.acronym");
        $welcome_message = "dilbo Server" . ((is_null($acronym)) ? "" : " of '" . $acronym->get_value() . "'") .
                 ". Version '" . $version . "'//" . i("BN4UxH|Connected as") . " '" . $username . "'.";
        $api_configuration .= ";server_welcome_message=" . str_replace(";", "_", $welcome_message);
        return $api_configuration;
    }

    /**
     * Return some configuration settings and a session renewer token
     * 
     * @param array $client_verified
     *            the verified client which requests the execution
     * @param String $request_session_id
     *            the ID of the session for $todo = regenerate and close.
     * @param String $todo
     *            the action to do. This is encoded in the transaction tablename.
     */
    public function api_session (array $client_verified, String $request_session_id, String $todo)
    {
        $user_id = $client_verified[$this->toolbox->users->user_id_field_name];
        if (strcasecmp($todo, "start") == 0) {
            // open the application session
            $response_session_id = $this->toolbox->app_sessions->api_session_start($user_id, $this->socket, 
                    $request_session_id);
            if ($response_session_id === false)
                return "406;Failed to open API session, possible cause: overload. You may try again later.";
        } elseif (strcasecmp($todo, "regenerate") == 0) {
            // open the application session
            $response_session_id = $this->toolbox->app_sessions->api_session_regenerate_id(
                    $request_session_id, $this->socket);
            if ($response_session_id === false)
                return "406;Failed to open API session, possible cause: overload. You may try again later.";
        } elseif (strcasecmp($todo, "close") == 0) {
            // open the application session
            $response_session_id = $this->toolbox->app_sessions->api_session_close(
                    i("uecpM1|Client request."), $request_session_id);
            if ($response_session_id === false)
                return "502;Failed to close API session, possible cause: Not existing.";
            else
                return "300;session closed.";
        }
        // provide the new API session.
        $tx_response = "300;api_session_id=" . $response_session_id . ";";
        $tx_response .= $this->api_configuration($client_verified, $tx_response);
        return $tx_response;
    }

    /**
     * Get the last access of every API client
     * 
     * @param bool $html
     *            set true to get html encoded output, false for plain text
     * @param bool $contentsize
     *            set true to get the contentsize table, false to omit.
     * @return string the last access information on every client.
     */
    public function get_last_accesses_API (bool $html, bool $contentsize)
    {
        $client_read_accesses = scandir("../log/lra");
        $active_clients_txt = "";
        $active_clients_html = "<ul>";
        $content_size_tables = "";
        $client_dirs = scandir("../log/contentsize");
        foreach ($client_dirs as $client_dir) {
            if (is_numeric($client_dir)) {
                $client_record = $this->socket->find_record("persons", "user_id", $client_dir);
                if ($client_record !== false) {
                    $active_clients_html .= "<li>";
                    $client_names[$client_dir] = $client_record["first_name"] . " " .
                             $client_record["last_name"];
                    $active_client_txt = "<b>" . $client_names[$client_dir] . "</b> (#" .
                             $client_record["user_id"] . ", " . $client_record["Rolle"] . ")";
                    if (file_exists("../log/lra/" . $client_dir))
                        $active_client_txt .= ", " . i("z1kw92|last activity:") . " " .
                                 date("d.m.Y H:i", strtotime(file_get_contents("../log/lra/" . $client_dir)));
                    else
                        $active_client_txt .= ", " . i("4fsAVc|last activity not known");
                    $active_clients_html .= $active_client_txt;
                    $active_clients_html .= "</li>";
                    $active_clients_txt .= "\n";
                } else {
                    // $active_clients_html .= i("mwgz0K|For the still existing c...", $client_dir);
                }
            }
        }
        if (strlen($content_size_tables) > 0)
            $content_size_tables = "<h4>" . i("52B267|Traffic volume of the la...") . "</h4>" .
                     $content_size_tables;
        return ($html) ? $active_clients_html . "</ul>" . $content_size_tables : $active_client_txt;
    }
}
