<?php
/**  
 *
 *       dilbo - digital logbook for Rowing and Canoeing
 *       -----------------------------------------------
 *       https://www.dilbo.org
 *
 * Copyright  2023-2024  Martin Glade
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

include_once "../classes/dilbo_uuids.php";
include_once '../classes/tfyh_list.php';

/**
 * class file for logbook reading, auditing and formatting capabilities.
 */
class Dilbo_logbook
{

    /**
     * The common toolbox used.
     */
    private $toolbox;

    /**
     * Tfyh_socket to data base.
     */
    private $socket;

    /**
     * uuid resolving utility class.
     */
    private $dilbo_uuids;

    /**
     * public Constructor. Runs the anonymization.
     */
    public function __construct (Tfyh_toolbox $toolbox, Tfyh_socket $socket)
    {
        $this->socket = $socket;
        $this->toolbox = $toolbox;
        $this->dilbo_uuids = new Dilbo_uuids($toolbox, $socket);
    }

    /**
     *
     * @return the end of the sport year which is the current sport year based on the sports_year_start
     *         configuration parameter as time (seconds since 1970).
     */
    private function get_sports_year_end (int $so_many_years_ago = 0)
    {
        include_once '../classes/dilbo_config.php';
        $sports_year_start = Dilbo_config::get_sports_year_start($this->toolbox->config->get_cfg(), 
                $so_many_years_ago - 1);
        return $sports_year_start - 86400;
    }

    /**
     * Get the statistics of the given name.
     * 
     * @param String $type
     *            type of statistics definition, like "total_distance"
     * @param String $field
     *            field of statistics definition, like "asset_uuid" or "crew"
     */
    public function get_statistics (String $type, String $field)
    {
        if (strcasecmp($type, "total_distance") == 0) {
            include_once '../classes/dilbo_config.php';
            $sports_year_start = Dilbo_config::get_sports_year_start($this->toolbox->config);
            $list_args = ["{sports_year_start}" => date("Y-m-d", $sports_year_start)
            ];
            $data_list = new Tfyh_list("../config/lists/statistics", 1, "", $this->socket, $this->toolbox, 
                    $list_args);
            $rows = $data_list->get_rows();
            $item_uuid_pos = $data_list->get_field_index($field);
            $is_list = (strcasecmp($field, "crew")) || (strcasecmp($field, "waters"));
            $distance_pos = $data_list->get_field_index("distance");
            $pivot = [];
            foreach ($rows as $row) {
                $pivot_line_items = ($is_list) ? explode("|", $row[$item_uuid_pos]) : [ $row[$item_uuid_pos] ];
                $distance = floatval($row[$distance_pos]);
                foreach ($pivot_line_items as $pivot_line_item) {
                    if (Tfyh_validate::is_uuid($pivot_line_item)) {
                        $name = $this->dilbo_uuids->uuid2name($pivot_line_item)[0];
                        if (! isset($pivot[$name]))
                            $pivot[$name] = ["distance" => 0,"sessions" => 0
                            ];
                        $pivot[$name]["distance"] += $distance;
                        $pivot[$name]["sessions"] += 1;
                    }
                }
            }
            $csv = "name;total_distance;sessions_count";
            foreach ($pivot as $name => $statistics)
                $csv .= "\n$name;" . $statistics["distance"] . ";" . $statistics["sessions"];
            return $csv;
        }
        return "";
    }
}
    
