<?php
/**  
 *
 *       dilbo - digital logbook for Rowing and Canoeing
 *       -----------------------------------------------
 *       https://www.dilbo.org
 *
 * Copyright  2023-2024  Martin Glade
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */
include_once '../classes/tfyh_list.php';

/**
 * class file for the dilbo data verification and modification. This class adds to the Efa_tables class whih
 * deifnes tables type semantics and contains static checker functions.
 */
class Dilbo_record
{

    /**
     * Column names of those columns that must not be empty. Forms may require more fields to be set.
     */
    public static $check_not_empty_fields = ["assets" => ["uuid","name","asset_type"
    ],"badges" => ["personid"
    ],"crews" => ["uuid","name"
    ],"damages" => ["number","asset_uuid","severity"
    ],"destinations" => ["uuid","name"
    ],"groups" => ["uuid","name"
    ],"logbook" => ["logbookname","number","asset_uuid"
    ],"messages" => ["number","to","subject"
    ],"persons" => ["uuid","first_name","last_name","user_id","Rolle","Workflows","Concessions"
    ],"reservations" => ["number","asset_uuid","type"
    ],"status" => ["uuid","name","type"
    ],"waters" => ["uuid","name"
    ],"workbook" => ["number","worbookname","uuid","date","description"
    ]
    ];

    public static $numbered_tables = ["logbook","workbook","damages","messages","reservations"
    ];

    public static $mode_name = [1 => "insert",2 => "update",3 => "delete"
    ];

    public static $mode_int = ["insert" => 1,"update" => 2,"delete" => 3
    ];

    /**
     * The list indices for name-to-uuid resolution referencing. Use $this->build_indices to initialize
     * respective associative arrays. for versionized tables only the last valid name will be used, no
     * previous ones. But uniqueness must be given over all, including the invalid ones, to ensure univque
     * selection in form fields.
     */
    private static $uuid2name_list_id = ["assets" => 1,"crews" => 2,"destinations" => 3,"groups" => 4,
            "persons" => 5,"status" => 6,"waters" => 7
    ];

    /**
     * The data base connection socket.
     */
    private $socket;

    /**
     * The generic toolbox.
     */
    private $toolbox;

    /**
     * For a bulk operation collect first all names and their ids to speed up uniqueness and references checks
     */
    private $uuids_for_names = array();

    /**
     * For a bulk operation with versionized data, make sure only the most recent names are used. This is done
     * during index creation.
     */
    private $invalidform_for_names = array();

    /**
     * collect the ecrids latest valid record per UUID for bulk updates.
     */
    private $table_names = array();

    private static $forever_limit_seconds = 1.0E11;

    /**
     * public Constructor.
     * 
     * @param Tfyh_toolbox $toolbox
     *            application toolbox
     * @param Tfyh_socket $socket
     *            the socket to connect to the database
     */
    public function __construct (Tfyh_toolbox $toolbox, Tfyh_socket $socket)
    {
        $this->toolbox = $toolbox;
        $this->socket = $socket;
    }

    /* --------------------------------------------------------------------------------------- */
    /* --------------- PREMODIFICATION CHECKS AND CORRECTIONS -------------------------------- */
    /* --------------------------------------------------------------------------------------- */
    
    /**
     * Initialize the arrays for checking in bulk operations.
     * 
     * @param int $list_id
     *            the id of the list within the set "../config/lists/uuid2names" which shall be used to
     *            identify the most recent record version.
     * @param bool $force_refresh
     *            set true to force a refeesh, even if the index was already build.
     */
    private function build_indices (int $list_id, bool $force_refresh)
    {
        if (! $force_refresh &&
                 (isset($this->uuids_for_names[$list_id]) && (count($this->uuids_for_names[$list_id]) > 0)))
            return;
        // clear index
        $this->uuids_for_names[$list_id] = [];
        $uuid_names = new Tfyh_list("../config/lists/uuid2names", $list_id, "", $this->socket, $this->toolbox);
        $this->table_names[$list_id] = $uuid_names->get_table_name();
        $col_uuid = $uuid_names->get_field_index("uuid");
        $col_uid = $uuid_names->get_field_index("uid");
        $col_invalid_from = $uuid_names->get_field_index("invalid_from");
        foreach ($uuid_names->get_rows() as $row) {
            $uuid = $row[0];
            // build name index
            if ($list_id == self::$uuid2name_list_id["persons"]) // Special case persons' name
                $name = $row[1] . " " . $row[2];
            else
                $name = $row[1]; // includes names_clubwork
            if ($col_invalid_from === false)
                $this->uuids_for_names[$list_id][$name] = $uuid;
            else {
                // in case of duplicate names (in versionized tables) use the most recent one
                $invalid_from = Tfyh_toolbox::timef($row[$col_invalid_from]);
                if (! isset($this->invalidform_for_names[$list_id][$name]) ||
                         ($invalid_from > $this->invalidform_for_names[$list_id][$name])) {
                    $this->uuids_for_names[$list_id][$name] = $uuid;
                    $this->invalidform_for_names[$list_id][$name] = $invalid_from;
                }
            }
        }
    }

    /**
     * Assert that unique fields are really unique. The result will be appended to the
     * Tfyh_validate::$validation_errors or Tfyh_validate::$validation_warnings.
     * 
     * @param array $record_as_strings
     *            the record to check as associative array of strings
     * @param String $tablename
     *            the tables name to know which fields are system and bool fields.
     * @return no value. The result will be appended to the Tfyh_validate::$validation_errors or
     *         Tfyh_validate::$validation_warnings.
     */
    private function check_name_uniqueness (array $record_as_strings, String $tablename)
    {
        if (! isset(self::$uuid2name_list_id[$tablename]))
            return;
        $this->build_indices(self::$uuid2name_list_id[$tablename], false);
        $name_to_assert = $record_as_strings["name"];
        if (strcmp($tablename, "persons") == 0)
            $name_to_assert = $record_as_strings["firstname"] . " " . $record_as_strings["lastname"];
        $referenced_uuid = $this->uuids_for_names[$tablename][$name_to_assert];
        if (! is_null($referenced_uuid) && strcasecmp($referenced_uuid, $record_as_strings["uuid"]) != 0)
            Tfyh_validate::add_finding(true, 
                    i("fu97I0|Name °%1° is already use...", $name_to_assert, $referenced_uuid));
    }

    /**
     * Check wether alle fields which shall be unique are really unique
     * 
     * @param String $tablename            
     * @param array $record_as_strings
     *            the record to check, all fields as SQL strings
     * @param String $id_field_name
     *            the name of the id field. Set this name for update transactions to identify, whether the
     *            detected overlap is only at record which will be updated.
     * @return true, if the record can be inserted without violation of the uniqueness.
     */
    public function write_check_unique_fields (String $tablename, array $record_as_strings, 
            String $id_field_name = null)
    {
        $record_cfg = $this->toolbox->config->get_by_path(".appTables.$tablename");
        if (is_null($record_cfg))
            return i("ynq0Ng|no table configuration a...", $tablename);
        $findings = "";
        foreach ($record_as_strings as $column => $value_str) {
            if ($record_cfg->has_child($column)) {
                $sql_indexed = $record_cfg->get_child($column)->get_descriptor_field("sql_indexed");
                if (isset($sql_indexed) && (strlen($sql_indexed) > 0) && (strpos($sql_indexed, "u") !== false)) {
                    $already_there = $this->socket->find_records($tablename, $column, $value_str, 2);
                    if ($already_there !== false) {
                        $finding = i("yxn8R3|A record with %1 °%2° al...", $column, $value_str, $tablename) .
                                 " ";
                        if (is_null($id_field_name)) // insert case, no overlap allowed
                            $findings .= i("AQ6jNh|Cannot insert record.") . " " . $finding;
                        elseif (strcmp($record_as_strings[$id_field_name], $already_there[0][$id_field_name]) !=
                                 0) {
                            // update case and different id
                            $ids = " $id_field_name for both: ";
                            $ids .= $record_as_strings[$id_field_name] . "/" .
                                     $already_there[0][$id_field_name];
                            $findings .= i("4vF8NA|Cannot update record.") . " " . $finding . $ids;
                        } elseif (count($already_there) > 1) {
                            // update case and multiple records with this id
                            $findings .= i("2A3jgB|Multiple records with %1...", $column, $value_str, 
                                    $tablename) . " ";
                        }
                    }
                }
            }
        }
        return $findings;
    }

    /**
     * Get a single record as csv, two lines - header and one row.
     * 
     * @param String $tablename
     *            the table from which the record stems. Used for proper formatting
     * @param array $record
     *            the record as was retrieved from the data base.
     * @return string the csv-two-liner.
     */
    public function get_csv (String $tablename, array $record)
    {
        $header = "";
        $values = "";
        $table_cfg = $this->toolbox->config->get_by_path(".appTables." . $tablename);
        foreach ($record as $key => $value) {
            $header .= ";" . $key;
            $type = $table_cfg->get_child($key)->get_type();
            $values .= ";" . Tfyh_toolbox::static_encode_entry_csv(
                    Tfyh_data::format(Tfyh_data::parse($value, $type), $type, "csv"));
        }
        return substr($header, 1) . "\n" . substr($values, 1);
    }

    /**
     * Get the next number based on the existing ones.
     * 
     * @param String $tablename
     *            the table to get the next number for
     * @param String $bookname
     *            the book name, if it is the logbook or the workbook table, else null
     * @return mixed|number
     */
    private function get_next_number (String $tablename, String $bookname = null)
    {
        include_once '../classes/dilbo_config.php';
        $sports_year_start = Dilbo_config::get_sports_year_start($this->toolbox->config);
        if (strcmp($tablename, "logbook") == 0)
            $matching = ["logbookname" => $bookname,"start" => date("Y-m-d", $sports_year_start)
            ];
        elseif (strcmp($tablename, "workbook") == 0)
            $matching = ["workbookname" => $bookname,"date" => date("Y-m-d", $sports_year_start)
            ];
        else
            $matching = [];
        $max_number_records = $this->socket->find_records_sorted_matched($tablename, $matching, 1, "=,>", 
                "number", false);
        if ($max_number_records !== false)
            return $max_number_records[0]["number"];
        return 1;
    }

    /**
     * Check whether needed fields are all set, and return an error, if not.
     * 
     * @param array $record_as_strings
     *            the record to check as associative array of strings
     * @param String $record_item
     *            the record's definition as configuration branch item
     * @param int $mode
     *            Set to mode of operations: 1 = insert, 2 = update.
     * @return no value. The result will be appended to the Tfyh_validate::$validation_errors or
     *         Tfyh_validate::$validation_warnings.
     */
    private function check_not_empty (array $record_as_strings, String $tablename, int $mode)
    {
        // a uid is in every table.
        $check_not_empty_fields = array_merge(self::$check_not_empty_fields[$tablename], ["uid"
        ]);
        // check for insertion of a new or a copy of a record after delimitation, that no needed fields are
        // empty.
        foreach ($check_not_empty_fields as $not_empty_field) {
            $is_missing = ! isset($record_as_strings[$not_empty_field]);
            $is_empty = ! $is_missing && (strlen($record_as_strings[$not_empty_field]) == 0);
            if (($mode == 1) && ($is_missing || $is_empty)) {
                Tfyh_validate::add_finding(true, i("NeU2jA|The required field °%1° ...", $not_empty_field));
            } elseif (($mode == 2) && ! $is_missing && $is_empty)
                Tfyh_validate::add_finding(true, i("nKI7OJ|The required field °%1° ...", $not_empty_field));
        }
    }

    /**
     * If validities are given as dates (valid_from_date, invalid_from_date) replace those by milliseconds
     * into valid_from and invalid_from.
     * 
     * @param array $record_as_strings
     *            the record in which the fields shall be replaced.
     * @return array the modified record
     */
    private function replace_validity_dates (array $record_as_strings)
    {
        $replacements = ["valid_from_date" => "valid_from","invalid_from_date" => "invalid_from"
        ];
        foreach ($replacements as $from => $to) {
            if ((isset($record_as_strings[$from])) && (strlen($record_as_strings[$from]) > 0)) {
                $time = Tfyh_data::parse($record_as_strings[$from], "date");
                $record_as_strings[$to] = $time->format("Uv");
                unset($record_as_strings[$from]);
            }
        }
        return $record_as_strings;
    }

    /**
     * resolve all names into uuids, if found. For single value felds and uuid/name-lists.
     * 
     * @param array $record_as_strings
     *            the record with the names which shall be resolved.
     * @param Tfyh_config_item $record_item
     *            the configuration branch of the table record with the field format and lookup information.
     * @return array the modified record
     */
    private function replace_names_by_uuids (array $record_as_strings, Tfyh_config_item $record_item)
    {
        // run through all fields provided
        foreach ($record_as_strings as $field => $value) {
            // check the fields data type
            if ($record_item->has_child($field)) {
                $field_item = $record_item->get_child($field);
                $data_type = $field_item->get_type();
                if ((strcmp($data_type, "uuid_or_name") == 0) && (strlen($value) > 0)) {
                    // find the reference for the lookup
                    $lookup_table = explode(".", $field_item->get_descriptor_field("column_reference"))[0];
                    $lookup_field = explode(".", $field_item->get_descriptor_field("column_reference"))[1];
                    $uuid2name_list_id = self::$uuid2name_list_id[$lookup_table];
                    // build the name index, or reuse it.
                    $this->build_indices($uuid2name_list_id, false);
                    // resolve all elements. This works for single elements and lists.
                    $resolved = "";
                    $value_parts = explode("|", $value);
                    foreach ($value_parts as $value_part) {
                        if (! Tfyh_validate::is_uuid($value_part)) {
                            $uuid = $this->uuids_for_names[$uuid2name_list_id][$value_part];
                            if (strlen($uuid) == 36)
                                $resolved .= "|" . $uuid;
                            else
                                $resolved .= "|" . $value_part;
                        }
                    }
                    // replace field
                    $record_as_strings[$field] = substr($resolved, 1);
                }
            }
        }
        return $record_as_strings;
    }

    /**
     * Get all reservations for the boat with the $reservation_record["BoatId"] value and check whether any
     * overlap.
     * 
     * @param array $reservation_record
     *            the reservation record to check
     * @return false in case of no overlap, else a String representation of the overlapping reservation.
     */
    // =============================
    // NOT TESTED; EFACLOUD RAW CODE
    // =============================
    private function has_overlap_reservation (array $reservation_record)
    {
        $all_reservations = $this->socket->find_records_sorted_matched("efa2boatreservations", 
                ["BoatId" => $reservation_record["BoatId"]
                ], 1000, "=", "DateFrom", false);
        if ($all_reservations === false)
            return false;
        if (strcasecmp($reservation_record["Type"], "WEEKLY") == 0) {
            $time_from_a = strtotime($reservation_record["DateTo"]);
            $time_to_a = strtotime($reservation_record["TimeTo"]);
        } else {
            $time_from_a = strtotime($reservation_record["DateFrom"] . " " . $reservation_record["TimeFrom"]);
            $time_to_a = strtotime($reservation_record["DateTo"] . " " . $reservation_record["TimeTo"]);
        }
        $is_weekly_record = (strcasecmp($reservation_record["Type"], "WEEKLY") == 0);
        foreach ($all_reservations as $reservation_to_check) {
            $is_this_reservation = (strcmp($reservation_to_check["ecrid"], $reservation_record["ecrid"]) == 0);
            $is_weekly_check = (strcasecmp($reservation_to_check["Type"], "WEEKLY") == 0);
            // must be different from the one provided, but must be of same type (see efa code for reference
            // of condiition
            if (! $is_this_reservation) {
                if ($is_weekly_record == $is_weekly_check) {
                    if ($is_weekly_record) {
                        if (strcasecmp($reservation_to_check["DayOfWeek"], $reservation_record["DayOfWeek"]) ==
                                 0) {
                            // compare weekly reservations, use time only
                            $time_from_b = strtotime($reservation_to_check["TimeFrom"]);
                            $time_to_b = strtotime($reservation_to_check["TimeTo"]);
                        } else {
                            // different days of the week, invalidate time for further check
                            $time_from_b = - 1;
                            $time_to_b = - 1;
                        }
                    } else {
                        // compare one-time reservations
                        $time_from_b = strtotime(
                                $reservation_to_check["DateFrom"] . " " . $reservation_to_check["TimeFrom"]);
                        $time_to_b = strtotime(
                                $reservation_to_check["DateTo"] . " " . $reservation_to_check["TimeTo"]);
                    }
                    $a_starts_within_b = ($time_from_a > $time_from_b) && ($time_from_a < $time_to_b);
                    $a_ends_within_b = ($time_to_a > $time_from_b) && ($time_to_a < $time_to_b);
                    $a_includes_b = ($time_from_a <= $time_from_b) && ($time_to_a >= $time_to_b);
                    if ($a_starts_within_b || $a_includes_b || $a_ends_within_b)
                        return i("AebG0R|The boat is occupied wit...", $reservation_to_check["Reason"], 
                                $reservation_to_check["DateFrom"], $reservation_to_check["TimeFrom"], 
                                $reservation_to_check["DateTo"], $reservation_to_check["TimeTo"]);
                }
            }
        }
        return false;
    }

    /**
     * Parse and validate the record. Perform all needed checks and if successfull execute the modification at
     * the data base.
     * 
     * @param array $record_as_strings
     *            the record to modify as associative array of strings
     * @param String $record_item
     *            the record's definition as configuration branch item
     * @param int $mode
     *            Set 1 for insert, 2 for update, 3 for delete
     * @param int $user_id
     *            The user ID which is put to the ecrown field in insert mode.
     * @return String the data base modification transaction result. The validation result will be appended to
     *         the Tfyh_validate::$validation_errors or Tfyh_validate::$validation_warnings.
     */
    public function modify_record (array $record_as_strings, Tfyh_config_item $record_item, int $mode, 
            int $user_id)
    {
        // prepare
        $tablename = $record_item->get_name();
        Tfyh_data::clear_findings();
        Tfyh_validate::clear_findings();
        
        // parse, validate and format the record, if it is to be inserted or updated.
        if ($mode < 3) {
            
            // replace dates by millis and names by uuids
            $this->replace_validity_dates($record_as_strings);
            $this->replace_names_by_uuids($record_as_strings, $record_item);
            // now perform data conversion into native and formatting to ensure proper syntax for sql
            $record_native = Tfyh_validate::parse_and_validate_record($record_as_strings, $record_item);
            $formatted = Tfyh_validate::format_parsed($record_native, $record_item, "sql");
            // apply duplicate name check for tables with names.
            if (array_key_exists($tablename, self::$uuid2name_list_id))
                $this->check_name_uniqueness($record_as_strings, $tablename);
            $formatted["modified"] = strval(floor(Tfyh_toolbox::timef()));
            
            // add fields on insert
            if ($mode == 1) {
                if (in_array($tablename, self::$numbered_tables)) {
                    // enforce the number uniqueness for numbered tables on insert
                    $bookname = (strcmp($tablename, "logbook") == 0) ? $record_as_strings["logbookname"] : $record_as_strings["workbookname"];
                    $number = $this->get_next_number($tablename, $bookname);
                    if (! isset($formatted["number"]) || (intval($formatted["number"]) <= 0))
                        $formatted["number"] = strval($number + 1);
                }
                $formatted["created_by"] = strval($this->toolbox->users->session_user["@id"]);
                $formatted["created_on"] = $formatted["modified"];
            }
            
            // remove or replace null values for not-null fields
            foreach ($record_item->get_children() as $fieldname => $field_definition) {
                $sql_null = $field_definition->get_descriptor_field("sql_null");
                $type_is_string = (strcmp(Tfyh_data::parse_as($field_definition->get_type()), "string") == 0);
                $value_is_null = ! $type_is_string && (strcasecmp($formatted[$fieldname], "NULL") == 0);
                if ($value_is_null)
                    if ($sql_null === false)
                        unset($formatted[$fieldname]);
                    else
                        // the formatter creates a String "NULL", but the socket needs a null-String to build
                        // the sql command `field` = NULL instead of `field` = "NULL".
                        $formatted[$fieldname] = null;
            }
            
            // apply not-empty checks
            $this->check_not_empty($formatted, $tablename, $mode);
            
            // now modify the record, all was fine.
            if (! Tfyh_validate::has_findings(false)) {
                if ($mode == 1) {
                    // insert record
                    $insert_result = $this->socket->insert_into($user_id, $tablename, $formatted);
                    if (! is_numeric($insert_result))
                        return i("tZT02P|Database error: %1. Fail...", $insert_result, $tablename);
                    return "";
                } elseif ($mode == 2) {
                    // update record
                    $matching_key = ["uid" => $formatted["uid"]
                    ];
                    $update_result = $this->socket->update_record_matched($user_id, $tablename, $matching_key, 
                            $formatted);
                    if (strlen($update_result) > 0)
                        return i("Jq2YEP|Database error: %1. Fail...", $update_result, $tablename);
                    return "";
                }
            } else {
                // errors were found. Return errors and warnings.
                return Tfyh_validate::get_findings(true);
            }
        } elseif ($mode == 3) {
            // special case persons: you must not delete your own user record.
            if (strcasecmp($tablename, "persons") == 0) {
                if (intval($record_as_strings["user_id"]) == intval(
                        $this->toolbox->users->session_user["@id"]))
                    return i("UgRFql|Handling error. You are ...");
            }
            
            // find the record
            $matching_key = ["uid" => $record_as_strings["uid"]
            ]; // it will be the same key for both the deleted record and the trash record.
            $record_to_delete = $this->socket->find_record_matched($tablename, $matching_key);
            if ($record_to_delete === false)
                return i("7lryij|Database error. Unable t...", $record_as_strings["uid"], $tablename);
            
            // create a trash record
            $rhistory_of_deleted = $record_to_delete["history"];
            unset($record_to_delete["history"]);
            $trashed_record = json_encode($record_to_delete);
            // limit size to 64k
            $cut_len = 65535 - 4096;
            while (strlen($trashed_record) > 65535) { // strlen == byte length
                foreach ($record_to_delete as $key => $value)
                    if (strlen($value) > $cut_len)
                        $record_to_delete[$key] = substr(strval($record_to_delete[$key]), 0, $cut_len);
                $trashed_record = json_encode($record_to_delete);
                $cut_len = $cut_len - 4096;
            }
            $trash_record = ["uid" => $record_to_delete["uid"],"author" => $user_id,"table" => $tablename,
                    "record" => $trashed_record,"rhistory" => $rhistory_of_deleted
            ]; // modified will be added by the data base.
               
            // insert the trash record
            $insert_result = $this->socket->insert_into($user_id, "trash", $trash_record);
            if (! is_numeric($insert_result))
                return i("zhlDnH|Database error: %1. Fail...", $insert_result);
            
            // delete the record
            $delete_result = $this->socket->delete_record_matched($user_id, $tablename, $matching_key);
            if (strlen($update_result) > 0) {
                // try to remove the trash_record, but ignore errors for this rubbish bin cleansing
                $this->socket->delete_record_matched($user_id, "trash", $matching_key);
                return i("6nchtK|Database error: %1. Fail...", $delete_result, $tablename);
            }
            // all went fine.
            return "";
        }
    }
}
