<?php

/**  
 *
 *       dilbo - digital logbook for Rowing and Canoeing
 *       -----------------------------------------------
 *       https://www.dilbo.org
 *
 * Copyright  2023-2024  Martin Glade
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */
include_once '../classes/dilbo_api.php';
include_once '../classes/dilbo_record.php';

/**
 * The transaction handling class. Clients are the dilbo web-Application or the dilbo Java application or
 * other.
 */
class Dilbo_tx_handler
{

    /**
     * The message separator String and its replacement, if hit within a transaction, as well as csv special
     * characters.
     */
    private static $ms = "\n-#|#-\n";

    private static $msr = "\n_#|#_\n";

    /**
     * The API versions supported. Version 1 through 3 are for efaCloud usage.
     */
    private $api_supported_versions = [4
    ];

    /**
     * The generic toolbox.
     */
    private $toolbox;

    /**
     * the Efa API class providung API transaction support.
     */
    private $dilbo_api;

    /**
     * The data base connection socket.
     */
    private $socket;

    /**
     * the currently handled request.
     */
    public $txc;

    /**
     * the transaction log
     */
    public static $api_version = 4;

    /**
     * the transaction log
     */
    public static $api_log_path = "../log/api_info.log";

    /**
     * the transaction error log
     */
    public static $api_error_log_path = "../log/api_errors.log";

    /**
     * the transaction warnings log
     */
    private static $api_warning_log_path = "../log/api_warnings.log";

    /**
     * the transaction warnings log, also used by Efa_api
     */
    private static $api_debug_log_path = "../log/debug_api.log";

    /**
     * The timestamp of the transaction start, created by init.php and posttx.php respectively
     */
    private $php_script_started_at;

    /**
     * Debug level to add mor information for support cases.
     */
    private $debug_on;

    /**
     * public Constructor.
     * 
     * @param Tfyh_toolbox $toolbox
     *            the application basic utilities
     * @param Tfyh_socket $socket
     *            the socket to connect to the data base
     * @param float $php_script_started_at
     *            the time the script started for performance analysis
     */
    public function __construct (Tfyh_toolbox $toolbox, Tfyh_socket $socket, float $php_script_started_at)
    {
        $this->php_script_started_at = $php_script_started_at;
        $this->toolbox = $toolbox;
        $this->socket = $socket;
        $cfg = $this->toolbox->config->get_cfg();
        $this->debug_on = $toolbox->config->debug_level > 0;
        $this->dilbo_api = new Dilbo_api($toolbox, $socket);
    }

    /**
     * Gather statistics on the exchanged content size for those who have bandwith limitations.
     * 
     * @param int $request_size
     *            the size of the request
     * @param int $response_size
     *            the size of the response
     * @param Tfyh_toolbox $toolbox
     *            the toolbox of this application
     * @param int $client_id
     *            the user ID of the client
     */
    public static function log_content_size (int $request_size, int $response_size, Tfyh_toolbox $toolbox, 
            int $client_id)
    {
        $size_filename = "../log/contentsize";
        if (! file_exists($size_filename))
            mkdir($size_filename);
        $size_filename .= "/" . $client_id;
        $size_table_header = "Date;requests;requSize;respSize";
        $today = date("Y-m-d");
        $latest_recent = time() - 1209600; // 14 days in seconds
        if (file_exists($size_filename))
            $sizes = $toolbox->read_csv_array($size_filename);
        else
            $sizes = [];
        $today_in = false;
        $start = (count($sizes) > 14) ? count($sizes) - 14 : 0;
        for ($i = $start; $i < count($sizes); $i ++) {
            $size = $sizes[$i];
            if ($sizes[$i]["Date"] == $today) {
                $sizes[$i]["requests"] = strval(intval($sizes[$i]["requests"]) + 1);
                $sizes[$i]["requSize"] = strval(intval($sizes[$i]["requSize"]) + $request_size);
                $sizes[$i]["respSize"] = strval(intval($sizes[$i]["respSize"]) + $response_size);
                $today_in = true;
            }
        }
        if (! $today_in) {
            $size["Date"] = $today;
            $size["requests"] = 1;
            $size["requSize"] = $request_size;
            $size["respSize"] = $response_size;
            $sizes[] = $size;
        }
        $out = $size_table_header . "\n";
        foreach ($sizes as $size) {
            if (strtotime($size["Date"]) >= $latest_recent)
                $out .= $size["Date"] . ";" . $size["requests"] . ";" . $size["requSize"] . ";" .
                         $size["respSize"] . "\n";
        }
        file_put_contents($size_filename, $out);
    }

    /**
     * Print a log string with all transaction container information for logging
     * 
     * @return the printed log String
     */
    private function txc_to_log ()
    {
        $log_string = "version:" . $this->txc["version"] . ", ";
        $log_string .= "cID:" . $this->txc["cID"] . ", ";
        $log_string .= "userID:" . $this->txc["userID"] . ", ";
        $log_string .= "password (length):" . strlen($this->txc["password"]) . ", ";
        $log_string .= "cresult_code:" . $this->txc["cresult_code"] . ", ";
        $log_string .= "cresult_message:" . $this->txc["cresult_message"] . "\n";
        return $log_string;
    }

    /**
     * Print a log string with all transaction information for logging
     * 
     * @param int $i
     *            the index of the transaction within the container
     * @param int $withMessageAndRecord
     *            set true to get a version with the transaction record and the result message (only error
     *            log).
     * @return the printed log String
     * @return string
     */
    public function tx_to_log (int $i, bool $withMessageAndRecord)
    {
        $tx_request = $this->txc["requests"][$i];
        $log_string = "(V" . $this->txc["version"] . ") ";
        $log_string = "client:" . $this->txc["userID"] . ", ";
        $log_string = "ID:" . $tx_request["ID"] . ", ";
        $log_string .= "type:" . $tx_request["type"] . ", ";
        // for the list transaction the table name is the list name, which is plain text rather than a
        // technical name.
        $log_string .= "tablename:" . i($tx_request["tablename"]) . ", ";
        $log_string .= "result_code:" . $tx_request["result_code"] . ", ";
        if ($withMessageAndRecord) {
            $contents = json_encode($tx_request["record"]);
            if (strlen($contents) > 2000)
                $contents = substr($contents, 0, 1997) . "...";
            $log_string .= "record:" . json_encode($tx_request["record"]) . " // ";
            $log_string .= "result_message:" . str_replace("\n", " // ", trim($tx_request["result_message"]));
        } else {
            $log_string .= "record length:" .
                     ((isset($tx_request["record"])) ? count($tx_request["record"]) : 0);
            $log_string .= ", result_message length:" . strlen($tx_request["result_message"]);
            $log_string .= ", lines:" . count(explode("\n", trim($tx_request["result_message"])));
        }
        
        return $log_string;
    }

    /**
     * Parse a transaction container according to the API format definition and put it to the $this->txc
     * variable for further processing. The header is checked for a version, containerID, user and password
     * entry and at least one transaction. All these elements must be present, the first three numeric. If the
     * check fails, the shall nevertheless be sent, providing the API client the failure reason.
     * 
     * @param String $txc_base64_api
     *            the transaction container, still in base64 API type encoding
     * @return nothing. See the txc["cresult_code"] for the parsing result.
     */
    public function parse_request_container (String $txc_base64_api)
    {
        if ($this->debug_on)
            file_put_contents(self::$api_debug_log_path, 
                    date("Y-m-d H:i:s") . ": parse_request_container of " . strlen($txc_base64_api) .
                             " characters length.\n", FILE_APPEND);
        
        // create container header array
        $this->txc = [];
        $this->txc["version"] = 0;
        $this->txc["cID"] = 0;
        $this->txc["userID"] = 0;
        $this->txc["password"] = 'none';
        $this->txc["cresult_code"] = 401;
        $this->txc["cresult_message"] = 'Syntax error in request.';
        $this->txc["requests"] = [];
        
        if (($txc_base64_api == null) || (strlen($txc_base64_api) < 5)) {
            $this->txc["cresult_message"] .= 'Transaction container missing, empty or too short.';
            file_put_contents(self::$api_log_path, 
                    "[" . date("Y-m-d H:i:s") . "] - Container: content empty or too short: '" .
                             json_encode($txc_base64_api) . "' // ", FILE_APPEND);
            if ($this->debug_on)
                file_put_contents(self::$api_debug_log_path, 
                        date("Y-m-d H:i:s") . ": parse_request_container failed. " .
                                 $this->txc["cresult_message"] . ".\n", FILE_APPEND);
            return;
        }
        
        // decode base64-API and split into header and requests
        $txc_plain = base64_decode(
                str_replace("_", "=", str_replace("-", "/", str_replace("*", "+", $txc_base64_api))));
        $celements = explode(";", $txc_plain, 5);
        // check container syntax
        if (count($celements) != 5) {
            $this->txc["cresult_message"] .= 'Decoded transaction container has too few elements: ' .
                     count($celements);
            file_put_contents(self::$api_log_path, 
                    "[" . date("Y-m-d H:i:s") . "] - Container: has too few elements.\n", FILE_APPEND);
            if ($this->debug_on)
                file_put_contents(self::$api_debug_log_path, 
                        date("Y-m-d H:i:s") . ": parse_request_container failed. " .
                                 $this->txc["cresult_message"] . ".\n", FILE_APPEND);
            return;
        }
        
        // check API version
        $api_request_version = intval($celements[0]);
        if ($api_request_version != 4) {
            $this->txc["cresult_code"] = 400;
            $this->txc["cresult_message"] = 'API version of container not supported.';
            file_put_contents(self::$api_log_path, 
                    "[" . date("Y-m-d H:i:s") .
                             "] - Container: API version $api_request_version of container not supported.\n", 
                            FILE_APPEND);
            if ($this->debug_on)
                file_put_contents(self::$api_debug_log_path, 
                        date("Y-m-d H:i:s") .
                                 ": API version $api_request_version of container not supported.\n", 
                                FILE_APPEND);
            return;
        }
        
        $this->txc["version"] = $api_request_version;
        $this->txc["cID"] = intval($celements[1]);
        if (! in_array($api_request_version, $this->api_supported_versions)) {
            $error_msg = 'Transaction container version ' . $api_request_version . ' not supported.';
            $this->txc["cresult_message"] .= $error_msg;
            $this->txc["version"] = $api_request_version;
            file_put_contents(self::$api_log_path, 
                    "[" . date("Y-m-d H:i:s") . "] - Container: " . $error_msg . ".\n", FILE_APPEND);
            if ($this->debug_on)
                file_put_contents(self::$api_debug_log_path, 
                        date("Y-m-d H:i:s") . ": parse_request_container failed. " . $error_msg . ".\n", 
                        FILE_APPEND);
            return;
        }
        
        // update container header array
        $user_id = intval($celements[2]);
        $this->txc["userID"] = $user_id;
        if ($this->txc["version"] * $this->txc["cID"] * $this->txc["userID"] == 0) {
            $this->txc["cresult_message"] .= "One of version, cID, or userID is either not numeric or missing.";
            file_put_contents(self::$api_log_path, 
                    "[" . date("Y-m-d H:i:s") .
                             "] - Container: One of version, cID, or userID is either not numeric or missing.\n", 
                            FILE_APPEND);
            if ($this->debug_on)
                file_put_contents(self::$api_debug_log_path, 
                        date("Y-m-d H:i:s") . ": parse_request_container failed. " .
                                 $this->txc["cresult_message"] . ".\n", FILE_APPEND);
            return;
        }
        $this->txc["password"] = $celements[3];
        $this->txc["cresult_code"] = 301;
        $this->txc["cresult_message"] = "Syntax ok. User to be verified.";
        
        // parse requests and add them to the container array
        $txc_requests = explode(self::$ms, $celements[4]);
        $container_description = "Container V" . $this->txc["version"] . ": Received from " .
                 $this->txc["userID"] . " with " . count($txc_requests) . " transaction requests.";
        file_put_contents(self::$api_log_path, "[" . date("Y-m-d H:i:s") . "] - $container_description\n", 
                FILE_APPEND);
        if ($this->debug_on)
            file_put_contents(self::$api_debug_log_path, 
                    date("Y-m-d H:i:s") . ": parse_request_container header completed: " .
                             $container_description . ".\n", FILE_APPEND);
        $this->txc["requests"] = [];
        foreach ($txc_requests as $tx_request) {
            // split request header and record
            $elements = str_getcsv($tx_request, ";", "\"");
            $tx = [];
            if ((count($elements) < 3) || (count($elements) % 2 != 1)) {
                // Note: this check above is different from efaCloud, since the retry field is no more
                // provided. It was ((count($elements) < 4) || (count($elements) % 2 != 0)).
                $tx["ID"] = (isset($elements[0])) ? $elements[0] : "0";
                $tx["type"] = (isset($elements[1])) ? $elements[1] : "parsing_error";
                $tx["tablename"] = (isset($elements[2])) ? $elements[2] : "undefined";
                $tx["result_code"] = 501;
                $tx["result_message"] = "invalid count of parameters in transaction request: " .
                         count($elements);
            } else {
                $tx["ID"] = $elements[0];
                $tx["type"] = $elements[1];
                $tx["tablename"] = $elements[2];
                $tx["result_code"] = 900;
                $tx["result_message"] = "not yet parsed nor processed";
                $tx_record = [];
                for ($i = 3; $i < count($elements); $i = $i + 2)
                    $tx_record[$elements[$i]] = $elements[$i + 1];
                $tx["record"] = $tx_record;
            }
            if ($this->debug_on)
                file_put_contents(self::$api_debug_log_path, 
                        date("Y-m-d H:i:s") . ": parse_request_container transaction completed: #" . $tx["ID"] .
                                 ": " . $tx["type"] . " " . $tx["tablename"] . ": " . $tx["result_code"] . " " .
                                 $tx["result_message"] . ".\n", FILE_APPEND);
            $this->txc["requests"][] = $tx;
        }
    }

    /**
     * This method verifies the transaction token and sets the current transaction array.
     * 
     * @param array $client_verified
     *            the verified client which requests the execution
     * @param Tfyh_menu $menu
     *            used to identify whether the enclosed transactions are permitted for the user
     * @return nothing. All results are put into the transaction container.
     */
    public function handle_request_container (array $client_verified, Tfyh_menu $menu)
    {
        if ($this->debug_on)
            file_put_contents(self::$api_debug_log_path, 
                    date("Y-m-d H:i:s") . ": handle_request_container started.\n", FILE_APPEND);
        for ($i = 0; $i < count($this->txc["requests"]); $i ++) {
            file_put_contents(self::$api_log_path, 
                    "[" . date("Y-m-d H:i:s") . "] - Transaction: to execute " . $this->tx_to_log($i, false) .
                             "\n", FILE_APPEND);
            $this->execute_transaction($client_verified, $i, $menu);
            $isError = (intval($this->txc["requests"][$i]["result_code"]) >= 400);
            if ($isError) {
                file_put_contents(self::$api_error_log_path, 
                        "[" . date("Y-m-d H:i:s") . "] - Transaction: did not execute " .
                                 $this->tx_to_log($i, true) . "\n", FILE_APPEND);
                file_put_contents(self::$api_log_path, 
                        "[" . date("Y-m-d H:i:s") . "] - Transaction: failed " . $this->tx_to_log($i, true) .
                                 "\n", FILE_APPEND);
            } else
                file_put_contents(self::$api_log_path, 
                        "[" . date("Y-m-d H:i:s") . "] - Transaction: executed " . $this->tx_to_log($i, false) .
                                 "\n", FILE_APPEND);
        }
    }

    /**
     * This method logs the dropped transactions of a failed container.
     */
    private function log_dropped_container_transactions ()
    {
        for ($i = 0; $i < count($this->txc["requests"]); $i ++)
            file_put_contents(self::$api_warning_log_path, 
                    "[" . date("Y-m-d H:i:s") . "] - Transaction: log_dropped_container_transactions. " .
                             $this->tx_to_log($i, true) . "\n", FILE_APPEND);
    }

    /**
     * Take the current transaction container, build an appropriate response container from it and send it
     * back. Transactions must be handled before building the response. The response will have the same
     * verison as the request, even if the server can do more.
     * 
     * @return this function will not return, but send a response back to the API user and exit.
     */
    public function send_response_and_exit ()
    {
        file_put_contents(self::$api_log_path, 
                "[" . date("Y-m-d H:i:s") . "] - Container: send response " . $this->txc_to_log(), FILE_APPEND);
        if ($this->debug_on)
            file_put_contents(self::$api_debug_log_path, 
                    date("Y-m-d H:i:s") . ": send_response - " . $this->txc_to_log(), FILE_APPEND);
        $cresult_message = $this->txc["cresult_message"];
        $cresult_message = str_replace(self::$ms, self::$msr, $cresult_message);
        if ((strpos($cresult_message, ";") !== false) || (strpos($cresult_message, '"') !== false))
            $cresult_message = '"' . str_replace('"', '""', $cresult_message) . '"';
        
        $api_user_id = intval($this->txc["userID"]);
        if (intval($this->txc["cresult_code"]) >= 400) {
            file_put_contents(self::$api_error_log_path, 
                    "[" . date("Y-m-d H:i:s") . "] - Container failed for user " . $api_user_id .
                             ". Result code: " . $this->txc["cresult_code"] . ". Result message: " .
                             $this->txc["cresult_message"] . "\n", FILE_APPEND); //
            $this->log_dropped_container_transactions();
        }
        
        $resp = $this->txc["version"] . ";" . $this->txc["cID"] . ";" . $this->txc["cresult_code"] . ";" .
                 $cresult_message . ";";
        $timestamp_txs = "";
        for ($i = 0; $i < count($this->txc["requests"]); $i ++) {
            $resp .= $this->txc["requests"][$i]["ID"] . ";" . $this->txc["requests"][$i]["result_code"] . ";";
            $result_message = $this->txc["requests"][$i]["result_message"];
            $result_message = str_replace(self::$ms, self::$msr, $result_message);
            if (intval($this->txc["requests"][$i]["result_code"]) >= 400) {
                file_put_contents(self::$api_error_log_path, 
                        "[" . date("Y-m-d H:i:s") . "] - Transaction: " . $this->txc["requests"][$i]["type"] .
                                 " transaction " . $this->txc["requests"][$i]["ID"] . " (" .
                                 $this->txc["requests"][$i]["tablename"] . ") failed for user " . $api_user_id .
                                 ". Result code: " . $this->txc["requests"][$i]["result_code"] .
                                 ". Result message: " . $this->txc["requests"][$i]["result_message"] . "\n", 
                                FILE_APPEND);
            }
            $timestamp_txs = (strlen($timestamp_txs) == 0) ? "api/" .
                     strtolower($this->txc["requests"][$i]["type"]) : "api/multiple";
            
            // the result message neither needs utf-8 encoding (the values are already encoded) nor
            // csv encoding (the values are as well already approporiately quotes).
            $resp .= $result_message . self::$ms;
        }
        if (count($this->txc["requests"]) > 0)
            $resp = substr($resp, 0, strlen($resp) - strlen(self::$ms));
        $resp = str_replace("=", "_", str_replace("/", "-", str_replace("+", "*", base64_encode($resp))));
        $content_size = (isset($_SERVER['CONTENT_LENGTH'])) ? intval($_SERVER['CONTENT_LENGTH']) : 0;
        self::log_content_size($content_size, strlen($resp), $this->toolbox, $api_user_id);
        if ($this->debug_on)
            file_put_contents(self::$api_debug_log_path, 
                    date("Y-m-d H:i:s") . ": send_response - start streaming " . strlen($resp) .
                             " characters.\n", FILE_APPEND);
        
        // this echo below is the resonse sending
        echo $resp;
        // this echo above is the resonse sending
        
        $user_id = (isset($this->txc["userID"])) ? intval($this->txc["userID"]) : - 1;
        $this->toolbox->logger->put_timestamp($user_id, $timestamp_txs, $this->php_script_started_at);
        if ($this->debug_on)
            file_put_contents(self::$api_debug_log_path, 
                    date("Y-m-d H:i:s") . ": send_response - completed.\n", FILE_APPEND);
        exit();
    }

    /**
     * This method executes a single transaction and returns the result message. The result_code and the
     * result_message fields of the transaction are also set according to the transaction result.
     * 
     * @param array $client_verified
     *            the verified client which requests the execution
     * @param array $transaction_request
     *            the transaction request which shall be parsed and processed. It will be changed.
     * @param Tfyh_menu $menu
     *            used to identify whether the enclosed transactions are permitted for the user
     * @return nothing. All results are put into the transaction container.
     */
    private function execute_transaction (array $client_verified, int $index, Tfyh_menu $menu)
    {
        if ($this->debug_on) {
            $log_string = date("Y-m-d H:i:s") . ": execute_transaction #" .
                     $this->txc["requests"][$index]["ID"] . " " . $this->txc["requests"][$index]["type"] . " " .
                     $this->txc["requests"][$index]["tablename"] . " " .
                     json_encode($this->txc["requests"][$index]["record"]) . "\n";
            file_put_contents(self::$api_debug_log_path, $log_string, FILE_APPEND);
        }
        
        $api_user_id = $client_verified[$this->toolbox->users->user_id_field_name];
        $api_version = $this->txc["version"];
        
        $txtype = $this->txc["requests"][$index]["type"];
        $txtablename = $this->txc["requests"][$index]["tablename"];
        $record = $this->txc["requests"][$index]["record"];
        $result_code = $this->txc["requests"][$index]["result_code"];
        $tx_response = "501;programming fault. Please raise a support request.";
        
        $type_recognized = true;
        // check user rights
        $tx_type = $this->txc["requests"][$index]["type"];
        $tx_tablename = $this->txc["requests"][$index]["tablename"];
        $transaction_path = "../api/" . $tx_type;
        $is_allowed = $menu->is_allowed_menu_item($transaction_path, $client_verified);
        if (! $is_allowed) {
            $tx_response = "502;Transaction '" . $this->txc["requests"][$index]["type"] .
                     "' not allowed in table " . $tx_tablename . " for role '" . $client_verified["Rolle"] .
                     "'";
            if ($this->debug_on)
                file_put_contents(self::$api_debug_log_path, 
                        date("Y-m-d H:i:s") .
                                 ":   aborting execute_transaction because of insufficient user rights.\n", 
                                FILE_APPEND);
        } elseif ($result_code != 900)
            $tx_response = $result_code . ";" . $this->txc["requests"][$index]["result_message"];
        
        // Write data (no keyfixing like in efaCloud)
        elseif (strcasecmp($txtype, "insert") == 0)
            $tx_response = $this->dilbo_api->api_modify($client_verified, $txtablename, $record, 1, 
                    self::$api_debug_log_path);
        elseif (strcasecmp($txtype, "update") == 0)
            $tx_response = $this->dilbo_api->api_modify($client_verified, $txtablename, $record, 2, 
                    self::$api_debug_log_path);
        elseif (strcasecmp($txtype, "delete") == 0)
            $tx_response = $this->dilbo_api->api_modify($client_verified, $txtablename, $record, 3, 
                    self::$api_debug_log_path);
        
        // Read data: no synch, no select. Only lists allowed
        elseif (strcasecmp($txtype, "list") == 0) {
            if (strcasecmp($txtablename, "statistics") == 0)
                $tx_response = $this->dilbo_api->api_statistics($client_verified, $record);
            else
                $tx_response = $this->dilbo_api->api_list($client_verified, $txtablename, $record);
        } //
          
        // Support functions - nop (dilbo clients), session (start, regenerate, close), config (get config
          // settings file) and info
        elseif (strcasecmp($txtype, "nop") == 0)
            $tx_response = $this->dilbo_api->api_nop($client_verified, $record);
        elseif (strcasecmp($txtype, "session") == 0)
            $tx_response = $this->dilbo_api->api_session($client_verified, $this->txc["session_id"], 
                    $txtablename);
        elseif (strcasecmp($txtype, "config") == 0)
            $tx_response = $this->dilbo_api->api_config($client_verified, $txtablename);
        elseif (strcasecmp($txtype, "info") == 0) {
            // include_once "../classes/dilbo_info.php";
            // $dilbo_info = new Dilbo_info($this->toolbox, $this->socket);
            // $tx_response = $dilbo_info->api_info($client_verified, $record);
            $tx_response = "300;not yet implemented";
        } elseif (strcasecmp($txtype, "parsing_error") == 0)
            // parsing error
            $tx_response = $this->txc["requests"][$index]["result_code"] . ";" .
                     $this->txc["requests"][$index]["result_message"];
        
        // Error to return on unrecognized transaction type
        if (! $type_recognized) {
            // 501 => "Transaction invalid."
            $tx_response = "501;Transaction invalid. Invalid type used: " . $txtype . " for table " .
                     $txtablename . " (API version of request = " . $api_version . ").";
            if ($this->debug_on)
                file_put_contents(self::$api_debug_log_path, 
                        date("Y-m-d H:i:s") . ":   execute_transaction with invalid function code.\n", 
                        FILE_APPEND);
        }
        
        // pass the result to the transaction
        $result_code = substr($tx_response, 0, 3);
        $this->txc["requests"][$index]["result_code"] = $result_code;
        $this->txc["requests"][$index]["result_message"] = substr($tx_response, 4);
        
        // time stamp all requests for the fast synch option and last activity display
        if (intval($this->txc["requests"][$index]["result_code"]) < 400) {
            $wasWriteAccess = ((strcasecmp($txtype, "insert") == 0) || (strcasecmp($txtype, "update") == 0) ||
                     (strcasecmp($txtype, "delete") == 0));
            if ($wasWriteAccess) {
                if ($this->debug_on)
                    file_put_contents(self::$api_debug_log_path, 
                            date("Y-m-d H:i:s") . ": execute_transaction remembered write access.\n", 
                            FILE_APPEND);
                file_put_contents("../log/lwa/" . strval($api_user_id), strval(time()));
                // TODO: Noticiations
                // notification of new reservations, damages or admin messages
                // include_once "../classes/dilbo_notifier.php";
                // $dilbo_notifier = new Dilbo_notifier($this->toolbox, $this->socket);
                // $dilbo_notifier->notify_api_write_event($this->txc["requests"][$index]);
            } else {
                file_put_contents("../log/lra/" . strval($api_user_id), date("Y-m-d H:i:s"));
            }
        }
        if ($this->debug_on)
            file_put_contents(self::$api_debug_log_path, 
                    date("Y-m-d H:i:s") . ": execute_transaction #" . $this->txc["requests"][$index]["ID"] .
                             " completed with result: " . $this->txc["requests"][$index]["result_code"] . ":" .
                             substr($this->txc["requests"][$index]["result_message"], 0, 250) . " ... .\n", 
                            FILE_APPEND);
    }
}