<?php
/**  
 *
 *       dilbo - digital logbook for Rowing and Canoeing
 *       -----------------------------------------------
 *       https://www.dilbo.org
 *
 * Copyright  2023-2024  Martin Glade
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

include_once '../classes/tfyh_list.php';

/**
 * class file for resolving the UUIDs into names.
 */
class Dilbo_uuids
{

    /**
     * The data base connection socket. 
     */
    private $socket;

    /**
     * The generic toolbox.
     */
    private $toolbox;

    /**
     * a set of all Names of boats, destinations, persons and waters as associative array with the UUID being
     * the key.
     */
    private $names;

    /**
     * a set of all UUIDs of boats, destinations, persons and waters as associative array with the name (for
     * persons full name = first name + last name) being the key.
     */
    private $uuids;

    /**
     * public Constructor.
     * 
     * @param Tfyh_toolbox $toolbox
     *            application toolbox
     * @param Tfyh_socket $socket
     *            the socket to connect to the database
     */
    public function __construct (Tfyh_toolbox $toolbox, Tfyh_socket $socket)
    {
        $this->socket = $socket;
        $this->toolbox = $toolbox;
        $this->names = [];
        $this->uuids = [];
    }

    /**
     * Collect for a single table all UUIDs and reference arrays
     * 
     * @param int $list_id
     *            the ID of the list to use
     * @return int the count of records found.
     */
    private function collect_arrays_per_table (int $list_id)
    {
        $list = new Tfyh_list("../config/lists/uuid2names", $list_id, "", $this->socket, $this->toolbox);
        $records_found = 0;
        $tablename = $list->get_table_name();
        $rows = $list->get_rows();
        $uuid_index = $list->get_field_index("uuid");
        $name_index = $list->get_field_index("name");
        $first_name_index = $list->get_field_index("first_name");
        $last_name_index = $list->get_field_index("last_name");
        $is_persons = (strpos($tablename, "persons") !== false);
        // the first 4 lists contain the last valid record for versionized tables
        // the other lists contain the only record for non-versionized tables
        foreach ($rows as $row) {
            $uuid = $row[$uuid_index];
            if (strlen($uuid) > 30) {
                $name = ($is_persons) ? $row[$first_name_index] . " " . $row[$last_name_index] : $row[$name_index];
                // this index provides a name for a UUID
                $this->names[$uuid] = [$name,$tablename
                ];
                // this index provides a UUID for a name, may be multiple
                if (! isset($this->uuids[$name]))
                    $this->uuids[$name] = [];
                $this->uuids[$name] = [$uuid,$tablename
                ];
                $records_found ++;
            }
        }
        return $records_found;
    }

    /**
     * Collect all UUIDs and names reference arrays
     */
    private function collect_arrays ()
    {
        $this->names = [];
        $this->uuids = [];
        $records_found = 0;
        $list = new Tfyh_list("../config/lists/uuid2names", 0, "", $this->socket, $this->toolbox);
        $count_of_lists = count($list->get_all_list_definitions());
        for ($i = 0; $i < $count_of_lists; $i ++)
            $records_found += $this->collect_arrays_per_table($i + 1);
        return $records_found;
    }

    /**
     * Find out to which table belongs the UUID and return the name for it. This will only return the last
     * valid name for versionized records.
     * 
     * @param String $uuid
     *            the uuid to get the name for
     * @return array Non-associative array with [ 0 = name, 1 = table ]. If the uuid is not found, name will
     *         be "unresolved" and table "unknown".
     */
    public function uuid2name (String $uuid)
    {
        if (count($this->names) == 0)
            $this->collect_arrays();
        if (! isset($this->names[$uuid]))
            return [i("ZV56RU|unresolved"),i("tWIACB|unknown")
            ];
        else
            return $this->names[$uuid];
    }

    /**
     * Find the first uuid for a name. This is a convenience option for the case, the table name is defined.
     * 
     * @param String $name
     *            the name value to look for.
     * @param String $table_name
     *            the name of the table in which the name shall be searched.
     * @return array Non-associative array with [ 0 = uuid, 1 = table ]. If the uuid is not found, it will
     *         return [ "unresolved", "unknown" ].
     */
    public function name2uuid (String $name, String $tablename)
    {
        return $this->name2uuids($name, $tablename)[0];
    }

    /**
     * Find the uuids for a name. May be multiple, e. g. for a name of a boat and a person being identical.
     * 
     * @param String $name
     *            the name value to look for.
     * @param String $table_name
     *            the name of the table in which the name shall be searched. Leave empty or set null for all
     *            tables
     * @return array Non-associative array of non-associative arrays with [ 0 = uuid, 1 = table ]. If the uuid
     *         is not found, it will return an array with a single element which contains [ "unresolved",
     *         "unknown" ].
     */
    public function name2uuids (String $name, String $tablename = null)
    {
        $unresolved = [i("TjA5DZ|unresolved"),i("OuG821|unknown")
        ];
        if (count($this->uuids) == 0)
            $this->collect_arrays();
        if (! isset($this->uuids[$name]))
            return [$unresolved
            ];
        if (is_null($tablename))
            return $this->uuids[$name];
        foreach ($this->uuids[$name] as $uuid_for_name)
            if (strcasecmp($uuid_for_name["table"], $tablename) == 0)
                return [$uuid_for_name
                ];
        return [$unresolved
        ];
    }
}
