<?php
/**  
 *
 *       dilbo - digital logbook for Rowing and Canoeing
 *       -----------------------------------------------
 *       https://www.dilbo.org
 *
 * Copyright  2023-2024  Martin Glade
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

include_once "../classes/tfyh_xml.php";
include_once "../classes/dilbo_config.php";

/**
 * class file for the specific handling of efa backup import and export.
 */
class Efa_import_export
{

    /**
     * The tenent meta-data: admins.efa2admins, configuration.efa2config, types.efa2types.
     */
    private static $efa_config_files = ["efa2admins","efa2config","efa2types"
    ];

    /**
     * The efa named data files.
     */
    private static $efa_named_config_files = ["efa2project"
    ];

    /**
     * The efa named data files.
     */
    private static $efa_named_data_files = ["efa2logbook","efa2clubwork"
    ];

    /**
     * The efa unique data files, which will be the sequence of import. BOATS MUST BE BEFORE BOATSTATUS.
     */
    private static $efa_data_files = ["efa2autoincrement","efa2boatdamages","efa2boatreservations",
            "efa2boats","efa2boatstatus","efa2crews","efa2destinations","efa2fahrtenabzeichen","efa2groups",
            "efa2messages","efa2persons","efa2sessiongroups","efa2statistics","efa2status","efa2waters"
    ];

    /**
     * The backup files, which shall be skipped.
     */
    private static $efa_skipped_files = ["meta"
    ];

    /**
     * The efaCloud data files, which will be the sequence of import. EFA2PERSONS above MUST BE IMPORTED
     * BEFORE EFACLOUDUSERS.
     */
    private static $efacloud_data_files = ["efaCloudUsers"
    ];

    /**
     * The data base connection socket.
     */
    private $socket;

    /**
     * The generic toolbox.
     */
    private $toolbox;

    /**
     * The root nod for the efa imported configuration
     */
    private $efa_import_root;

    /**
     * The efa2 imported tables
     */
    private $efa2tables = [];

    /**
     * The efa2 imported table headers
     */
    private $efa2tableheaders = [];

    /**
     * The directory into which the uploaded zip file is unzipped.
     */
    private $zip_dir_path;

    /**
     * The directory into which the uploaded zip file is unzipped.
     */
    private $progress_file;

    /**
     * The directory into which the uploaded zip file is unzipped.
     */
    private $log_file;

    /**
     * public Constructor.
     */
    public function __construct (Tfyh_toolbox $toolbox, Tfyh_socket $socket)
    {
        $this->socket = $socket;
        $this->toolbox = $toolbox;
        $this->efa_import_root = $this->toolbox->config->get_cfg_root_item()->get_child("efa2import", 
                "branch", null, "efa2 import");
        $this->progress_file = "../log/executionProgress." . session_id();
        $this->log_file = "../log/efa2import.log";
        // add all extra configuration
        include_once '../classes/dilbo_config.php';
        foreach (Dilbo_config::$extra_config_files as $cfg_file => $object_if_missing)
            $toolbox->config->include_settings($cfg_file, $object_if_missing);
    }

    /**
     * Add a statement to the progress file.
     * 
     * @param String $progress_statement
     *            the statement to add.
     * @param bool $append
     *            set false to create a new progress log.
     */
    public function add_progress (String $progress_statement, bool $append = true)
    {
        file_put_contents($this->progress_file, $progress_statement, (($append) ? FILE_APPEND : null));
        file_put_contents($this->log_file, 
                date("Y-m-d H:i:s") . ": progess statement " . $progress_statement . "\n", FILE_APPEND);
        usleep(10000);
    }

    /**
     * Clear the progress file and get its contents.
     * 
     * @param bool $clear
     *            set true to rempove the progress file.
     * @return string the progress file contents.
     */
    public function get_progress (bool $clear = false)
    {
        $progress_contents = file_get_contents($this->progress_file);
        if ($clear)
            unlink($this->progress_file);
        return $progress_contents;
    }

    /**
     * Remove all settings affected by the import. This removes the settings files: appSettings, clubSettings,
     * efa2import, and efaCloudImport in the ../config/settings folder and clears and initializes the
     * root.appSettings and root.club nodes in the configuration,
     */
    public function clear_settings ()
    {
        $this->add_progress(i("7hyOxK|Removing existing config...") . "<br>");
        // Remove previous import results
        if (file_exists("../config/settings/efa2import"))
            unlink("../config/settings/efa2import");
        if (file_exists("../config/settings/efaCloudImport"))
            unlink("../config/settings/efaCloudImport");
        // remove and reinitialize the application configuration
        $this->toolbox->config->get_cfg_root_item()->remove_branch("appSettings");
        $app_cfg = $this->toolbox->config->get_cfg_root_item()->get_child("appSettings", 
                ".appObjects.Application");
        // restore application defaults to include available translations
        $this->toolbox->config->include_settings("appSettings", null, true);
        file_put_contents("../config/settings/appSettings", $app_cfg->to_string(false, "csv", 99, false));
        // remove and reinitialize the club configuration
        $this->toolbox->config->get_cfg_root_item()->remove_branch("clubSettings");
        $club_cfg = $this->toolbox->config->get_cfg_root_item()->get_child("clubSettings", ".appObjects.Club");
        file_put_contents("../config/settings/clubSettings", $club_cfg->to_string(false, "csv", 99, false));
    }

    /**
     * Unzip the efa-backup archive into the file system and create a list of files to be imported.
     * 
     * @param String $zip_path
     *            path to the eFa-zip archive
     */
    public function load_zip (String $zip_path, bool $reset_db)
    {
        // unzip the archive
        $this->zip_dir_path = mb_substr($zip_path, 0, mb_strrpos($zip_path, "."));
        $file_list = $this->toolbox->unzip($zip_path, true);
        $this->add_progress(i("w96wp3|Expanding %1 files from ...", count($file_list)) . "<br><br>");
        // create the import metadata
        $_SESSION["efaImport"] = [];
        $_SESSION["efaImport"]["zipDirPath"] = $this->zip_dir_path;
        $_SESSION["efaImport"]["fileList"] = $file_list;
        $_SESSION["efaImport"]["backupFileName"] = mb_substr($zip_path, 
                mb_strrpos($zip_path, DIRECTORY_SEPARATOR) + 1);
        // use two different filelists for configuration and data
        $_SESSION["efaImport"]["fileListConfig"] = [];
        $data_files = []; // will be sorted by extension.
        foreach (self::$efa_data_files as $efa2extension)
            $data_files[$efa2extension] = [];
        // split the file list
        foreach ($file_list as $filepath) {
            // config files are unsorted
            $fileextension = mb_substr($filepath, mb_strrpos($filepath, ".") + 1);
            file_put_contents($this->log_file, date("Y-m-d H:i:s") . ": unzipping " . $fileextension . "\n", 
                    FILE_APPEND);
            if (in_array($fileextension, self::$efa_config_files) ||
                     in_array($fileextension, self::$efa_named_config_files)) {
                $_SESSION["efaImport"]["fileListConfig"][] = $filepath;
            } elseif (! in_array($fileextension, self::$efa_skipped_files)) {
                // data files get an extension based sorting, as defined in self::$efa_data_files, see above.
                $data_files[$fileextension][] = $filepath;
            }
        }
        // Convert the data files into a flat list, sort by first extension, than name
        $_SESSION["efaImport"]["fileListData"] = [];
        foreach ($data_files as $efa2extension => $filepaths) {
            sort($filepaths);
            foreach ($filepaths as $filepath)
                // if data base was not reset, only logbook files are accepted for update.
                if ($reset_db)
                    $_SESSION["efaImport"]["fileListData"][] = $filepath;
                elseif (strcasecmp($efa2extension, "efa2logbook") == 0)
                    $_SESSION["efaImport"]["fileListData"][] = $filepath;
                else
                    $this->add_progress(i("TkfnPp|No update for non-logboo...", $efa2extension) . "<br>");
        }
        
        unlink($zip_path);
        if (! is_array($file_list))
            $this->add_progress($file_list . ".<br>");
        file_put_contents($this->log_file, date("Y-m-d H:i:s") . ": unzipping completed.\n", FILE_APPEND);
    }

    /**
     *
     * @return boolean true, if a valid efa confg branch exists
     */
    private function valid_efa_config_branch ()
    {
        if (! isset($this->efa_values)) {
            $this->add_progress(i("pZ1nbk|No efa config files prov...") . "<br>");
            return false;
        }
        $efa_branch = $this->efa_values;
        if (! isset($this->efa_values["config"])) {
            $this->add_progress(i("cNDfYs|No efa config section pr...") . "<br>");
            return false;
        }
        $efaconfig_branch = $this->efa_values["config"];
        return true;
    }

    /**
     * Parse the efa tenant information and map it onto the dilbo tenant configuration. Store the tenant
     * configuration.
     */
    private function import_admins ()
    {
        $this->add_progress(i("LKteOp|Importing the admins con...") . ": ");
        if (! $this->valid_efa_config_branch())
            return;
        
        $xml_parser = new Tfyh_xml();
        $efaadmins_branch = $this->efa_values["config"]["admins"];
        $efaadmins_branch_array = $xml_parser->get_array("data", "record", $efaadmins_branch);
        $config_branch = $this->efa_import_root->get_child("config", "branch", null, 
                i("H9tW8m|configuration"));
        $admins_branch = $config_branch->get_child("admins", "branch", null, i("mIJlp8|administrators"));
        $cnt = 0;
        // the flat types structure is mapped to a tree of categories nd types within
        foreach ($efaadmins_branch_array as $admin_record) {
            // get the category and type and remove the respective information from the record
            $admin_name = $admin_record["Name"];
            $admins_branch_admin = $admins_branch->get_child($admin_name, "branch", null, $admin_name);
            foreach ($admin_record as $rkey => $rvalue) {
                $admins_branch_admin_child = $admins_branch_admin->get_child($rkey, null, $rvalue, $rkey);
                $cnt ++;
            }
            file_put_contents($this->log_file, date("Y-m-d H:i:s") . ": imported admin " . $admin_name . "\n", 
                    FILE_APPEND);
        }
        $this->add_progress($cnt . " " . i("UtwTNu|type values imported") . "<br>");
    }

    /**
     * Parse the efa tenant information and map it onto the dilbo tenant configuration. Store the tenant
     * configuration.
     */
    private function import_configuration ()
    {
        $this->add_progress(i("MVk9OQ|Importing the configurat...") . ": ");
        if (! $this->valid_efa_config_branch())
            return;
        $xml_parser = new Tfyh_xml();
        $efaconfig_branch = $this->efa_values["config"]["config"];
        $configuration_branch_array = $xml_parser->get_array("data", "record", $efaconfig_branch);
        
        $config_branch = $this->efa_import_root->get_child("config", "branch", null, 
                i("3DtcXL|configuration"));
        $config_configuration_branch = $config_branch->get_child("configuration", "branch", null, 
                i("c2S7Zr|configuration"));
        $cnt = 0;
        // the flat types structure is mapped to a tree of categories nd types within
        foreach ($configuration_branch_array as $configuration_record) {
            // get the category and type and remove the respective information from the record
            $rname = $configuration_record["Name"];
            // add the respective type branch. This will
            $configuration_branch_child = $config_configuration_branch->get_child($rname, "branch", null, 
                    $rname);
            file_put_contents($this->log_file, 
                    date("Y-m-d H:i:s") . ": added config'" . $configuration_branch_child->get_path() . "'\n", 
                    FILE_APPEND);
            // add the value, position and LastModified values
            $changecount = $configuration_branch_child->get_child("ChangeCount", "int", 
                    $configuration_record["ChangeCount"], i("iFLIy6|ChangeCount"));
            $last_modified = $configuration_branch_child->get_child("LastModified", "microtime", 
                    $configuration_record["LastModified"], i("kcKoJv|LastModified"));
            $record_value = (isset($configuration_record["Value"])) ? $configuration_record["Value"] : "[" .
                     i("uLDW4W|empty") . "]";
            $value = $configuration_branch_child->get_child("Value", null, $record_value, i("p31A4j|value"));
            $cnt ++;
        }
        $this->add_progress($cnt . " " . i("7BVGfL|configuration values imp...") . "<br>");
    }

    /**
     * Parse the efa tenant information and map it onto the dilbo tenant configuration. Store the tenant
     * configuration.
     */
    private function import_types ()
    {
        $this->add_progress(i("O19Vr8|Importing the types conf...") . ": ");
        if (! $this->valid_efa_config_branch())
            return;
        $xml_parser = new Tfyh_xml();
        $efatypes_branch = $this->efa_values["config"]["types"];
        $efatypes_branch_array = $xml_parser->get_array("data", "record", $efatypes_branch);
        
        $types_branch = $this->efa_import_root->get_child("config")->get_child("types", "branch", null, 
                i("BZTGSR|types"));
        $cnt = 0;
        // the flat types structure is mapped to a tree of categories nd types within
        foreach ($efatypes_branch_array as $type_record) {
            // get the category and type and remove the respective information from the record
            $rcategory = $type_record["Category"];
            unset($type_record["Category"]);
            $rtype = $type_record["Type"];
            $rvalue = $type_record["Value"];
            unset($type_record["Type"]);
            // add the respective category branch, if missing.
            $types_category = $types_branch->get_child($rcategory, "branch", null, strtolower($rcategory));
            // add the respective type branch. This will
            $types_category_type = $types_category->get_child($rtype, "branch", null, $rvalue);
            file_put_contents($this->log_file, 
                    date("Y-m-d H:i:s") . ": added type '" . $types_category_type->get_path() . "'\n", 
                    FILE_APPEND);
            // add the position and LastModified values
            $types_category_type->get_child("Position", "int", $type_record["Position"], i("AWUl7C|Position"));
            $types_category_type->get_child("LastModified", "millis", $type_record["LastModified"], 
                    i("6lp1WX|LastModified"));
            $cnt ++;
        }
        $this->add_progress($cnt . " " . i("44XRtA|type values imported") . "<br>");
    }

    /**
     * copy all imported types into the dilbo configuration tree.
     */
    private function copy_types ()
    {
        $type_lists = ["BOAT" => "asset_subtypes","COXING" => "boat_coxing",
                "GENDER" => "person_gender","NUMSEATS" => "boat_seating","RIGGING" => "boat_rigging",
                "SESSION" => "session_types","STATUS" => "person_status"
        ];
        $efa_config_root = $this->efa_import_root->get_child("config");
        $efa_types_root = $efa_config_root->get_child("types");
        $dilbo_types_root = $this->toolbox->config->get_by_path(".appSettings.catalogs");
        
        foreach ($type_lists as $efa_type_list => $dilbo_type_list) {
            $efa_list_root = $efa_types_root->get_child($efa_type_list);
            $efa_list_root->set_descriptor(["text_local_name" => $efa_type_list
            ]);
            if (is_null($efa_list_root)) // no types file provided.
                return;
            $items = $efa_list_root->get_children();
            $sorted = [];
            foreach ($items as $item_name => $item) {
                $position = $item->get_child("Position")->get_value();
                $sorted[$position] = $item;
            }
            
            $dilbo_list_root = $dilbo_types_root->get_child($dilbo_type_list);
            ksort($sorted);
            foreach ($sorted as $position => $item) {
                $item_name = $item->get_name();
                $text_local_name = $item->get_descriptor_field("text_local_name");
                $dilbo_child = $dilbo_list_root->get_child($item_name, ".appObjects.CatalogItem", null, 
                        $text_local_name);
                file_put_contents($this->log_file, 
                        date("Y-m-d H:i:s") . ": copied type to '" . $dilbo_child->get_path() . "'\n", 
                        FILE_APPEND);
            }
        }
    }

    /**
     * Parse the efa projects information and map it onto the dilbo projects configuration. Store the projects
     * configuration.
     */
    private function import_projects ()
    {
        $this->add_progress(i("8c3xMn|Importing the projects c..."));
        if (! is_array($this->efa_values["projects"])) {
            $this->add_progress(" - " . i("d4yKBL|No projects provided."));
            return;
        }
        $xml_parser = new Tfyh_xml();
        $projects_branch = $this->efa_import_root->get_child("projects", "branch", null, i("M3LyZI|projects"));
        $cnt = 0;
        foreach ($this->efa_values["projects"] as $project_name => $project_branch) {
            $projects_branch_project = $projects_branch->get_child($project_name);
            $projects_branch_project->set_descriptor(["text_local_name" => $project_name
            ]);
            $this->add_progress(" - " . $project_name . ": ");
            $project_branch_array = $xml_parser->get_array("data", "record", $project_branch);
            foreach ($project_branch_array as $project_branch_record) {
                // read the type and open a branch, if not existing
                $rtype = $project_branch_record["Type"];
                if (strcasecmp($rtype, "Project") == 0) {
                    $projects_branch_project_main = $projects_branch_project->get_child("Properties");
                    $projects_branch_project_main->set_descriptor(
                            ["text_local_name" => "Properties"
                            ]);
                    foreach ($project_branch_record as $rkey => $rvalue) {
                        $projects_branch_project_main_child = $projects_branch_project_main->get_child($rkey, 
                                null, $rvalue, $rkey);
                        file_put_contents($this->log_file, 
                                date("Y-m-d H:i:s") . ": added to project main '" .
                                         $projects_branch_project_main_child->get_path() . "'\n", FILE_APPEND);
                        $cnt ++;
                    }
                } else {
                    $rtype_branch_name = $rtype . "s";
                    unset($project_branch_record["Type"]);
                    $projects_branch_project_type = $projects_branch_project->get_child($rtype_branch_name, 
                            "branch", null, $rtype_branch_name);
                    file_put_contents($this->log_file, 
                            date("Y-m-d H:i:s") . ": added/recalled to project types '" .
                            $projects_branch_project_type->get_path() . "'\n", FILE_APPEND);
                    // read the name and open a record branch, if not existing
                    if (isset($project_branch_record["Name"]))
                        $namefield = "Name";
                    elseif (isset($project_branch_record["ClubName"]))
                        $namefield = "ClubName";
                    $rname = Tfyh_data::to_identifier($project_branch_record[$namefield]);
                    unset($project_branch_record[$namefield]);
                    $projects_branch_project_type_name = $projects_branch_project_type->get_child($rname, 
                            "branch", null, $rname);
                    file_put_contents($this->log_file, 
                            date("Y-m-d H:i:s") . ": added/recalled to project records '" .
                                     $projects_branch_project_type_name->get_path() . "'\n", FILE_APPEND);
                    $cnt ++;
                    // add all values to the record branch
                    foreach ($project_branch_record as $rkey => $rvalue) {
                        $projects_branch_project_type_name_key = $projects_branch_project_type_name->get_child(
                                $rkey, null, $rvalue, $rkey);
                        file_put_contents($this->log_file, 
                                date("Y-m-d H:i:s") . ": added to project records keys '" .
                                         $projects_branch_project_type_name_key->get_path() . "'\n", 
                                        FILE_APPEND);
                        $cnt ++;
                    }
                }
            }
            $this->add_progress($cnt . " " . i("lTdLce|configuration values imp..."));
        }
        $this->add_progress("<br>");
    }

    /**
     * Copy the project configuration data into the dilbo configuration
     */
    private function copy_project (String $project_name)
    {
        $efa_projects_branch = $this->efa_import_root->get_child("projects");
        $efa_project_branch = $efa_projects_branch->get_child($project_name);
        $efa_clubs_branch = $efa_project_branch->get_child("Clubs");
        // use the very first club, if multiple are listed
        $efa_clubs = array_keys($efa_clubs_branch->get_children());
        $efa_club_branch = $efa_clubs_branch->get_child($efa_clubs[0]);
        if (is_null($efa_club_branch)) // no project file provided.
            return;
        $efa_club_cfg_path = $efa_club_branch->get_path() . ".";
        
        // copy club location properties. Assume this is a boat house
        $efa_addressStreet = $efa_club_branch->get_child("AddressStreet");
        $efa_addressCity = $efa_club_branch->get_child("AddressCity");
        $efa_storageUser = $efa_project_branch->get_child("Properties")->get_child("StorageUsername");
        $dilbo_locations = $this->toolbox->config->get_by_path(".clubSettings.locations");
        $new_location = $dilbo_locations->get_child($efa_clubs[0], ".appObjects.Location", null, 
                $efa_clubs[0]);
        file_put_contents($this->log_file, 
                date("Y-m-d H:i:s") . ": copied location '" . $new_location->get_path() . "'\n", FILE_APPEND);
        $new_location_properties = ["type" => "BOATHOUSE","dilbo_user" => $efa_storageUser->get_value(),
                "street" => ((isset($efa_addressStreet)) ? $efa_addressStreet->get_value() : ""),
                "city" => ((isset($efa_addressCity)) ? $efa_addressCity->get_value() : "")
        ];
        foreach ($new_location_properties as $name => $value) {
            $property = $new_location->get_child($name);
            file_put_contents($this->log_file, 
                    date("Y-m-d H:i:s") . ": copied location property '" . $property->get_path() . "'\n", 
                    FILE_APPEND);
            $property->set_descriptor(["value_current" => $value
            ]);
        }
        
        // dilbo club and its location and associations
        $dilbo_club = $this->toolbox->config->get_by_path(".clubSettings");
        $dilbo_club->set_descriptor(["text_local_name" => $efa_clubs[0]
        ]);
        $dilbo_club->get_child("address")->set_descriptor(
                ["value_current" => $new_location->get_name()
                ]);
        $dilbo_associations = $this->toolbox->config->get_by_path(".clubSettings.associations");
        $associations = ["GlobalAssociationName" => "global","RegionalAssociationName" => "regional",
                "MemberOfDRV" => "member_of_drv"
        ];
        foreach ($associations as $efa_val => $dilbo_val) {
            $efa_association = $efa_club_branch->get_child($efa_val);
            $dilbo_association = $dilbo_associations->get_child($dilbo_val);
            file_put_contents($this->log_file, 
                    date("Y-m-d H:i:s") . ": copied association '" . $dilbo_association->get_path() . "'\n", 
                    FILE_APPEND);
            $dilbo_association->set_descriptor(
                    ["value_current" => $efa_association->get_value()
                    ]);
        }
        
        // read logbooks
        $efa_logbooks_branch = $efa_project_branch->get_child("Logbooks");
        if (is_object($efa_logbooks_branch) && ! is_null($efa_logbooks_branch)) {
            $efa_logbooks = $efa_logbooks_branch->get_children();
            $logbook_names = [];
            $start_month = - 1;
            foreach ($efa_logbooks as $name => $item)
                $logbook_names[] = Dilbo_config::normalize_bookname($name, "efa2logbook");
        }
        $dilbo_logbooks_branch = $this->toolbox->config->get_by_path(".clubSettings.logbooks");
        foreach ($logbook_names as $logbook_name) {
            $dilbo_logbook = $dilbo_logbooks_branch->get_child($logbook_name, ".appObjects.Logbook", null, 
                    $logbook_name);
            file_put_contents($this->log_file, 
                    date("Y-m-d H:i:s") . ": copied logbook '" . $dilbo_logbook->get_path() . "'\n", 
                    FILE_APPEND);
        }
        
        // use the starting month as sports year start value
        $dilbo_sports_year_start = $this->toolbox->config->get_by_path(
                ".clubSettings.habits.sports_year_start");
        $dilbo_logbook->set_descriptor(["value_current" => $start_month
        ]);
        
        // read club workbooks
        $efa_workbooks_branch = $efa_project_branch->get_child("ClubworkBooks");
        if (is_object($efa_workbooks_branch) && ! is_null($efa_workbooks_branch)) {
            $efa_workbooks = $efa_workbooks_branch->get_children();
            $workbook_names = [];
            foreach ($efa_logbooks as $name => $item) {
                $start_date = Tfyh_data::parse($item->get_child("StartDate")->get_value(), "date");
                $start_month = intval($start_date->format("m"));
                $start_year = intval($start_date->format("Y"));
                $end_date = Tfyh_data::parse($item->get_child("EndDate")->get_value(), "date");
                $end_year = intval($end_date->format("Y"));
                $workbook_names[] = Dilbo_config::normalize_bookname($name, "efa2clubwork");
            }
        }
        $dilbo_workbooks_branch = $this->toolbox->config->get_by_path(".clubSettings.workbooks");
        foreach ($workbook_names as $workbook_name) {
            $dilbo_workbook = $dilbo_workbooks_branch->get_child($workbook_name, ".appObjects.Workbook", null, 
                    $workbook_name);
            file_put_contents($this->log_file, 
                    date("Y-m-d H:i:s") . ": copied workbook '" . $dilbo_workbook->get_path() . "'\n", 
                    FILE_APPEND);
        }
    }

    /**
     * Import the project. This information will not be written to the data base.
     * 
     * @param String $xml
     *            the xml encoded table which shall be imported
     */
    private function read_config_tree (String $file_path)
    {
        if (is_dir($file_path))
            return " (" . $file_path . ")";
        $file_extension = mb_substr($file_path, mb_strrpos($file_path, ".") + 1);
        $datatype = mb_substr($file_extension, 4);
        $filename = mb_substr($file_path, mb_strrpos($file_path, DIRECTORY_SEPARATOR) + 1);
        $objectname = mb_substr($filename, 0, mb_strrpos($filename, "."));
        
        $xml = file_get_contents($file_path);
        $tree = new Tfyh_xml();
        $read_info = $filename;
        if (in_array($file_extension, Efa_import_export::$efa_named_config_files)) {
            $this->efa_values[$datatype . "s"][$objectname] = $tree->read_xml($xml);
        } elseif (in_array($file_extension, Efa_import_export::$efa_config_files)) {
            $this->efa_values["config"][$datatype] = $tree->read_xml($xml);
        }
        // data files and the backup.meta file are not read this way.
        return $read_info . " ... ";
    }

    /**
     * Read the data branch for import
     * 
     * @param String $values_branch_name
     *            branch name, i.e. one of config, meta, named, data
     */
    public function import_config ()
    {
        
        // read all unzipped files
        $this->add_progress(i("aveC01|Parsing configuration fi...") . ": ");
        $fcnt = 0;
        foreach ($_SESSION["efaImport"]["fileListConfig"] as $filepath)
            $this->read_config_tree($filepath);
        $this->add_progress(
                count($_SESSION["efaImport"]["fileListConfig"]) . " " . i("5gYbSn|files parsed") . "<br>");
        
        // add defaults
        Dilbo_config::add_type_list_defaults($this->toolbox);
        
        // import admins, config and types
        $this->import_admins();
        $this->import_configuration();
        $this->import_types();
        $this->copy_types();
        
        // import project and club settings
        $this->import_projects();
        $projects_branch = $this->efa_import_root->get_child("projects");
        if (! is_null($projects_branch)) {
            $project_name = array_keys($this->efa_import_root->get_child("projects")->get_children())[0];
            $this->copy_project($project_name);
        }
        // write result
        $this->write_efa2import_config_file($_SESSION["efaImport"]["backupFileName"]);
        $appSettings_node = $this->toolbox->config->get_by_path(".appSettings");
        file_put_contents("../config/settings/appSettings", 
                $appSettings_node->to_string(false, "csv", 99, false));
        $clubSettings_node = $this->toolbox->config->get_by_path(".clubSettings");
        file_put_contents("../config/settings/clubSettings",
                $clubSettings_node->to_string(false, "csv", 99, false));
        
        // reload the configuration
        $this->toolbox->config = new Tfyh_config($this->toolbox);
    }

    /**
     * Execute an api SELECT command to efacloud. This will use the configured efaCloud settings, but require
     * the password to be manually entered.
     * 
     * @param String $project_name
     *            the name of the efa project to get the URL and user name from
     * @param String $efacloud_password
     *            the efacloud password for the configured user
     * @param String $efacloud_tablename
     *            the table to select. Always a full table select. Note that "efaCloudConfig" as table name
     *            will select the configuration settings.
     * @return boolean|mixed
     */
    private function api_select_from_efacloud (String $project_name, String $efacloud_password, 
            String $efacloud_tablename, array $conditions_record)
    {
        $storage_type = $this->toolbox->config->get_by_path(
                ".efa2import.projects." . $project_name . ".Properties.StorageType")->get_value();
        if (strcasecmp($storage_type, "file/efaCloud") != 0) {
            $this->add_progress(i("8QpdLl|No efacloud project") . "<br>");
            return false;
        } elseif (strlen($efacloud_password) >= 8) {
            $efaCloud_URL = $this->toolbox->config->get_by_path(
                    ".efa2import.projects." . $project_name . ".Properties.EfaCloudURL")->get_value();
            if (mb_substr($efaCloud_URL, mb_strlen($efaCloud_URL) - 1) != "/")
                $efaCloud_URL .= "/";
            $efaCloud_user = $this->toolbox->config->get_by_path(
                    ".efa2import.projects." . $project_name . ".Properties.StorageUsername")->get_value();
            include_once "../classes/efacloud_api.php";
            $efacloud_api = new Efacloud_api($efaCloud_URL, $efaCloud_user, $efacloud_password);
            $tx_id_append = $efacloud_api->append_transaction("select", $efacloud_tablename, 
                    $conditions_record);
            $efacloud_api->send_container();
            $result = $efacloud_api->get_result($tx_id_append);
            if (intval($result[0]) >= 400) {
                $this->add_progress(i("JwyWs2|efaCloud server connnect...", $result[0], $result[1]) . "<br>");
                return false;
            }
            return $result[1];
        } else {
            $this->add_progress(i("5gZmqR|No efacloud credentials ...") . "<br>");
            return false;
        }
    }

    /**
     * Import all data configuration data from the respective efaCloud server.
     */
    public function import_efacloud_config (String $project_name, String $efacloud_password)
    {
        $this->add_progress(i("L4Mkdc|Importing the efaCloud c...") . ": ");
        $cnt = 0;
        $efaCloud_cfg_app_base64 = $this->api_select_from_efacloud($project_name, $efacloud_password, 
                "efaCloudConfig", [""
                ]);
        $efaCloud_cfg_app = unserialize(base64_decode($efaCloud_cfg_app_base64));
        if ($efaCloud_cfg_app !== false) {
            Dilbo_config::map_efaCloud_config($efaCloud_cfg_app, $this->toolbox->config);
            $efaCloud_config = $this->toolbox->config->get_cfg_root_item()->get_child("efaCloudConfig", 
                    "branch", null, "efaCloud import");
            file_put_contents($this->log_file, 
                    date("Y-m-d H:i:s") . ": read efaCloud configuration '" . $efaCloud_config->get_path() .
                             "'\n", FILE_APPEND);
            foreach ($efaCloud_cfg_app as $name => $value)
                if (! is_array($value)) {
                    $cnt ++;
                    $new_child = $efaCloud_config->get_child($name, null, $value, $name);
                }
            file_put_contents("../config/settings/efaCloudImport", 
                    $efaCloud_config->to_string(false, "csv", 99, false));
            $this->copy_efacloud_config();
            $this->add_progress($cnt . " " . i("UETSE9|configuration values imp...") . "<br>");
            
            // write result
            $appSettings_node = $this->toolbox->config->get_by_path(".appSettings");
            file_put_contents("../config/settings/appSettings",
                    $appSettings_node->to_string(false, "csv", 99, false));
            $clubSettings_node = $this->toolbox->config->get_by_path(".clubSettings");
            file_put_contents("../config/settings/clubSettings",
                    $appSettings_node->to_string(false, "csv", 99, false));
        }
    }


    /**
     * Copy all efcloud configuration values to the dilbo configuration
     */
    private function copy_efacloud_config ()
    {
        if (! $this->toolbox->config->get_cfg_root_item()->has_child("efaCloudConfig"))
            return;
        $efaCloud_config = $this->toolbox->config->get_cfg_root_item()->get_child("efaCloudConfig");
        foreach ($efaCloud_config->get_children() as $name => $item) {
            if (isset(Dilbo_config::$efaCloud_to_dilbo_config[$name])) {
                $dilbo_mapped_config = $this->toolbox->config->get_by_path(
                        Dilbo_config::$efaCloud_to_dilbo_config[$name]);
                if (! is_null($dilbo_mapped_config)) {
                    $dilbo_mapped_config->set_descriptor(
                            ["value_current" => $item->get_value(),
                                    "text_local_name" => $item->get_descriptor_field("text_local_name")
                            ]);
                    file_put_contents($this->log_file, 
                            date("Y-m-d H:i:s") . ": copied efaCloud configuration '" .
                                     $dilbo_mapped_config->get_path() . "'\n", FILE_APPEND);
                }
            }
        }
    }

    /**
     * Parse the efa2boats data tables and compie the diufferen boat variants detected.
     */
    public function compile_boat_variants ()
    {
        // get the boats data file.
        $error_message = i("geDnMi|Failed to load efa2boats...");
        if (! isset($_SESSION["efaImport"]["fileListData"])) {
            $this->add_progress($error_message . "<br>");
            return;
        }
        $boats_file_path = false;
        foreach ($_SESSION["efaImport"]["fileListData"] as $table_file)
            if (mb_strpos($table_file, "boats.efa2boats") !== false)
                $boats_file_path = $table_file;
        if (! $boats_file_path) {
            $this->add_progress($error_message . "<br>");
            return;
        }
        $xml = file_get_contents($boats_file_path);
        if (! $xml) {
            $this->add_progress($error_message . "<br>");
            return;
        }
        // this will always read the complete file for ach chunk, but step-by-step can only store data between
        // steps in the $_SESSSION var and this is neither very effective.
        $xml_parser = new Tfyh_xml();
        $xml_parser->read_xml($xml);
        $table_branch_array = $xml_parser->get_array("data", "record");
        $cnt = count($table_branch_array);
        $this->add_progress(i("2trQTF|Extracting boat variant ...", $cnt) . "<br>");
        $type_fields = ["TypeType" => "built","TypeSeats" => "seating","TypeRigging" => "rigging",
                "TypeCoxing" => "coxing"
        ];
        $boat_variants_branch = $this->toolbox->config->get_by_path(".appSettings.boat_variants");
        $boat_subtypes_branch = $this->toolbox->config->get_by_path(".appSettings.catalogs.asset_subtypes");
        foreach ($table_branch_array as $table_branch_record) {
            file_put_contents($this->log_file,
                    date("Y-m-d H:i:s") . ": compiling boat variants for '" .
                    $table_branch_record["Name"] . "'\n", FILE_APPEND);
            Dilbo_config::compile_boat_variants($table_branch_record, $boat_variants_branch);
            Dilbo_config::compile_boat_subtypes($table_branch_record, $boat_subtypes_branch);
        }
        
        // write result
        $appSettings_node = $this->toolbox->config->get_by_path(".appSettings");
        file_put_contents("../config/settings/appSettings", 
                $appSettings_node->to_string(false, "csv", 99, false));
    }

    /**
     * Boat damages may have duplicate numbers. This needs resolution. all damages will be renumbered for this
     * reason. This function creates the number mapping.
     */
    public function map_damage_numbers ()
    {
        if (! isset($_SESSION["efaImport"]["fileListData"]))
            return;
        $damagesfilepath = false;
        foreach ($_SESSION["efaImport"]["fileListData"] as $filepath) {
            $filename = mb_substr($filepath, mb_strrpos($filepath, DIRECTORY_SEPARATOR) + 1);
            $efa_tablename = mb_substr($filename, mb_strrpos($filename, ".") + 1);
            if (strcasecmp($efa_tablename, "efa2boatdamages") == 0)
                $damagesfilepath = $filepath;
        }
        if (! $damagesfilepath)
            return;
        $xml = file_get_contents($damagesfilepath);
        if ($xml === false)
            return;
        $xml_parser = new Tfyh_xml();
        $xml_parser->read_xml($xml);
        $table_branch_array = $xml_parser->get_array("data", "record");
        $cnt = count($table_branch_array);
        $mapping_of_damage_numbers = [];
        for ($i = 0; $i < $cnt; $i ++)
            $mapping_of_damage_numbers[$table_branch_array[$i]["Damage"] . "." .
                     $table_branch_array[$i]["BoatId"]] = $i + 100;
        $this->add_progress(i("T1s1x8|Mapped %1 damage numbers...", $cnt) . "<br>");
        $_SESSION["efaImport"]["map_damage_numbers"] = $mapping_of_damage_numbers;
    }

    /**
     * Read a chunk of a full table of records.
     * 
     * @param int $chunk
     *            the size of the chunk to import
     * @param int $from
     *            the index of the first record of the chunk
     * @param int $user_id
     *            the id of the user triggering the import
     * @param String $filename
     *            the efa or efaCloud file to which the record belongs.
     * @param String $efa_tablename
     *            the efa or efaCloud table the records belong to
     * @param String $bookname
     *            the bookname if the table represents a logbook or a clubwork book (not relevant for efaCloud
     *            import).
     * @param String $record_cfg_path
     *            the path to the records definition
     * @param bool $reset_db
     *            true, if the data base was reset. A uniqueness check of trips in the logbook is than
     *            skipped.
     * @return number the remainder which is to be imported.
     */
    private function import_records_chunk (int $chunk, int $from, int $user_id, String $filename, 
            String $efa_tablename, String $bookname, String $record_cfg_path, bool $reset_db)
    {
        $cnt = count($_SESSION["efaImport"]["data_records"]);
        if (($chunk == 0) || ($from == 0))
            $this->add_progress(i("elFKdn|Importing %1. %2 records...", $filename, $cnt) . " ");
        
        // read and validate efa2 records and store it to the data base.
        include_once "../classes/tfyh_validate.php";
        include_once "../classes/dilbo_config.php";
        include_once "../classes/dilbo_record.php";
        $dilbo_record_obj = new Dilbo_record($this->toolbox, $this->socket);
        $to = ($chunk == 0) ? $cnt : min($from + $chunk, $cnt);
        $record_cfg = $this->toolbox->config->get_by_path($record_cfg_path);
        if (is_null($record_cfg))
            $this->add_progress(i("RFuXPT|no import rule for table...", $efa_tablename) . " ");
        else {
            $findings = "";
            $err_prefix = "<br> --! ";
            $warn_prefix = "<br> -- ";
            for ($i = $from; $i < $to; $i ++) {
                // parse the record
                Tfyh_data::clear_findings();
                $efa_record_cleansed = [];
                foreach ($_SESSION["efaImport"]["data_records"][$i] as $key => $value) {
                    if ($record_cfg->has_child($key)) {
                        // parse the value
                        $type = $record_cfg->get_child($key)->get_type();
                        $parsed = Tfyh_data::parse($value, $type);
                        // format to an SQL compatible String. (Limits check will be done before insertion
                        // into dilbo table)
                        $efa_record_cleansed[$key] = Tfyh_data::format($parsed, $type, "sql");
                    } else {
                        $record_cfg->get_child($key, "string", $value, $key);
                        $this->add_progress(
                                i("nytswg|unknown column °%1° adde...", $key, $efa_tablename) . " ");
                        $efa_record_cleansed[$key] = $value;
                    }
                }
                foreach (Tfyh_data::get_errors() as $error)
                    $this->add_progress($err_prefix . $error);
                foreach (Tfyh_data::get_warnings() as $warning)
                    $this->add_progress($warn_prefix . $warning);
                
                // Map the record to dilbo format
                Dilbo_config::clear_findings();
                $dilbo_record_as_str = Dilbo_config::map_record_efa_to_dilbo($efa_tablename, $bookname, 
                        $efa_record_cleansed, $this->socket);
                foreach (Dilbo_config::get_errors() as $error)
                    $this->add_progress($err_prefix . $error);
                foreach (Dilbo_config::get_warnings() as $warning)
                    $this->add_progress($warn_prefix . $warning);
                // on failures which prohibit insertion or update, $dilbo_record_str will be empty.
                
                // add the record to the data base
                $dilbo_tablename = Dilbo_config::dilbo_table($efa_tablename);
                if (($dilbo_record_as_str !== false) && ($dilbo_tablename !== false)) {
                    
                    // apply dilbo limits
                    $dilbo_record_cfg = $this->toolbox->config->get_by_path(".appTables." . $dilbo_tablename);
                    Tfyh_data::clear_findings();
                    foreach ($dilbo_record_as_str as $key => $value_str) {
                        // parse the value
                        if (! $dilbo_record_cfg->has_child($key))
                            $this->add_progress(
                                    i("oFp0pA|Field °%1° of record is ...", $key, $dilbo_tablename) . " ");
                        $type = $dilbo_record_cfg->get_child($key)->get_type();
                        $parsed = Tfyh_data::parse($value_str, $type);
                        $min = $dilbo_record_cfg->get_child($key)->get_min();
                        $max = $dilbo_record_cfg->get_child($key)->get_max();
                        // check limits and format to an SQL compatible String
                        $validated = Tfyh_data::check_limits($parsed, $type, $min, $max);
                        $dilbo_record_as_str[$key] = Tfyh_data::format($validated, $type, "sql");
                    }
                    foreach (Tfyh_data::get_errors() as $error)
                        $this->add_progress($err_prefix . $error);
                    foreach (Tfyh_data::get_warnings() as $warning)
                        $this->add_progress($warn_prefix . $warning);
                    
                    // check for existence in logbook update mode
                    if (strcasecmp($dilbo_tablename, "logbook") == 0) {
                        $matching = ["number" => $dilbo_record_as_str["number"],
                                "logbookname" => $dilbo_record_as_str["logbookname"],
                                "start" => $dilbo_record_as_str["start"]
                        ];
                        $already_there = $this->socket->find_records_matched($dilbo_tablename, $matching, 3);
                        if ($already_there !== false) {
                            if (count($already_there) > 1) {
                                // There must not be multiple trips in the same logbook at the same date with
                                // the same number. Skip record.
                                $this->add_progress(
                                        i("3fFURY|Multiple trips for logbo...", 
                                                $dilbo_record_as_str["logbookname"], 
                                                $dilbo_record_as_str["number"], $dilbo_record_as_str["start"]) .
                                                 " ");
                                $dilbo_record_as_str = false;
                            } else
                                $dilbo_record_as_str["uid"] = $already_there[0]["uid"];
                        }
                    }
                    if ($dilbo_record_as_str !== false) {
                        // check for uniqueness
                        // insert or update depending on provided uid
                        $is_update = (isset($dilbo_record_as_str["uid"]) &&
                                 (strlen($dilbo_record_as_str["uid"]) >= 6));
                        // check for uniqueness
                        $is_unique = $dilbo_record_obj->write_check_unique_fields($dilbo_tablename, 
                                $dilbo_record_as_str, ($is_update) ? "uid" : null);
                        
                        // finally write to data base
                        if ((strlen($is_unique) == 0) && ($dilbo_record_as_str !== false)) {
                            if ($is_update) {
                                $update_result = $this->socket->update_record_matched($user_id, 
                                        $dilbo_tablename, 
                                        ["uid" => $dilbo_record_as_str["uid"]
                                        ], $dilbo_record_as_str);
                                if (strlen($update_result) > 0)
                                    $this->add_progress($update_result . " ");
                            } else {
                                $dilbo_record_as_str["uid"] = Tfyh_toolbox::create_uid(6);
                                $insert_result = $this->socket->insert_into($user_id, $dilbo_tablename, 
                                        $dilbo_record_as_str);
                                if (! is_numeric($insert_result))
                                    $this->add_progress($insert_result . " ");
                            }
                        } else
                            $this->add_progress($err_prefix . $is_unique);
                    }
                }
            }
        }
        
        $remainder = ($cnt - $to);
        $progress_char = (($remainder == 0) || ($to < $chunk)) ? "&#x2713;" : ((($to % (10 * $chunk)) < $chunk) ? "!" : ((($to %
                 (5 * $chunk)) < $chunk) ? ":" : ".")); // "&#x2713;" = '✓'
        $this->add_progress($progress_char . (($remainder == 0) ? "<br>" : ""));
        return $remainder;
    }

    /**
     * Parse the efa data tables and map it onto the dilbo data tables. Store the data tables.
     * 
     * @param int $file_index
     *            the idex of the file path within the $_SESSION["efaImport"]["fileListData"] array.
     * @param int $chunk
     *            the chunk of records to be imported, set to 0 to import all.
     * @param int $from
     *            the row index of the first record to import.
     * @param bool $reset_db
     *            true, if the data base was reset. A uniqueness check of trips in the logbook is than
     *            skipped.
     * @return the remainder of records left.
     */
    public function import_data (int $file_index, int $chunk, int $from, bool $reset_db)
    {
        if (! isset($_SESSION["efaImport"]["fileListData"][$file_index]))
            return;
        $filepath = $_SESSION["efaImport"]["fileListData"][$file_index];
        if (is_null($filepath) || (strlen($filepath) == 0))
            return;
        
        $filename = mb_substr($filepath, mb_strrpos($filepath, DIRECTORY_SEPARATOR) + 1);
        $efa_tablename = mb_substr($filename, mb_strrpos($filename, ".") + 1);
        $bookname = mb_substr($filename, 0, mb_strrpos($filename, "."));
        $user_id = $this->toolbox->users->session_user["@id"];
        
        // this will always read the complete file for ach chunk, but step-by-step can only store data between
        // steps in the $_SESSSION var and this is neither very effective.
        $xml = file_get_contents($filepath);
        if ($xml === false) {
            $this->add_progress(i("hVTaaQ|Failed ") . $filename . ": ");
            return;
        }
        $xml_parser = new Tfyh_xml();
        $xml_parser->read_xml($xml);
        
        // use records cache
        if (! isset($_SESSION["efaImport"]["data_records"]))
            $_SESSION["efaImport"]["data_records"] = $xml_parser->get_array("data", "record");
        // import chunk
        $remainder = $this->import_records_chunk($chunk, $from, $user_id, $filename, $efa_tablename, 
                $bookname, ".efa2tables." . $efa_tablename, $reset_db);
        // delete cache after completion
        if ($remainder == 0)
            unset($_SESSION["efaImport"]["data_records"]);
        return $remainder;
    }

    /**
     * Parse the efa data tables and map it onto the dilbo data tables. Store the data tables.
     * 
     * @param int $file_index
     *            the idex of the file path within the $_SESSION["efaImport"]["fileListData"] array.
     * @param int $chunk
     *            the chunk of records to be imported, set to 0 to import all.
     * @param int $from
     *            the row index of the first record to import.
     * @param bool $reset_db
     *            true, if the data base was reset. A uniqueness check of trips in the logbook is than
     *            skipped.
     * @param String $project_name
     *            the efa project name to read URL and efaCloud user
     * @param String $efacloud_password
     *            the efacloud password to download the data.
     * @return the remainder of records left.
     */
    public function import_efacloud_data (int $file_index, int $chunk, int $from, bool $reset_db, 
            String $project_name, String $efacloud_password)
    {
        // data will be stored in the same directory as the efa-files.
        $dirpath = $_SESSION["efaImport"]["zipDirPath"];
        $user_id = $this->toolbox->users->session_user["@id"];
        
        $efacloud_tablename = self::$efacloud_data_files[$file_index];
        if (! file_exists($dirpath . "/efacloud"))
            mkdir($dirpath . "/efacloud");
        $efacloud_data_filepath = $dirpath . "/efacloud/" . $efacloud_tablename . ".csv";
        if ($from == 0) {
            // first chunk. get csv data from efacloud server.
            $efacloud_data_csv = $this->api_select_from_efacloud($project_name, $efacloud_password, 
                    $efacloud_tablename, ["LastModified" => "0","?" => ">"
                    ]);
            if ($efacloud_data_csv !== false)
                file_put_contents($efacloud_data_filepath, $efacloud_data_csv);
            else {
                $this->add_progress(i("GP3RV9|Failed to retrieve data ...", $efacloud_tablename) . " ");
                return 0;
            }
        }
        
        // use records cache
        if (! isset($_SESSION["efaImport"]["data_records"]))
            $_SESSION["efaImport"]["data_records"] = Tfyh_toolbox::static_read_csv_array(
                    $efacloud_data_filepath);
        // import chunk
        $remainder = $this->import_records_chunk($chunk, $from, $user_id, $efacloud_tablename, 
                $efacloud_tablename, "", ".efaCloudTables." . $efacloud_tablename, $reset_db);
        // delete cache after completion
        if ($remainder == 0)
            unset($_SESSION["efaImport"]["data_records"]);
        return $remainder;
    }

    /**
     * Write the import result into the efa2import configuration file
     */
    private function write_efa2import_config_file (String $backup_filename)
    {
        $this->add_progress(i("yqXP9E|Writing settings/efa2con...") . "<br>");
        $backup_file_item = $this->efa_import_root->get_child("backup_filename", "string", $backup_filename, 
                i("7itWPt|backup filename"));
        $csv = $this->efa_import_root->to_string(false, "csv", 99, false);
        $write_to = "../config/settings/efa2import";
        file_put_contents($write_to, $csv);
    }
}
