<?php
/**  
 *
 *       the tools-for-your-hobby framework
 *       ----------------------------------
 *       https://www.tfyh.org
 *
 * Copyright  2023-2024  Martin Glade
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

include_once "../classes/tfyh_data.php";
include_once "../classes/tfyh_data_item.php";

/**
 * A utility class to hold a configuration item
 */
class Tfyh_config_item
{

    /**
     * The configuration item's data item containing all information, including the name.
     */
    private $data_item;

    /**
     * The configuration item parent.
     */
    private $parent;

    /**
     * The configuration item children array.
     */
    private $children;

    /**
     * The output String for branch write
     */
    private static $out;

    /**
     * The array of descriptor fields for output
     */
    private static $out_fields;

    /**
     * set true to trace all actions on the item in ../log/trace.<uid>
     */
    private $trace;

    private $trace_file_name;

    /**
     * Enable or disable the tracing. Enablings resets the trace file, disabling returns and deletes it.
     * 
     * @param bool $enabled
     *            set true to enable, false to disable.
     * @return string the trace file, either a single line at start or the full file before deletion at end.
     */
    public function set_trace (bool $enabled)
    {
        $this->trace = false; // exclude the following statements from tracing.
        $uid = $this->data_item->get_uid();
        $this->trace_file_name = "../log/trace_" . str_replace(".", "_", $this->get_path()) . "." .
                 $this->data_item->get_uid();
        if ($enabled) {
            $_SESSION["traced"][$uid]["data_item_clone"] = $this->data_item->get_clone(
                    $this->data_item->get_name() . "_clone");
            $_SESSION["traced"][$uid]["children"] = [];
            foreach ($this->children as $cname => $child)
                $_SESSION["traced"][$uid]["children"][$child->data_item->get_uid()] = $cname;
            file_put_contents($this->trace_file_name, 
                    "\n" . date("Y-m-d H:i:s") . ": Starting trace for " . $this->get_path() .
                             ".\n--------------------------------------\n", FILE_APPEND);
            file_put_contents($this->trace_file_name, 
                    " - Descriptor: " . $this->data_item->get_descriptor_json() . "\n", FILE_APPEND);
            file_put_contents($this->trace_file_name, 
                    " - Children: " . json_encode($_SESSION["traced"][$uid]["children"]) . "\n", FILE_APPEND);
            $this->trace = true;
        } else {
            $this->trace = false;
            unset($_SESSION["traced"][$uid]);
            unlink($this->trace_file_name);
        }
    }

    /**
     * Add a trace point to the trace file.
     * 
     * @param String $function_name
     *            the function calling the point to be recorded.
     * @param bool $at_end
     *            set true to add a "completed" to the trace point.
     */
    private function add_trace (String $function_name, bool $at_end = false)
    {
        if (! $this->trace)
            return;
        
        $uid = $this->data_item->get_uid();
        file_put_contents($this->trace_file_name, 
                date("Y-m-d H:i:s") . ": " . $function_name . (($at_end) ? " (completed)" : "") . ".\n", 
                FILE_APPEND);
        
        if (isset($_SESSION["traced"][$uid]["data_item_clone"]))
            $diff_descriptor = $this->data_item->diff_descriptor($_SESSION["traced"][$uid]["data_item_clone"], 
                    true, false);
        if (strlen($diff_descriptor) > 0) {
            file_put_contents($this->trace_file_name, 
                    " - Descriptor old A > new B: " . $diff_descriptor . "\n", FILE_APPEND);
            file_put_contents($this->trace_file_name, 
                    " - Descriptor: " . $this->data_item->get_descriptor_json() . "\n", FILE_APPEND);
        }
        $_SESSION["traced"][$uid]["data_item_clone"] = $this->data_item->get_clone(
                $this->data_item->get_name() . "_clone");
        
        $children = [];
        foreach ($this->children as $cname => $child)
            $children[$child->data_item->get_uid()] = $cname;
        $diff_children = Tfyh_data::diff_arrays($_SESSION["traced"][$uid]["children"], $children);
        if (strlen($diff_children) > 0) {
            file_put_contents($this->trace_file_name, " - Children old A > new B: " . $diff_children . "\n", 
                    FILE_APPEND);
            file_put_contents($this->trace_file_name, " - Children: " . json_encode($children) . "\n", 
                    FILE_APPEND);
        }
        $_SESSION["traced"][$uid]["children"] = $children;
        
        if ((strlen($diff_descriptor) == 0) && (strlen($diff_children) == 0))
            file_put_contents($this->trace_file_name, " - no changes.\n", FILE_APPEND);
    }

    /**
     * Unlink (delete) all trace files from the log directory
     */
    public static function clear_traces ()
    {
        foreach (glob('../log/trace.*') as $file)
            unlink($file);
    }

    /**
     * Construct a configuration item. It becomes child to $parent, replacing any existing child of this
     * $name. For the root set parent = null.
     * 
     * @param String $name
     *            the name of the item
     * @param String $type
     *            the type of the item
     * @param Tfyh_config_item $parent
     *            the parent item, set null for root items
     * @param String $uid
     *            the uid of the item. Omit or set null to generate a new one.
     * @param String $text_local_name
     *            the local name of the item, i. e. the name which is used for display. Set null for default.
     */
    private function __construct (String $name, String $type, Tfyh_config_item $parent = null, String $uid = null, 
            String $text_local_name = null)
    {
        $this->parent = $parent;
        if (! is_null($parent))
            $parent->children[$name] = $this;
        $this->children = [];
        if (substr($type, 0, 1) == ".") {
            // object type
            if (isset(Tfyh_data::$object_templates[$type])) {
                // an object template exists. Copy it, but use new uids
                $this->data_item = Tfyh_data::$object_templates[$type]->data_item->get_clone($name);
                if (! is_null($uid))
                    $this->data_item->set_uid($uid);
                if (! is_null($text_local_name))
                    $this->data_item->set_descriptor(
                            ["text_local_name" => $text_local_name
                            ]);
                // the templates data type is "object", but the data type of the new child shall be the path
                // of the object template
                $this->set_descriptor(["value_type" => $type
                ]);
                Tfyh_data::$object_templates[$type]->copy_children($this, 99);
                // if the children contain themselves objects without their own branches, the default tree is
                // not complete. Expand all objects to fix this.
            } else {
                // no object template exists. text_local_name is the error message
                $this->data_item = new Tfyh_data_item($name, "unknown", null, 
                        i("kUFgHW|object type %1 unknown", $type));
            }
        } else {
            // simple data type
            $this->data_item = new Tfyh_data_item($name, $type, $uid, $text_local_name);
        }
        $this->add_trace(__FUNCTION__, true);
    }
    
    /**
     * Get a new configuration root. You can provide a type to read the full object branch. Default type is
     * "branch". If the root is an object, it should be initialized like it, if you want objects to be
     * expanded.
     * 
     * @param String $uid
     *            you can provide a uid, but default is null to create a new one.
     * @param String $type
     *            the roots data type, default is branch
     * @return Tfyh_config_item the new root.
     */
    public static function get_new_root (String $uid = null, String $type = null)
    {
        return new Tfyh_config_item("root", ((isset($type)) ? $type : "branch"), null, $uid, i("Z4RVra|root node"));
    }

    /**
     * Read a configuration file and attach it to a node. Any record must have a parent before it in the file.
     * The root node as first line is the only exception. Its parentId will be ignored. When reading an object
     * item, either its full branch is created with defaults, or just its data descriptor used. Set
     * $expand_objects to control the behaviour.
     * 
     * @param String $cfg_filename
     *            the name of the file to read from
     * @param bool $expand_objects
     *            when reading an object item, it is built by its template, i. e. including its full branch.
     *            Although this is helpful in initial readings to manage object changes, it updates an item
     *            partly with its defaults, when updating an item with an update description that does not
     *            contain the entire branch of the object. Set expand objects, to initialize the full object
     *            template while readiung. Set it to false to only regard the items own descriptor.
     * @return Tfyh_config_item|String the root node of the read branch or an error String.
     */
    public static function read_branch (String $cfg_filename, bool $expand_objects)
    {
        $csv_string = (file_exists($cfg_filename)) ? file_get_contents($cfg_filename) : "";
        return self::read_branch_csv($csv_string, $expand_objects);
    }

    /**
     * Read a configuration csv-table and attach it to a node. Details see Tfyh_config::read_branch()
     * 
     * @param String $csv_string
     *            the csv String to be read
     * @param bool $expand_objects
     *            Set true, to initialize the full object template while reading an object item.
     * @return Tfyh_config_item|String the root node of the read branch or an error String.
     */
    public static function read_branch_csv (String $csv_string, bool $expand_objects)
    {
        // read the branch.
        $cfg_array = Tfyh_toolbox::static_read_csv_string_array($csv_string);
        
        // the node uid index is used to find the parent.
        $uid_index = [];
        
        // check whether the first node is root
        $root_node = null;
        for ($i = 0; $i < count($cfg_array); $i ++) {
            if (! isset($cfg_array[$i]["name"]) || ! isset($cfg_array[$i]["uid"]) ||
                     ! isset($cfg_array[$i]["value_type"]))
                return "Missing name, uid or value type in configuration. Please correct the configuration.";
            $name = $cfg_array[$i]["name"];
            $uid = $cfg_array[$i]["uid"];
            $type = $cfg_array[$i]["value_type"];
            if ($i == 0) {
                $root_node = self::get_new_root($uid, $type);
                $root_node->data_item->parse_descriptor($cfg_array[$i]);
                // add the root node to the uid index (PHP only)
                $uid_index[$root_node->get_uid()] = $root_node;
            } else {
                // if no parent id is provided, return error.
                if (! isset($cfg_array[$i]["parent"]))
                    return "Missing parent Id for $name in configuration. Please correct the configuration.";
                $parent = $uid_index[$cfg_array[$i]["parent"]];
                if (! isset($parent)) {
                    return "Invalid parent Id '" . $cfg_array[$i]["parent"] .
                             "' for $name in configuration. Please correct the configuration.";
                }
                // Get the new item by the type template.
                $new_item = new Tfyh_config_item($name, $type, $parent, $uid, 
                        $cfg_array[$i]["text_local_name"]);
                // remove all children, if the object shall not be expanded.
                if (! $expand_objects && (strpos($type, ".") !== false))
                    foreach ($new_item->get_children() as $cname => $child)
                        $new_item->remove_branch($cname);
                // parse the own descriptor and add the item to the uid index.
                $new_item->parse_descriptor($cfg_array[$i]);
                $uid_index[$new_item->get_uid()] = $new_item;
            }
        }
        // if an empüty or missing cfg-file was read, the root node is null.
        if (! is_null($root_node))
            $root_node->sort_children(99);
        return $root_node;
    }

    /**
     * Get a clone of this configuration item with a selectable name, always a new uid and neither parent, nor
     * children.
     * 
     * @param String $name_of_clone
     *            the name of the cloned data item.
     */
    public function get_clone (String $name_of_clone)
    {
        $this->add_trace(__FUNCTION__);
        $clone = Tfyh_config_item::get_new_root();
        $clone->data_item = $this->data_item->get_clone($name_of_clone);
        return $clone;
    }

    /**
     * Copy the children of a node to another node. This will replace those children of the to_node which have
     * a name of a child of the from_node.
     * 
     * @param Tfyh_config_item $to_node
     *            the node to copy the children to
     * @param int $drilldown
     *            the levels to drill down. To copy only this config item's children, set it to 0.
     */
    public function copy_children (Tfyh_config_item $to_node, int $drilldown)
    {
        $this->add_trace(__FUNCTION__);
        foreach ($this->children as $childname => $child) {
            // the children will have the same name, but a different uid.
            $clone = $child->get_clone($childname);
            $clone->parent = $to_node;
            $to_node->children[$clone->get_name()] = $clone;
            if ($drilldown > 0)
                $child->copy_children($clone, $drilldown - 1);
        }
    }

    /**
     * Attach a branch to a parent item, e. g. when read from a file. This will replace an existing branch of
     * the branch name.
     * 
     * @param String $branchname
     *            the name of the branch in the tree, to which it is attached.
     * @param Tfyh_config_item $branch_root
     *            the root node of the branch to attach.
     */
    public function attach_branch (String $branchname, Tfyh_config_item $branch_root)
    {
        $this->add_trace(__FUNCTION__);
        $this->children[$branchname] = $branch_root;
        $branch_root->parent = $this;
        $branch_root->data_item->set_name($branchname);
        $this->add_trace(__FUNCTION__, true);
    }

    /**
     * Update a branch using the input form another branch. The descriptor of $this will be replaced be a
     * descriptor clone of $update_branch. The same happens for children that exist on both sides. Children of
     * $update_branch that do not exist at $this will be attached. Then drilldown is performed for all
     * children asf. as deep as $update_branch provides. Use a full descriptor for upate, always with
     * defaults.
     * 
     * @param String $branchname
     *            the name of the branch in the tree, to which it is attached.
     * @param Tfyh_config_item $update_branch
     *            the root node of the branch to attach.
     */
    public function update_branch (String $branchname, Tfyh_config_item $update_branch)
    {
        $this->add_trace(__FUNCTION__);
        $this->data_item->replace_descriptor($update_branch->get_descriptor_clone());
        foreach ($update_branch->get_children() as $cname => $cdefinition) {
            if ($this->has_child($cname))
                $this->get_child($cname)->update_branch($cname, $cdefinition);
            else
                $this->attach_branch($cname, $cdefinition);
        }
        $this->add_trace(__FUNCTION__, true);
    }

    /**
     * Remove a child branch from a this item. This removes te child entry in $this and the parent entry in
     * the child item.
     * 
     * @param Tfyh_config_item $branch_root
     *            the root item of the branch to remove
     * @return bool true in case of success
     */
    public function remove_branch (String $name)
    {
        $this->add_trace(__FUNCTION__);
        if (! $this->has_child($name))
            return false;
        unset($this->children[$name]->parent);
        unset($this->children[$name]);
        $this->add_trace(__FUNCTION__, true);
        return true;
    }

    /**
     * Move a child branch up or down in the children sequence. PHP associative arrays are ordered, See:
     * https://stackoverflow.com/questions/10914730/are-php-associative-arrays-ordered
     * 
     * @param String $item_name
     *            the name of the child item to move
     * @param int $by
     *            the number of places to move. +1 to move down, -1 to move up
     * @return bool true in case of success
     */
    public function move_child (String $item_name, int $by)
    {
        $this->add_trace(__FUNCTION__);
        if ($by == 0) // nothing to move
            return true;
        
        // rearrange the chldren names
        $children_names = array_keys($this->children);
        // identify the current and new item position
        $from_position = array_search($item_name, $children_names);
        $to_position = $from_position + $by;
        // do not move, if target position is beyond the ends
        if (($to_position >= count($children_names)) || ($to_position < 0))
            return false;
        
        // cache name of object at current position
        $cached_name = $children_names[$from_position];
        // now move the names in between $from_position and $to_position
        // this will duplicate the name at the $to_position
        $end = abs($by);
        $fwd = $by / $end;
        for ($i = 1; $i <= $end; $i ++)
            $children_names[$from_position + (($i - 1) * $fwd)] = $children_names[$from_position + ($i * $fwd)];
        // replace the name at the $to_position by the cached name
        $children_names[$to_position] = $cached_name;
        
        // now rearrange the children according to the rearranged names.
        $children_cached = $this->children;
        $this->children = [];
        foreach ($children_names as $child_name)
            $this->children[$child_name] = $children_cached[$child_name];
        
        $this->add_trace(__FUNCTION__, true);
        return true;
    }

    /**
     * Sort all children to get the branches first.
     * 
     * @param int $drilldown
     *            the levels to drill down. To sort only this item's children set it to 0.
     * @return boolean
     */
    public function sort_children (int $drilldown)
    {
        $this->add_trace(__FUNCTION__);
        // split children into branches and leafs
        $branch_names = [];
        $leaf_names = [];
        foreach ($this->children as $childname => $child) {
            if ($child->has_children())
                $branch_names[] = $childname;
            else
                $leaf_names[] = $childname;
        }
        
        // now rearrange the children according to the rearranged names.
        $children_cached = $this->children;
        $this->children = [];
        foreach ($branch_names as $childname)
            $this->children[$childname] = $children_cached[$childname];
        foreach ($leaf_names as $childname)
            $this->children[$childname] = $children_cached[$childname];
        
        // go for further levels, if required.
        if ($drilldown > 0)
            foreach ($this->children as $childname => $child)
                $child->sort_children($drilldown - 1);
        
        $this->add_trace(__FUNCTION__, true);
    }

    /**
     * Return the child of the given name. If no such child exits, create a new one. If $type is null and
     * $value is not null, the type will be guessed via Tfyh_data::guess_type().
     * 
     * @param String $name
     *            the name of the item
     * @param String $type
     *            only, if new: the type of the item. Omit to guess the type, based on the value.
     * @param String $value_str
     *            only, if new: the string formatted value of the item. May be null.
     * @param String $text_local_name
     *            only, if new: the local name of the item, i. e. the name which is used for display. Set null
     *            for default.
     * @return Tfyh_config_item the child of this item with the provided name. A new item on no match.
     */
    public function get_child (String $name, String $type = null, String $value_str = null, 
            String $text_local_name = null)
    {
        $this->add_trace(__FUNCTION__);
        if (isset($this->children[$name]))
            return $this->children[$name];
        // guess type, if null
        if (is_null($type))
            $type = Tfyh_data::guess_type($value_str);
        // create child
        $new_child = new Tfyh_config_item($name, $type, $this, null, $text_local_name);
        // set value
        $value = Tfyh_data::parse($value_str, $type);
        $new_child->set_descriptor(["value_current" => $value
        ]);
        $this->add_trace(__FUNCTION__, true);
        return $new_child;
    }

    /**
     * Get the difference of two configuration branches. The report is a string, empty on no difference.
     * 
     * @param Tfyh_config_item $to_compare
     *            the root item of the branch to compare
     * @param bool $exclude_current_value
     *            set true to exclude the respective current value from the comparison.
     * @param bool $exclude_defaults
     *            set true to exclude the fields which have the default value of the data type
     * @param String $new_line
     *            set "\n" for text and "<br>" for html.
     * @return String The collected differences for the entire branch.
     */
    public function diff (Tfyh_config_item $to_compare, bool $exclude_current_value, bool $exclude_defaults, 
            String $new_line)
    {
        $this->add_trace(__FUNCTION__);
        $diff_collected = $this->get_name() . " ?= " . $to_compare->get_name() . ": ";
        $diff_data_item = $to_compare->data_item->diff($this->data_item, $exclude_current_value, 
                $exclude_defaults);
        $value = Tfyh_data::format($to_compare->get_value(), $to_compare->get_type());
        $diff_collected .= ((strlen($diff_data_item) == 0) ? i("zk0z7u|equal") . ". " : $diff_data_item) . $new_line;
        $names_collected = [];
        foreach ($this->children as $name => $child) {
            $names_collected[] = $name;
            if (! $to_compare->has_child($name))
                $diff_collected .= i("mqIFYg|Missing child item") . " $name. ";
            else
                $diff_collected .= $child->diff($to_compare->get_child($name), $exclude_current_value, 
                        $exclude_defaults, $new_line);
        }
        foreach ($to_compare->children as $name => $child) {
            if (! $this->has_child($name))
                $diff_collected .= i("fBJNgp|Extra child item ") . " $name. ";
        }
        return $diff_collected;
    }

    /**
     * Output a branch to a String, csv or html type. Result in self::$out.
     * 
     * @param bool $as_html
     *            set true to get html code instead of csv.
     * @param String $language_code
     *            the language to use for formatting. Default is "csv" which represents the formatting used by
     *            the data base.
     * @param int $drilldown
     *            the levels to drill down. To get only this config item, set it to 0.
     * @param bool $include_defaults
     *            set true to include all fields with default values as well, needed for data type descriptor
     *            writing.
     * @return String the String representing the branch of this item.
     */
    public function to_string (bool $as_html, String $language_code = "csv", int $drilldown = 99, 
            bool $include_defaults = false)
    {
        $this->add_trace(__FUNCTION__);
        // collect and sort all fields which are used in the branch
        self::$out_fields = [];
        self::collect_fields($this, $drilldown, $include_defaults);
        sort(self::$out_fields);
        // output of headline for of the fields used
        self::$out = ($as_html) ? "<table><tr><th>uid</th><th>parent</th><th>name</th>" : "uid;parent;name";
        foreach (self::$out_fields as $field_name)
            self::$out .= ($as_html) ? "<th>$field_name</th>" : ";" . $field_name;
        self::$out .= ($as_html) ? "</tr>\n" : "\n";
        // collect data of the fields used
        self::out_item($this, $as_html, $language_code, $drilldown, $include_defaults);
        self::$out .= ($as_html) ? "</table>\n" : "";
        return self::$out;
    }

    /**
     * Collect all descriptor fields which have a non-null value in any item of the branch.
     * 
     * @param Tfyh_config_item $cfg_item
     *            the item to get the fields for
     * @param int $drilldown
     *            the levels to drill down. To only collect from this config item, set it to 0.
     * @param bool $include_defaults
     *            set true to include all fields with default values as well, needed for data type descriptor
     *            writing.
     */
    private static function collect_fields (Tfyh_config_item $cfg_item, int $drilldown = 99, 
            bool $include_defaults = false)
    {
        if ($include_defaults) {
            // if you include the defaults also include all fields.
            // this is needed to get a descriptor which allows field deletion.
            self::$out_fields = Tfyh_data::get_descriptor_field_names();
            return;
        }
        
        $descriptor = $cfg_item->data_item->get_descriptor_clone();
        $data_type = $descriptor["value_type"];
        if (! $include_defaults)
            $descriptor = Tfyh_data::clone_without_defaults($descriptor);
        
        foreach ($descriptor as $name => $field) {
            $field_data_type = Tfyh_data::get_descriptor_field_data_type($name, $descriptor["value_type"]);
            $df_parseAs = Tfyh_data::parse_as($field_data_type);
            $is_string = (strcasecmp($df_parseAs, "string") == 0) && ! is_array($descriptor[$name]);
            $is_empty_String = $is_string && (strlen($descriptor[$name]) == 0);
            if (! is_null($field) && ! $is_empty_String && ! in_array($name, self::$out_fields) &&
                     (! in_array($name, Tfyh_data::$item_keys))) {
                self::$out_fields[] = $name;
            }
        }
        // drilldown
        if ($drilldown > 0)
            foreach ($cfg_item->children as $childname => $child) {
                self::collect_fields($child, $drilldown - 1, $include_defaults);
            }
    }

    /**
     * Output a branch to a String, csv or html type. Result in self::$out.
     * 
     * @param Tfyh_config_item $cfg_item
     *            the config item to write
     * @param bool $as_html
     *            set true to get html code instead of csv.
     * @param String $language_code
     *            the language to use for formatting. Default is "csv" which represents the formatting used by
     *            the data base.
     * @param int $drilldown
     *            the levels to drill down. To get only this config item, set it to 0.
     * @param bool $include_defaults
     *            set true to include all fields with default values as well, needed for data type descriptor
     *            writing.
     * @return String the String to write
     */
    private static function out_item (Tfyh_config_item $cfg_item, bool $as_html, String $language_code, 
            int $drilldown, bool $include_defaults)
    {
        // define separators
        $line_start = ($as_html) ? "<tr><td>" : "";
        $separator = ($as_html) ? "</td><td>" : ";";
        $line_end = ($as_html) ? "</td></tr>\n" : "\n";
        // write uid, parent uid and name
        self::$out .= $line_start . strval($cfg_item->get_uid()) . $separator .
                 ((isset($cfg_item->parent)) ? strval($cfg_item->parent->get_uid()) : "") . $separator .
                 strval($cfg_item->get_name()) . $separator;
        // get descriptor for export and write it
        $descriptor_out = ($include_defaults) ? $cfg_item->data_item->get_descriptor_clone() : $cfg_item->data_item->get_descriptor_without_defaults();
        self::$out .= self::write_descriptor($descriptor_out, $as_html, self::$out_fields, $language_code);
        // close line
        self::$out .= $line_end;
        
        // drilldown
        if ($drilldown > 0) {
            // start with own properties (children without grandchildren) first for result readability reasons
            $own_property_names = [];
            foreach ($cfg_item->children as $childname => $child) {
                if (! $child->has_children()) {
                    self::out_item($child, $as_html, $language_code, $drilldown - 1, $include_defaults);
                    $own_property_names[] = $childname;
                }
            }
            // continue with the remainder, i.e. the child branches.
            foreach ($cfg_item->children as $childname => $child) {
                if (! in_array($childname, $own_property_names))
                    self::out_item($child, $as_html, $language_code, $drilldown - 1, $include_defaults);
            }
            // Note that the sequence in the configuration provider is opposite and therefore does a
            // resprting.
        }
    }

    /**
     * Return a String representation of a descriptor
     * 
     * @param array $descriptor
     *            the descriptor to be written
     * @param bool $as_html
     *            set true to get an html type String, else get a csv type String
     * @param array $field_names
     *            the field nams to be chosen, exactly those will be written, regradless of the descriptor
     *            contents
     * @param String $language_code
     *            the language to use for formatting. Default is "csv" which represents the formatting used by
     *            the data base.
     * @return string a csv encoded String with the descriptor field values according to the $field_names
     *         array
     */
    private static function write_descriptor (array $descriptor, bool $as_html, array $field_names, 
            String $language_code = "csv")
    {
        $separator = ($as_html) ? "</td><td>" : ";";
        if (count($field_names) == 0) {
            $field_names = [];
            foreach ($descriptor as $name => $field)
                $field_names[] = $name;
        }
        $out = "";
        // write descriptor fields
        foreach ($field_names as $name) {
            $descriptor_data_type = (isset($descriptor["value_type"])) ? $descriptor["value_type"] : "string";
            $type_field = Tfyh_data::get_descriptor_field_data_type($name, $descriptor_data_type);
            $out .= $separator;
            // start with isset() to avoid "Undefined array key" warnings in the apache log
            if (isset($descriptor[$name])) {
                if (is_null($descriptor[$name]))
                    $out .= Tfyh_data::format(null, $type_field, $language_code);
                elseif (Tfyh_data::is_descriptor_field_name($name)) {
                    $text = Tfyh_data::format($descriptor[$name], $type_field, $language_code);
                    $out .= (($as_html) ? htmlspecialchars($text) : Tfyh_toolbox::static_encode_entry_csv(
                            $text));
                }
            }
        }
        return mb_substr($out, mb_strlen($separator));
    }

    /**
     * Parse and set the descriptor values. This starts always with the defaults for the data type and
     * overwrites any default by the given descriptor information. NOTA BENE: The previous descriptor, if
     * there was one, is dropped.
     * 
     * @param array $descriptor_record
     *            the descriptor as record of all Strings to be used.
     */
    public function parse_descriptor (array $descriptor_record)
    {
        $this->add_trace(__FUNCTION__);
        $this->data_item->parse_descriptor($descriptor_record, $this);
        $this->add_trace(__FUNCTION__, true);
    }

    /**
     * Simple setter. Copy the fields of the source items descriptor and keep other fields.
     * 
     * @param Tfyh_config_item $source
     *            the config item to write
     */
    public function copy_descriptor (Tfyh_config_item $source)
    {
        $this->add_trace(__FUNCTION__);
        $this->data_item->set_descriptor($source->data_item->get_descriptor_clone());
        $this->add_trace(__FUNCTION__, true);
    }

    /**
     * Simple setter. Only set provided fields and keeps other fields.
     * 
     * @param array $descriptor
     *            the descriptor as an array of parsed values
     */
    public function set_descriptor (array $descriptor)
    {
        $this->add_trace(__FUNCTION__);
        
        if ($this->trace)
            file_put_contents($this->trace_file_name, " - Descriptor: " . json_encode($descriptor) . "\n", 
                    FILE_APPEND);
        
        $this->data_item->set_descriptor($descriptor);
        
        if ($this->trace)
            file_put_contents($this->trace_file_name, 
                    " - this Descriptor: " . json_encode($this->get_descriptor_clone()) . "\n", FILE_APPEND);
        
        $this->add_trace(__FUNCTION__, true);
    }

    /**
     * Get a descriptor field.
     * 
     * @param String $name
     *            the name of the field. If this field is not set, null is returned.
     * @return array the descriptor of the related data item.
     */
    public function get_descriptor_field ($name)
    {
        $this->add_trace(__FUNCTION__);
        if (strpos($name, "text_") === 0)
            return i($this->data_item->get_descriptor_field($name));
        else
            return $this->data_item->get_descriptor_field($name);
    }

    /**
     *
     * @return array the descriptor of the related data item.
     */
    public function get_descriptor_clone ()
    {
        $this->add_trace(__FUNCTION__);
        return $this->data_item->get_descriptor_clone();
    }

    /**
     *
     * @return String the name of this item.
     */
    public function get_name ()
    {
        $this->add_trace(__FUNCTION__);
        return $this->data_item->get_name();
    }

    /**
     *
     * @return String the name of this item.
     */
    public function get_uid ()
    {
        $this->add_trace(__FUNCTION__);
        return $this->data_item->get_uid();
    }

    /**
     *
     * @return String the data type.
     */
    public function get_type ()
    {
        $this->add_trace(__FUNCTION__);
        return $this->data_item->get_type();
    }

    /**
     *
     * @return String the node_addable_type for this item. THis is for completeness only, because the twin
     *         code in Javascript needs it.
     */
    public function get_addable_type ()
    {
        $this->add_trace(__FUNCTION__);
        return (is_null($this->data_item->get_descriptor_field("node_addable_type"))) ? "" : $this->data_item->get_descriptor_field(
                "node_addable_type");
    }

    /**
     *
     * @return mixed the value in the native type.
     */
    public function get_value ()
    {
        $this->add_trace(__FUNCTION__);
        return $this->data_item->get_value();
    }

    /**
     * Get the current value formatted as String. Only needed at the Javascript twin code side.
     * 
     * @return String the formatted value using the default formatting for the given type.
     */
    public function get_value_as_string (String $language_code = "csv")
    {
        $this->add_trace(__FUNCTION__);
        return Tfyh_data::format($this->get_value(), $this->get_type(), $language_code);
    }

    /**
     *
     * @return mixed the lower limit in the native type.
     */
    public function get_min ()
    {
        $this->add_trace(__FUNCTION__);
        return $this->data_item->get_min();
    }

    /**
     *
     * @return mixed the upper limit as integer, float or Date; for types which are parsed as string the
     *         maximum size.
     */
    public function get_max ()
    {
        $this->add_trace(__FUNCTION__);
        return $this->data_item->get_max();
    }

    /**
     * Get the absolute path of this item, e.g. ".app.view.width". The path is an empty Sting for the root
     * item and never ends with a trailing dot, but starts with a dot for other items than the root.
     * 
     * @return String the absolute path of this item. This is an empty Sting for the root item. The path never
     *         ends with a trailing dot.
     */
    public function get_path ()
    {
        $this->add_trace(__FUNCTION__);
        $path = "";
        $current = $this;
        while (! is_null($current->parent)) {
            $path = "." . $current->get_name() . $path;
            $current = $current->parent;
        }
        return $path;
    }

    /**
     *
     * @return Tfyh_config_item[] the children of this item.
     */
    public function get_children ()
    {
        $this->add_trace(__FUNCTION__);
        return $this->children;
    }

    /**
     *
     * @return Tfyh_config_item the parent of this item.
     */
    public function get_parent ()
    {
        $this->add_trace(__FUNCTION__);
        return $this->parent;
    }

    /**
     * Return true, if the child of the given name exists.
     * 
     * @param String $name
     *            the name of the child to get
     * @return bool true, if the child of the given name exists.
     */
    public function has_child (String $name)
    {
        $this->add_trace(__FUNCTION__);
        return (isset($this->children[$name]));
    }

    /**
     * Return true, if the config item has children at all.
     * 
     * @param String $name
     *            the name of the child to get
     * @return bool true, if the child of the given name exists.
     */
    public function has_children ()
    {
        $this->add_trace(__FUNCTION__);
        return (is_array($this->children) && count($this->children) > 0);
    }
}
