<?php
/**  
 *
 *       the tools-for-your-hobby framework
 *       ----------------------------------
 *       https://www.tfyh.org
 *
 * Copyright  2023-2024  Martin Glade
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */
include_once "../classes/tfyh_validate.php";

/**
 * A fully static utility class to provide data parsing, validation and formatting. Each data object has both
 * a String and an inMemory representation. Parsing converts String to inMemory, validation checks the value
 * against its limits, and formatting converts inMemory to String using the proper local formats. An empty
 * String is parsed as an empty String for String data types, a boolean false for boolean data types else as
 * NULL. The very last parsing, cleansing and validation result is cached in memory and must be read
 * immediately after parsing and cleansing, in particular, if null is returned, because this may be either
 * really null or a parsing, cleansing or validation error indication. Formatting applies rules according to
 * the language chosen. Two special languages are "csv", which formats for data base write actions and "csv"
 * which is the default and formats for file write. Quotation will nor be provided.
 */
class Tfyh_data
{

    /**
     * The absolute limits for size constraints. They assume 64 Bit operation. They will be adjusted to the
     * config/settings/dataTypes definitions in init():
     */
    // for 32 bit - 2147483648
    private static $int_min = - 9223372036854775808;

    // for 32 bit 2147483647
    private static $int_max = 9223372036854775807;

    // for 32 bit: - 3.40E+38
    private static $float_min = - 1.79E+308;

    // for 32 bit: 3.40E+38
    private static $float_max = 1.79E+308;

    // Date and time values are parsed to DateTimeImmutable and int within the init() function
    private static $date_min;

    private static $date_max;

    private static $time_min;

    private static $time_max;

    private static $datetime_min;

    private static $datetime_max;

    // this is an arbitrary limit which will fit into a MySQL data field
    private static $string_size = 4096;

    // this is an arbitrary limit which will fit into a MySQL (medium =
    // DEFAULT) Text blob
    private static $text_size = 65536;
    
    // eternity in efa is 9223372036854775807 (19 characters) = 9.2E+18 millis
    public const forever_seconds = 9.223372e+15;

    /**
     * The "value_parse_as" collection. All data types will be parsed or formatted using one of the keys of
     * this array. The array values are the corresponding PHP gettype() or get_class() (only
     * DateTimeImmutable) result. Note that any float value will return a "double" as gettype() because the
     * 32-Bit float type does not exist in PHP.
     */
    public static $parse_as_to_native = ["boolean" => "boolean","date" => "DateTimeImmutable",
            "datetime" => "DateTimeImmutable","float" => "double","int" => "integer","none" => "string",
            "string" => "string","time" => "integer"
    ];

    /**
     * The native types collection. All data are represented in memory using one of of the keys of this array.
     * The array values are the corresponding "value_parse_as" values. Note that there are multiple options
     * for "DateTimeImmutable", "integer", and "String". This is the best precision fit.
     */
    public static $native_to_parse_as = ["boolean" => "boolean","DateTimeImmutable" => "datetime",
            "double" => "float","integer" => "int","string" => "string"
    ];

    /**
     * The formatting rules for date and datetime fields depending on the language. The language code sql
     * reppresents the formatting for data base write actions. See
     * https://en.wikipedia.org/wiki/Date_format_by_country
     */
    private static $date_formats = ["sql" => ["date" => "Y-m-d","datetime" => "Y-m-d H:i:s"
    ],"csv" => ["date" => "Y-m-d","datetime" => "Y-m-d H:i:s"
    ],"de" => ["date" => "d.m.Y","datetime" => "d.m.Y H:i:s"
    ],"en" => ["date" => "Y-m-d","datetime" => "Y-m-d H:i:s"
    ],"fr" => ["date" => "d/m/Y","datetime" => "d/m/Y H:i:s"
    ],"it" => ["date" => "d/m/Y","datetime" => "d/m/Y H:i:s"
    ],"nl" => ["date" => "d-m-Y","datetime" => "d.m.Y H:i:s"
    ] // NL officially: "d-m-Y H:i:s", but that will result in parsing abiguities
    ];

    /**
     * The formatting rules: list of language codes for which the decimal point is a point. see
     * https://de.wikipedia.org/wiki/Dezimaltrennzeichen#Punkt-L%C3%A4nder. Note that arab countries use the
     * Momayyez (٫) \u066b.
     */
    private static $decimal_point_for = ["sql","csv","en"
    ];

    /**
     * The formatting rules: list of language codes for which a null value shall be represented as "NULL"
     * instead of an empty String
     */
    private static $write_null_as_NULL_for = ["sql"
    ];

    /**
     * The formatting rules: list of language codes for which a null value shall be represented as "NULL"
     * instead of an empty String
     */
    private static $write_boolean_as_true_false_for = ["csv"
    ];

    /**
     * The formatting rules: list of language codes for which a null value shall be represented as "NULL"
     * instead of an empty String
     */
    private static $write_boolean_as_1_0_for = ["sql"
    ];

    /**
     * Error count for last parsing, cleansing or validation.
     */
    private static $errors = [];

    /**
     * Warnings count for last parsing, cleansing or validation.
     */
    private static $warnings = [];

    /**
     * Type descriptor provides a set of fields for all purposes, ordered in category_field name like
     * "value_type" The definition of these fields is only used by this class.
     */
    private static $descriptor_fields = [];

    /**
     * The reserved names which must never be set for any item itself.
     */
    public static $item_keys = ["uid","name","parent"
    ];

    /**
     * Names of properties and functions of the javascript Array type. Using these names may lead to issues.
     */
    public static $discouraged_names_array = ["length","at","concat","copyWithin","entries","every",
            "fill","filter","find","findIndex","findLast","findLastIndex","flat","flatMap","forEach","from",
            "fromAsync","includes","indexOf","isArray","join","keys","lastIndexOf","map","of","pop","push",
            "reduce","reduceRight","reverse","shift","slice","some","sort","splice","toLocaleString",
            "toReversed","toSorted","toSpliced","toString","unshift","values","with"
    ];

    /**
     * Names of properties and functions of the javascript Object type. Using these names may lead to issues.
     */
    public static $discouraged_names_object = ["__defineGetter__","__defineSetter__","__lookupGetter__",
            "__lookupSetter__","assign","create","defineProperties","defineProperty","entries","freeze",
            "fromEntries","getOwnPropertyDescriptor","getOwnPropertyDescriptors","getOwnPropertyNames",
            "getOwnPropertySymbols","getPrototypeOf","groupBy","hasOwn","hasOwnProperty","is","isExtensible",
            "isFrozen","isPrototypeOf","isSealed","keys","preventExtensions","propertyIsEnumerable","seal",
            "setPrototypeOf","toLocaleString","toString","valueOf","values"
    ];

    /**
     * Type templates provide all default information on a data type, including the constraints, the SQL table
     * column definition, the form input default layout asf. The set of information provided at maximum is
     * defined in the descriptor fields.
     */
    public static $type_templates = [];

    /**
     * The set of templates for any object initialization.
     */
    public static $object_templates = [];

    /**
     * the value_parse_as property of all data types
     */
    private static $value_parse_as = [];

    /**
     * clear the self:$errors and self:$findings
     */
    public static function clear_findings ()
    {
        self::$errors = [];
        self::$warnings = [];
    }

    /**
     * Set the self::$validation_error according to the reason of validation failure
     * 
     * @param int $reason_code.
     *            0 = success, 1 = Format error; 2 = Numeric value required; 3 = lower limit exceeded; 4 =
     *            upper limit exceeded, 5 = unknown type for validation, 6 = unfit native type, 7 = data type
     *            w/o parse_as definition, 8 = string too long
     * @param unknown $violating_value_str
     *            the violating value, formatted as String, may be null or left out
     * @param unknown $violated_limit_str
     *            the violated limit, formatted as String, may be null or left out
     */
    private static function add_finding (int $reason_code, String $violating_value_str = null, 
            String $violated_limit_str = null)
    {
        if ($reason_code == 1)
            self::$errors[] = i("T0EpKy|Format error in °%1°", $violating_value_str);
        elseif ($reason_code == 2)
            self::$errors[] = i("4OrTAN|Numeric value required i...", $violating_value_str);
        elseif ($reason_code == 3)
            self::$warnings[] = i("XtsZfd|°%1° is too small. Repla...", $violating_value_str, 
                    $violated_limit_str);
        elseif ($reason_code == 4)
            self::$warnings[] = i("82Hnmk|°%1° is too big. Replace...", $violating_value_str, 
                    $violated_limit_str);
        elseif ($reason_code == 5)
            self::$errors[] = i("U7ppLY|Unknown data type / vali...", $violating_value_str);
        elseif ($reason_code == 6)
            self::$errors[] = i("QD9ESj|The value°s native type ...");
        elseif ($reason_code == 7)
            self::$errors[] = i("dSXCuP|The value°s data type °%...", $violating_value_str);
        elseif ($reason_code == 8)
            self::$warnings[] = i("fLIzFW|String °%1° too long. Cu...", $violating_value_str, 
                    $violated_limit_str);
    }

    /**
     * Simple getter
     */
    public static function get_errors ()
    {
        return self::$errors;
    }

    /**
     * Simple getter
     */
    public static function get_warnings ()
    {
        return self::$warnings;
    }

    /**
     * Initialization of the default data type constraints. Is called by the file inclusion, see very bottom
     * of this file.
     */
    public static function init ()
    {
        // no duplicate initialization
        if (count(self::$descriptor_fields) > 0)
            return;
        // initialize the $descriptor_fields first.
        $descriptor_settings = file_get_contents("../config/settings/descriptor");
        if ($descriptor_settings === false) {
            echo ("Desriptor settings missing. Aborting.");
            exit(); // Really exit, no test case left over.
        }
        $descriptor_fields = Tfyh_toolbox::static_read_csv_string_array($descriptor_settings);
        self::$descriptor_fields = [];
        foreach ($descriptor_fields as $field) {
            $field_name = $field["name"];
            self::$descriptor_fields[$field_name] = [];
            foreach ($field as $key => $value)
                // descriptor field values are all Strings, no parsing required.
                self::$descriptor_fields[$field_name][$key] = $value;
        }
        // remove the structure definition fields which shall never be parsed.
        unset(self::$descriptor_fields["uid"]);
        unset(self::$descriptor_fields["parent"]);
        unset(self::$descriptor_fields["name"]);
        
        // continue with creating the appropriate internal absolute date & time limits
        // all these limits will apply when doing the following parsing.
        self::$date_min = date_create_immutable("1000-01-01");
        self::$date_max = date_create_immutable("2999-12-31");
        self::$time_min = - 359999; // -99:59:59
        self::$time_max = 359999; // 99:59:59
        self::$datetime_min = date_create_immutable("1000-01-01 00:00:00");
        self::$datetime_max = date_create_immutable("2999-12-31 23:59:59");
        
        // initialize the $type_templates next.
        $type_templates = file_get_contents("../config/settings/dataTypes");
        if ($type_templates === false) {
            echo ("dataTypes settings missing. Aborting.");
            exit(); // Really exit, no test case left over.
        }
        $type_template_records = Tfyh_toolbox::static_read_csv_string_array($type_templates);
        // initialize using the "value_type" and "value_parse_as" fields first to enable parsing.
        // both fields are String, thus will not need parsing themselves.
        self::$type_templates = [];
        foreach ($type_template_records as $type_template_record) {
            $type_name = $type_template_record["name"];
            self::$type_templates[$type_name] = [];
            self::$type_templates[$type_name]["value_type"] = $type_name;
            self::$value_parse_as[$type_name] = $type_template_record["value_parse_as"];
            // configuration check at server side, no need to reflect this in Javascript.
            if (! array_key_exists($type_template_record["value_parse_as"], self::$parse_as_to_native)) {
                echo "data type template error for type $type_name. Please correct the ../config/settings/dataTypes file. Aborting.";
                exit(); // real exit, no test case left over.
            }
        }
        // now parse all data type templates
        foreach ($type_template_records as $type_template_record) {
            $type_name = $type_template_record["name"];
            self::$type_templates[$type_name] = self::parse_descriptor($type_template_record);
        }
        
        // add "unknown" templates as default for undefined types
        self::$type_templates["_of_value"] = self::clone_descriptor(self::$type_templates["unknown"]);
        self::$type_templates[""] = self::clone_descriptor(self::$type_templates["unknown"]);
        
        // now take the passed limits as validation parameters.
        self::$int_min = self::$type_templates["int"]["value_min"];
        self::$int_max = self::$type_templates["int"]["value_max"];
        self::$float_min = self::$type_templates["float"]["value_min"];
        self::$float_max = self::$type_templates["float"]["value_max"];
        self::$date_min = self::$type_templates["date"]["value_min"];
        self::$date_max = self::$type_templates["date"]["value_max"];
        self::$time_min = self::$type_templates["time"]["value_min"];
        self::$time_max = self::$type_templates["time"]["value_max"];
        self::$datetime_min = self::$type_templates["datetime"]["value_min"];
        self::$datetime_max = self::$type_templates["datetime"]["value_max"];
        self::$string_size = self::$type_templates["string"]["value_size"];
        self::$text_size = self::$type_templates["text"]["value_size"];
    }

    /**
     * Get the parsing rule for a data type
     * 
     * @param String $data_type
     *            the data type to get the parsing rule for.
     * @return string the parsing rule. String for objects and unmatched types.
     */
    public static function parse_as (String $data_type)
    {
        // valid data type
        if (isset(self::$value_parse_as[$data_type]))
            return self::$value_parse_as[$data_type];
        // object or any other. Use string as the most tolerant parsing rule.
        return "string";
    }

    /**
     * check whether the name is a reserved name, i. e. one of the descriptor fields or one of "name", "uid",
     * "parent" or within the discouraged names that represent properties of the Object or array prototype in
     * javascript
     * 
     * @param String $name
     *            the name to check.
     * @return boolean true, if the name is reserved.
     */
    public static function is_reserved_name (String $name)
    {
        return array_key_exists($name, self::$descriptor_fields) || in_array($name, self::$item_keys) ||
                 in_array($name, self::$discouraged_names_array) ||
                 in_array($name, self::$discouraged_names_object);
    }

    /**
     * Check whether a value is "empty", i. e. either null, or a String of length 0.
     * 
     * @param mixed $value
     *            the value to check.
     * @return boolean true, if the value is empty.
     */
    public static function is_empty ($value)
    {
        // Note: is_string(null) = false.
        return (is_string($value)) ? (strlen($value) > 0) : ! is_null($value);
    }

    /**
     * Check whether the name is a name of a descriptor field.
     * 
     * @param String $name
     *            the name to check.
     * @return boolean true, if the name is a name of a descriptor field.
     */
    public static function is_descriptor_field_name (String $name)
    {
        return array_key_exists($name, self::$descriptor_fields);
    }

    /**
     * Check whether the data type or object template exists.
     * 
     * @param String $data_type
     *            the data type to check
     * @return boolean true, if the condition is met.
     */
    private static function valid_type (String $data_type)
    {
        return (isset(self::$value_parse_as[$data_type]) || isset(self::$object_templates[$data_type]));
    }

    /* ------------------------------------------------------------------------ */
    /* ----- DESCRIPTOR HANDLING ---------------------------------------------- */
    /* ------------------------------------------------------------------------ */
    
    /**
     * get a default descriptor clone for a given data type, either the simple type descriptor or the objects
     * root item descriptor. Returns the "unknown" type descriptor, if the data type is not matched.
     * 
     * @param String $data_type
     *            the data type to get the template descriptor for.
     * @return array cloned descriptor template for the type or the object root, or the unknown type, if not
     *         matched.
     */
    public static function get_descriptor_template (String $data_type)
    {
        // get default descriptor
        if (isset(self::$type_templates[$data_type])) {
            // known simple data type
            return self::clone_descriptor(self::$type_templates[$data_type]);
        } elseif (isset(self::$object_templates[$data_type])) {
            // known object data type
            return self::$object_templates[$data_type]->get_descriptor_clone();
        } else {
            // unknown data type
            return self::clone_descriptor(self::$type_templates["unknown"]);
        }
    }

    /**
     * Get a limit for the respective descriptor template.
     * 
     * @param String $data_type
     *            the data type to get the template descriptor for.
     * @param int $limit_index
     *            0 for min,, 1 for max, 2 for maximum size
     * @return int|float the respective limit
     */
    public static function get_descriptor_template_limit (String $data_type, int $limit_index)
    {
        if ($limit_index == 0) {
            if (isset(self::$type_templates[$data_type]) &&
                     isset(self::$type_templates[$data_type]["value_min"]))
                return self::$type_templates[$data_type]["value_min"];
        } elseif ($limit_index == 1) {
            if (isset(self::$type_templates[$data_type]) &&
                     isset(self::$type_templates[$data_type]["value_max"]))
                return self::$type_templates[$data_type]["value_max"];
        } elseif ($limit_index == 2) {
            if (isset(self::$type_templates[$data_type]) &&
                     isset(self::$type_templates[$data_type]["value_size"]))
                return self::$type_templates[$data_type]["value_size"];
        }
        // no value returned, i. e. no value existing. return null.
        return null;
    }

    /**
     * Get the data type for this descriptor field. If the data type is "_of_value", return $of_value. If the
     * $descriptor_field_name is not matched, return "string" (in particular this can haappen for the reserved
     * names uid, parent, name).
     * 
     * @return array the names of all descriptor fields
     */
    public static function get_descriptor_field_names ()
    {
        $names = [];
        foreach (self::$descriptor_fields as $name => $field)
            $names[] = $name;
        return $names;
    }

    /**
     * Get the data type for this descriptor field. If the data type is "_of_value", return $of_value. If the
     * $descriptor_field_name is not matched, return "string" (in particular this can haappen for the reserved
     * names uid, parent, name).
     * 
     * @param String $descriptor_field_name
     *            the name of the descriptor field.
     * @param String $of_value
     *            this will be returned, if the descriptor field data type is "_of_value" (case sensitive).
     * @return string the data type of this descriptor field
     */
    public static function get_descriptor_field_data_type (String $descriptor_field_name, String $of_value)
    {
        if (isset(self::$descriptor_fields[$descriptor_field_name])) {
            $dt = self::$descriptor_fields[$descriptor_field_name]["value_type"];
            return (strcmp($dt, "_of_value") == 0) ? $of_value : $dt;
        }
        return "string";
    }

    /**
     * Simple getter.
     * 
     * @param String $descriptor_field_name
     *            the name of the descriptor field.
     * @return string the local name of this descriptor field
     */
    public static function get_descriptor_field_local_name (String $descriptor_field_name)
    {
        if (isset(self::$descriptor_fields[$descriptor_field_name]))
            return self::$descriptor_fields[$descriptor_field_name]["text_local_name"];
        return i("OcBU6p|unknown descriptor field");
    }

    /**
     * Parse a descriptor with only String values as read from a csv definition and return it with all fields.
     * Start with default descriptor for the given data type and overwrite those fields defined by the
     * $descriptor_record and not being empty.
     * 
     * @param array $descriptor_record
     *            the descriptor as a record of all Strings which shall be parsed
     * @return array the parsed descriptor, with default values where the $descriptor_record has no value set.
     */
    public static function parse_descriptor (array $descriptor_record)
    {
        $descriptor_data_type = (isset($descriptor_record["value_type"])) ? $descriptor_record["value_type"] : "string";
        if ((substr($descriptor_data_type, 0, 1) != '.') &&
                 ! isset(self::$type_templates[$descriptor_data_type]))
            // this will only happen during initialization
            self::$type_templates[$descriptor_data_type] = [];
        
        // initialize with default. Note that this was already parsed.
        $parsed = self::get_descriptor_template($descriptor_data_type);
        // parse all known descriptor fields and overwrite the default, if
        // matched
        foreach (self::$descriptor_fields as $name => $field) {
            $field_data_type = (strcasecmp($field["value_type"], "_of_value") == 0) ? $descriptor_data_type : $field["value_type"];
            // Parse all fields except the empty ones and the item keys uid, parent, and name
            if (isset($descriptor_record[$name]) && (strlen($descriptor_record[$name])) &&
                     ! in_array($name, self::$item_keys)) {
                $parsed_value = self::parse($descriptor_record[$name], $field_data_type);
                // the parse result will be false on failures, but in these cases the value
                // shall rather be null except for boolean parsing.
                $parsed[$name] = ((strcasecmp($field_data_type, "boolean") != 0) && ($parsed_value === false)) ? null : $parsed_value;
            }
        }
        // ensure that at least the data type is defined
        if (! isset($parsed["value_type"]))
            $parsed["value_type"] = "string";
        // remove min/max = false defaults for boolean created by parsing
        $is_boolean = (strcasecmp($parsed["value_type"], "boolean") == 0);
        if ($is_boolean || ! isset($parsed["value_min"])) // isset() includes is_null() to avoid
                                                          // an "Undefined array key" PHP log warning
            unset($parsed["value_min"]);
        if ($is_boolean || ! isset($parsed["value_max"]))
            unset($parsed["value_max"]);
        return $parsed;
    }

    /**
     * Clone a descriptor by deep array copy. Date/datetime values are copied by object cloning. Because they
     * are objects a simple passing as parameter may not create a full copy.
     * 
     * @param array $descriptor
     *            the original
     * @return array[] the clone
     */
    public static function clone_descriptor (array $descriptor)
    {
        $clone = [];
        foreach ($descriptor as $name => $value) {
            if (is_object($value))
                $clone[$name] = clone $value;
            else
                $clone[$name] = $value;
        }
        return $clone;
    }

    /**
     * Cleanse the descriptor by removing all elements which have a vaue that is identical to the default for
     * the chosen data type. Upon construction these defaults are the starting point and therefore always
     * included. But for writing a descriptor, defaults carry redundancy around.
     * 
     * @param array $descriptor
     *            the descriptor to cleanse as two level array of native values.
     */
    public static function clone_without_defaults (array $descriptor)
    {
        // get default descriptor
        $data_type = $descriptor["value_type"];
        $defaults = self::get_descriptor_template($data_type);
        // For "unknown" data type also no defaults to remove.
        if (strcmp($defaults["value_type"], "unknown") == 0)
            // unknown data type
            return self::clone_descriptor($descriptor);
        // clone descriptor
        $cleansed_descriptor = self::clone_descriptor($descriptor);
        // remove equal values
        foreach ($cleansed_descriptor as $name => $field) {
            // do not remove the data type, current value and text local name.
            $is_current = (strcasecmp($name, "value_current") == 0);
            $is_type = (strcasecmp($name, "value_type") == 0);
            $is_local_name = (strcasecmp($name, "text_local_name") == 0);
            $keep = $is_current || $is_type || $is_local_name;
            if (! $keep && self::is_equal_values($field, (isset($defaults[$name])) ? $defaults[$name] : null)) {
                unset($cleansed_descriptor[$name]);
            }
        }
        return $cleansed_descriptor;
    }

    /* ------------------------------------------------------------------------ */
    /* ----- DATA EQUALITY ---------------------------------------------------- */
    /* ------------------------------------------------------------------------ */
    
    /**
     * Drilldown for differnce check in arrays. Keys must also be identical, but not in their sequence.
     * 
     * @param unknown $a
     *            array of values of any type (values may also be an array)
     * @param unknown $b
     *            array of values of any type (values may also be an array)
     * @return @return String empty if equal, else a difference report.
     */
    public static function diff_arrays ($a, $b)
    {
        $diff = "";
        $keys_checked = [];
        foreach ($a as $k => $v) {
            $diff_single = self::diff_single($v, (isset($b[$k])) ? $b[$k] : null);
            $keys_checked[] = $k;
            $diff .= (strlen($diff_single) > 0) ? $k . ": " . $diff_single . ", " : "";
        }
        foreach ($b as $k => $v) {
            if (! in_array($k, $keys_checked))
                $diff .= $k . ": " . i("XPWCVX|Extra field in B") . ",";
        }
        return $diff;
    }

    /**
     * Create a difference statement for two values.
     * 
     * @param $a any
     *            value of any type
     * @param $b any
     *            value of any type
     * @return String empty, if they are equal, a difference statement, if not.
     */
    private static function diff_single ($a, $b)
    {
        $diff = "";
        // echo "diff_single: "; var_dump($a); echo " ?= "; var_dump($b);echo "<br>";
        
        // start with simple cases: null equality
        if (is_null($a))
            $diff .= (is_null($b)) ? "" : i("EMt2mp|A is null, but B is not ...") . " ";
        // start with simple cases: array type equality
        elseif (is_array($a) && ! is_array($b))
            $diff .= i("B8udF2|A is an array, but B not...") . " ";
        elseif (! is_array($a) && is_array($b))
            $diff .= i("sF4MkY|A is a single value, but...") . " ";
        
        // drilldown in case of two arrrays
        elseif (is_array($a))
            $diff .= self::diff_arrays($a, $b);
        
        // single values
        // boolean
        elseif (is_bool($a))
            $diff .= (is_bool($b)) ? (($a == $b) ? "" : i("4igrr0|boolean A is not(boolean...")) : i(
                    "xitScV|A is boolean, B not.");
        // integer or time
        elseif (is_int($a))
            $diff .= (is_int($b)) ? (($a == $b) ? "" : i("jCvlcK|integer A != integer B.")) : i(
                    "YA9xLD|A is integer, B not.");
        // floating point, maximum requested precision is 1E-11
        elseif (is_float($a))
            $diff .= (is_float($b)) ? (($a == $b) ? "" : i("YgV1U0|float A != float B.")) : i(
                    "qQXEIR|A is float, B not.");
        // date, time, datetime
        elseif (is_object($a)) {
            // only DateTimImmutable objects are allowed in the Tfyh data context as value objects
            if (! is_object($b))
                $diff .= i("8qn1n1|A is object, B not.");
            elseif (strcasecmp(get_class($a), "DateTimeImmutable") != 0)
                $diff .= i("OHYGer|A is object, but not a D...");
            elseif (strcasecmp(get_class($b), "DateTimeImmutable") != 0)
                $diff .= i("K8RkEM|A is DateTimeImmutable, ...");
            $diff .= ((strcmp($a->format("Y-m-d H:i:s"), $b->format("Y-m-d H:i:s")) == 0)) ? "" : i(
                    "z5G34s|datetime A != datetime B...");
        } elseif (is_string($a)) // String
            $diff .= (is_string($b)) ? ((strcmp($a, $b) == 0) ? "" : i("fkVBQK|string A differs from st...")) : i(
                    "y5iS0B|A is a string, B not.");
        // no other values suported. They are always regarded as unequal.
        else
            $diff .= i("Wxhd4A|equality check failed du...") . json_encode($a) . ".";
        
        // echo " result: " . $diff . "<br>";
        return $diff;
    }

    /**
     * Drilldown for equality check in arrays. Keys must also be identical, but not in their sequence.
     * is_null($a[$k]) is regarded as equal to both $b[$k] not set and $b[$k] = null. The same vice versa.
     * 
     * @param unknown $a
     *            array of values of any type (values may also be an array)
     * @param unknown $b
     *            array of values of any type (values may also be an array)
     * @return boolean true, if equality is given, false else. No error option.
     */
    private static function is_equal_arrays ($a, $b)
    {
        $equal = true;
        $keys_collected = [];
        foreach ($a as $k => $v) {
            $equal &= self::is_equal_values($a[$k], $b[$k]);
            $keys_collected[] = $k;
        }
        foreach ($b as $k => $v) {
            if (! in_array($k, $keys_collected))
                $equal = false;
        }
        return $equal;
    }

    /**
     * Check whether two values of data are equal. Special cases: null == null, type safety (i. e. corresponds
     * to the ==== operand), array drilldown and equality means equality on all levels, date/time/datetime
     * equality as "Y-m-d H:i:s - equality", floating point numbers equal to PHP_FLOAT_EPSILON precision only.
     * 
     * @param $a any
     *            value of any type
     * @param $b any
     *            value of any type
     * @return boolean true, if equality is given, false else. No error option.
     */
    private static function is_equal_values ($a, $b, bool $echo = false)
    {
        if ($echo) {
            var_dump($a);
            echo " --- ";
            var_dump($b);
            echo ": ";
        }
        $equal = true;
        // start with simple cases: null equality
        if (is_null($a))
            $equal = is_null($b);
        // start with simple cases: array type equality
        elseif (is_array($a) xor is_array($b))
            $equal = false;
        
        // drilldown in case of two arrrays
        elseif (is_array($a))
            $equal &= self::is_equal_arrays($a, $b);
        
        // single values
        // boolean
        elseif (is_bool($a))
            $equal &= is_bool($b) && ($a == $b);
        // integer or time
        elseif (is_int($a))
            $equal &= is_int($b) && ($a == $b);
        // floating point, maximum provided precision is PHP_FLOAT_EPSILON
        elseif (is_float($a))
            $equal &= is_float($b) && ((($a == 0) && ($b == 0)) || (abs(($a - $b) / $a) <= PHP_FLOAT_EPSILON));
        // date, time, datetime
        elseif (is_object($a))
            // only DateTimImmutable objects are allowed in the Tfyh data context as value objects
            $equal &= is_object($b) && (strcasecmp(get_class($a), "DateTimeImmutable") == 0) &&
                     (strcasecmp(get_class($b), "DateTimeImmutable") == 0) &&
                     (strcmp($a->format("Y-m-d H:i:s"), $b->format("Y-m-d H:i:s")) == 0);
        // String
        elseif (is_string($a))
            $equal &= is_string($b) && (strcmp($a, $b) == 0);
        // no other values suported. They are always regarded as unequal.
        else
            $equal = false;
        if ($echo) {
            echo ($equal) ? "equal" : "not equal";
            echo "<br>";
        }
        return $equal;
    }

    /* ------------------------------------------------------------------------ */
    /* ----- DATA VALIDATION -------------------------------------------------- */
    /* ----- The result is a native type value, null in case of errors. ------- */
    /* ----- Errors can occur on parsing and limit or rules check. ------------ */
    /* ----- Errors are documented in the public $validation_error. ----------- */
    /* ------------------------------------------------------------------------ */
    
    /**
     * Check whether the native type of $value matches the data type expectations. The value must not be an
     * array.
     * 
     * @param String $data_type
     *            thge data type of the data item
     * @param unknown $value
     *            the value to check, if its native type matches the data type
     * @return bool true, if the types match, in case of arrays if the types match for each element, false
     *         else. A $value of null always returns true. If no "parsing_as" property exists for the given
     *         $data_type, it will return false.
     */
    private static function is_matching_type_single (String $data_type, $value)
    {
        if (is_null($value))
            return true;
        if (! self::valid_type($data_type))
            return false;
        if (is_array($value))
            return false; // arrays cannot be matched to a single value
        if (substr($data_type, 0, 1) == ".")
            return false; // objects can ot be matched to a single value
        $value_parse_as = self::parse_as($data_type);
        $native = (is_object($value)) ? get_class($value) : gettype($value);
        return (strcasecmp(self::$parse_as_to_native[$value_parse_as], $native) == 0);
    }

    /**
     * Check whether a value fits the native PHP type matching the type constraints and its min/max limits.
     * Single values only, no arrays allowed. Values exceeding limits are adjusted to the exceeded limit.
     * 
     * @param String $value
     *            the single value to validate
     * @param String $type
     *            the type to assume for validation. This also definesd the expected PHP native type
     * @param mixed $min
     *            the lower limit to apply for validation. Set null to use type defaults.
     * @param mixed $max
     *            the upper limit or maximum length for String to apply for validation. Set null to use type
     *            defaults.
     * @return mixed the value using the native PHP type. false in case of errors. If the parsing rule is
     *         boolean, an error can only be distinguished from the validated value by checking the
     *         self::$validation_errors_count
     */
    private static function check_limits_single ($value, String $type, $min, $max)
    {
        if (! self::valid_type($type)) {
            self::add_finding(7, $type);
            return false;
        }
        if (! self::is_matching_type_single($type, $value)) {
            self::add_finding(6, Tfyh_data::format($value));
            return false;
        }
        $parse_as = self::parse_as($type);
        // at that point the data type is one of the programmatically defined basic types
        if (strcasecmp($parse_as, "int") == 0) {
            $llimit = max($min, self::$int_min);
            $ulimit = min($max, self::$int_max);
        } elseif (strcasecmp($parse_as, "float") == 0) {
            $llimit = max($min, self::$float_min);
            $ulimit = min($max, self::$float_max);
        } elseif (strcasecmp($parse_as, "boolean") == 0) {
            return $value; // a boolean value never has limits
        } elseif (strcasecmp($parse_as, "date") == 0) {
            // min/max may not work with objects
            $llimit = (self::$date_min < $min) ? $min : self::$date_min;
            $ulimit = (self::$date_max > $max) ? $max : self::$date_max;
        } elseif (strcasecmp($parse_as, "datetime") == 0) {
            $llimit = (self::$datetime_min < $min) ? $min : self::$datetime_min;
            $ulimit = (self::$datetime_max > $max) ? $max : self::$datetime_max;
        } elseif (strcasecmp($parse_as, "time") == 0) {
            $llimit = max($min, self::$time_min);
            $ulimit = min($max, self::$time_max);
        } elseif (strcasecmp($parse_as, "string") == 0) {
            $ulimit = (strcasecmp($type, "text") == 0) ? min($max, self::$text_size) : min($max, 
                    self::$string_size);
            if (mb_strlen($value) > $ulimit) {
                // shorten String, if too long
                self::add_finding(8, 
                        mb_substr($value, 0, min(mb_strlen($value), 20)) . "[" . $type . "(" .
                                 mb_strlen($value) . "])", strval($ulimit));
                return ($ulimit > 12) ? mb_substr($value, 0, $ulimit - 4) . " ..." : mb_substr($value, 0, 
                        $ulimit);
            } else
                return $value;
        } else {
            // unknown type
            self::add_finding(7, $type);
            return false;
        }
        
        // adjust value to not exceed the limits
        if ($value < $llimit) {
            self::add_finding(3, self::format($value), self::format($llimit, $type));
            return $llimit;
        } elseif ($value > $ulimit) {
            self::add_finding(4, self::format($value), self::format($ulimit, $type));
            return $ulimit;
        } else
            return $value;
    }

    /**
     * Check whether a value fits the native PHP type matching the type constraints and its min/max limits.
     * Values exceeding limits are adjusted to the exceeded limit.
     * 
     * @param String $value
     *            the value to validate
     * @param String $type
     *            the type to assume for validation. This also definesd the expected PHP native type
     * @param mixed $min
     *            the lower limit to apply for validation. Set null to use type defaults.
     * @param mixed $max
     *            the upper limit or maximum length for String to apply for validation. Set null to use type
     *            defaults.
     * @return mixed the value using the native PHP type - including null => null - or false, if the
     *         validation fails. If the parsing rule is boolean, an error can only be distinguished from the
     *         validated value by checking the self::$validation_errors_count. For arrays an array is returned
     *         with each single value being validated accordingly. See self::$errors_count or
     *         self::$validation_warnings_count to assess, if errors or warnings occurred. $value == null is
     *         valid in all cases.
     */
    public static function check_limits ($value, String $type, $min = null, $max = null)
    {
        // null is always valid
        if (is_null($value))
            return null;
        // identify constraints which shall apply.
        if (is_null($min) || (! is_object($min) && (strlen($min) == 0)))
            $min = self::get_descriptor_template_limit($type, 0);
        if (is_null($max) || (! is_object($max) && (strlen($max) == 0)))
            if (strcmp(self::parse_as($type), "string") != 0)
                $max = self::get_descriptor_template_limit($type, 1);
            else
                $max = self::get_descriptor_template_limit($type, 2);
        // validate single
        if (! is_array($value))
            return self::check_limits_single($value, $type, $min, $max);
        // ... or array
        $checked = [];
        foreach ($value as $element)
            $checked[] = self::check_limits_single($element, $type, $min, $max);
        return $checked;
    }

    /**
     * Guess the type of the value. value, usually a String, but may as well be another native type. If $value
     * is an array, the type of the $value[0] will be guessed. Null values return a "none" data type.
     * 
     * @param $value the
     *            value for the item, usually a String, but may as well be another native type. If $value is
     *            an array, the type of the $value[0] will be guessed. Null values return a "string" data
     *            type.
     * @return String the data type guessed.
     */
    public static function guess_type ($value)
    {
        if (is_null($value))
            return "string";
        if (is_array($value))
            $value = $value[0];
        if (is_object($value)) {
            $class_name = get_class($value);
            if (strcasecmp($class_name, "DateTimeImmutable") == 0)
                return "datetime";
            else
                return "object";
        }
        
        // for a non-string value, use the self::$native_to_parse_as mapping
        $value_native_type = gettype($value);
        if (strcmp($value_native_type, "string") != 0)
            return self::$native_to_parse_as[$value_native_type];
        
        // if String, check null and for enclosing [] as array marker.
        if ((strcmp($value, "NULL") == 0) || (strcmp($value, "null") == 0))
            return "string";
        $trimmed = trim($value);
        if ((strcmp(substr($trimmed, 0, 1), "[") == 0) &&
                 (strcmp(mb_substr($trimmed, mb_strlen($trimmed) - 1, 1), "]") == 0)) {
            $array = str_getcsv(mb_substr($trimmed, 1, mb_strlen($trimmed) - 2), ",", "\"");
            $value = $array[0];
        }
        
        // an empty String is always a String
        if (strlen($value) == 0)
            return "string";
        // a long String is always text
        if (strlen($value) > 4096)
            return "text";
        // check for numbers
        if (is_numeric($value)) {
            if (intval($value) == intval(floatval($value))) {
                $int = intval($value);
                if (($int > 100000000000) || ($int < 3000000000000))
                    return "microtime";
                elseif (($int > 2147483647) || ($int < - 2147483648))
                    return "string";
                else
                    return "int";
            }
            return "float";
        }
        // check for boolean
        if ((strcasecmp($value, "on") == 0) || (strcasecmp($value, "true") == 0) ||
                 (strcasecmp($value, "false") == 0))
            return "boolean";
        // check for date and time separately
        if ((strlen(trim($value)) < 12) && (strpos(trim($value), " ") === false)) {
            // if a colon is contained, it may be a time format
            if (strpos($value, ":") !== false) {
                // Short String without space character and with colon
                $time = self::cleanse_time($value);
                if ($time === false)
                    return "string";
                return "time";
            }
            
            // if a dot or dash is contained, it may be a date format
            if ((strpos($value, "-") !== false) || (strpos($value, "-") !== false) ||
                     (strpos($value, "/") !== false)) {
                // Short String without space character and with dot or dash
                $date = self::cleanse_date($value);
                if ($date === false)
                    return "string";
                return "date";
            }
        }
        // check for datetime
        if ((strlen(trim($value)) < 20) && (count(explode(" ", trim($value))) == 2) &&
                 (strpos($value, ":") !== false)) {
            // Short String with single space character and with colon
            $datetime = self::cleanse_datetime($value);
            if ($datetime === false)
                return "string";
            return "datetime";
        }
        // no specific format detected
        return "string";
    }

    /* ------------------------------------------------------------------------ */
    /* ----- DATA PARSING ----------------------------------------------------- */
    /* ----- The result is a native type value or null, in case of errors. ---- */
    /* ----- Errors are documented in the public $validation_error. ----------- */
    /* ------------------------------------------------------------------------ */
    
    /**
     * Convert a single or multiple value String to a PHP native single value or array. In case of errors the
     * return value will be 0 or its equivalent (false, new Date(0)), either single or as array element. For
     * self::parse_as($type) == "none" the return value is always null.
     * 
     * @param String $value_string            
     * @param String $type
     *            the data type to be used. Set null to guess
     * @return mixed the native PHP value. Collect errors in
     */
    public static function parse (String $value_string = null, String $type = null)
    {
        // Guess type, if not given.
        if (is_null($type))
            $type = Tfyh_data::guess_type($value_string);
        // parse_as "none" returns always null.
        if (strcmp(self::parse_as($type), "none") == 0)
            return null;
        
        // initial clearing and null checks
        if (is_null($value_string))
            return (strcmp($type, "boolean") == 0) ? false : null;
        
        if ((strcmp($value_string, "NULL") == 0) || (strcmp($value_string, "null") == 0))
            return null;
        
        // trim string for correct array identification
        $trimmed = trim($value_string);
        // arrays always start and end with a square bracket
        if ((strcmp(substr($trimmed, 0, 1), "[") != 0) ||
                 (strcmp(mb_substr($trimmed, mb_strlen($trimmed) - 1, 1), "]") != 0))
            // no array, return the parsed singe value
            return self::parse_single($trimmed, $type);
        $trimmed = mb_substr($trimmed, 1, mb_strlen($trimmed) - 2);
        $array = str_getcsv($trimmed, ",", "\"");
        // parse each element of the array
        $parsed = [];
        // the type must be always the same, no mixed arrays.
        foreach ($array as $element)
            $parsed[] = self::parse_single(trim($element), $type);
        return $parsed;
    }

    /**
     * Convert a single value String to a PHP native value. 'NULL' or 'null' will return null. In case of
     * errors the return value will be 0 or its equivalent (false, new DateTimeImmutable), either single or as
     * array element.
     * 
     * @param String $value_string            
     * @param String $type
     *            the data type to be used. Set null to guess
     * @return mixed the parsed value as native PHP variable. In case of errors the return value will be 0 or
     *         its equivalent (false, new Date(0)). For self::parse_as($type) == "none" the return value is
     *         always null.
     */
    private static function parse_single (String $value_string, String $type)
    {
        // initial clearing and null checks
        if (is_null($value_string))
            return (strcmp($type, "boolean") == 0) ? false : null;
        if ((strcmp($value_string, "NULL") == 0) || (strcmp($value_string, "null") == 0))
            return null;
        // parse boolean and String according to the rule
        $parse_as = self::parse_as($type);
        if (strcasecmp($parse_as, "string") == 0)
            return $value_string;
        if (strcasecmp($parse_as, "boolean") == 0) {
            return self::parse_boolean($value_string);
        }
        
        // Neither String, nor boolean: an empty String represents NULL.
        if (strlen($value_string) == 0)
            return null;
        
        // parse other according to the rule
        if (strcasecmp($parse_as, "int") == 0)
            return self::parse_number($value_string, false);
        if (strcasecmp($parse_as, "float") == 0)
            return self::parse_number($value_string, true);
        if (strcasecmp($parse_as, "date") == 0)
            return self::parse_date($value_string);
        if (strcasecmp($parse_as, "datetime") == 0)
            return self::parse_datetime($value_string);
        if (strcasecmp($parse_as, "time") == 0)
            return self::parse_time($value_string);
        
        // unknown type
        self::add_finding(5, $type . ": " . $parse_as);
        return false;
    }

    /**
     * Convert a String to boolean. Returns false, if bool_string = null, "", "FALSE", "false" (csv) or "0"
     * (sql tinyint), else this wil return true. Note that null or undefined are also converted to false, see
     * #parseSingle().
     * 
     * @param String $bool_string            
     * @return bool In case of errors the return value will be false.
     */
    private static function parse_boolean (String $bool_string = null)
    {
        return (! is_null($bool_string) && (strlen($bool_string) > 0) && (strcmp($bool_string, "false") != 0) &&
                 (strcmp($bool_string, "FALSE") != 0) && (strcmp($bool_string, "0") != 0));
    }

    /**
     * Convert a not-empty String to a number. null or undefined will return null, 'NULL' or 'null' will
     * return null. If not is_numeric($number_string) this will return 0.
     * 
     * @param String $number_string            
     * @param bool $floating_point
     *            set true to return a float, false to return an int
     * @return int|float the integer or float value. In case of errors this will be 0. For null, empty Strings
     *         and 'NULL', or 'null' null will be returned.
     */
    private static function parse_number (String $number_string, bool $floating_point)
    {
        // check for ,/. replacement need
        if (strpos($number_string, ",") !== false) {
            if ((strrpos($number_string, ".") !== false) &&
                     (strrpos($number_string, ".") > strrpos($number_string, ",")))
                // UK style like 1,234.567 for 1234.567
                $number_string = str_replace(",", "", $number_string);
            elseif (count(explode(",", $number_string)) > 1)
                // UK style like 1,234,567 for 1234567
                $number_string = str_replace(",", "", $number_string);
            else
                // DE style like 1234,567 or 1.234,567 for 1234.567. Note: 1,234 will be taken for DE style
                // 1.234, not for UK style 1234
                $number_string = str_replace(",", ".", str_replace(".", "", $number_string));
        } elseif ((count(explode(".", $number_string)) > 2))
            // DE style like 1.234.567 for 1234567
            $number_string = str_replace(".", "", $number_string);
        
        // check whether the cleansed string is numeric
        if (! is_numeric($number_string)) {
            self::add_finding(2, $number_string);
            return 0;
        }
        
        // parse the string
        return ($floating_point) ? floatval($number_string) : intval($number_string);
    }

    /**
     * Convert a String to a number of seconds. no limits to the number of hours apply. null or undefined will
     * return null, 'NULL' or 'null' will return null. If time doesn't contain a ":" this will be 0.
     * 
     * @param String $time_string
     *            the string to parse
     * @param bool $no_hours
     *            set true to interpret xx:yy as mm:ss rather than hh:mm.
     * @return int timestring in seconds. On errors 0.
     */
    private static function parse_time (String $time_string, bool $no_hours = false)
    {
        $cleansed = Tfyh_data::cleanse_time($time_string, $no_hours);
        if ($cleansed === false) {
            self::add_finding(2, $time_string);
            return 0;
        }
        $sign = (strcmp(substr(trim($time_string), 0, 1), "-") == 0) ? - 1 : 1;
        $hms = explode(":", $cleansed);
        return $sign * (abs(intval($hms[0])) * 3600 + intval($hms[1]) * 60 + intval($hms[2]));
    }

    /**
     * Convert a String to a DateTimeImmutable. If the year is two digits only, It will be assumed to be in
     * the range of this year -89years .. +10years. null or undefined will return null, 'NULL' or 'null' will
     * return null. In case of errors this wll be new DateTimeImmutable().
     * 
     * @param String $date_string
     *            a date String either dd.mm.yyyy or yyyy-mm-dd.
     * @return DateTimeImmutable the parsed date. On errors false is returned and null for $date_string ==
     *         "NULL" (not case sensitive).
     */
    private static function parse_date (String $date_string)
    {
        $datestr = self::cleanse_date($date_string);
        if ($datestr === false) {
            self::add_finding(1, $date_string);
            return new DateTimeImmutable();
        }
        $dti = date_create_immutable($datestr);
        if ($dti === false) {
            self::add_finding(2, $date_string);
            return new DateTimeImmutable();
        }
        return $dti;
    }

    /**
     * Convert a datetime String to a DateTimeImmutable Object. If no date is given, the current date is
     * inserted. If no time is given, the current time is inserted.
     * 
     * @param String $datetime_string
     *            the String to be converted.
     * @return DateTimeImmutable the parsed datetime. On errors false is returned and null for $date_string ==
     *         "NULL" (not case sensitive).
     */
    private static function parse_datetime (String $datetime_string)
    {
        $datetime = self::cleanse_datetime($datetime_string);
        if (is_null($datetime)) {
            self::add_finding(1, $datetime_string);
            return false;
        }
        $dti = date_create_immutable($datetime);
        if ($dti === false) {
            self::add_finding(2, $datetime_string);
            return false;
        }
        return $dti;
    }

    /* ------------------------------------------------------------------------ */
    /* ----- STRING CLEANSING ------------------------------------------------- */
    /* ----- Errors are documented in the public $validation_error. ----------- */
    /* ------------------------------------------------------------------------ */
    
    /**
     * Cleanse a date string to YYYY-MM-DD format.
     * 
     * @param String $date_string
     *            a date String either dd.mm.yyyy or yyyy-mm-dd.
     * @return String a date in YYYY-MM-DD format. On errors false is returned and "NULL" for $date_string ==
     *         "NULL" (not case sensitive).
     */
    private static function cleanse_date (String $date_string)
    {
        if (strpos($date_string, ".") !== false) {
            // de, nl (for simplicity)
            $parts = explode(".", $date_string);
            $s = [2,1,0
            ];
        } elseif (strpos($date_string, "/") !== false) {
            // fr, it
            $parts = explode("/", $date_string);
            $s = [2,1,0
            ];
        } else {
            // en, csv, sql
            $parts = explode("-", $date_string);
            $s = [0,1,2
            ];
        }
        
        $current_year = intval(date("Y"));
        if (count($parts) == 1) {
            // single integer is taken to be a year. Add first of Jánuary
            if ((intval($parts[0]) > 1000) && (intval($parts[0]) < 2999))
                return $parts[0] . "-01-01";
            self::add_finding(1, $date_string);
            return false;
        }
        
        $y = (count($parts) < 3) ? intval(date("Y")) : intval($parts[$s[0]]);
        if (count($parts) == 2) {
            // duplicate integer is taken to be day + month
            $m = ($s[0] > $s[2]) ? intval($parts[0]) : intval($parts[1]);
            $d = ($s[0] > $s[2]) ? intval($parts[1]) : intval($parts[0]);
            if (($m >= 1) && ($m <= 12) && ($d >= 1) && ($d >= 31))
                return $y . "-" . sprintf("%'.02d", $m) . "-" . sprintf("%'.02d", $d);
            self::add_finding(1, $date_string);
            return false;
        }
        
        if ($y < 100) {
            // extend two digits. Will only work fine until 2090.
            $year_now = $current_year % 100;
            $saeculum_now = intval(date("Y")) - $year_now;
            $y = ($y > ($year_now + 10)) ? ($saeculum_now - 100) + $y : $saeculum_now + $y;
        }
        $datestr = trim(
                sprintf("%'.04d", $y) . "-" . sprintf("%'.02d", intval($parts[$s[1]])) . "-" .
                         sprintf("%'.02d", intval($parts[$s[2]])));
        return $datestr;
    }

    /**
     * Cleanse a time string to HH:MM:SS format. Milliseconds are dropped.
     * 
     * @param String $time_string
     *            the String to be converted.
     * @param bool $no_hours
     *            set true to convert xx:xx is 00:MM:SS instead of HH:SS:00. xxx:xx is anyway converted to
     *            HHH:MM:00
     * @return String|bool the formatted time. On errors false is returned and "NULL" for $time_string ==
     *         "NULL" (not case sensitive).
     */
    private static function cleanse_time (String $time_string, bool $no_hours = false)
    {
        // split off the "minus", if existing.
        $sign = "";
        if (substr($time_string, 0, 1) == "-") {
            $time_string = trim(substr($time_string, 1));
            $sign = "-";
        }
        // cleanse the remainder
        $hms = explode(":", $time_string);
        if ((count($hms) < 2) || (count($hms) > 3))
            return false;
        $hms0 = sprintf("%'.02d", intval($hms[0]));
        $hms1 = sprintf("%'.02d", intval($hms[1]));
        if (count($hms) == 2)
            return ($no_hours) ? $sign . "00:" . $hms0 . ":" . $hms1 : $sign . $hms0 . ":" . $hms1 . ":00";
        $hms2 = sprintf("%'.02d", intval($hms[2]));
        return $sign . $hms0 . ":" . $hms1 . ":" . $hms2;
    }

    /**
     * Cleanse a datetime string to YYYY-MM-DD HH:MM:SS format. Milliseconds are dropped. If no date is given,
     * the current date is inserted. If no time is given, the current time is inserted.
     * 
     * @param String $datetime_string
     *            the String to be converted.
     * @return String datetime String (Y-m-d H:i:s). On errors false is returned and "NULL" for
     *         $datetime_string == "NULL" (not case sensitive).
     */
    private static function cleanse_datetime (String $datetime_string)
    {
        if (strcasecmp($datetime_string, "NULL") == 0)
            return "NULL";
        $dt = explode(" ", trim($datetime_string));
        if (count($dt) == 1) {
            // try both, date or time
            $date = self::cleanse_date($dt[0]);
            $time = self::cleanse_time($dt[0], false); // always with hours
            if ($date !== false)
                return $date . " 00:00:00";
            elseif ($time !== false)
                return date("Y-m-d") . " " . $time;
            else {
                self::add_finding(1, trim($datetime_string));
                return false;
            }
        } else {
            $date = self::cleanse_date($dt[0]);
            $time = self::cleanse_time($dt[1], false); // always with hours
            if (($date === false) || ($time === false)) {
                self::add_finding(1, trim($datetime_string));
                return false;
            }
            return $date . " " . $time;
        }
    }

    /* ------------------------------------------------------------------------ */
    /* ----- DATA FORMATTING -------------------------------------------------- */
    /* ----- No errors are documented ----------------------------------------- */
    /* ------------------------------------------------------------------------ */
    
    /**
     * Format a boolean value for storage in files and the data base.
     * 
     * @param float $date
     *            the floating point number to format
     * @param String $language_code
     *            the language to use for formatting. Default is "csv" which represents the formatting used by
     *            the data base.
     * @return String the formatted date. "NULL" for $float_or_double == null.
     */
    private static function format_boolean ($bool = null, String $language_code = "csv")
    {
        if (in_array($language_code, self::$write_boolean_as_1_0_for))
            return ($bool) ? "1" : "0";
        if (in_array($language_code, self::$write_boolean_as_true_false_for))
            return ($bool) ? "true" : "false";
        return ($bool) ? "on" : "";
    }

    /**
     * Format a float or double value for storage in files and the data base.
     * 
     * @param float $date
     *            the floating point number to format
     * @param String $language_code
     *            the language to use for formatting. Default is "csv" which represents the formatting used by
     *            the data base.
     * @return String the formatted date. "NULL" for $float_or_double == null.
     */
    private static function format_int ($int = null, String $language_code = "csv")
    {
        return strval($int);
    }

    /**
     * Format a float or double value for storage in files and the data base.
     * 
     * @param float $float
     *            the floating point number to format
     * @param String $language_code
     *            the language to use for formatting. Default is "csv" which represents the formatting used by
     *            the data base.
     * @return String the formatted date. "NULL" for $float_or_double == null.
     */
    private static function format_float ($float = null, String $language_code = "csv")
    {
        $number_string = strval($float);
        if (in_array($language_code, self::$decimal_point_for))
            return $number_string;
        $number_string = str_replace(".", ",", $number_string);
        return $number_string;
    }

    /**
     * Format a date value for storage in files and the data base.
     * 
     * @param DateTimeImmutable $date
     *            the date to format
     * @param String $language_code
     *            the language to use for formatting. Default is "csv" which represents the formatting used by
     *            the data base.
     * @return String the formatted date. "NULL" for $date == null.
     */
    private static function format_date (DateTimeImmutable $date = null, String $language_code = "csv")
    {
        return $date->format(self::$date_formats[$language_code]["date"]);
    }

    /**
     * Convert a time int to HH:MM:SS format for values -360000 .. 360000 (-100 .. 100 hours). Convert to date
     * and time for values > 360000. CAUTION: this is different from Tfyh_toolbox::format_time()!
     * 
     * @param int $time_int
     *            the integer number to be converted.
     * @param String $language_code
     *            the language to use for formatting. Default is "csv" which represents the formatting used by
     *            the data base.
     * @return String the formatted time. "NULL" for $time_int == null.
     */
    private static function format_time (int $time_int = null, String $language_code = "csv")
    {
        $sign = ($time_int < 0) ? "-" : "";
        $sign_int = ($time_int < 0) ? - 1 : 1;
        $time_int = abs($time_int);
        if ($time_int < 360000) {
            $s = $time_int % 60;
            $m = (($time_int - $s) / 60) % 60;
            $h = (($time_int - $m * 60 - $s) / 3600);
            return $sign . sprintf("%'.02d", $h) . ":" . sprintf("%'.02d", $m) . ":" . sprintf("%'.02d", $s);
        } else {
            if ($time_int > 2147483647)
                $time_int = 2147483647;
            return date("Y-m-d H:i:s", $sign_int * $time_int);
        }
    }

    /**
     * Format a datetime value for storage in files and the data base.
     * 
     * @param DateTimeImmutable $datetime
     *            the date to format
     * @param String $language_code
     *            the language to use for formatting. Default is "csv" which represents the formatting used by
     *            the data base.
     * @return String the formatted datetime. "NULL" for $datetime == null.
     */
    private static function format_datetime (DateTimeImmutable $datetime = null, String $language_code = "csv")
    {
        return $datetime->format(self::$date_formats[$language_code]["datetime"]);
    }

    /**
     * Format a value for storage in files and the data base.
     * 
     * @param mixed $value
     *            the value to format as mixed type: PHP bool for dataType boolean, int for int & time, float
     *            for float, DateTimeImmutable for date & datetime, string for all other
     * @param String $type
     *            the type to use for formatting
     * @param String $language_code
     *            the language to use for formatting. Default is "csv" which represents the formatting used by
     *            the data base.
     * @return String the formatted value. For $value == null, an empty string is returned or "NULL" depending
     *         on the language code. For boolean "true"/"false" and "on"/"" is returned depending on the
     *         language code
     */
    private static function format_single ($value, String $type, String $language_code = "csv")
    {
        if (is_null($value))
            return in_array($language_code, self::$write_null_as_NULL_for) ? "NULL" : "";
        // the following if / elseif / else sequence is ordered in the guessed
        // sequence of occurrence frequency for performance reasons
        $format_value_as = self::parse_as($type);
        if (strcasecmp($format_value_as, "none") == 0)
            $formatted = ""; // no display of these values
        elseif (strcasecmp($format_value_as, "string") == 0)
            // $value is already a string.
            $formatted = strval($value);
        elseif (strcasecmp($format_value_as, "int") == 0)
            $formatted = self::format_int($value, $language_code);
        elseif (strcasecmp($format_value_as, "float") == 0)
            $formatted = self::format_float($value, $language_code);
        elseif (strcasecmp($format_value_as, "boolean") == 0)
            $formatted = self::format_boolean($value, $language_code);
        elseif (strcasecmp($format_value_as, "date") == 0)
            $formatted = self::format_date($value, $language_code);
        elseif (strcasecmp($format_value_as, "datetime") == 0)
            $formatted = self::format_datetime($value, $language_code);
        elseif (strcasecmp($format_value_as, "time") == 0)
            $formatted = self::format_time($value);
        else
            $formatted = strval($value);
        return $formatted;
    }

    /**
     * Format a value for storage in files and the data base. Be aware that arrays will always be formatted
     * into a list type String, i. e. a date array will not conform to the MySQL date parser expectation, only
     * its members.
     * 
     * @param mixed $value
     *            the value to format as String
     * @param String $type
     *            the type to use for formatting. Set null to guess the type.
     * @param String $language_code
     *            the language to use for formatting. Default is "csv" which represents the formatting used by
     *            the data base.
     * @return String the formatted value. "NULL" for $value == null.
     */
    public static function format ($value = null, String $type = null, String $language_code = "csv")
    {
        if (is_null($value))
            return in_array($language_code, self::$write_null_as_NULL_for) ? "NULL" : "";
        if (is_null($type))
            $type = self::guess_type($value);
        
        if (is_array($value)) {
            $out = "[";
            foreach ($value as $fsingle) {
                $fsingle = self::format_single($fsingle, $type);
                if ((strpos($fsingle, "\n") !== false) || (strpos($fsingle, ",") !== false) ||
                         (strpos($fsingle, "\"") !== false))
                    $fsingle = "\"" . str_replace("\"", "\"\"", $fsingle) . "\"";
                $out .= $fsingle . ",";
            }
            $len = mb_strlen($out);
            if ($len > 1)
                $out = mb_substr($out, 0, $len - 1);
            $out .= "]";
            return $out;
        } else {
            return self::format_single($value, $type, $language_code);
        }
    }

    /* ------------------------------------------------------------------------ */
    /* ----- SPECIAL FORMATTING ----------------------------------------------- */
    /* ----- No errors are documented ----------------------------------------- */
    /* ------------------------------------------------------------------------ */
    
    /**
     * Convert a String into an Identifier by replacing forbidden characters by an underscore and cutting the
     * length to 64 characters maximum.
     * 
     * @param String $str
     *            the String to convert.
     * @return string the vonverted String
     */
    public static function to_identifier (String $str)
    {
        $identifier = "";
        $firstAllowed = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_";
        $subsequentAllowed = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_0123456789";
        $first = mb_substr($str, 0, 1);
        if (strpos($firstAllowed, $first) === false)
            $identifier .= "_";
        for ($i = 0; ($i < mb_strlen($str)) && (strlen($identifier) < 64); $i ++) {
            $c = mb_substr($str, $i, 1);
            $identifier .= ((strpos($subsequentAllowed, $c) === false)) ? (($c == " ") ? "_" : "") : $c;
        }
        return $identifier;
    }
}
