<?php
/**  
 *
 *       the tools-for-your-hobby framework
 *       ----------------------------------
 *       https://www.tfyh.org
 *
 * Copyright  2023-2024  Martin Glade
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

include_once "../classes/tfyh_data.php";

/**
 * A utility class to hold a configuration item
 */
class Tfyh_data_item
{

    /**
     * the properties for a form layout line to be read from the "input" part of the descriptor
     */
    private static $form_input_properties = ["label","type","class","size","maxlength"
    ];

    /**
     * A shortcut list for the form table tags.
     * 
     * @var array
     */
    private static $html_input_style = ["%r" => "<div class='w3-row'>","%c1" => "<div class='w3-col l1'>",
            "%c2" => "<div class='w3-col l2'>","%c3" => "<div class='w3-col l3'>",
            "%c4" => "<div class='w3-col l4'>","%c6" => "<div class='w3-col l6'>","%/" => "</div>"
    ];

    /**
     * The configuration item descriptor.
     */
    private $uid;

    /**
     * The configuration item descriptor.
     */
    private $name;

    /**
     * The data item descriptor.
     */
    private $descriptor;

    /**
     * Construct a data item.
     * 
     * @param String $name
     *            the name of the item
     * @param String $type
     *            the type of the item
     * @param String $uid
     *            the uid of the item. Omit or set null to generate a new one.
     * @param String $text_local_name
     *            the local name of the item, i. e. the name which is used for display. Set null for default.
     */
    public function __construct (String $name, String $type, String $uid = null, String $text_local_name = null)
    {
        $this->name = $name;
        $this->uid = (is_null($uid)) ? Tfyh_toolbox::create_uid(6) : $uid;
        if (! isset(Tfyh_data::$type_templates[$type]))
            $this->descriptor = Tfyh_data::get_descriptor_template("unknown");
        else
            $this->descriptor = Tfyh_data::get_descriptor_template($type);
        if (! is_null($text_local_name))
            $this->descriptor["text_local_name"] = $text_local_name;
    }

    /**
     * Get a clone of this data item with a selectable name and always a new uid.
     * 
     * @param String $name_of_clone
     *            the name of the cloned data item.
     * @return Tfyh_data_item the data item's clone
     */
    public function get_clone (String $name_of_clone)
    {
        // create the smallest new item with a new uid
        $clone = new Tfyh_data_item($name_of_clone, "none");
        $clone->descriptor = Tfyh_data::clone_descriptor($this->descriptor);
        return $clone;
    }

    /**
     * Get a descriptor field.
     * 
     * @param String $name
     *            the name of the field. If this field is not set, null is returned.
     * @return array the descriptor
     */
    public function get_descriptor_field (String $name)
    {
        return (isset($this->descriptor[$name])) ? $this->descriptor[$name] : null;
    }

    /**
     * Get a clone of the descriptor.
     * 
     * @return array the descriptor
     */
    public function get_descriptor_clone ()
    {
        return Tfyh_data::clone_descriptor($this->descriptor);
    }

    /**
     * Get a json-String of the descriptor.
     * 
     * @return String the json-String of the descriptor
     */
    public function get_descriptor_json ()
    {
        return json_encode($this->descriptor);
    }

    /**
     * Get the descriptor, but without the default values.
     * 
     * @return array the descriptor
     */
    public function get_descriptor_without_defaults ()
    {
        return Tfyh_data::clone_without_defaults($this->descriptor);
    }

    /**
     * Parse and set the descriptor values. Starts always with the defaults for the data type and overwrites
     * any default by the given descriptor information. The previous descriptor, if there was one, is dropped.
     * 
     * @param array $descriptor_record
     *            the descriptor as a record of strings to be parsed.
     */
    public function parse_descriptor (array $descriptor_record)
    {
        $this->descriptor = Tfyh_data::parse_descriptor($descriptor_record);
        if (array_key_exists("name", $descriptor_record))
            $this->uid = $descriptor_record["name"];
        if (array_key_exists("uid", $descriptor_record))
            $this->uid = $descriptor_record["uid"];
    }

    /**
     *
     * @return String the name of this item.
     */
    public function get_name ()
    {
        return $this->name;
    }

    /**
     * Simple setter
     */
    public function set_name (String $name)
    {
        $this->name = $name;
    }

    /**
     *
     * @return String the uid of this item.
     */
    public function get_uid ()
    {
        return $this->uid;
    }

    /**
     * Simple setter. The uid is always created anew upon item instantiation to ensure, that it is never
     * empty. Use the setter whgen reading the item from a file.
     */
    public function set_uid (String $uid)
    {
        $this->uid = $uid;
    }

    /**
     *
     * @return mixed the value in the native type.
     */
    public function get_value ()
    {
        return (isset($this->descriptor["value_current"])) ? $this->descriptor["value_current"] : null;
    }

    /**
     *
     * @return String the data type.
     */
    public function get_type ()
    {
        return (isset($this->descriptor["value_type"])) ? $this->descriptor["value_type"] : null;
    }

    /**
     *
     * @return mixed the lower limit as integer or float.
     */
    public function get_min ()
    {
        return (isset($this->descriptor["value_min"])) ? $this->descriptor["value_min"] : null;
    }

    /**
     *
     * @return mixed the upper limit as integer, float or DateTimeImmutable; for types which are parsed as
     *         string the maximum size.
     */
    public function get_max ()
    {
        if (strcmp(Tfyh_data::parse_as($this->descriptor["value_type"]), "string") != 0)
            return (isset($this->descriptor["value_max"])) ? $this->descriptor["value_max"] : null;
        return (isset($this->descriptor["value_size"])) ? $this->descriptor["value_size"] : null;
    }

    /**
     * Simple setter. Only set provided fields and keeps the rest. Writes empty values to non-empty ones. Set
     * $clear_with_empty = false to keep the existing values in the case the descriptor's value is empty.
     * 
     * @param array $descriptor
     *            the descriptor as an array of parsed values
     */
    public function set_descriptor (array $descriptor)
    {
        // clone to unlink arrays and objects.
        $descriptor_clone = Tfyh_data::clone_descriptor($descriptor);
        foreach ($descriptor_clone as $name => $value)
            $this->descriptor[$name] = $value;
    }

    /**
     * Simply replaces the current descriptor by the new one. $descriptor is taken as is, use a cloen for it.
     * 
     * @param array $descriptor            
     */
    public function replace_descriptor (array $descriptor)
    {
        $this->descriptor = $descriptor;
    }

    /**
     * Get the difference of two configuration branches. The difference is as well a branch
     * 
     * @param Tfyh_config_item $item_to_compare
     *            the root item of the branch to compare
     * @param bool $exclude_current_value
     *            set true to exclude the current value from the comparison.
     * @param bool $exclude_defaults
     *            set true to exclude the fields which have the default value of the data type
     * @return String a collection of all differences.
     */
    public function diff (Tfyh_data_item $item_to_compare, bool $exclude_current_value, bool $exclude_defaults)
    {
        $diff = $this->get_name() . " ?= " . $item_to_compare->get_name() . ": ";
        if (strcmp($this->get_name(), $item_to_compare->get_name()) != 0)
            $diff .= i("2pnXeO|Name A not equal to name...") . " ";
        $diff .= $this->diff_descriptor($item_to_compare, $exclude_current_value, $exclude_defaults);
    }

    /**
     * Check whether the $item_to_compare has a descriptor different from the own. Defaults are ignored, only
     * those values are compared which are different from default.
     * 
     * @param array $item_to_compare
     *            the descriptor to compare with the descriptor of $this
     * @param bool $exclude_current_value
     *            set true to exclude the current value from the comparison.
     * @param bool $exclude_defaults
     *            set true to exclude the fields which have the default value of the data type
     * @return String a collection of differences. An empty String, if both are equal.
     */
    public function diff_descriptor (Tfyh_data_item $item_to_compare, bool $exclude_current_value, 
            bool $exclude_defaults)
    {
        $this_to_compare = ($exclude_defaults) ? Tfyh_data::clone_without_defaults($this->descriptor) : Tfyh_data::clone_descriptor(
                $this->descriptor);
        $other_to_compare = ($exclude_defaults) ? Tfyh_data::clone_without_defaults(
                $item_to_compare->descriptor) : Tfyh_data::clone_descriptor($item_to_compare->descriptor);
        if ($exclude_current_value) {
            unset($this_to_compare["value_current"]);
            unset($other_to_compare["value_current"]);
        }
        return Tfyh_data::diff_arrays($this_to_compare, $other_to_compare);
    }
}
