<?php
/**  
 *
 *       the tools-for-your-hobby framework
 *       ----------------------------------
 *       https://www.tfyh.org
 *
 * Copyright  2023-2024  Martin Glade
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */
include_once "../classes/tfyh_socket.php";

/**
 * Class file to adjust the data base layout to the version required by the current code.
 */
class Tfyh_db_setup implements Tfyh_socket_trigger, Tfyh_socket_read_listener
{

    /**
     * Log path for specific data base audit logging
     */
    private static $db_audit_log = "../log/sys_db_audit.log";

    /**
     * The sql command to add missing table columns (way #2. Start with adding missing columns, adjust null to
     * 0 and end with adjusting the collumn types.).
     */
    public static $sql_add_column_command = "ALTER TABLE `{table}` ADD `{column}` {type} {null} {default};";

    /**
     * A statment to change a field from nullable to not nullable, providing a default replacing all existing
     * NULL values.
     */
    public static $sql_column_null_to_default_adjustment = "UPDATE `{table}` SET `{column}` = '{default}' WHERE ISNULL(`{column}`);";

    /**
     * The commands to adjust a specific column to use the correct layout. (way #2. Start with adding missing
     * columns, adjust null to 0 and end with adjusting the collumn types.)
     */
    public static $sql_change_column_command = "ALTER TABLE `{table}` CHANGE `{column}` `{column}` {type} {null} {default};";

    /**
     * The data base permissions cache, to be read from the .appTables configuration. Fields will be filled on
     * demand.
     */
    private static $column_permissions = [];

    private static $record_permissions = [];

    /**
     * The data base connection socket.
     */
    private $socket;

    /**
     * The generic toolbox.
     */
    private $toolbox;

    /**
     * Read the permissions for a specific column from the cache or from the configuration and fill the cache.
     * 
     * @param String $tablename
     *            the table to get the configuration for
     * @param String $columnname
     *            the column to get the configuration for
     * @return int the permissions mask
     */
    private function get_column_permissions (String $tablename, String $columnname)
    {
        // read cache, if existing
        if (! isset(self::$column_permissions[$tablename]))
            self::$column_permissions[$tablename] = [];
        elseif (isset(self::$column_permissions[$tablename][$columnname]))
            return self::$column_permissions[$tablename][$columnname];
        // if not, fill cache
        $column_cfg_root = $this->toolbox->config->get_by_path(".appTables.$tablename.$columnname");
        self::$column_permissions[$tablename][$columnname] = $column_cfg_root->get_descriptor_field(
                "node_permissions");
        return self::$column_permissions[$tablename][$columnname];
    }

    /**
     * Read the maximum permissions for a table record from the cache or from the configuration and fill the
     * cache.
     * 
     * @param String $tablename
     *            the table to get the configuration for
     * @return int the permissions mask
     */
    private function get_record_permissions (String $tablename)
    {
        // read cache, if existing
        if (isset(self::$record_permissions[$tablename]))
            return self::$record_permissions[$tablename];
        
        // if not, fill cache
        self::$record_permissions[$tablename] = 0;
        $table_cfg_root = $this->toolbox->config->get_by_path(".appTables.$tablename");
        if (! is_null($table_cfg_root))
            return 0;
        $rpermissions = 0;
        foreach ($table_cfg_root->get_children() as $cname => $cdefinition) {
            if (substr($cname, 0, 1) != "_") {
                $cpermissions = $this->get_column_permissions($tablename, $cname);
                $rpermissions = $rpermissions | $cpermissions;
            }
        }
        self::$record_permissions[$tablename] = $rpermissions;
        return $rpermissions;
    }

    /**
     * check, whether the $permission String allows the request access
     * 
     * @param String $tablename
     *            the name of the table
     * @param String $columnname
     *            the name of the column
     * @param int $access
     *            4 = read, 2 = update, 1 = insert
     * @param int $group
     *            4 = admin, 3 = referred, 2 = conceded, 1 = other, 0 = public
     * @return boolean true, if the access is permitted.
     */
    public function is_allowed_field (String $tablename, String $columnname, int $access, int $group)
    {
        $mask = $access << ($group * 3);
        $permissions = $this->get_column_permissions($tablename, $columnname);
        return ($permissions & $mask) > 0;
    }

    /**
     * check, whether the $permission String allows the request access for any of the fields of the record.
     * 
     * @param String $tablename
     *            the name of the table
     * @param String $column_name
     *            the name of the column
     * @param int $access
     *            4 = read, 2 = update, 1 = insert
     * @param int $group
     *            4 = admin, 3 = referred, 2 = conceded, 1 = other, 0 = public
     * @return boolean true, if the access is permitted.
     */
    public function is_allowed_record (String $tablename, int $access, int $group)
    {
        $mask = $access << ($group * 3);
        $permissions = $this->get_record_permissions($tablename);
        return ($permissions & $mask) > 0;
    }

    /**
     * public Constructor.
     * 
     * @param Tfyh_toolbox $toolbox
     *            application toolbox
     * @param Tfyh_socket $socket
     *            the socket to connect to the database
     */
    public function __construct (Tfyh_toolbox $toolbox, Tfyh_socket $socket)
    {
        $this->toolbox = $toolbox;
        $this->socket = $socket;
        $tfyh_cfg = $toolbox->config->settings_tfyh["config"];
    }

    /**
     *
     * {@inheritdoc}
     *
     * @see Tfyh_socket_read_listener::post_read_transaction()
     */
    public function post_read_transaction (String $tablename, array $records)
    {
        $no_changes = true;
        $collected = [];
        foreach ($records as $record) {
            // TODO: find the users rights, may be by role (admin), by referral or by concession
        }
        return $no_changes;
    }

    /**
     *
     * {@inheritdoc}
     *
     * @see Tfyh_socket_trigger::pre_write_transaction()
     */
    public function pre_write_transaction (int $user_id, int $mode, String $tablename, array $record)
    {
        // TODO: permissions check.
        return $this->add_trigger_fields($user_id, $mode, $tablename, $record);
    }

    /**
     * Resolve a reference value into the correct record or value list for ~cfg-resolution.
     * 
     * @param String $tablename
     *            the table name of the record for which a field value shall be resolved
     * @param String $columnname
     *            the field name of the value that shall be resolved
     * @param String $value_str
     *            the value that shall be resolved, as String for an SQL-where statement
     * @return mixed the configuration item for a configuration search and the record for a table search.
     */
    public function resolve (String $tablename, String $columnname, String $value_str)
    {
        $column_cfg_root = $this->toolbox->config->get_by_path(".appTables.$tablename.$columnname");
        if (is_null($column_cfg_root))
            return null; // no column configuration
        $column_reference = $column_cfg_root->get_descriptor_field("column_reference");
        if (is_null($column_reference) || (strlen($column_reference) == 0))
            return null; // no reference to other object
        if (strpos($column_reference, ".") === 0) {
            // if the reference starts with a "." it refers to a configuration item.
            return $this->toolbox->config->get_by_path($column_reference . "." . $value_str);
        } else {
            // else it refers to a table record's field.
            $reference = explode(".", $column_reference);
            return $this->socket->find_record($reference[0], $reference[1], $value_str);
        }
    }

    /**
     * Convert a record into a json string, omitting the history field, if it exists. Cuts the text down to
     * less than 64k characters.
     * 
     * @param array $record
     *            the record to be stringyfied
     * @param String $history_field
     *            the history field to exclude. Set "", if the table has no history field.
     * @return string the stringyfied record
     */
    private static function record_to_text (array $record, String $history_field)
    {
        // remove record history
        if ((strlen($history_field) > 0) &&
                 (isset($record[$history_field]) || is_null($record[$history_field])))
            unset($record[$history_field]);
        $record_to_text = json_encode($record);
        // limit size to 64k
        $cut_len = 65535 - 4096;
        while (strlen($record_to_text) > 65535) { // strlen == byte length
            foreach ($record as $key => $value)
                if (strlen($value) > $cut_len)
                    $record[$key] = substr(strval($record[$key]), 0, $cut_len);
            $archive_entry = json_encode($record);
            $cut_len = $cut_len - 4096;
        }
        return $record_to_text;
    }

    /**
     * Check whether a person belongs to the owners of a record. This includes the lookup of the user_id
     * 
     * @param array $record
     *            the record for which the ownership shall be checked
     * @param String $tablename
     *            the table to which the recird belongs
     * @param array $owner_ids
     *            an array of ids to indentify the owner, may be a user id and a uuid
     * @param Tfyh_toolbox $toolbox
     *            the generic toolbox. It has to be passed for list splitting
     * @return boolean
     */
    public function owns (array $record, String $tablename, array $owner_ids, Tfyh_toolbox $toolbox)
    {
        $table_cfg_root = $this->toolbox->config->get_by_path(".appTables.$tablename");
        foreach ($table_cfg_root->get_children() as $cname => $cdefinition) {
            if (substr($cname, 0, 1) != "_") {
                // iterate through all table columns to look for owner fields
                $handling = $cdefinition->get_descriptor_field("column_handling");
                if (strpos($handling, "o") !== false) {
                    // check, whether the respective owner field has any value at all
                    if (isset($record[$cname]) && (strlen($record[$cname]) > 0)) {
                        // may be a list (e.g. a crew). Split list or convert entry to array
                        $is_list = (strpos($handling, "l") !== false);
                        $owner_field_values = ($is_list) ? $toolbox->read_csv_line($record[$cname], "|") : array(
                                $record[$cname]
                        );
                        // now check whether any of these entries matches the possible owner Ids
                        foreach ($owner_field_values as $owner_field_value)
                            if (in_array($owner_field_value, $owner_ids))
                                // it is sufficient to find one match
                                return true;
                    }
                }
            }
        }
        // no match found
        return false;
    }

    /**
     * add all system fields to the record. The "owner" and the "created" types will only be added for insert.
     * Set the system fields in the settings_tfyh file, in the module record & db_layout; system fields are
     * type=name, e.g uuid=Id records.uid=uid.
     * 
     * @param int $user_id
     *            the user of the transaction. Will be added as owner field.
     * @param int $mode
     *            the mode of the transaction, being either of 1 for insert, 2 for update or 3 for delete.
     * @param String $tablename
     *            the tablen to whicht hte record belongs
     * @param array $record
     *            the record to which the system fields shall be added
     * @return array|bool the record, if modified, else true.
     */
    private function add_trigger_fields (int $user_id, int $mode, String $tablename, array $record)
    {
        $changed = false;
        $table_cfg_root = $this->toolbox->config->get_by_path(".appTables.$tablename");
        foreach ($table_cfg_root->get_children() as $cname => $cdefinition) {
            if (substr($cname, 0, 1) != "_") {
                // Iterate through all columns for an app trigger. ignore the "socket-generated"
                // values, because the Tfyh_socket will take care for the change log and record history.
                $event = $cdefinition->get_descriptor_field("sql_trigger_event");
                $trigger_value = $cdefinition->get_descriptor_field("sql_trigger_value");
                $is_update_trigger = (strcasecmp($event, "app_update") == 0);
                $is_insert_trigger = (strcasecmp($event, "app_insert") == 0);
                $is_update = ($mode == 2);
                $is_insert = ($mode == 1);
                $is_triggered = ($is_insert_trigger && $is_insert) ||
                         ($is_update_trigger && ($is_insert || $is_update));
                $ignore = (strcmp($trigger_value, "auto-generated") == 0);
                if ($is_triggered && ! $ignore) {
                    $value_to_use = "";
                    if (strcmp($trigger_value, "user_id") == 0)
                        $value_to_use = $user_id;
                    elseif (strcmp($trigger_value, "microtime") == 0)
                        $value_to_use = Tfyh_toolbox::timef();
                    // use value for insert trigger, if insert, or for update trigger if insert or update
                    if (($is_insert_trigger && ($mode == 1)) || ($is_update_trigger && ($mode <= 2))) {
                        $changed = true;
                        $record[$cname] = $value_to_use;
                    }
                }
            }
        }
        return ($changed) ? $record : true;
    }

    /**
     * insert the default admin record for user id 1142. The user table must be empty for that. If not, the
     * function will return without doing anything.
     */
    private function insert_admin_record (array $admin_record)
    {
        $user_records_cnt = $this->socket->count_records($this->toolbox->users->user_table_name);
        if ($user_records_cnt > 0) {
            $error_message = i("N2Co3F|Failed to add default ad...", $this->toolbox->users->user_table_name);
            file_put_contents(self::$db_audit_log, date("Y-m-d H:i:s") . ": " . $error_message . "\n", 
                    FILE_APPEND);
            return $error_message;
        }
        $admin_record["Rolle"] = $this->toolbox->users->useradmin_role;
        $success = $this->socket->insert_into($admin_record[$this->toolbox->users->user_id_field_name], 
                $this->toolbox->users->user_table_name, $admin_record);
        if (is_numeric($success)) {
            $success_message = i("kx8iiD|Added admin record.");
            file_put_contents(self::$db_audit_log, 
                    date("Y-m-d H:i:s") . ": " . i("tNKKVB|Added admin record.") . "\n", FILE_APPEND);
            return $success_message;
        } else {
            $error_message = i("7vRDMJ|Failed to add default ad...") . $success;
            file_put_contents(self::$db_audit_log, date("Y-m-d H:i:s") . ": " . $error_message . "\n", 
                    FILE_APPEND);
            return $error_message;
        }
    }

    /**
     * Build the sql command based on the definition here and the template like self::$sql_add_column_command
     * 
     * @param String $tablename
     *            table to use
     * @param String $coluumnname
     *            column to use
     * @param String $template
     *            template to use
     */
    private function build_sql_column_command (String $tablename, String $columnname, String $template)
    {
        $sql_cmd = str_replace("{table}", $tablename, str_replace("{column}", $columnname, $template));
        
        $cdefinition = $this->toolbox->config->get_by_path(".appTables.$tablename.$columnname");
        $type = ($cdefinition->get_descriptor_field("sql_type"));
        $size = ($cdefinition->get_descriptor_field("sql_size"));
        if (intval($size) > 0)
            $type .= "(" . $size . ")";
        $null = ($cdefinition->get_descriptor_field("sql_null")) ? "NULL" : "NOT NULL";
        $default = $cdefinition->get_descriptor_field("sql_default");
        // when reading the table definition, "NULL" will be converted to native null. Revert this.
        if (is_null($default))
            $default = "NULL";
        // notice special cases of fields without default value (e.g. autoincrement fields, first & last name
        // of user) or of fields using a trigger rather than a default for bigint unix timestamp.
        $default_str = "DEFAULT '$default'";
        if (strcasecmp($default, "(kein)") == 0)
            $default_str = "";
        elseif ((strcasecmp($default, "NULL") == 0) || (strcasecmp($default, "CURRENT_TIMESTAMP") == 0))
            $default_str = "DEFAULT $default";
        else
            $default_str = "DEFAULT '$default'";
        // BLOB, TEXT, GEOMETRY or JSON columns can't have a default value
        if (strcasecmp($type, "text") == 0)
            $default_str = "";
        return str_replace("{type}", $type, 
                str_replace("{null}", $null, str_replace("{default}", $default_str, $sql_cmd)));
    }

    /**
     * Build the sql command set (multiple commands) to create a table: DROP TABLE, CREATE TABLE (with all
     * columns), ALTER ... ADD UNIQUE, and ALTER ... MODIFY ... AUTO_INCREMENT
     * 
     * @param String $tablename
     *            table to use
     */
    private function build_sql_add_table_commands (String $tablename)
    {
        // create an array of SQL commands to create the table.
        $sql_cmds = [];
        
        // build the create table statement with all columns
        $sql_cmds[0] = "CREATE TABLE `" . $tablename . "` ( ";
        $table_cfg_root = $this->toolbox->config->get_by_path(".appTables.$tablename");
        $uniques = "";
        $autoincrements = "";
        $i = 1;
        foreach ($table_cfg_root->get_children() as $cname => $cdefinition) {
            if (substr($cname, 0, 1) != "_") {
                $col_cmd = "`" . $cname . "` {type} {null} {default}, ";
                $sql_cmds[0] .= self::build_sql_column_command($tablename, $cname, $col_cmd);
                // build all the needed add unique statements on the way.
                $sql_indexed = $cdefinition->get_descriptor_field("sql_indexed");
                if (strpos($sql_indexed, "u") !== false) {
                    $sql_cmds[$i] = "ALTER TABLE `" . $tablename . "` ADD UNIQUE(`" . $cname . "`)";
                    $i ++;
                }
                // build the autoincrement statements on the way.
                if (strpos($sql_indexed, "a") !== false) {
                    $sql_cmds[$i] = "ALTER TABLE `" . $tablename . "` MODIFY `" . $cname .
                             "` INT UNSIGNED NOT NULL AUTO_INCREMENT";
                    $i ++;
                }
            }
        }
        // close the columns list of the create table statement
        $sql_cmds[0] = mb_substr($sql_cmds[0], 0, mb_strlen($sql_cmds[0]) - 2) . " )";
        return $sql_cmds;
    }

    /**
     * Execute and log an $sql_cmd. THIS MUST ONLY BE CALLED, IF THE USER HAS BEEN PROVED BEING AN ADMIN USER,
     * BECAUSE IT CIRCUMVENTS ALL PERMISSION CHECKS AND APPLICATION DATA HANDLING TRIGGERS.
     * 
     * @param String $appUserID
     *            the user issueing the command
     * @param String $sql_cmd
     *            the command issued
     * @param String $log_message
     *            a message to include in the log
     * @return boolean the actiities success
     */
    private function execute_and_log (String $appUserID, String $sql_cmd, String $log_message)
    {
        // this circumvents any permission check, because it is only called by an admin user.
        $success = $this->socket->query($sql_cmd, $this);
        if ($success === false) {
            $fail_message = i("sHmiya|Failed data base stateme...", $appUserID, $log_message, 
                    $this->socket->get_last_mysqli_error(), $sql_cmd);
            file_put_contents(self::$db_audit_log, date("Y-m-d H:i:s") . ": $fail_message.\n", FILE_APPEND);
            return false;
        } else {
            $success_message = i("sBBcUy|Executed data base state...", $appUserID, $log_message);
            file_put_contents(self::$db_audit_log, date("Y-m-d H:i:s") . ": $success_message.\n", FILE_APPEND);
            return true;
        }
    }

    /**
     * Delete all existing tables and build the data base from scratch. Keep the current user as single admin.
     */
    public function init_data_base ()
    {
        $result = "";
        // check user priviledges
        if (strcasecmp($this->toolbox->users->useradmin_role, $this->toolbox->users->session_user["Rolle"]) !=
                 0)
            return i("vB1jX7|Error: User does not hav...");
        $appUserID = $this->toolbox->users->session_user["@id"];
        $admin_record = $this->socket->find_record($this->toolbox->users->user_table_name, 
                $this->toolbox->users->user_id_field_name, $appUserID);
        if ($admin_record === false) {
            $admin_record = $this->toolbox->users->session_user;
            foreach ($admin_record as $key => $value)
                if (strpos($key, "@") === 0)
                    unset($admin_record[$key]);
        }
        // now reset all tables
        $this->toolbox->logger->log(0, $appUserID, 
                date("Y-m-d H:i:s") . ": " . i("mSMaVo|Starting init_data_base"));
        
        // Drop all Triggers
        $sql_cmd = "SELECT Concat('DROP TRIGGER ', Trigger_Name, ';') FROM information_schema.TRIGGERS WHERE TRIGGER_SCHEMA = '" .
                 $this->toolbox->config->get_cfg_db()["db_name"] . "';";
        
        // build all tables
        $tables_cfg_root = $this->toolbox->config->get_by_path(".appTables");
        foreach ($tables_cfg_root->get_children() as $tablename => $unused_definition) {
            $log_message = i("2p7kEh|Dropping table °%1°... ", $tablename);
            $sql_cmd = "DROP TABLE `" . $tablename . "`;";
            $drop_success = $this->execute_and_log($appUserID, $sql_cmd, $log_message);
            $result .= $log_message . (($drop_success) ? i("vI1rIC|ok.") : i("VTLYoN|no such table.")) . "<br>";
            // this will abort execution after first failure.
            $log_message = i("WRvThr|Creating empty new table...", $tablename);
            $sql_cmds = $this->build_sql_add_table_commands($tablename);
            $reset_success = true;
            foreach ($sql_cmds as $sql_cmd)
                // this will abort execution after first failure.
                $reset_success = $reset_success && $this->execute_and_log($appUserID, $sql_cmd, $log_message);
            $result .= $log_message .
                     (($reset_success) ? i("sQuH9Y|ok.") : i("uhcKky|aborted, see sys_db_audi...")) . "<br>";
        }
        // if also the users table was dropped and rebuilt, insert now the admin user.
        $log_message = i("wwM2JD|Inserting admin %1 recor...", 
                $admin_record[$this->toolbox->users->user_id_field_name], 
                $this->toolbox->users->user_table_name) . ": ";
        $admin_insert_success = $this->insert_admin_record($admin_record);
        $result .= $log_message . $admin_insert_success;
        return $result;
    }

    /**
     * Check the current session user whether it has admin rights.
     * 
     * @param bool $log_violations
     *            log a violation, if the user has not sufficient priviledges, set false to skip logging.
     * @return boolean|array the user record. if the user has admin rights, or false, if not.
     */
    private function is_admin_session_user (bool $log_violations)
    {
        if (! isset($this->toolbox->users->session_user) || ! isset(
                $this->toolbox->users->session_user["@id"])) {
            $this->toolbox->logger->log(1, 0, i("NUSssM|Data base manipulation c..."));
            return false;
        }
        // cache the user record for re-insertion after table reset.
        $appUserID = $this->toolbox->users->session_user["@id"];
        $app_user_record = $this->socket->find_record($this->toolbox->users->user_table_name, 
                $this->toolbox->users->user_id_field_name, $appUserID);
        if ($app_user_record === false) {
            if ($log_violations)
                $this->toolbox->logger->log(1, $appUserID, 
                        i("5Ca1BA|Data base manipulation p...", $appUserID));
            return false;
        }
        if (! isset($app_user_record["Rolle"]) ||
                 (strcasecmp($app_user_record["Rolle"], $this->toolbox->users->useradmin_role) != 0)) {
            if ($log_violations)
                $this->toolbox->logger->log(1, $appUserID, 
                        i("mMnBAj|Data base manipulation p...", 
                                $app_user_record[$this->toolbox->users->user_id_field_name]));
            return false;
        }
        return $app_user_record;
    }

    /**
     * Update all tables and columns to comply to the selected data base layout.
     * 
     * @param bool $verify_only
     *            If this flag is true, nothing will be done, but just checked whether something would be
     *            done. Only exception: default value setting is not checked.
     * @return boolean false is returned for 1. for $verify_only == true and a deviation from the layout or a
     *         previous verification happened less than 10 minutes ago, 2. for $verify_only == false and
     *         failure to correct the data base. Else true is returned.
     */
    public function update_database_layout (bool $verify_only)
    {
        // check user priviledges
        $admin_record = $this->is_admin_session_user(false);
        $appUserID = $admin_record[$this->toolbox->users->user_id_field_name];
        file_put_contents(self::$db_audit_log, 
                date("Y-m-d H:i:s") . ": " . i("8hPspQ|Starting") . " update_database_layout [" .
                         json_encode($verify_only) . ", " . $appUserID . "].\n");
        if (($admin_record === false) && ! $verify_only) {
            file_put_contents(self::$db_audit_log, 
                    date("Y-m-d H:i:s") . ": " . i("a9d5U7|User is no admin and thu...") . "\n", FILE_APPEND);
            return false;
        }
        
        // now adjust all tables
        $correction_success = true;
        $verification_success = true;
        $table_names_existing = $this->socket->get_table_names();
        $lower_case_tablenames = true;
        foreach ($table_names_existing as $table_name_existing)
            if (strcmp(strtolower($table_name_existing), $table_name_existing) != 0)
                $case_sensitive_names = false;
        
        if (! $verify_only)
            $this->toolbox->logger->log(0, $appUserID, 
                    date("Y-m-d H:i:s") . ": " . i("mrY3PQ|Starting") . " update_database_layout.");
        // adjust or add tables according to the expected layout
        $tables_cfg_root = $this->toolbox->config->get_by_path(".appTables");
        $tables_of_layout = [];
        foreach ($tables_cfg_root->get_children() as $tablename => $unused_definition) {
            // cache all table names for later
            // Microsoft mySQL implementations on MS Azure use all lower case table names.
            $tables_of_layout[] = $tablename;
            $tables_of_layout[] = strtolower($tablename);
            // adjust existing table
            if (in_array($tablename, $table_names_existing)) {
                if (! $this->update_table_layout($appUserID, $tablename, $verify_only)) {
                    file_put_contents(self::$db_audit_log, 
                            date("Y-m-d H:i:s") . ": " . i("rSrSm0|Update failed for table ...", $tablename) .
                                     "\n", FILE_APPEND);
                    $correction_success = false;
                    $verification_success = false;
                }
            } else {
                // add missing table
                if ($verify_only) {
                    file_put_contents(self::$db_audit_log, 
                            date("Y-m-d H:i:s") . ": " . i("KprqM0|Verification failed. Tab...", $tablename) .
                                     "\n", FILE_APPEND);
                    $verification_success = false;
                } else {
                    $log_message = i("wpf4AI|Create table °%1° with a...", $tablename);
                    $sql_cmds = $this->build_sql_add_table_commands($tablename);
                    foreach ($sql_cmds as $sql_cmd)
                        // this will continue execution after failure.
                        $correction_success = $this->execute_and_log($appUserID, $sql_cmd, $log_message) &&
                                 $correction_success;
                }
            }
        }
        // and drop the obsolete ones
        foreach ($table_names_existing as $tablename) {
            if (! in_array($tablename, $tables_of_layout)) {
                if ($verify_only) {
                    file_put_contents(self::$db_audit_log, 
                            date("Y-m-d H:i:s") . ": " . i("9g3u60|Verification failed. Tab...", $tablename) .
                                     "\n", FILE_APPEND);
                    $verification_success = false;
                } else {
                    $sql_cmd = "DROP TABLE `" . $tablename . "`";
                    $log_message = i("TQrqJw|Drop table °%1°.", $tablename);
                    $correction_success = $this->execute_and_log($appUserID, $sql_cmd, $log_message) &&
                             $correction_success;
                }
            }
        }
        
        // notify and register activity.
        if (! $verify_only) {
            $this->toolbox->logger->log(0, $appUserID, 
                    date("Y-m-d H:i:s") . ": " . i("qiUbiD|database_layout upgraded"));
        }
        return ($verify_only) ? $verification_success : $correction_success;
    }

    /**
     * Update the layout of a single table. This way multiple layout versions can be supported. The access
     * permission of the session user must be checked by the calling function.
     * 
     * @param String $appUserID
     *            the ID of the application user of the user who performs the update. For change logging.
     * @param bool $verify_only
     *            If this flag is true, nothing will be done, but just checked whether something would be
     *            done. Only exception: default value setting is not checked.
     * @return boolean false is returned for 1. for $verify_only == true and a deviation from the layout or a
     *         previous verification happened less than 10 minutes ago, 2. for $verify_only == false and
     *         failure to correct the data base. Else true is returned.
     */
    private function update_table_layout (String $appUserID, String $tablename, bool $verify_only)
    {
        // start process and logging.
        file_put_contents(self::$db_audit_log, 
                date("Y-m-d H:i:s") . ": " . i("YPmPF9|Starting") . " update_table_layout(" . $tablename .
                         ").\n", FILE_APPEND);
        
        // read existing schema information.
        $column_names_existing = $this->socket->get_column_names($tablename);
        $column_nullables_existing = $this->socket->get_column_nullables($tablename);
        $column_type_descriptions_existing = $this->socket->get_column_types($tablename);
        $column_types_existing = [];
        $column_sizes_existing = [];
        foreach ($column_type_descriptions_existing as $cname => $column_type_description) {
            $has_size = strpos($column_type_description, "(") !== false;
            $column_types_existing[] = ($has_size) ? explode("(", $column_type_description)[0] : $column_type_description;
            $column_sizes_existing[] = ($has_size) ? intval(
                    str_replace(")", "", explode("(", $column_type_description)[1])) : 0;
        }
        $indexes_existing = $this->socket->get_indexes($tablename, false);
        $autoincrements_existing = $this->socket->get_autoincrements($tablename);
        
        // adjust or add columns according to the expected layout
        $verification_success = true;
        $correction_success = true;
        $table_cfg_root = $this->toolbox->config->get_by_path(".appTables.$tablename");
        
        // Collect what is expoected
        // =========================
        $columns_expected = [];
        $indexes_expected = [];
        foreach ($table_cfg_root->get_children() as $cname => $cdefinition) {
            if (substr($cname, 0, 1) != "_")
                $columns_expected[] = $cname;
            if (strcmp(strtolower($cname), $cname) != 0)
                // for windows environments
                $columns_expected[] = strtolower($cname);
            $is_unique_handling = (strpos($cdefinition->get_descriptor_field("column_handling"), "u") !== false);
            $is_autoinc_handling = (strpos($cdefinition->get_descriptor_field("column_handling"), "a") !==
                     false);
            if ($is_unique_handling)
                $indexes_expected[] = $cname;
            if ($is_autoinc_handling)
                $indexes_expected[] = $cname;
            $is_unique_sql_indexed = (strpos($cdefinition->get_descriptor_field("sql_indexed"), "u") !== false);
            $is_autoinc_sql_indexed = (strpos($cdefinition->get_descriptor_field("sql_indexed"), "a") !== false);
            if ($is_unique_handling != $is_unique_sql_indexed)
                file_put_contents(self::$db_audit_log, 
                        date("Y-m-d H:i:s") . ": " . i("XfN4qO|Configuration mismatch i...", $tablename) .
                                 ".\n", FILE_APPEND);
            if ($is_autoinc_handling != $is_autoinc_sql_indexed)
                file_put_contents(self::$db_audit_log, 
                        date("Y-m-d H:i:s") . ": " . i("nbJ0aO|Configuration mismatch i...", $tablename) .
                                 ".\n", FILE_APPEND);
        }
        
        // Add or modify all columns of this table
        // =======================================
        foreach ($table_cfg_root->get_children() as $cname => $cdefinition) {
            if (substr($cname, 0, 1) != "_") {
                // find the column within the set of exiting ones
                $c = array_search($cname, $column_names_existing, true);
                if ($c !== false) {
                    // get existing parameters of column
                    $ctype = trim($column_types_existing[$c]);
                    $csize = ((strcasecmp($ctype, "text") == 0) || (strcasecmp($ctype, "mediumtext") == 0)) ? 0 : intval(
                            trim($column_sizes_existing[$c]));
                    $sql_type = $cdefinition->get_descriptor_field("sql_type");
                    $sql_size = intval($cdefinition->get_descriptor_field("sql_size"));
                    
                    // column is existing. Adjust default first
                    $cnull = $column_nullables_existing[$c];
                    $sql_null = $cdefinition->get_descriptor_field("sql_null");
                    $default = $cdefinition->get_descriptor_field("sql_default");
                    $default_available = ((strcasecmp($default, "(n.r.)") != 0) &&
                             (strcasecmp($default, "(kein)") != 0));
                    
                    // null to not null adjustment
                    if (! $verify_only && ($cnull && ! $sql_null) && $default_available) {
                        $sql_cmd = self::build_sql_column_command($tablename, $cname, 
                                self::$sql_column_null_to_default_adjustment);
                        $log_message = i("LjCmY9|Set Default for °%1°.°%2...", $tablename, $cname, $default);
                        $correction_success = $this->execute_and_log($appUserID, $sql_cmd, $log_message) &&
                                 $correction_success;
                    }
                    
                    // now change the column.
                    if (! $verify_only) {
                        $sql_cmd = self::build_sql_column_command($tablename, $cname, 
                                self::$sql_change_column_command);
                        if ((strcasecmp($ctype, $sql_type) != 0) || ($csize != $sql_size) ||
                                 ($cnull != $sql_null)) {
                            $log_message = i("sNh3bq|Changed column °%1°.°%2°...", $tablename, $cname);
                            $correction_success = $this->execute_and_log($appUserID, $sql_cmd, $log_message) &&
                                     $correction_success;
                        }
                    } else {
                        // or check type, size or default, if just the verification shall happen
                        if (strcasecmp($ctype, $sql_type) != 0) {
                            file_put_contents(self::$db_audit_log, 
                                    date("Y-m-d H:i:s") . ": " . i("mwaK4O|Verification failed. Col...", 
                                            $tablename, $cname, $ctype, $sql_type) . "\n", FILE_APPEND);
                            $verification_success = false;
                        }
                        if ($csize != $sql_size) {
                            file_put_contents(self::$db_audit_log, 
                                    date("Y-m-d H:i:s") . ": " . i("3EWY5V|Verification failed. Col...", 
                                            $tablename, $cname, $csize, $sql_size) . ".\n", FILE_APPEND);
                            $verification_success = false;
                        }
                    }
                } else {
                    // column is NOT existing.
                    if ($verify_only) {
                        file_put_contents(self::$db_audit_log, 
                                date("Y-m-d H:i:s") . ": " .
                                         i("rxMN98|Verification failed. Col...", $tablename, $cname) . "\n", 
                                        FILE_APPEND);
                        $verification_success = false;
                    } else {
                        // add column
                        $activity_template = self::$sql_add_column_command;
                        $log_message = i("xw01aj|Added column °%1°.°%2° w...", $tablename, $cname);
                        $sql_cmd = self::build_sql_column_command($tablename, $cname, $activity_template);
                        $correction_success = $this->execute_and_log($appUserID, $sql_cmd, $log_message) &&
                                 $correction_success;
                    }
                }
            }
            
            // add a unique quality, if not yet existing
            if (strpos($cdefinition->get_descriptor_field("column_handling"), "u") !== false) {
                if (! in_array($cname, $indexes_existing)) {
                    // sometimes indices are not recognized by the socket. It may be there. Drop it first and
                    // restore it.
                    $sql_cmd_1 = "ALTER TABLE `" . $tablename . "` DROP INDEX `" . $cname . "`;";
                    $sql_cmd_2 = "ALTER TABLE `" . $tablename . "` ADD UNIQUE `" . $cname . "` (`" . $cname .
                             "`); ";
                    $log_message = i("S1nQio|Added unique property to...", $tablename, $cname);
                    if (! $verify_only) {
                        $correction_success = $this->execute_and_log($appUserID, $sql_cmd_1, $log_message) &&
                                 $correction_success;
                        $correction_success = $this->execute_and_log($appUserID, $sql_cmd_2, $log_message) &&
                                 $correction_success;
                    } else {
                        file_put_contents(self::$db_audit_log, 
                                date("Y-m-d H:i:s") . ": " . i("9mUNZn|Verification failed. Ind...", 
                                        $tablename, $cname) . "\n", FILE_APPEND);
                        $verification_success = false;
                    }
                }
            }
            
            // add an autoincrement quality, if needed
            // e.g. ALTER TABLE `persons` CHANGE `id` `id` INT NOT NULL AUTO_INCREMENT;
            if (strpos($cdefinition->get_descriptor_field("column_handling"), "a") !== false) {
                if (! array_key_exists($cname, $autoincrements_existing)) {
                    $sql_cmd = "ALTER TABLE `" . $tablename . "` CHANGE `" . $cname . "` `" . $cname .
                             "` INT UNSIGNED NOT NULL AUTO_INCREMENT";
                    $log_message = i("H5p5xA|Added auto increment pro...", $tablename, $cname);
                    if (! $verify_only) {
                        $correction_success = $this->execute_and_log($appUserID, $sql_cmd, $log_message) &&
                                 $correction_success;
                    } else {
                        file_put_contents(self::$db_audit_log, 
                                date("Y-m-d H:i:s") . ": " . i("rSGs5f|Verification failed. aut...", 
                                        $tablename, $cname) . "\n", FILE_APPEND);
                        $verification_success = false;
                    }
                }
            }
        }
        
        // delete what is obsolete in the expected layout
        // ==============================================
        // columns
        foreach ($column_names_existing as $cname_existing) {
            if (! in_array($cname_existing, $columns_expected)) {
                if ($verify_only) {
                    file_put_contents(self::$db_audit_log, 
                            date("Y-m-d H:i:s") . ": " . i("ulIoAv|Verification failed. Col...", $tablename, 
                                    $cname_existing) . "\n", FILE_APPEND);
                    $verification_success = false;
                } else {
                    // e.g. ALTER TABLE `efaCloudPartners` DROP `UseEcrm`;
                    $sql_cmd = "ALTER TABLE `" . $tablename . "` DROP `" . $cname_existing . "`;";
                    $log_message = i("TuOggl|Dropped obsolete column ...", $tablename, $cname_existing);
                    $correction_success = $this->execute_and_log($appUserID, $sql_cmd, $log_message) &&
                             $correction_success;
                }
            }
        }
        // indices
        foreach ($indexes_existing as $iname_existing) {
            if (! in_array($iname_existing, $indexes_expected)) {
                if ($verify_only) {
                    file_put_contents(self::$db_audit_log, 
                            date("Y-m-d H:i:s") . ": " .
                                     i("nPmhjO|Verification failed. Ind...", $tablename, $iname_existing) .
                                     "\n", FILE_APPEND);
                    $verification_success = false;
                } else {
                    $sql_cmd = "ALTER TABLE `" . $tablename . "` DROP INDEX `" . $iname_existing . "`;";
                    $log_message = i("zQY162|Dropped obsolete index  ...", $tablename, $iname_existing);
                    $correction_success = $this->execute_and_log($appUserID, $sql_cmd, $log_message) &&
                             $correction_success;
                }
            }
        }
        
        // if the $verify_only flag is set and this point reached, all is fine
        if ($verify_only) {
            if ($verification_success) {
                file_put_contents(self::$db_audit_log, 
                        date("Y-m-d H:i:s") . ": " . i("gLdeJY|Verification successful ...", $tablename) . "\n", 
                        FILE_APPEND);
            } else {
                file_put_contents(self::$db_audit_log, 
                        date("Y-m-d H:i:s") . ": " . i("A1tKnO|Verification failed for ...", $tablename) . "\n", 
                        FILE_APPEND);
            }
        } else {
            if ($correction_success) {
                file_put_contents(self::$db_audit_log, 
                        date("Y-m-d H:i:s") . ": " . i("6QCftb|Completed update_table_l...", $tablename) . "\n", 
                        FILE_APPEND);
            } else {
                file_put_contents(self::$db_audit_log, 
                        date("Y-m-d H:i:s") . ": " . i("URJSm4|Correction failed for °%...", $tablename) . "\n", 
                        FILE_APPEND);
            }
        }
        return ($verify_only) ? $verification_success : $correction_success;
    }
}
    
