<?php
/**  
 *
 *       the tools-for-your-hobby framework
 *       ----------------------------------
 *       https://www.tfyh.org
 *
 * Copyright  2023-2024  Martin Glade
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

include_once "../classes/tfyh_data.php";
include_once "../classes/tfyh_toolbox.php";

/**
 * A fully static utility class to provide data validation against rules.
 */
class Tfyh_validate
{

    /**
     * Error count for last parsing, cleansing or validation.
     */
    private static $errors = [];

    /**
     * Warnings count for last parsing, cleansing or validation.
     */
    private static $warnings = [];

    /**
     * Set the self::$validation_error according to the reason of validation failure
     * 
     * @param String $violated_value            
     */
    public static function clear_findings ()
    {
        self::$errors = [];
        self::$warnings = [];
    }

    /**
     * Add the findings from data validation to the Tfyh_Validate findings
     * 
     * @param String $violated_value            
     */
    private static function add_data_validation_findings ()
    {
        self::$errors = array_merge(self::$errors, Tfyh_data::get_errors());
        self::$warnings = array_merge(self::$warnings, Tfyh_data::get_warnings());
    }

    /**
     * Add a finding from local validation to the Tfyh_Validate findings
     * 
     * @param bool $is_error set true to indicate that the finding represents an error, else a warning
     * @param String $finding the finding's text
     */
    public static function add_finding (bool $is_error, String $finding)
    {
        if ($is_error)
            self::$errors[] = $finding;
        else
            self::$warnings[] = $finding;
    }

    /**
     * Quick check whether findings exist.
     * 
     * @param bool $include_warnings
     *            Set true to also check for warnings
     * @return bool true if errors (or warnings) exist, false else.
     */
    public static function has_findings (bool $include_warnings = false)
    {
        return (($include_warnings && (count(self::$warnings) > 0)) || (count(self::$errors) > 0));
    }

    /**
     * Get a list of all findings, one finding per line, using \n als line break.
     * 
     * @param bool $include_warnings
     *            Set true to also include warnings
     * @return String the findings list if errors (or warnings) exist, else an empty String.
     */
    public static function get_findings (bool $include_warnings = false)
    {
        $findings = "";
        $error_text = i("nQeqPT|Error") . ": ";
        foreach (self::$errors as $error)
            $findings .= $error_text . $error . "\n";
        if (! $include_warnings)
            return $findings;
        $warning_text = i("yjp3yF|Warning") . ": ";
        foreach (self::$warnings as $warning)
            $findings .= $warning_text . $warning . "\n";
        return $findings;
    }

    /**
     * Validate a String according to the data type and rules provided in the template item. Get the validated
     * value back as native value.
     * 
     * @param Tfyh_config_item $template_item
     *            the configuration item holding all information to parse and validate the value.
     * @param String $value_as_string
     *            the value to be parsed and validated, may be null
     * @return mixed the value parsed and validated.
     */
    public static function parse_and_validate_value (Tfyh_config_item $template_item, String $value_as_string = null)
    {
        $data_type = $template_item->get_type();
        $parsed = Tfyh_data::parse($value_as_string, $data_type);
        Tfyh_validate::add_data_validation_findings();
        $type_and_limits_checked = Tfyh_data::check_limits($parsed, $data_type, $template_item->get_min(), 
                $template_item->get_max());
        Tfyh_validate::add_data_validation_findings();
        if (!is_null($template_item->get_descriptor_field("value_rules")) &&
                 (strlen($template_item->get_descriptor_field("value_rules")) > 0)) {
                     $rules = explode(",", $template_item->get_descriptor_field("value_rules"));
            foreach ($rules as $rule) {
                $rule_applied_finding = self::apply_rule($type_and_limits_checked, $rule);
                if (strlen($rule_applied_finding) > 0)
                    Tfyh_validate::add_finding(true, $rule_applied_finding);
            }
        }
        return $type_and_limits_checked;
    }

    /**
     * Validate a data record according to the data type and rules provided in the record item. Get the
     * validated record back as native values. If the $record_as_strings contains a field without template
     * definition, it is returned unchanged.
     * 
     * @param array $record_as_strings
     *            the record as associative array, such as it is provided by a form or by the socket.
     * @param Tfyh_config_item $record_item
     *            the configuration branch root of the record holding all information to parse and validate
     *            the records values.
     * @return array the record as associative array parsed and validated.
     */
    public static function parse_and_validate_record (array $record_as_strings, Tfyh_config_item $record_item)
    {
        $errors = "";
        $record_validated = [];
        foreach ($record_as_strings as $field => $value) {
            if (! $record_item->has_child($field)) {
                $record_validated[$field] = $value;
            } else {
                $field_item = $record_item->get_child($field);
                $validated = Tfyh_validate::parse_and_validate_value($field_item, $value);
                $record_validated[$field] = $validated;
            }
        }
        return $record_validated;
    }

    /**
     * Format the full record according to the rules of its configuration
     * 
     * @param array $record_native
     *            the record as associative array, such as it is provided by parsing.
     * @param Tfyh_config_item $record_item
     *            the configuration branch root of the record holding all information to format the records
     *            values.
     * @param String $language_code
     *            the language to use for formatting. Default is "csv" which represents the formatting used by
     *            the data base.
     */
    public static function format_parsed (array $record_native, Tfyh_config_item $record_item, 
            String $language_code)
    {
        $record_formatted = [];
        foreach ($record_native as $field => $value) {
            $data_type = ($record_item->has_child($field)) ? $record_item->get_child($field)->get_type() : "string";
            $record_formatted[$field] = Tfyh_data::format($value, $data_type, $language_code);
        }
        return $record_formatted;
    }

    /**
     * This will apply a validation rule to the value. Return value ist true, if compliant or an error String,
     * if not compliant.
     */
    public static function apply_rule ($value, String $rule)
    {
        if (strcasecmp($rule, "identifier") == 0) {
            return self::validate_identifier($value);
        } elseif (strcasecmp($rule, "password") == 0) {
            return self::validate_password($value);
        } else
            return "";
    }

    /**
     * An identifier is a String consisting of [_a-zA-Z] followed by [_a-zA-Z0-9] and of 1 .. 64 characters
     * length
     * 
     * @param String $identifier
     *            the string to check
     * @return no return value. The result is provided via Tfyh_validate::add_validation_result
     */
    private static function validate_identifier (String $identifier)
    {
        if (strlen($identifier) < 1)
            Tfyh_validate::add_finding(true, i("HE2ICg|Empty identifier"));
        if (strlen($identifier) > 64)
            Tfyh_validate::add_finding(true, i("VfEQj7|The maximum identifier l..."));
        $first = substr($identifier, 0, 1);
        $remainder = str_replace("_", "A", substr($identifier, 1));
        if (! ctype_alpha($first) && (strcmp($first, "_") != 0)) {
            if (ctype_digit($first))
                Tfyh_validate::add_finding(false, 
                        i("cVYtkK|Numeric start character ...", $identifier));
            else
                Tfyh_validate::add_finding(true, i("va4j2V|invalid character in ide...", $identifier));
        }
        if (! ctype_alnum($remainder))
            Tfyh_validate::add_finding(true, i("WVta4w|Invalid identifier: %1.", $identifier));
    }

    /**
     * Check whether $to_check represents a uid
     *
     * @param String $to_check
     *            the String to check
     * @param bool $or_shorter
     *            set true to allow also for partial uid-Strings (<= 8 characters)
     * @return bool true if, and only if $to_check represents a valid uid. Format: 'xxxxxxxx', each x
     *         representing a character within Tfyh_toolbox::$base62.
     */
    public static function is_uid (String $to_check, bool $or_shorter)
    {
        if ((strlen($to_check) > 8) || (! $or_shorter && (strlen($to_check) < 8)))
            return false;
            for ($i = 0; $i < strlen($to_check); $i ++)
                if (strpos(base62chars, substr($to_check, $i, $i + 1)) === false)
                    return false;
                    return true;
    }
    
    /**
     * Check whether $to_check represents a UUID
     * 
     * @return bool true if, and only if $to_check represents a valid uuid. Format:
     *         'xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx', each block representing a hex number.
     */
    public static function is_uuid (String $to_check)
    {
        if (strlen($to_check) != 36)
            return false;
        $parts = explode("-", $to_check);
        if (count($parts) != 5)
            return false;
            $sizes = [8,4,4,4,12
        ];
        for ($i = 0; $i < 5; $i ++) {
            if (strlen($parts[$i]) != $sizes[$i])
                return false;
            if (! ctype_xdigit($parts[$i]))
                return false;
        }
        return true;
    }

    /**
     * Check, whether the pwd complies to password rules. Copy of Tfyh_toolbox::check_password.
     * 
     * @param String $pwd
     *            password to be checked
     * @return no return value. The result is provided via Tfyh_validate::add_validation_result
     */
    private static function validate_password (String $pwd)
    {
        $errors = "";
        if ((strlen($pwd) < 8) || (strlen($pwd) > 32))
            Tfyh_validate::add_finding(true, i("aJ5Cy9|The password must be bet..."));
        $numbers = (preg_match("#[0-9]+#", $pwd)) ? 1 : 0;
        $lowercase = (preg_match("#[a-z]+#", $pwd)) ? 1 : 0;
        $uppercase = (preg_match("#[A-Z]+#", $pwd)) ? 1 : 0;
        // Four ASCII blocks: !"#$%&'*+,-./ ___ :;<=>?@ ___ [\]^_` ___ {|}~
        $specialchars = (preg_match("#[!-/]+#", $pwd) || preg_match("#[:-@]+#", $pwd) ||
                 preg_match("#[\[-`]+#", $pwd) || preg_match("#[{-~]+#", $pwd)) ? 1 : 0;
        if (($numbers + $lowercase + $uppercase + $specialchars) < 3)
            Tfyh_validate::add_finding(true, i("iJUmCH|The password must contai..."));
    }
}
