<?php
/**  
 *
 *       the tools-for-your-hobby framework
 *       ----------------------------------
 *       https://www.tfyh.org
 *
 * Copyright  2023-2024  Martin Glade
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

include_once "../classes/tfyh_validate.php";

/**
 * Class file to maintain a word index, referencing to data base records. The index itself is a dta base
 * table.
 */
class Tfyh_word_index
{

    /**
     * In order to separate the words and create an index of the text we need to parse the text, separate the
     * words and transcribe the characters to ASCII to avoid sorting and misspeling problems. Characters which
     * are not listed are used as in-word characters and replaced by an undescore '_'.
     */
    private static $char_blocks = ["stay" => "abcdefghijklmnopqrstuvwxyz0123456789_-",
            // characters which stay one character
            "from_1" => "ABCDEFGHIJKLMNOPQRSTUVWXYZÁÀÂÄÇÉÈÊËÍÌÎÏÑÓÒÔÚÙÛŸáàâçéèêëíìîïñóòôúùûÿ",
            "to_1" => "abcdefghijklmnopqrstuvwxyzaaaaceeeeiiiinooouuuyaaaceeeeiiiinooouuuy",
            // special characters which become two characters
            "from_2" => "ÅÄÆØÖŒÜåäæøöœüß","to_2" => "AaAeAeOeOeOeUeaaaeaeoeoeoeuess",
            // word separator characters
            "separate" => " ',;.:#+*/=§$%&@€|<>(){}[]?!`\"\f\n\r\t\v\\"
    ];

    /**
     * The minimum length of a wor to become part of the index, usually 3+
     * 
     * @var integer
     */
    const MIN_WORD_LENGTH = 3;

    /**
     * The data base connection socket.
     */
    private $socket;

    /**
     * The generic toolbox.
     */
    private $toolbox;

    /**
     * The index which leads from the uid to the record.
     */
    private $uids;

    /**
     * cache for the words and uuids of a record
     */
    private $extracted;

    /**
     * cache for the uids of the search result;
     */
    private $search_result_all_words;

    /**
     * cache for the uids of the search result;
     */
    private $search_result_this_word;

    /**
     * cache for the uids of the search result;
     */
    public $find_result;

    /**
     * The index.
     */
    private $words;

    private $index_dir = "../log/index";

    private $build_log = "../log/index/word_index.log";

    private $words_file = "../log/index/words";

    private $uids_file = "../log/index/uids";

    /**
     * public Constructor.
     * 
     * @param Tfyh_toolbox $toolbox
     *            application toolbox
     * @param Tfyh_socket $socket
     *            the socket to connect to the database
     */
    public function __construct (Tfyh_toolbox $toolbox, Tfyh_socket $socket)
    {
        $this->toolbox = $toolbox;
        $this->socket = $socket;
    }

    /**
     * Find the words and return all uids of records matching. This will not look at any user permission. The
     * result ist stored at "$this->find_result"
     * 
     * @param String $words
     *            a String with all words. They will be split the same way, as the index is built.
     * @param bool $anded
     *            set true to ask for all words to be matched, false to ask for any word to be matched.
     * @param bool $like
     *            set true to ask for words to be matched using the LIKE '%word' SQL operator, false to use
     *            exact equality (faster)
     */
    public function find (String $words, bool $anded, bool $like)
    {
        $words = $this->extract_words($words);
        $this->read_index();
        $this->search_result_all_words = [];
        // iterate over all words
        foreach ($words as $word) {
            $this->search_result_this_word = [];
            // uids have a separate index for not to be case sensitive
            if (Tfyh_validate::is_uid($word, true)) {
                foreach ($this->uids as $uid => $tablename) {
                    $use_it = (strcmp($uid, $word) == 0) || ($like && (strpos($uid, $word) !== false));
                    if ($use_it && ! in_array($uid, $this->search_result_this_word))
                        $this->search_result_this_word[] = $uid;
                }
            }
            $word_lower_ascii = self::to_lower_ascii($word);
            foreach ($this->words as $word_of_index => $uids) {
                $use_it = (strcmp($word_of_index, $word_lower_ascii) == 0) ||
                         ($like && (strpos($word_of_index, $word_lower_ascii) === 0));
                if ($use_it)
                    $this->search_result_this_word = array_unique(
                            array_merge($this->search_result_this_word, $uids));
            }
            if ($anded && (count($this->search_result_all_words) > 0))
                // if anded ntersect, but starting from the second word. If not this will always intersect
                // with an empty array.
                $this->search_result_all_words = array_intersect($this->search_result_all_words, 
                        $this->search_result_this_word);
            else
                $this->search_result_all_words = array_unique(
                        array_merge($this->search_result_all_words, $this->search_result_this_word));
        }
        
        // get the result in a useable way
        $this->find_result = [];
        foreach ($this->search_result_all_words as $uid) {
            // the $this->find_result is a two level array: per table and per record
            $tablename = $this->uids[$uid];
            if (! isset($this->find_result[$tablename]))
                $this->find_result[$tablename] = [];
            if (isset($tablename)) {
                // read the relevant fields of the record
                $record_cfg = $this->toolbox->config->get_by_path(".appTables." . $tablename);
                $sql_cmd = "SELECT `uid`";
                foreach ($record_cfg->get_children() as $cname => $cdefinition) {
                    $handling = $cdefinition->get_descriptor_field("column_handling");
                    if (strpos($handling, "w") !== false)
                        $sql_cmd .= ", `" . $cname . "`";
                }
                $sql_cmd = $sql_cmd . " FROM `" . $tablename . "` WHERE `uid` = '" . $uid . "';";
                $res = $this->socket->query($sql_cmd, $this);
                if (isset($res->num_rows) && (intval($res->num_rows) > 0))
                    $row = $res->fetch_row();
                $c = 0;
                $record = [];
                $record["@in_fields"] = "";
                $record["@display_as"] = "";
                foreach ($record_cfg->get_children() as $cname => $cdefinition) {
                    $handling = $cdefinition->get_descriptor_field("column_handling");
                    if (strpos($handling, "w") !== false) {
                        $c ++;
                        $record[$cname] = $row[$c];
                        if (! is_null($row[$c])) {
                            $field_local = $cdefinition->get_descriptor_field("text_local_name");
                            foreach ($words as $word)
                                if (strpos(self::to_lower_ascii($row[$c]), $word) !== false)
                                    $record["@in_fields"] .= $field_local . ": " . str_replace($word, 
                                            "<b>" . $word . "</b>", self::to_lower_ascii($row[$c])) . ", ";
                            $value_native = Tfyh_data::parse($row[$c], $cdefinition->get_type());
                            $display_as = Tfyh_data::format($value_native, $cdefinition->get_type(), 
                                    $this->toolbox->config->language_code);
                            $record["@display_as"] .= $field_local . ":" . $display_as . ", ";
                        }
                    }
                }
            }
            $this->find_result[$tablename][$uid] = $record;
        }
    }

    /**
     * Read the index from file.
     */
    private function read_index ()
    {
        $this->words = [];
        $this->uids = [];
        $words_file = file_get_contents($this->words_file);
        if ($words_file === false)
            return;
        $uids_file = file_get_contents($this->uids_file);
        if ($uids_file === false)
            return;
        $words_lines = explode("\n", $words_file);
        foreach ($words_lines as $words_line)
            if (strlen($words_line) > 0) {
                $pos_commma = strpos($words_line, ",");
                $word = substr($words_line, 0, $pos_commma);
                $this->words[$word] = explode(",", substr($words_line, $pos_commma + 1));
            }
        $uids_lines = explode("\n", $uids_file);
        foreach ($uids_lines as $uids_line)
            if (strlen($uids_line) > 0) {
                $parts = explode(",", $uids_line);
                $this->uids[$parts[0]] = $parts[1];
            }
    }

    /**
     * Filter a word and add it to either the uuid-index or the word-index.
     * 
     * @param String $word            
     */
    private function add_word (String $word)
    {
        $is_uuid = Tfyh_validate::is_uuid($word);
        $is_uid = Tfyh_validate::is_uid($word, false);
        if ($is_uuid) {
            // for UUID and uid stay case sensitive
            if (! in_array($word, $this->extracted))
                $this->extracted[] = $word;
        } elseif ($is_uid && isset($this->uids[$uid])) {
            // for UUID and uid stay case sensitive
            if (! in_array($word, $this->extracted))
                $this->extracted[] = $word;
        } else {
            $snippets = explode("-", $word);
            foreach ($snippets as $snippet)
                if ((strlen($snippet) >= self::MIN_WORD_LENGTH) && ! is_numeric($snippet) && ! in_array(
                        $snippet, $this->extracted))
                    $this->extracted[] = $snippet;
        }
    }

    /**
     * Transform a word into its ASCII representation by replacing special characters like ä, é with ae, e and
     * puttin the whole word into lower case.
     * 
     * @param String $word            
     */
    public static function to_lower_ascii (String $word)
    {
        $len = mb_strlen($word);
        $word_ascii = "";
        for ($i = 0; $i < $len; $i ++) {
            $c = mb_substr($word, $i, 1);
            // check character. THose which need no handling first.
            if (mb_strpos(self::$char_blocks["stay"], $c) !== false)
                $word_ascii .= $c;
            else {
                // check character. Special characters which are extended to two characters
                $pos_from_2 = mb_strpos(self::$char_blocks["from_2"], $c);
                if ($pos_from_2 !== false)
                    $word_ascii .= mb_substr(self::$char_blocks["to_2"], $pos_from_2 * 2, 2);
                else {
                    // check character. Special characters which are replaced by a single character
                    $pos_from_1 = mb_strpos(self::$char_blocks["from_1"], $c);
                    if ($pos_from_1 !== false)
                        $word_ascii .= mb_substr(self::$char_blocks["to_1"], $pos_from_1, 1);
                    else
                        // not in the characters list of characters with known handling. replace by
                        // underscore.
                        $word_ascii .= "_";
                }
            }
        }
        return $word_ascii;
    }

    /**
     * Extract all word from a text. Lower case characters are kept, Upper ase charatcers converted to lower
     * case and special characters transcoded. Words are separated and collected into a deduplicated array.
     * 
     * @param String $text
     *            the text to parse
     * @return string[] the array of lower case ASCII-transcode wirds.
     */
    private function extract_words (String $text)
    {
        $this->extracted = [];
        $word = "";
        $len = mb_strlen($text);
        for ($i = 0; $i < $len; $i ++) {
            $c = mb_substr($text, $i, 1);
            if (mb_strpos(self::$char_blocks["separate"], $c) !== false) {
                $this->add_word(self::to_lower_ascii($word));
                $word = "";
            } else
                $word .= $c;
        }
        // add the last word, if not empty (e.g. because the text ended with a separator)
        if (strlen(self::to_lower_ascii($word)) > 0)
            $this->add_word(self::to_lower_ascii($word));
        return $this->extracted;
    }

    /**
     * Build or rebuild the word index. This will delete the respective index files
     */
    public function rebuild ()
    {
        if (! file_exists($this->index_dir))
            mkdir($this->index_dir);
        file_put_contents($this->build_log, date("Y-m-d H:i:s") . ": Starting word index build.\n");
        $start_time = time();
        
        // iterate over all tables
        $this->words = [];
        $this->uids = [];
        $app_tables = $this->toolbox->config->get_by_path(".appTables");
        foreach ($app_tables->get_children() as $tablename => $record_cfg) {
            file_put_contents($this->build_log, date("Y-m-d H:i:s") . ": Reading $tablename: ", FILE_APPEND);
            // collect all word index relevant fields from the configuration
            $record_cfg = $this->toolbox->config->get_by_path(".appTables." . $tablename);
            $sql_cmd = "SELECT `uid`, ";
            $c = 0;
            foreach ($record_cfg->get_children() as $cname => $cdefinition) {
                $handling = $cdefinition->get_descriptor_field("column_handling");
                if (strpos($handling, "w") !== false) {
                    $c ++;
                    $sql_cmd .= "`" . $cname . "`, ";
                }
            }
            if ($c > 0) {
                file_put_contents($this->build_log, date("Y-m-d H:i:s") . ": Analysing $tablename: ", 
                        FILE_APPEND);
                $sql_cmd = mb_substr($sql_cmd, 0, mb_strlen($sql_cmd) - 2) . " FROM `" . $tablename .
                         "`WHERE 1;";
                $res = $this->socket->query($sql_cmd, $this);
                $r = 0;
                if (isset($res->num_rows) && (intval($res->num_rows) > 0)) {
                    $row = $res->fetch_row();
                    while ($row) {
                        $r ++;
                        if (($r % 100) == 0)
                            file_put_contents($this->build_log, ".", FILE_APPEND);
                        if (($r % 500) == 0)
                            file_put_contents($this->build_log, " ", FILE_APPEND);
                        $uid = $row[0];
                        if (! isset($this->uids[$uid]))
                            $this->uids[$uid] = $tablename;
                        for ($i = 1; $i <= $c; $i ++) {
                            $text = $row[$i];
                            if (isset($text)) {
                                $words = $this->extract_words($text);
                                foreach ($words as $word) {
                                    if (! isset($this->words[$word]))
                                        $this->words[$word] = [];
                                    if (! in_array($uid, $this->words[$word]))
                                        $this->words[$word][] = $uid;
                                }
                            }
                        }
                        $row = $res->fetch_row();
                    }
                }
                file_put_contents($this->build_log, 
                        "\n$tablename completed. $r records analysed. Words so far: " . count($this->words) .
                                 "\n", FILE_APPEND);
            } else
                file_put_contents($this->build_log, date("Y-m-d H:i:s") . ": Skipping $tablename.\n", 
                        FILE_APPEND);
        }
        
        // store the result
        $words_file = "";
        foreach ($this->words as $word => $uids) {
            $words_file .= $word;
            foreach ($uids as $uid)
                $words_file .= "," . $uid;
            $words_file .= "\n";
        }
        file_put_contents($this->words_file, $words_file);
        $uids_file = "";
        foreach ($this->uids as $uid => $tablename)
            $uids_file .= $uid . "," . $tablename . "\n";
        file_put_contents($this->uids_file, $uids_file);
        
        file_put_contents($this->build_log, 
                date("Y-m-d H:i:s") . ": Completed word index build in " . (time() - $start_time) .
                         " seconds.\n", FILE_APPEND);
    }
}


