<?php

/**  
 *
 *       dilbo - digital logbook for Rowing and Canoeing
 *       -----------------------------------------------
 *       https://www.dilbo.org
 *
 * Copyright  2023-2024  Martin Glade
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

/**
 * The response page to any javascript form post. Provides configuration and data insert, update and delete
 * capabilities.
 * 
 * @author mgSoft
 */
function return_result (int $code, String $text)
{
    echo $code . ";" . $text;
    end_script(false);
    exit(); // really exit. No test case left over.
}

// ===== initialize toolbox and socket and start session.
$user_requested_file = __FILE__;
include_once "../classes/init.php";

// === APPLICATION LOGIC ==============================================================
$mode = intval($_POST["mode"]);
if ($mode <= 0)
    return_result(501, "Invalid mode.");
$uid = $_POST["uid"];

$is_config = ($mode < 10);
if ($is_config) {
    $csv = $_POST["csv"];
    $path = $_POST["path"];
    // Modifications are only allowed in two branches.
    if ((strlen($path) < 1) || (substr($path, 0, 1) != '.'))
        return_result(502, "Invalid path: '$path'");
} else {
    $mode = $mode - 10;
    $tablename = $_POST["tablename"];
}

if ($is_config) {
    
    // ===========================
    // MODIFY A CONFIGURATION ITEM
    // ===========================
    
    // get the top branch to check whether the change is allowed.
    $parts = explode(".", $path);
    $top_branch_name = $parts[1];
    if (! in_array($top_branch_name, ["appSettings","appTables","clubSettings","uiSettings"
    ]))
        return_result(502, "Invalid top branch for insert: '$top_branch_name'");
    
    include_once '../classes/dilbo_config.php';
    // load the specific configuration, if it was optional.
    if (! $toolbox->config->get_cfg_root_item()->has_child($top_branch_name)) {
        // add the used extra configuration
        foreach (Dilbo_config::$extra_config_files as $cfg_file => $object_if_missing)
            if (strcasecmp($top_branch_name, $cfg_file) == 0)
                $toolbox->config->include_settings($cfg_file, $object_if_missing);
    }
    
    // set variables for later use
    $top_branch = $toolbox->config->get_cfg_root_item()->get_child($top_branch_name);
    if (is_null($top_branch_name))
        return_result(502, "Could not initialize $top_branch_name");
    
    $config_item = $toolbox->config->get_by_path($path);
    $parent_path = substr($path, 0, strrpos($path, "."));
    $item_name = substr($path, strrpos($path, ".") + 1);
    $config_parent = $toolbox->config->get_by_path($parent_path);
    
    // execute configuration change
    // ============================
    
    include_once "../classes/dilbo_config.php";
    if (in_array($top_branch_name, Dilbo_config::$editable_config_files)) {
        
        if ($mode == 1) {
            // insert. A configuration parent must be there.
            if (is_null($config_parent))
                return_result(502, "No config item at parent path given on insert: '$parent_path'");
            $parent_addable_type = $config_parent->get_descriptor_field("node_addable_type");
            if (strlen($parent_addable_type) == 0)
                return_result(502, "Adding not permitted for: '$parent_path'");
            // check the name validity
            include_once "../classes/tfyh_validate.php";
            $identifier_check = Tfyh_validate::apply_rule($item_name, "identifier");
            if (strlen($identifier_check) > 0)
                return_result(502, $identifier_check);
            // A full branch is posted for insert. But in any case expand the objects.
            $branch_to_add = Tfyh_config_item::read_branch_csv($csv, true);
            if (is_string($branch_to_add))
                return_result(502, "Configuration parsing error: '$branch_to_add'");
            // attach the branch to the parent and store the result.
            $config_parent->attach_branch($item_name, $branch_to_add);
            $success = file_put_contents("../config/settings/$top_branch_name", 
                    $top_branch->to_string(false, "csv", 99, false));
            if ($success)
                return_result(300, "insert ok.");
            else
                return_result(503, "$top_branch_name file write failed.");
        } else {
            // Update, delete or move. A configuratio item to modify must be there.
            if (is_null($config_item))
                return_result(502, "Configuration item '$path' not found in $top_branch_name.");
            if ($mode == 2) {
                // A partial branch is posted for update, usually the configuration item plus children only.
                $branch_to_update = Tfyh_config_item::read_branch_csv($csv, false);
                // replace branch
                $config_item->update_branch($config_item->get_name(), $branch_to_update);
                // write settings file
                $success = file_put_contents("../config/settings/$top_branch_name", 
                        $top_branch->to_string(false, "csv", 99, false));
                if ($success)
                    return_result(300, "config update ok.");
                else
                    return_result(503, "$top_branch_name file write failed.");
            } elseif ($mode == 3) {
                // Delete. Check first whether deletion is possible. Therefore the item must be of the addable
                // type for its parent.
                $parent_addable_type = $config_parent->get_descriptor_field("node_addable_type");
                if (strcasecmp($parent_addable_type, $config_item->get_type()) != 0)
                    return_result(503, "Delete forbidden for this configuration item.");
                else {
                    // remove and store.
                    $config_item->get_parent()->remove_branch($config_item->get_name());
                    $success = file_put_contents("../config/settings/$top_branch_name", 
                            $top_branch->to_string(false, "csv", 99, false));
                    if ($success)
                        return_result(300, "delete ok.");
                    else
                        return_result(503, "$top_branch_name file write failed.");
                }
            } elseif (($mode == 4) || ($mode == 5)) {
                // Move up or down. Check first whether parent has addables. Therefore the item must be of the
                // addable type for its parent.
                $parent_addable_type = $config_parent->get_descriptor_field("node_addable_type");
                if (strcasecmp($parent_addable_type, $config_item->get_type()) != 0)
                    return_result(503, "Move forbidden for this configuration item.");
                else {
                    // move and store.
                    $success = $config_item->get_parent()->move_child($item_name, ($mode == 4) ? - 1 : 1);
                    $success = $success && file_put_contents("../config/settings/$top_branch_name", 
                            $top_branch->to_string(false, "csv", 99, false));
                    if ($success)
                        return_result(300, "move ok.");
                    else
                        return_result(503, "move of $item_name or $top_branch_name file write failed.");
                }
            } else
                return_result(502, "Invalid mode.");
        }
    }
} else {
    
    // ====================
    // MODIFY A DATA RECORD
    // ====================
    
    $record = [];
    $tablename = $_POST["tablename"];
    $record_cfg = $toolbox->config->get_by_path(".appTables." . $tablename);
    foreach ($_POST as $key => $value)
        if ((strcasecmp($key, "mode") != 0) && (strcasecmp($key, "tablename") != 0) &&
                 $record_cfg->has_child($key))
            $record[$key] = $value;
    
    // execute modification
    $mode_str = ($mode == 1) ? "insert" : (($mode == 2) ? "update" : "delete");
    $user_id = $toolbox->users->session_user["@id"];
    include_once "../classes/dilbo_record.php";
    $dilbo_record = new Dilbo_record($toolbox, $socket);
    Tfyh_validate::clear_findings();
    $record_cfg = $toolbox->config->get_by_path(".appTables.$tablename");
    $findings = $dilbo_record->modify_record($record, $record_cfg, $mode, $user_id);
    
    // Return response
    if (strlen($findings) == 0)
        return_result(300, "$mode_str ok.");
    else
        return_result(502, $findings);
}
// Tis shall not be reached, therefore return an error.
return_result(502, "Nothing to do.");
