/**  
 *
 *       dilbo - digital logbook for Rowing and Canoeing
 *       -----------------------------------------------
 *       https://www.dilbo.org
 *
 * Copyright  2023-2024  Martin Glade
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

/**
 * Handle the display of forms, and apply he necessary logic to the results
 * provided. For each form the is a form_do and a form_done function. The
 * form_do prepares the form and shows it to the user, the form_done is called
 * when the form is submitted. It provides all data postprocessing, writes the
 * data to the local lists and sends a write transaction to the server.
 */

class DilboFormHandler {
	
	#modal = null;
	#formTemplateCache = "";
	#languageCode = "en";
	
	// configuration item editing
	#config = null;
	#configItem = null;
	#configCache = null;
	
	// data record editing
	#recordTablename = "";
	#recordConfig = "";
	#recordModifyErrors = "";
	#recordUid = "";
	#recordEditMode = 2;
	#recordEntered = {};
	#lookupSet = [];
	#lookupLoadIndex = -1;
	
	#editItemMinimumForm = 
		`tags;name;label
		_R;_no_input;_headline
		_//_R;text_local_name|value_current;
		_//_R;submit;Submit
		_//;_no_input;`;
			
	#createItemMinimumForm = 
		`tags;name;label
		_R;_no_input;_headline
		_//_R;name|;
		_//_r;text_local_name|~value_type;
		_//_R;submit;Submit
		_//;_no_input;`;
	
	#listElementFrame = 
		`<div class='w3-row' style='display:flex;flex-direction:row;overflow: hidden;' id='listElementBox-{index}'>
		<div class='closeModal removeElement' id='listElementClose-{index}' style='text-align:center;width:5%'> &times; </div>
		<div style='width:94.9%;'>{listElement}</div>
		</div>`;
	
	#listElementAdd = 
		`<div class='w3-row' style='display:flex;flex-direction:row;overflow: hidden;' id='listElementBox-add'>
		<div class='closeModal addElement' id='listElement-add' style='text-align:center;width:100%'> + </div>
		</div>`;

	validityModifications = { 
			editValidfrom : _("gutWog|Edit period start"), 
			extendPeriod : _("yjZn8H|Extend period end"), 
			limitPeriod : _("5YBrdm|Limit period end"), 
			newPeriod : _("ms8K9B|Create a new period reco...")
	}
	
	constructor() {
		this.#languageCode = userPreferences.languageCode;
	}
	
	// =====================
	// Configuration editing
    // =====================
	
	/**
	 * Open Form "editItem".
	 */
	editItem_do (item) {

		// set the configuration item for the form
		this.#configItem = item;
		// Cache the default template.
		this.#formTemplateCache = $_formTemplates.editItem;
		
		// get the item specific template or a minimum one.
		$_formTemplates.editItem = item.getEditForm();
		if (!$_formTemplates.editItem) 
			$_formTemplates.editItem = this.#editItemMinimumForm;
		// the default template includes a mandatory "name" field for item
		// creation. This must be removed for item editing.
		$_formTemplates.editItem = $_formTemplates.editItem.replace(";*name|", ";")
			.replace(";*name\n", ";\n").replace("|*name|", "|").replace("|*name\n", "\n");

		// now that the template is prepared, initialize the form
		dilboForm.init("editItem", item);
		dilboForm.headline = "<h3>" + item.get_descriptor_field("text_local_name") + "<br><small>" 
				+ item.get_path() + " (uid: " + item.get_uid() + ")</small></h3>";
		
		// preset uid
		dilboForm.presetValue("uid", item.get_uid());
		// preset descriptor fields
		dilboForm.presetValues(item.getDescriptor_asStrings(this.#languageCode));
		// preset children values
		let childrenValues = {};
		for (let childname in item.get_children()) 
			childrenValues[childname] = item.get_child(childname).get_value_as_string(this.#languageCode);
		dilboForm.presetValues(childrenValues);

		// show form
		dilboModal.showForm(true, this, dilboForm);
		
		// special hack for the text_local_name and the text_explanation which
		// shall no be edited, if they have a translation token.
		let textLocalName = $("#cFormInput-text_local_name").val()
		if ((textLocalName && textLocalName.indexOf("|") == 6) && (_(textLocalName) != textLocalName)) {
			$("#cFormInput-text_local_name").val(_(textLocalName));
			$("#cFormInput-text_local_name").addClass("display-bold");
		}
		let textExplanation = $("#cFormInput-text_explanation").val()
		if (textExplanation && (textExplanation.indexOf("|") == 6) && (_(textExplanation) != textExplanation)) {
			$("#cFormInput-text_explanation").val(_(textExplanation));
			$("#cFormInput-text_explanation").addClass("display-bold");
		}
	}

	/**
	 * Form "editItem" is submitted. Editing was either for the item itself or
	 * its children.
	 */
	editItem_done () {
		
		// check for duplicate or forbidden name
		if (dilboForm.getChanged("text_local_name"))
			if (! this.#isOkNames(this.#configItem.get_parent(), null, dilboForm.getParsed("text_local_name")))
				return;
		
		// split own descriptor and child values
		let itemDescriptorFields = {};
		let childValues = {};
		for (let field in dilboForm.getParsed())
			if (TfyhData.is_descriptor_field_name(field))
				itemDescriptorFields[field] = dilboForm.getParsed(field);
			else
				childValues[field] = dilboForm.getParsed(field);
		
		// put the item change to memory
		if (Object.keys(itemDescriptorFields).length > 0) 
			this.#configItem.set_descriptor(itemDescriptorFields);
		// put the children changes to memory
		let children = this.#configItem.get_children();
		for (let childname in childValues) {
			let childValue = childValues[childname];
			if (this.#configItem.has_child(childname)) {
				// descriptor fields must include the data type for a post
				// request data bundle
				this.#configItem.get_child(childname).set_descriptor({ 
					value_current : childValue, 
					value_type : this.#configItem.get_child(childname).get_type() 
				});
			}
		}

		// restore the default template
		$_formTemplates.editItem  = this.#formTemplateCache;
		
		// post the change to the server. You must include the defaults to be
		// able to reset a non default value to a default value. One level drill
		// down is sufficient
		let modificationCsv = this.#configItem.to_string(false, "csv", 1, true); 
		dilboConfig.postModify(2, this.#configItem, modificationCsv, this.onPostItemCallback);
	}

	/**
	 * Open Form "createChild".
	 */
	createChild_do (item) {
		
		// identify the object to create
		if (item.get_addable_type().length == 0) {
			dilboModal.showHtml("<h3>" + _("mliI2K|No data type declared to...") + "</h3>");
			return;
		}
		if (! item.get_addable_type().startsWith(".")) {
			dilboModal.showHtml("<h3>" + _("1tpaxz|No basic data types addi...") + "</h3>");
			return;
		}
		let node_addable_type = item.get_descriptor_field("node_addable_type");
	    if (! TfyhData.typeTemplates[node_addable_type] && ! TfyhData.objectTemplates[node_addable_type])  {
			alert("Application configuration error: " + node_addable_type + " not found");
			return;
		}

	    // get a new item with a forbidden name, to ensure there will be no name
		// conflicts
		let newChild = item.get_child("NEW!", node_addable_type);
		
		// Cache the default template.
		this.#formTemplateCache = $_formTemplates.createChild;
		// get the item specific template or a minimum one.
		$_formTemplates.createChild = newChild.getEditForm();
		if (!$_formTemplates.createChild) 
			$_formTemplates.createChild = this.#createItemMinimumForm;

		// create an extra 'name' child for the form initialization
		newChild.get_child("name", "string", "", _("paPlYi|Name"));
		// now that the template is prepared, initialize the form
		dilboForm.init("createChild", newChild);
		dilboForm.presetValue("value_type", node_addable_type);
		
		let localName = item.get_descriptor_field("text_local_name");
		dilboForm.headline = "<h3>" + _("yOMZoX|Create a child item for") + " " + localName + "<br>(" + item.get_path() + ")</h3>";
		// dilboForm.selectCatalog = item;
		this.#configItem = item;
		// remove temporary child from object template after form creation
		newChild.remove_branch("name");
		// show form
		dilboModal.showForm(true, this, dilboForm);
	}
	
	/**
	 * Form "createChild" is submitted.
	 */
	createChild_done () {
		// remove the temporary child item.
		this.#configItem.remove_branch("NEW!");
		// check for duplicate or forbidden name
		let childName = dilboForm.getParsed("name");
		if (! this.#isOkNames(this.#configItem, childName, dilboForm.getParsed("text_local_name")))
			return;
		let dataType = this.#configItem.get_addable_type();
		if (dataType.substring(0, 1) == '.') {
			// get the objects template. The parent layer will be used for
			// access
			let newChild = this.#configItem.get_child(childName, dataType, null, childName);
			dilboForm.useParsedFor(newChild);
			// move new branches up the leafs section
			this.#configItem.sort_children();
			// post the change to the server. You must include the defaults to
			// be able to reset a non default value to a default value. One
			// level drilldown is sufficient.
			let newChildCsv = newChild.to_string(false, "csv", 1, true);
			dilboConfig.postModify(1, newChild, newChildCsv, this.onPostItemCallback);
		} else {
			// Should actually not occur, only objects can be added. No
			// raw data adding.
			alert("Application configuration error: Only objects can be added.");
		}
		// restore the default template
		$_formTemplates.createChild  = this.#formTemplateCache;
	}
	
	// ===============
	// session editing
	// ===============
	
	/**
	 * Set input autocompletion ranges for the session input fields
	 */
	#setSessionInputAutocompletion() {
		let input = $("#cFormInput-asset_uuid")[0];
		TfyhAutoComplete.set(input, dilboInputValidator, dilboTables.validNames.assets, "assets", this);
		input = $("#cFormInput-cox")[0];
		TfyhAutoComplete.set(input, dilboInputValidator, dilboTables.validNames.persons, "persons", this);
		for (var i = 1; i <= 8; i++) {
			input = $("#cFormInput-crew_" + i)[0];
			TfyhAutoComplete.set(input, dilboInputValidator, dilboTables.validNames.persons, "persons", this);
		}
		input = $("#cFormInput-destination")[0];
		TfyhAutoComplete.set(input, dilboInputValidator, dilboTables.validNames.destinations, "destinations", this);
		input = $("#cFormInput-waters")[0];
		TfyhAutoComplete.set(input, dilboInputValidator, dilboTables.validNames.persons.waters, "waters", this);
	}
	
	/**
	 * Show or hide the session end date and time depending on the situation
	 */
	#showSessionEnd(close) {
		let sessionType = $('#cFormInput-session_type').find(":selected").val();
		if (!close && (sessionType != 'LATEENTRY')) {
			// hide the end time and date for new sessions.
			dilboForm.presetValue("endDate", "");
			dilboForm.presetValue("endTime", "");
			$('#cFormInput-endDate').parent().hide();
			$('#cFormInput-endTime').parent().hide();
		} else {
			$('#cFormInput-endDate').parent().show();
			$('#cFormInput-endTime').parent().show();
		}
	}
	
	/**
	 * resolve a uuid for a boat, destination, waters or person record. Return
	 * the uuid on success and the original value on failure.
	 */
	#resolveUuid(name, table, tripEntered) {
		if (tripEntered[name] && dilboTables.tables[table].iNames[tripEntered[name]]) {
			let uuid = dilboTables.tables[table].iNames[tripEntered[name]].uuid;
			return ((uuid) ? uuid : tripEntered[name]);
		}
		return tripEntered[name];
	}

	/**
	 * resolve Uuids and datetime values
	 */
	#presetTripData(trip) {
		
		dilboForm.presetValue("number", trip.number);
		dilboForm.presetValue("logbookname", trip.logbookname);
		let start = TfyhData.format(trip.start, "datetime");
		dilboForm.presetValue("startDate", start.substring(0,10));
		dilboForm.presetValue("startTime", start.substring(11,16));
		dilboForm.presetValue("asset_uuid", (TfyhToolbox.isUUID(trip.asset_uuid)) ? dilboTables.tables.assets.iUuids[trip.asset_uuid].name : trip.asset_uuid);
		dilboForm.presetValue("asset_variant", trip.asset_variant);
		
		// split crew
		let crew = trip.crew.split(/\|/g);
		let cox = "";
		let offset = 0;
		// get cox name
		if (trip.coxed) {
			cox = crew[0];
			offset = 1;
			if (crew[0] && TfyhToolbox.isUUID(crew[0])) {
				let person = dilboTables.tables.persons.iUuids[crew[0]];
				if (person)
					cox = person.first_name + " " + person.last_name;
			}
			dilboForm.presetValue("cox", cox);
		}
		// get crew names
		let crewMember;
		for (let i = 1; i <= 8; i++) {
			crewMember = crew[i - 1 + offset];
			if (crewMember && TfyhToolbox.isUUID(crewMember)) {
				let person = dilboTables.tables.persons.iUuids[crewMember];
				if (person)
					crewMember = person.first_name + " " + person.last_name;
			}
			dilboForm.presetValue("crew_" + i, crewMember);
		}
		
		// set the remainder
		let destination = (TfyhToolbox.isUUID(trip.destination)) ? dilboTables.tables.destinations.iUuids[trip.destination].name : trip.destination;
		let waters = (TfyhToolbox.isUUID(trip.waters)) ? dilboTables.tables.waters.iUuids[trip.waters].name : trip.waters;
		dilboForm.presetValue("destination", destination);
		dilboForm.presetValue("distance", trip.distance);
		dilboForm.presetValue("session_type", trip.session_type);
		dilboForm.presetValue("waters", waters);
		dilboForm.presetValue("comments", trip.comments);
	}

	/**
	 * all generic presets for a new session
	 */
	#presetNewSession() {
		dilboForm.presetValue("uid", "" + TfyhToolbox.generateUid(6));
		dilboForm.presetValue("number", "" + dilboTables.getNextNumber("logbook"));
		dilboForm.presetValue("logbookname", userPreferences.logbook);
		dilboForm.presetValue("session_type", "NORMAL");
		let start = TfyhData.format(new Date(), "datetime");
		dilboForm.presetValue("startDate", start.substring(0, 10));
		dilboForm.presetValue("startTime", start.substring(11, 16));
	}

	/**
	 * Open Form "editSession". This will create an empty form, if uid is
	 * undefined, a form for a new session, if uid is a asset, the same form to
	 * edit the session if uid refers to a session, and a session close form, if
	 * uid refers to a session and clos == true.
	 */
	editSession_do(uid, close) {
		// prepare form with defaults and select options
		dilboForm.selectCatalog["session_type"] = dilboConfig.get_by_path(".appSettings.catalogs.session_types");
		dilboForm.selectCatalog["asset_variant"] = dilboConfig.get_by_path(".appSettings.boat_variants");
		dilboForm.headline = "<h3>" + _("BlgYNu|Session") + "</h3>";
		let tripCfg = dilboConfig.get_by_path(".appTables.logbook");
		dilboForm.init("editSession", tripCfg, this.editSession_redo);
		dilboForm.mode = "insert";
		if (uid) {
			this.#recordUid = uid;
			let tablename = dilboTables.iUids[uid];
			if (tablename == 'logbook') {
				// the provided uid points to a session. This is when editing or
				// closing the session
				dilboForm.presetValue("uid", uid);
				let session = dilboTables.tables.logbook.rows[uid];
				dilboForm.mode = "update";
				this.#presetTripData(session);
				if (close) {
					let end = TfyhData.format(new Date(), "datetime");
					dilboForm.presetValue("endDate", end.substring(0, 10));
					dilboForm.presetValue("endTime", end.substring(11, 16));
				}
			} else if (tablename == 'assets') {
				// the uid points to an asset. This is when starting a session
				// from the boats list.
				this.#presetNewSession();
				dilboForm.presetValue("asset_uuid", dilboTables.tables.assets.rows[uid].name);
				dilboForm.presetValue("asset_variant", dilboTables.tables.assets.rows[uid].variant_options.split("|")[0]);
			}
		} else {
			// No uid, i. e. start of a session with unknown boat
			this.#presetNewSession();
		}
		dilboModal.showForm(true, this, dilboForm);
		// always hide the uid
		$('#cFormInput-uid').parent().hide();
		// show or hide the session end fields
		this.#showSessionEnd(close);
		let that = this;
		$('#cFormInput-session_type').change(function() {
			that.#showSessionEnd(close);
		});
		this.#setSessionInputAutocompletion();
		dilboInputValidator.validateEntry($("#cFormInput-asset_uuid")[0], "assets");
		$("#cFormInput-asset_variant").on("change", function() {
			dilboInputValidator.handleValue(this, "assets");
		});
	}

	/**
	 * resolve Uuids and datetime values
	 */
	#processTripData(tripEntered) {
		// resolve asset uuid
		if (tripEntered.asset_uuid) {
			let asset = dilboTables.tables.assets.iNames[tripEntered.asset_uuid];
			if (asset) {
				// set uuid
				tripEntered.asset_uuid = asset.uuid;
				// set variant
				let assetVariantOptions = asset.variant_options.split(/\|/g);
				if (! tripEntered.asset_variant)
					tripEntered.asset_variant = assetVariantOptions[0]; 
				
			} 
			// else keep name.
		}

		delete tripEntered.number;

		// handle crew
		tripEntered.crew = "";
		for (var i = 1; i <= 8; i++) {
			let inputName = "crew_" + i;
			if (tripEntered[inputName]) 
				tripEntered.crew += "|" + this.#resolveUuid(inputName, "persons", tripEntered);
			delete tripEntered[inputName];
		}

		// handle cox
		tripEntered.coxed = "";
		if (tripEntered.cox) {
			tripEntered.crew = this.#resolveUuid("cox", "persons", tripEntered) + tripEntered.crew;
			tripEntered.coxed = "on";
		} else {
			tripEntered.crew = tripEntered.crew.substring(1);
			tripEntered.coxed = "";
		}
		delete tripEntered.cox;

		// resolve destination and waters
		if (tripEntered.destination)
			tripEntered.destination = this.#resolveUuid("destination", "destinations", tripEntered);
		if (tripEntered.waters)
			tripEntered.waters = this.#resolveUuid("waters", "waters", tripEntered);
		
		// add missing end date, if an end time is provided
		if (tripEntered.endTime && !tripEntered.endDate) 
			tripEntered.end = tripEntered.startDate + " " + tripEntered.endTime;
		// but delete end date, if no end time is provided
		else if (!tripEntered.endTime)
			delete tripEntered.endDate;
		
		// also delete start date, if no start time is provided
		if (!tripEntered.startTime)
			delete tripEntered.startDate;
		
		// join date and time for start and end
		if (tripEntered.startDate) 
			tripEntered.start = tripEntered.startDate + " " + tripEntered.startTime;
		if (tripEntered.endDate) 
			tripEntered.end = tripEntered.endDate + " " + tripEntered.endTime;
		for (let part of [ "startDate", "startTime", "endDate", "endTime"])
			delete tripEntered[part];
		
		return tripEntered;
	}

	/**
	 * Form "editSession" is submitted. Parse result and store it. Distinguish a
	 * new trip start from a change or the trip end entry by checking for the
	 * existence of the used entry id.
	 */
	editSession_done() {
		// even in case of update use the full record, because times are preset
		// for convenience and may therefore no have been changed by the user.
		let tripEntered = dilboForm.getParsedReformatted(true);
		// make sure the entry Id is computed at the server side to avoid
		// conflicts.
		if (dilboForm.mode == "insert")
			tripEntered.uid = TfyhToolbox.generateUid(6);
		else 
			tripEntered.uid = this.#recordUid;
		// process the data, i. e. replace
		tripEntered = this.#processTripData(tripEntered);
		// merge trip to local table (without trip number)
		dilboTables.mergeRecord(tripEntered, dilboConfig.get_by_path(".appTables.logbook"), 
				dilboTables.tables.logbook, true);
		// autonotification works for all listeners to "logbook", but add
		// notification to listeners on "boats", because their status will also
		// change.
		dilboTables.notifyListeners("assets");
		// send trip to server.
		if (dilboTxQueue)
			dilboTxQueue.addNewTxToPending(dilboForm.mode, "logbook", tripEntered, null, this.editSession_redo);
		else 
			this.editRecord_done();
	}

	/**
	 * Form "startSession" had errors. Try again
	 */
	editSession_redo(tx) {
		dilboModal.showForm(true, dilboFormHandler, dilboForm, (tx) ? tx.resultMessage : dilboForm.formErrors);
		// always hide the uid
		$('#cFormInput-uid').parent().hide();
		dilboFormHandler.#setSessionInputAutocompletion();
	}

	/**
	 * Do all name checks: name is valid identifier, not reserved and not
	 * duplicate. Local name is not duplicate. If the name argument is not set,
	 * the respective check is skipped.
	 */
	#isOkNames(parent, nameToCheck, localNameToCheck) {
		// check validity of name
		if (nameToCheck) {
			if (TfyhValidate.apply_rule(nameToCheck, "identifier").length > 0) {
				dilboModal.showForm(true, this, dilboForm, _("GWrjW8|Invalid child name %1", nameToCheck));
				return false;
			}
			if (parent.has_child(nameToCheck)) {
				dilboModal.showForm(true, this, dilboForm, _("uTPBBb|Duplicate child name %1", nameToCheck));
				return false;
			}
			if (TfyhData.is_reserved_name(nameToCheck)) {
				dilboModal.showForm(true, this, dilboForm, _("S81unZ|Forbidden to use reserve...", nameToCheck));
				return false;
			}
		}
		// check for duplicate local name
		if (localNameToCheck) {
			for (let childname in parent.get_children()) {
				if (localNameToCheck.localeCompare(
						parent.get_child(childname).get_descriptor_field("text_local_name")) == 0) {
					dilboModal.showForm(true, this, dilboForm, _("2IFi0b|Duplicate local name %1", localNameToCheck));
					return false;
				}
			}
		}
		// all was fine, return true.
		return true;
	}

	// ==============
	// damage editing
	// ==============
	
	/**
	 * Set basic fields for damage form
	 */
	#setDamageInputAutocompletion() {
		let input = $("#cFormInput-asset_uuid")[0];
		TfyhAutoComplete.set(input, dilboInputValidator, dilboTables.validNames.assets, "assets", this);
		input = $("#cFormInput-reported_by")[0];
		TfyhAutoComplete.set(input, dilboInputValidator, dilboTables.validNames.persons, "persons", this);
	}

	/**
	 * resolve Uuids and datetime values
	 */
	#processDamageData(damageEntered) {
		// resolve Uuids
		let asset = dilboTables.tables.assets.iNames[damageEntered.asset_uuid];
		if (asset) 
			damageEntered.asset_uuid = asset.uuid;
		// else keep name.
		damageEntered.reported_by = this.#resolveUuid("reported_by", "persons", damageEntered);
		return damageEntered;
	}

	/**
	 * Open Form "reportDamage".
	 */
	reportDamage_do(uid) {
		// prepare form with defaults and select options
		dilboForm.selectCatalog["severity"] = dilboConfig.get_by_path(".appSettings.catalogs.damage_severities");
		dilboForm.headline = "<h3>" + _("oZkMb8|Damage") + "</h3>";
		let damageCfg = dilboConfig.get_by_path(".appTables.damages");
		dilboForm.init("reportDamage", damageCfg);
		dilboForm.presetValue("number", "" + dilboTables.getNextNumber("damages"));
		dilboForm.presetValue("reported_on", TfyhData.format(new Date()), "datetime", "csv");
		if (uid) 
			dilboForm.presetValue("asset_uuid", dilboTables.tables.assets.rows[uid].name);
		dilboModal.showForm(true, this, dilboForm);
		this.#setDamageInputAutocompletion();
		dilboInputValidator.validateEntry($("#cFormInput-asset_uuid")[0], "assets");
	}

	/**
	 * Form "reportDamage" is submitted. Parse result and store it.
	 */
	reportDamage_done() {
		let damageEntered = dilboForm.getParsedReformatted(true);
		// make sure the entry Id is computed at the server side to avoid
		// conflicts.
		damageEntered.uid = TfyhToolbox.generateUid(6);
		// process the data, i. e. replace
		damageEntered = this.#processDamageData(damageEntered);
		// merge trip to local table (without trip number)
		dilboTables.mergeRecord(damageEntered, dilboConfig.get_by_path(".appTables.damages"), 
				dilboTables.tables.damages, true);
		// autonotification works for all listeners to "damages", but add
		// notification to listeners on "boats", because their status will also
		// change.
		dilboTables.notifyListeners("assets");
		// send trip to server.
		let tx = dilboTxQueue.addNewTxToPending("insert", "damages", damageEntered, null, this.reportDamage_redo);
	}

	/**
	 * Form "reportDamage" had errors. Try again
	 */
	reportDamage_redo(tx) {
		dilboModal.showForm(true, dilboFormHandler, dilboForm, tx.resultMessage);
		dilboFormHandler.#setDamageInputAutocompletion();
		dilboInputValidator.validateEntry($("#cFormInput-asset_uuid")[0], "assetVariants");
	}

	// ===============
	// message editing
	// ===============
	
	/**
	 * Set basic fields for damage form
	 */
	#setMessageInputAutocompletion() {
		let input = $("#cFormInput-from")[0];
		TfyhAutoComplete.set(input, dilboInputValidator, dilboTables.validNames.persons, "persons", this);
	}

	/**
	 * resolve Uuids and datetime values
	 */
	#processMessageData(messageEntered) {
		messageEntered.from = this.#resolveUuid("from", "persons", messageEntered);
		return messageEntered;
	}

	/**
	 * Open Form "sendMessage".
	 */
	sendMessage_do() {
		// prepare form with defaults and select options
		dilboForm.selectCatalog["to"] = dilboConfig.get_by_path(".appSettings.catalogs.message_destinations");
		dilboForm.headline = "<h3>" + _("3lfLrL|Message") + "</h3>";
		let messageCfg = dilboConfig.get_by_path(".appTables.messages");
		dilboForm.init("sendMessage", messageCfg);
		dilboForm.presetValue("number", "" + dilboTables.getNextNumber("messages"));
		dilboForm.presetValue("sent", TfyhData.format(new Date()), "datetime", "csv");
		dilboModal.showForm(true, this, dilboForm);
		this.#setMessageInputAutocompletion();
	}

	/**
	 * Form "sendMessage" is submitted. Parse result and store it.
	 */
	sendMessage_done() {
		let messageEntered = dilboForm.getParsedReformatted(true);
		// make sure the entry Id is computed at the server side to avoid
		// conflicts.
		messageEntered.uid = TfyhToolbox.generateUid(6);
		// process the data, i. e. replace
		messageEntered = this.#processMessageData(messageEntered);
		// merge trip to local table (without trip number)
		dilboTables.mergeRecord(messageEntered, dilboConfig.get_by_path(".appTables.messages"), 
				dilboTables.tables.messages, true);
		// send message to server.
		let tx = dilboTxQueue.addNewTxToPending("insert", "messages", messageEntered, null, this.sendMessage_redo);
	}

	/**
	 * Form "sendMessage" had errors. Try again
	 */
	sendMessage_redo(tx) {
		dilboModal.showForm(true, dilboFormHandler, dilboForm, tx.resultMessage);
		dilboFormHandler.#setMessageInputAutocompletion();
		dilboInputValidator.validateEntry($("#cFormInput-asset_uuid")[0], "assetVariants");
	}

	// ===================
	// reservation editing
	// ===================
	
	/**
	 * Set basic fields for reservation form
	 */
	#setReservationInputAutocompletion() {
		let input = $("#cFormInput-asset_uuid")[0];
		TfyhAutoComplete.set(input, dilboInputValidator, dilboTables.validNames.assets, "assets", this);
		input = $("#cFormInput-person_uuid")[0];
		TfyhAutoComplete.set(input, dilboInputValidator, dilboTables.validNames.persons, "persons", this);
	}

	/**
	 * resolve Uuids and datetime values
	 */
	#processReservationData(reservationEntered) {
		// resolve Uuids
		let asset = dilboTables.tables.assets.iNames[reservationEntered.asset_uuid];
		if (asset) 
			reservationEntered.asset_uuid = asset.uuid;
		// else keep name.
		reservationEntered.person_uuid = this.#resolveUuid("person_uuid", "persons", reservationEntered);
		if (reservationEntered.startDate)
			reservationEntered.start = reservationEntered.startDate + " 0" + reservationEntered.startTime + ":00";
		if (reservationEntered.endDate) 
			reservationEntered.end = reservationEntered.endDate + " 0" + reservationEntered.endTime + ":00";
		delete reservationEntered.startDate;
		delete reservationEntered.startTime;
		delete reservationEntered.endDate;
		delete reservationEntered.endTime;
		return reservationEntered;
	}

	/**
	 * Open Form "requestReservation".
	 */
	requestReservation_do(uid) {
		// prepare form with defaults and select options
		let reservationCfg = dilboConfig.get_by_path(".appTables.reservations");
		dilboForm.headline = "<h3>" + _("oW68Gp|Reservation") + "</h3>";
		dilboForm.init("requestReservation", reservationCfg);
		dilboForm.presetValue("number", "" + dilboTables.getNextNumber("reservations"));
		dilboForm.presetValue("type", "ONETIME");
		if (uid) 
			dilboForm.presetValue("asset_uuid", dilboTables.tables.assets.rows[uid].name);
		dilboModal.showForm(true, this, dilboForm);
		this.#setReservationInputAutocompletion();
		dilboInputValidator.validateEntry($("#cFormInput-asset_uuid")[0], "assets");
	}

	/**
	 * Form "requestReservation" is submitted. Parse result and store it.
	 */
	requestReservation_done() {
		let reservationEntered = dilboForm.getParsedReformatted(true);
		// make sure the entry Id is computed at the server side to avoid
		// conflicts.
		reservationEntered.uid = TfyhToolbox.generateUid(6);
		// process the data, i. e. replace
		reservationEntered = this.#processReservationData(reservationEntered);
		// merge trip to local table (without trip number)
		dilboTables.mergeRecord(reservationEntered, dilboConfig.get_by_path(".appTables.reservations"), 
				dilboTables.tables.reservations, true);
		// autonotification works for all listeners to "reservations", but add
		// notification to listeners on "boats", because their status will also
		// change.
		dilboTables.notifyListeners("assets");
		// send damage to server.
		let tx = dilboTxQueue.addNewTxToPending("insert", "reservations", reservationEntered, null, this.requestReservation_redo);
	}

	/**
	 * Form "requestReservation" had errors. Try again
	 */
	requestReservation_redo(tx) {
		dilboModal.showForm(true, dilboFormHandler, dilboForm, tx.resultMessage);
		dilboFormHandler.#setReservationInputAutocompletion();
		dilboInputValidator.validateEntry($("#cFormInput-asset_uuid")[0], "assets");
	}

	// ========================
	// user preferences editing
	// ========================

	/**
	 * Open Form "editPreferences".
	 */
	editPreferences_do(uid) {
		// prepare form with defaults and select options
		dilboForm.headline = "<h3>" + _("v2J1hT|User preferences") + "</h3>";
		dilboForm.selectCatalog["logbook"] = dilboConfig.get_by_path(".clubSettings.logbooks");
		dilboForm.selectCatalog["workbook"] = dilboConfig.get_by_path(".clubSettings.workbooks");
		dilboForm.selectCatalog["profile"] = dilboConfig.get_by_path(".uiSettings.profiles");
		dilboForm.init("editPreferences");
		dilboForm.presetValues(userPreferences);
		dilboModal.showForm(true, this, dilboForm);
	}

	/**
	 * Form "editPreferences" is submitted. Parse result and store it.
	 */
	editPreferences_done() {
		let preferencesEntered = dilboForm.getParsedReformatted(true);
		let preferencesString = "";
		for (let preferenceEntered in preferencesEntered) {
			preferencesString += ";" + preferenceEntered + "=" + preferencesEntered[preferenceEntered]; 
		}
		let userRecord = {
				uid : user.uid,
				preferences : preferencesString.substring(1)
		}
		// send preferences to server.
		let tx = dilboTxQueue.addNewTxToPending("update", "persons", userRecord, this.editPreferences_stored, this.editPreferences_redo);
		tx = dilboTxQueue.addNewTxToPending("session", "close", null, null, null);
	}
	
	/**
	 * After change of the user settings, the application has to reload
	 */
	editPreferences_stored() {
		window.location.reload();
	}

	/**
	 * Form "editPreferences" had errors. Try again
	 */
	editPreferences_redo(tx) {
		dilboModal.showForm(true, dilboFormHandler, dilboForm, tx.resultMessage);
	}

	// ====================
	// table record editing
	// ====================

	/**
	 * create an empty record with all mandatory fields for this table.
	 */
	createRecord(tablename) {
		this.#recordEntered = { uid : TfyhToolbox.generateUid(6) }; 
		this.#recordConfig = dilboConfig.get_by_path(".appTables." + tablename);
		if (this.#recordConfig.has_child("uuid"))
			this.#recordEntered.uuid = TfyhToolbox.generateUUID();
		if (this.#recordConfig.has_child("valid_from"))
			this.#recordEntered.valid_from = Math.floor(Date.now() / 1000);
		if (this.#recordConfig.has_child("invalid_from"))
			this.#recordEntered.invalid_from = TfyhData.forever_seconds;
		return this.#recordEntered.uid;
	}
	
	/**
	 * Open Form as defined in the configuration.
	 */
	editRecord(tablename, uid, addRecord) {

		this.#recordUid = uid;
		this.#recordEditMode = (addRecord) ? 1 : 2;
		this.#recordTablename = tablename;
		try {
			if (php_record_as_csv) 
				this.#recordEntered = TfyhToolbox.csvToObject(php_record_as_csv)[0];
		} catch (e) {}
		// Cache the default template.
		this.#formTemplateCache = $_formTemplates.editRecord;
		// get the item specific template or a minimum one.
		this.#recordConfig = dilboConfig.get_by_path(".appTables." + tablename);
		$_formTemplates.editRecord = this.#recordConfig.getEditForm();
		if (!$_formTemplates.editRecord) 
			$_formTemplates.editRecord = this.#formTemplateCache;
		let tablenameLocal = dilboConfig.get_by_path(".appTables." + tablename).get_descriptor_field("text_local_name");
		// prepare form with defaults and select options
		dilboForm.headline = "<h3>" + _("wNsRmC|Edit record of %1", tablenameLocal) + "</h3>";
		if (this.#recordEntered.valid_from)
			dilboForm.headline += this.#validity_notice(); 
		dilboForm.selectCatalog["logbook"] = dilboConfig.get_by_path(".clubSettings.logbooks");
		dilboForm.selectCatalog["workbook"] = dilboConfig.get_by_path(".clubSettings.workbooks");
		dilboForm.selectCatalog["profile"] = dilboConfig.get_by_path(".uiSettings.profiles");
		dilboForm.init("editRecord", this.#recordConfig);
		if (this.#recordEntered)
			dilboForm.presetValues(this.#recordEntered);
		// show the form, or redo the form with the modification errors.
		dilboModal.showForm(true, this, dilboForm, this.#recordModifyErrors);
		// once shown the modificaltion errors are cleared.
		this.#recordModifyErrors = "";
		// add autocompletion options for validation
		// use the form definition as field collection, not the record.
		// The record is empty for new record insertion.
		this.#lookupSet = [];
		for (let field of dilboForm.getFieldNames()) {
			if (this.#recordConfig.has_child(field)) {
				let fieldConfig = this.#recordConfig.get_child(field);
				let inputType = fieldConfig.get_descriptor_field("input_type");
				if (inputType.startsWith("autocomplete") || inputType.startsWith("autoclist")) {
					let fromWhat = inputType.substring(inputType.indexOf("from") + 5).trim();
					let tablename = fromWhat.replace(".uuid", "");
					this.#lookupSet.push({ tablename : tablename, field : field, table : dilboTables.tables[tablename] });
				}
			}
		}
		// The form is prepared, now add the lookups, load asynchronously.
		if (this.#lookupSet.length > 0) {
			this.#lookupLoadIndex = this.#lookupSet.length; 
			this.#loadLookup();
		}
		else
			this.#presetRecordUuidsToNames();
	}

	/**
	 * create a validity notice used for versionized records for display on top
	 * of the form.
	 */
	#validity_notice() {
		let validFromDate = TfyhToolbox.timef2string(this.#recordEntered.valid_from, languageCode);
		let invalidFromDate = TfyhToolbox.timef2string(this.#recordEntered.invalid_from, languageCode);
		return "<p>" + _("dY3WT2|valid from") + ": <b>" + validFromDate + "</b> " + _("FkesLE|until") + " <b>" + invalidFromDate + "</b><p>";
	}
	
	/**
	 * The uuid replacement to names. Can only happen, if the respective
	 * uuid2names list is loaded. THis will therefore be executed after loading.
	 */
	#presetRecordUuidsToNames() {
		// all lookups have been loaded, now replace uuids where necessary
		for (let field in this.#recordEntered) {
			if (this.#recordConfig.has_child(field)) {
				let fieldConfig = this.#recordConfig.get_child(field);
				let inputType = fieldConfig.get_descriptor_field("input_type");
				if (this.#recordEntered[field] && inputType.startsWith("autocomplete")) {
					let fromWhat = inputType.substring(inputType.indexOf("from") + 5).trim();
					let tablename = fromWhat.replace(".uuid", "");
					let uuid = this.#recordEntered[field];
					let row = dilboTables.tables[tablename].iUuids[uuid];
					let input = $('#cFormInput-' + field)[0];
					// replace input value by the name, if the uuid could be
					// resolved
					if (row) {
						let name = (tablename == 'persons') ? row.first_name + ' ' + row.last_name : row.name;
						dilboForm.presetValue(field, name);
						$(input).val(name);
					}
					TfyhAutoComplete.set(input, dilboInputValidator, dilboTables.validNames[tablename], tablename, dilboFormHandler);
				}
			}
		}
	}
	
	/**
	 * Load a lookup list for form validation. This will be called
	 * asynchronously
	 */
	#loadLookup() {
		// a lookup has been loaded. On the first call of the sequence, he if
		// clause is skipped
		if (dilboFormHandler.#lookupLoadIndex < dilboFormHandler.#lookupSet.length) {
			let tablename = dilboFormHandler.#lookupSet[dilboFormHandler.#lookupLoadIndex].tablename;
			let input = $("#cFormInput-" + dilboFormHandler.#lookupSet[dilboFormHandler.#lookupLoadIndex].field)[0];
			TfyhAutoComplete.set(input, dilboInputValidator, dilboTables.validNames[tablename], tablename, dilboFormHandler);
		}
		dilboFormHandler.#lookupLoadIndex --;
		// return from loop when all lookups habe been loaded.
		if (dilboFormHandler.#lookupLoadIndex < 0) {
			dilboFormHandler.#presetRecordUuidsToNames();
			return;
		}

		if (!dilboFormHandler.#lookupSet[dilboFormHandler.#lookupLoadIndex].table) {
			// asynchronous loading will return to this function after
			// completion
			let tablename = dilboFormHandler.#lookupSet[dilboFormHandler.#lookupLoadIndex].tablename;
			dilboTables.getNameList(tablename, dilboFormHandler.#loadLookup);
		}
		else
			// if the table was loaded, skip it.
			dilboFormHandler.#loadLookup();
	}

	/**
	 * The names resolving to uuids.
	 */
	#resolveRecordNamesToUuids() {
		// all lookups have been loaded, now replace uuids where necessary
		for (let field in this.#recordEntered) {
			if (this.#recordConfig.has_child(field)) {
				let fieldConfig = this.#recordConfig.get_child(field);
				let inputType = fieldConfig.get_descriptor_field("input_type");
				if (this.#recordEntered[field] && inputType.startsWith("autocomplete")) {
					let fromWhat = inputType.substring(inputType.indexOf("from") + 5).trim();
					let tablename = fromWhat.replace(".uuid", "");
					let row = dilboTables.tables[tablename].iNames[this.#recordEntered[field]];
					if (row && row.uuid)
						this.#recordEntered[field] = row.uuid;
				}
			}
		}
	}
	
	/**
	 * Modify the records validity
	 */
	editValidity(action, tablename, uid) {
		let formTemplate = $_formTemplates[action];
		this.#recordTablename = tablename;
		this.#recordConfig = dilboConfig.get_by_path(".appTables." + tablename);
		this.#recordEntered = TfyhToolbox.csvToObject(php_record_as_csv)[0];
		dilboForm.init(action, this.#recordConfig);
		let validFrom = (action == "newPeriod") ? Math.floor(Date.now() / 1000) : parseInt(this.#recordEntered.valid_from);
		this.#recordEntered.valid_from_date = TfyhToolbox.timef2string(validFrom).substring(0, 10);
		this.#recordEntered.invalid_from_date = TfyhToolbox.timef2string().substring(0, 10);
		dilboForm.presetValues(this.#recordEntered);
		dilboForm.headline = "<h4>" + this.validityModifications[action] + "</h4>";
		// show the form, or redo the form with the modification errors.
		dilboModal.showForm(true, this, dilboForm, this.#recordModifyErrors);
	}
	
	/**
	 * Return paths are different for the different validity editing actions.
	 * Redirect them here.
	 */
	editValidfrom_done() { this.#editValidityDone(false); } 
	extendPeriod_done() { this.#editValidityDone(false); } 
	limitPeriod_done() { this.#editValidityDone(false); } 
	newPeriod_done() { this.#editValidityDone(true); } 

	/**
	 * execute th requested validity change
	 */
	#editValidityDone(createNew) {
		let recordEntered = dilboForm.getParsedReformatted(true);
		recordEntered.uid = this.#recordEntered.uid 
		if (recordEntered.valid_from_date) {
			// editValidfrom
			recordEntered.valid_from = Math.floor(new Date(recordEntered.valid_from_date).getTime() / 1000);
			delete recordEntered.valid_from_date;
		} 
		else if (recordEntered.invalid_from_date) {
			// limitPeriod or newPeriod
			recordEntered.invalid_from = Math.floor(new Date(recordEntered.invalid_from_date).getTime() / 1000);
			delete recordEntered.invalid_from_date;
		} else {
			// extendPeriod
			recordEntered.invalid_from = TfyhData.forever_seconds;
		}
		if (createNew && recordEntered.valid_from) {
			let periodSplit = recordEntered.valid_from;
			// limit the current period, no callback.
			this.#recordEntered.invalid_from = periodSplit;
			this.postRecordModify(2, this.#recordTablename, this.#recordEntered);
			// insert a new period
			this.#recordEntered.valid_from = recordEntered.valid_from;
			this.#recordEntered.invalid_from = TfyhData.forever_seconds;
			this.#recordEntered.history = "REMOVE!";
			this.#recordEntered.uid = TfyhToolbox.generateUid(6);
			this.postRecordModify(1, this.#recordTablename, this.#recordEntered, this.newPeriod_stored);
		}
		this.postRecordModify(this.#recordEditMode, this.#recordTablename, recordEntered, this.editValidity_stored)
	}
	
	/**
	 * Delete a record. This just shows a confirmation dialog.
	 */
	deleteRecord(tablename, uid) {
		this.#recordEntered = TfyhToolbox.csvToObject(php_record_as_csv)[0];
		this.#recordTablename = tablename;
		dilboForm.init("deleteRecord", this.#recordConfig);
		let recordname = (tablename == 'persons') ? this.#recordEntered.first_name + " " + this.#recordEntered.last_name : this.#recordEntered.name;  
		this.#recordConfig = dilboConfig.get_by_path(".appTables." + tablename);
		let tablenameLocal = this.#recordConfig.get_descriptor_field("text_local_name"); 
		
		dilboForm.headline = "<h4>" + _("Really delete record '%1' from '%2'?", recordname, tablenameLocal) + "</h4>";
		// show the form, or redo the form with the modification errors.
		dilboModal.showForm(true, this, dilboForm, this.#recordModifyErrors);
	}
	
	/**
	 * Form "editRecord" is submitted. Parse result and store it.
	 */
	editRecord_done() {
		let recordEntered = dilboForm.getParsedReformatted(false);
		recordEntered.uid = this.#recordUid;
		// merge into the exiting record, because the existing record may not
		// have displayed all fields within the form
		this.#recordEntered = { ...this.#recordEntered, ...recordEntered };
		this.#resolveRecordNamesToUuids();
		this.#recordEntered.uid = this.#recordUid;
		this.postRecordModify(this.#recordEditMode, this.#recordTablename, this.#recordEntered, this.editRecord_stored)
	}
	
	/**
	 * Execute record deletion
	 */
	deleteRecord_done() {
		this.postRecordModify(3, this.#recordTablename, this.#recordEntered, this.editRecord_stored);
	}
	
	/**
	 * Form "editRecord" is stored.
	 */
	editRecord_stored(success, responseText) {
		if (success)
			window.location.reload();
		else {
			dilboFormHandler.#recordModifyErrors = responseText;
			dilboFormHandler.editRecord(dilboFormHandler.#recordTablename, dilboFormHandler.#recordUid);
		}
	}
	
	/**
	 * Form "editRecord" is stored.
	 */
	editValidity_stored(success, responseText) {
		if (success)
			window.location.reload();
		else 
			dilboDialog.showHtml ("<h4>" + _("Failed") + "</h4><p>" + responseText + "</p>", [ i("Noted") ]);
	}
	
	/**
	 * Form "editRecord" is stored.
	 */
	newPeriod_stored(success, responseText) {
		if (success)
			window.location.href("../pages/view_record.php?table=" + this.#recordTablename + "&uid=" + this.#recordUid);
		else 
			dilboDialog.showHtml ("<h4>" + _("Failed") + "</h4><p>" + responseText + "</p>", [ i("Noted") ]);
	}
	
	/**
	 * After change of the user settings, the application has to reload
	 */
	editPreferences_stored() {
		window.location.reload();
	}

	/**
	 * Form "editPreferences" had errors. Try again
	 */
	editPreferences_redo(tx) {
		dilboModal.showForm(true, dilboFormHandler, dilboForm, tx.resultMessage);
	}

	// =========================
	// edit a list in the dialog
	// =========================
	
	/**
	 * Show the dialog with the option to edit the list
	 */
	editList(fieldname) {
		let fieldListHtml = "";
		let list = dilboForm.getValue(fieldname).split(/\|/g);
		for (let i = 1; i <= list.length; i++) {
			let fieldHtml = dilboForm.getListElementField(fieldname, list, i);
			fieldListHtml += this.#listElementFrame.replace(/\{index\}/g, i).replace(/\{listElement\}/g, fieldHtml);
		}
		dilboDialog.showHtml(fieldListHtml + this.#listElementAdd, [ _("Done"),  _("Cancel") ], this.editListDone, fieldname);
		// add lookup and replace uuids by names
		let tablename = false;
		for (let lookup of this.#lookupSet) {
			if (lookup.field == fieldname) {
				tablename = lookup.tablename; 
				for (let i = 1; i <= list.length; i++) {
					let input = $("#cFormInput-" + fieldname + "_" + i)[0];
					let uuid = input.value;
					let row = dilboTables.tables[tablename].iUuids[uuid];
					// replace input value by the name, if the uuid could be
					// resolved
					if (row) {
						let name = (tablename == 'persons') ? row.first_name + ' ' + row.last_name : row.name;
						$(input).val(name);
					}
					TfyhAutoComplete.set(input, dilboInputValidator, dilboTables.validNames[tablename], tablename, dilboFormHandler);
				}
			}
		}
		// add removeElement events
		$('.removeElement').click(function() {
			// remove both the htnl-form input and the dilboForm field.
			let id = $(this).attr("id");
			let i = parseInt(id.split("-")[1]);
			dilboForm.removeField(fieldname + "_" + i);
			$("#listElementBox-" + i).remove();
		});
		// add addElement event
		this.bindAddElelementEvent(fieldname, tablename, list);
	}
	
	/**
	 * Bind the add element event. Because the event will triggr renewal of the
	 * add-Button, ths needs to be separate.
	 */
	bindAddElelementEvent(fieldname, tablename, list) {
		let that = this;
		$('.addElement').click(function() {
			let boxes = $('div[id^="listElementBox-"]');
			// get the one before last, since last is the add box
			let lastInputBox = boxes[boxes.length - 2]; 
			let id = $(lastInputBox).attr("id");
			let next = parseInt(id.split("-")[1]) + 1;
			let fieldHtml = dilboForm.getListElementField(fieldname, list, next);
			fieldHtml = that.#listElementFrame.replace(/\{index\}/g, next).replace(/\{listElement\}/g, fieldHtml);
			$('#listElement-add').remove();
			$(lastInputBox).parent().append(fieldHtml + that.#listElementAdd);
			that.bindAddElelementEvent(fieldname, tablename, list);
			let input = $("#cFormInput-" + fieldname + "_" + next)[0];
			if (tablename)
				TfyhAutoComplete.set(input, dilboInputValidator, dilboTables.validNames[tablename], tablename, dilboFormHandler);
		});
	}

	/**
	 * The list editing is completed, get the value.
	 */
	editListDone(button, fieldname) {
		if (button == "buttonRight")
			return;
		let list = "";
		let inputs = $('select[name^="' +fieldname + '_"], input[name^="' +fieldname + '_"]');
		let fieldConfig = ((dilboFormHandler.#recordConfig.has_child(fieldname))) 
							? dilboFormHandler.#recordConfig.get_child(fieldname) : false;
		let inputType = fieldConfig.get_descriptor_field("input_type");
		inputs.each(function(input) {
			let id = $(this).attr("id");
			let i = id.replace("cFormInput-" + fieldname + "_", "");
			let element = dilboForm.readField(fieldname + "_" + i);
			// resolve names to uuids, if needed
			if (inputType.startsWith("autoclist")) {
				let fromWhat = inputType.substring(inputType.indexOf("from") + 5).trim();
				let tablename = fromWhat.replace(".uuid", "");
				let row = dilboTables.tables[tablename].iNames[element];
				if (row && row.uuid)
					element = row.uuid;
			}
			list += "|" + element; 
			dilboForm.removeField(fieldname + "_" + i);
		});
		dilboForm.presetValue(fieldname, list.substring(1));
		dilboForm.setChanged(fieldname);
		dilboForm.updateField(fieldname);
	}
	
	// ========================
	// submission and handshake
	// ========================

	/**
	 * Submit the form. Usually this will be handled by the modal. But in case
	 * he autocomplete class triggers the submission, it uses the form handler.
	 */
	submitForm() {
		let formErrors = dilboForm.evaluate(false);  // values will never end
												// unquoted in SQL-statements;
		if (formErrors && formErrors.length > 1) {
			dilboModal.showForm(true, this, dilboForm, formErrors);
		} else {
			dilboModal.hide();
			// select the handling function by using the form name.
			this[form.getName() + "_done"]();
		}
	}
	
	/**
	 * post a modification transaction to the server. mode is 11, 12, 13 for
	 * insert, update, delete a record. The record must contain all necessary
	 * information, including the uid.
	 */
	postRecordModify(mode, tablename, record, callback) {
		// Assign handlers immediately after making the request,
		// and remember the jqxhr object for this request
		let that = this;
		let postData = {
				// for record manipulation add 10 the the mode
				mode : (10 + mode),
				tablename : tablename,
		}
		for (let field in record)
			if ((record[field] != null) && (typeof record[field] !== 'undefined'))
				postData[field] = record[field];
		console.log("Data modify post: mode = " + (10 + mode) 
				+ ", table = " + tablename + "(" + record.uid + ")");
		var jqxhr = $
			.post( "../forms/jspost.php", postData)
			.done(function(done) {
				that.#onPostDone(done, callback);
			})
			.fail(function(fail) {
				that.#onPostFail(fail, callback);
			});
	}
	
	/**
	 * called after successful completion.
	 */
	#onPostDone(done, callback) {
		let responseParts = done.split(";");
		let success = ((parseInt(responseParts[0]) < 400) && (parseInt(responseParts[0]) > 0)); 
		if (success)
			console.log("Data modifification completed.");
		else
			console.log("Data modifification error " + responseParts[0] + ": " + responseParts[1]);
		if (typeof callback == 'function')
			callback(success, done);
	}

	/**
	 * called after failed competeion.
	 */
	#onPostFail(fail, callback) {
		console.log("Configuration modify failed. Error " + fail.responseText);
		if (typeof callback == 'function')
			callback(false, fail.responseText);
	}
	
	/**
	 * Callback-function after a local item modification form had been posted.
	 */
	onPostItemCallback(success, responseText) {
		let responseParts = responseText.split(";");
		let responsecode = (responseParts && (responseParts.length > 1)) ? parseInt(responseParts[0]) : 500; 
		if (! success) {
			// go for alert instead of modal display to wait for the users
			// repsonse
			alert(_("HdQh9C|Form submission failure.") + "\n" + responseText);
			// and reload afterwards the page to display what has reached
			// through to the server.
			window.location.reload();
		}
		else
			// in case of success it is sufficient to refresh the panel.
			dilboConfig.refreshPanel();
	}
	
}
