/**  
 *
 *       dilbo - digital logbook for Rowing and Canoeing
 *       -----------------------------------------------
 *       https://www.dilbo.org
 *
 * Copyright  2023-2024  Martin Glade
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

/**
 * Copyright (c) 2024 by Martin Glade, see https://www.dilbo.org/. License: GNU
 * General Public License v2
 */

// core modules
var dilboConfig = false;
var dilboTables = false;

// configuration manager
var dilboCfgPanel;

// dilbo-webUI
var dilboContainer = false;
var dilboMenu;
var dilboModal;
var dilboDialog;
var dilboInputValidator;
var dilboForm;

// dilbo forms and events
var dilboFormHandler;

// the API for server access
var dilboTxQueue = false;
var apiTxHandler = false;

// session based settings
var modeWebApp = false;
var modeConfigEdit = false;
var user = {};
var userPreferences = {};
var languageCode = "de";
var api_session_id;

// global varable definitions for the application
const globalSelectOptions = {
	providerTypes : "none=please select;menu=Menu;eventList=Event list;boatList=Boat list;statistics=Statistics;htmlView=HTML view;clock=Clock",
	sportsTypes : "rowing=Rowing;canoeing=Canoeing",
	assetTypes : "BOAT=boat;TRAILER=trailer;OTHER=other",
	nameFormats : "FIRSTLAST=first last;LASTFIRST=last, first",
	sportsYearStarts : "1=1;4=4;7=7;10=10",
	languageCodes : "en=English;de=Deutsch",
	timezones : "1=Berlin",
	reservationHours : "5=5:00;6=6:00;7=7:00;8=8:00;9=9:00;10=10:00;11=11:00;12=12:00;13=13:00;14=14:00;15=15:00;16=16:00;17=17:00;18=18:00;19=19:00;20=20:00;21=21:00;22=22:00;23=23:00"
}

/**
 * Read the user preferences for the UI.
 */
function includeUserSettings() {
	identifyPreferencesToUse();
	languageCode = (userPreferences.language) ? userPreferences.language : "de";
	// initialize internationalization.
	TfyhI18n.i18n_init(languageCode, initModules);
}

/**
 * Initialize all global variable and in particular server side passed
 * information.
 */
function initGlobals() {
	modeConfigEdit = ($('#tfyhCfg-container').length > 0);
	modeWebApp = !modeConfigEdit && ($('#tfyhMain').length > 0);
	try {
		let userData = php_user_record.split(/\|/g);
		for (let userField of userData) 
			user[userField.split("=")[0]] = userField.substring(userField.indexOf("=") + 1);
	} catch (e) {
		user = {};
	}
	try {
		api_session_id = php_api_session_id;
	} catch (e) {
		api_session_id = "";
	}
}

/**
 * Initialize the modules one by one. This is a chain of partly asynchronous
 * activities.
 */
function initModules() {
	if (dilboConfig === false) {
		// load configuration first.
		dilboConfig = new TfyhConfig();
		dilboConfig.init(); // calls initModules() after asynchronous completion
	} else if (TfyhI18n.loaded === false) {
		// load the user settings and the language pack
		includeUserSettings();
	} else if (modeWebApp && (dilboTables === false)) {
		dilboTxQueue = new DilboTxQueue(user.user_id, api_session_id);
		dilboTables = new DilboTables(true);
		// dilboTables calls init_modules upon completion
	} else if (modeWebApp && (dilboContainer === false)) {
		// initialize the UI afterwards.
		dilboContainer = TfyhContainer.buildLayout(dilboConfig.get_by_path(".uiSettings"));
		initModules();
	} else {
		$('#dilbo-load').hide();
		// for server side form editing a dilboTables instance is needed to
		// provide lookup and period validation functionality.
		if (dilboTables === false)
			dilboTables = new DilboTables(false);
		// initialize the modal user interface
		dilboModal = new TfyhModal();
		dilboDialog = new TfyhDialog();
		dilboInputValidator = new DilboInputValidator();
		dilboForm = new TfyhForm(dilboConfig);
		dilboFormHandler = new DilboFormHandler();
		// initialize the configuration edit/show panel, if requested
		if (modeConfigEdit) {
			let panelSettings = $('#tfyhCfg-top').attr("value").split(":");
			dilboCfgPanel = new TfyhCfgPanel(dilboConfig, dilboModal, 
					panelSettings[0], panelSettings[1]);
			dilboConfig.setCfgPanel(dilboCfgPanel);
			dilboCfgPanel.refresh();
		}
		// start progress control, if a progressUrl is provided.
		try {
			let progressChunk = 100;
			if ((typeof progressUrl !== 'undefined') && (progressUrl.length > 0)) 
				dilboModal.showProgress(progressUrl, 1, progressChunk, 0);
		} catch (e) {}
	}
}

/**
 * Refresh the configuration panel.
 */
function refresh() {
	if (dilboCfgPanel)
		dilboCfgPanel.refresh();
}

/**
 * Events are called by menu items and list items. The DOM-element id contains
 * the iformation what to do. Here is just the top level distribvution to either
 * the form handler or the menu.
 * 
 * @param itemId
 */
function handleEvent(itemId) {
	let action = DilboUIprovider.splitId(itemId);
	try {
		// event type: form submission
		if ((action.length == 1) && dilboFormHandler && (itemId.localeCompare("submit") == 0))
			dilboFormHandler.onSubmit();
		// event type: server form, edit a data record 
		else if (action[0] == "editRecord")
			dilboFormHandler.editRecord(action[1], action[2], false);
		// event type: server form, add a data record 
		else if (action[0] == "addRecord") {
			let newUid = dilboFormHandler.createRecord(action[1]);
			dilboFormHandler.editRecord(action[1], newUid, true);
		}
		// event type: server form, delete a data record. Show warning first 
		else if (action[0] == "deleteRecord")
			dilboFormHandler.deleteRecord(action[1], action[2]);
		// event type: server form, add a data record 
		else if (Object.keys(dilboFormHandler.validityModifications).indexOf(action[0]) >= 0)
			dilboFormHandler.editValidity(action[0], action[1], action[2]);
		// event type: server form, edit a list, like group members 
		else if (action[0] == "editList")
			dilboFormHandler.editList(itemId.substring(itemId.indexOf("_") + 1));
		// event type: request to display a form for entry
		else if (dilboFormHandler && dilboFormHandler[action[0] + "_" + action[1]])
			dilboFormHandler[action[0] + "_" + action[1]](action[2]);
		// event type: menu item or list item click (action[0] is changeItem or
		// onListItemClick)
		else if (dilboMenu)
			dilboMenu.handleEvent(itemId);
		else
			alert(_("RwEWVR|Unable to execute action..." + itemId));
	} catch (e) {
		dilboModal.showException(e);
	}
}

/**
 * Toggle the sidebar menu sub-options
 * 
 * @param idSuffix
 * @returns
 */
function openSubMenu(idSuffix) {
	var submenu_items = document.getElementsByClassName("subMenu" + idSuffix);
	for (var i = 0; i < submenu_items.length; i++) {
		if (submenu_items[i].className.indexOf("w3-show") == -1) {
			submenu_items[i].className += " w3-show";
		} else {
			submenu_items[i].className = submenu_items[i].className.replace(
					" w3-show", "");
		}
	}
}

// Open and close sidebar
function w3_open() {
	document.getElementById("menuSidebar").style.display = "block";
	document.getElementById("menuOverlay").style.display = "block";
}

function w3_close() {
	document.getElementById("menuSidebar").style.display = "none";
	document.getElementById("menuOverlay").style.display = "none";
}

$(document).ready(function() {
	initGlobals();
	initModules();
	$(window).resize(function() {
		TfyhContainer.onResize();
	});
});
