/**  
 *
 *       dilbo - digital logbook for Rowing and Canoeing
 *       -----------------------------------------------
 *       https://www.dilbo.org
 *
 * Copyright  2023-2024  Martin Glade
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

/**
 * The collection of tables for dilbo
 */

class DilboTables {
	
	#listDefs = {};
	tables = {};
	iUids = {}; // the overall index uid > tablename
	iUuids = {}; // the overall index uuid > tablename
	iNames = {}; // the overall index name > [tablenames] (names may occur in
					// multiple tables)
	
	// special lists for boat status display
	openTrips = [];
	openDamages = [];
	openSevereDamages = [];
	openReservations = [];
	currentReservations = [];
	
	// special name lists for autocompletion
	validNames =  {
			assets : [], crews : [], destinations : [], groups : [], persons : [], status : [], waters : [] 
	};
	// the sequence of the uuid2names list set definitions
	#nameListIds =  {
			assets : 1, crews : 2, destinations : 3, groups : 4, persons : 5, status : 6, waters : 7 
	};

	// for the moment only the status table is cached in local storage rather to
	// test the functionality.
	#cached = [ "assets", "badges", "crews", "damages", "destinations", "groups", 
		"logbook", "messages", "reservations", "status", "waters", "workbook" ]; 

	#loadingIndex = 0; 
	#initializing = true; 
	
	/**
	 * Build all tables and load them from local storage
	 */
	constructor(loadOnInstantiation) {
		if (loadOnInstantiation)
			this.load(true);
	}

	/**
	 * Create a new empty table.
	 */
	#createTable() {
		let newTable = {
				refreshing : false,
				config : null,
				csv : "", // the server input
				mtime : 0, // the download time (microtime)
				iUuids : {}, // the index uuid > table row
				iNames : {}, // the index name > table row
				columns : [], // the column names
				rows : {}, // the table data, records are objects indexed by
							// the uid
				listeners : {} // the listeners which will be called on any
								// list change
			}
		return newTable;
	}
	
	/**
	 * return true, if at least one table is refreshing or the list definitions
	 * downlod is busy.
	 */
	loadingBusy() {
		return (this.#loadingIndex > 0);
	}
	
	/**
	 * Download the server side tables as lists. Set firstRequest =
	 * "@definitions" to initialize. Definititons will always be fetched from
	 * the server. For the rest see #onListResponse(tx).
	 */
	load(fullRefresh) {
		this.#loadingIndex = 1;
		let synchLimit = (fullRefresh) ? 0 : dilboTxQueue.getSynchLimit(); 
		dilboTxQueue.addNewTxToPending("list", "@definitions", { 
			setname : "dilboWeb", 
			modified_after : synchLimit
		}, 0, null, null);
	}

	/**
	 * Activities on receiving a transaction response.
	 */
	onTxResponse(tx) {
		const modifications = [ "insert", "update", "delete" ];
		if (tx.type == 'list') {
			if (tx.tablename == 'statistics') {
				let type = tx.record[0].split(";")[1];
				let field = tx.record[1].split(";")[1];
				let providerName = tx.record[2].split(";")[1];
				// for statistcs the callback is not a function, but the
				// uiProvider
				tx.callback.buildStatistics(providerName, tx.resultMessage);
				// remove callback after execution
				tx.callback = null;
			}
			else
				this.#onListResponse(tx);
		}
		else if (modifications.indexOf(tx.type) >= 0) {
			this.#onModifyResponse(tx, modifications.indexOf(tx.type));
		}
		// do callback, if still to be done.
		if (tx.callback && (typeof tx.callback == 'function'))
			tx.callback();
	}
	
	/**
	 * get a record by its uid
	 */
	getRecord(uid) {
		let tablename = this.iUids[uid];
		return this.tables[tablename].rows[uid];
	}
	
	/**
	 * get the name to display for a uuid
	 */
	getName(uuid) {
		if (!this.iUuids[uuid])
			return "???" + uuid;
		let tablename = this.iUuids[uuid];
		if (!this.tables[tablename].iUuids[uuid])
			return "??" + uuid;
		let record = this.tables[tablename].iUuids[uuid];
		if (tablename == "persons")
			return record.first_name + " " + record.last_name;
		if (!record.name)
			return "?" + uuid;
		return record.name
	}

	/**
	 * get the current asset status based on the open session, reservations and
	 * damages.
	 */
	getAssetStatus(assetUid) {
		let asset = this.tables.assets.rows[assetUid]; 
		let status = "AVAILABLE";
		for (let reservationUid of dilboTables.openReservations)
			if (dilboTables.getRecord(reservationUid).asset_uuid == asset.uuid)
				status = "BOOKED";
		for (let damageUid of dilboTables.openSevereDamages)
			if (dilboTables.getRecord(damageUid).asset_uuid == asset.uuid)
				status = "DAMAGED";
		for (let tripUid of dilboTables.openTrips)
			if (dilboTables.getRecord(tripUid).asset_uuid == asset.uuid)
				status = "ONTHEWATER";
		return status;
	}
	
	/**
	 * Notify all listeners on a completed change
	 */
	notifyListeners(tablename) {
		for (let listenername in this.tables[tablename].listeners)
			this.tables[tablename].listeners[listenername].notify();
	}
	
	/**
	 * Notify all listeners on a completed change
	 */
	addListener(tablename, listenername, listener) {
		this.tables[tablename].listeners[listenername] = listener;
	}
	
	/**
	 * Get resolved names
	 */
	getValueToDisplay(value, type, column_reference) {
		let toDisplay = "";
		if (type == 'microtime')
			toDisplay = TfyhToolbox.timef2string(value, userPreferences.languageCode);
		else if (TfyhData.parseAs(type) == "string") {
			let values = (Array.isArray(value)) ? value : value.split(/\|/g);
			for (let element of values) {
				if (TfyhToolbox.isUUID(element))
					element = this.getName(element);
				else if (column_reference && column_reference.startsWith(".")) {
					let referenced = dilboConfig.get_by_path(column_reference + "." + element);
					if (referenced)
						element = referenced.get_descriptor_field("text_local_name");
					else
						element = value;
				} else if (element.length > 0) 
					element = TfyhData.format(element, type);
				toDisplay += element + "; ";
			}
			if (toDisplay.length > 1)
				toDisplay = toDisplay.substring(0, toDisplay.length - 2);
		} else
			toDisplay = TfyhData.format(value, type);
		return toDisplay;
	}
	
	/**
	 * Get the next numeric id for a table.
	 */
	getNextNumber(tablename) {
		if (!this.tables[tablename] || !this.tables[tablename].rows)
			return 0; // no such table
		let maxNumber = 0;
		for (let uid in this.tables[tablename].rows) {
			if (!this.tables[tablename].rows[uid].number)
				return 0; // table has no number field
			let number = this.tables[tablename].rows[uid].number;
			if (number > maxNumber)
				maxNumber = number;
		}
		return maxNumber + 1;
	}
	
	/**
	 * After record modification a table shall be updated, but there are only
	 * lists to get. get the first list for that table and call #getList with
	 * it.
	 */
	#getTable(tablename, modified_after) {
		for (let listname in this.#listDefs)
			if (this.#listDefs[listname].from == tablename)
				this.#getList(listname, modified_after);
	}
	
	/**
	 * Get a name list which shall be merged into the tables. This is used by
	 * server side forms for lookup and validation of values
	 */
	getNameList(tablename, callback) {
		const set = "uuid2names";
		let listid = this.#nameListIds[tablename];
		if (!this.tables[tablename]) {
			this.tables[tablename] = this.#createTable();
			this.tables[tablename].config = dilboConfig.get_by_path(".appTables." + tablename);
		}
		var that = this;
    	$.get('../pages/jsget.php?set=' + set + '&listid=' + listid, function(data) {
    		that.#mergeCsv(tablename, data, false);
    		that.#refreshIndices(tablename);
    		callback();
    	});
	}
	
	/**
	 * Append a transaction to the queue to get the named list.
	 */
	#getList(listname, modified_after) {
		if (!this.#listDefs[listname])
			return;

		let tablename = this.#listDefs[listname].from;
		// if the table is empty, cached, and available in local storage, load
		// cache first and wait for the synchronisation to fetch the server data
		if ((this.#cached.indexOf(tablename) >= 0) 
				&& window.localStorage.getItem(tablename + ".mtime")
				&& (Object.keys(this.tables[tablename].rows).length == 0)) {
			dilboTxQueue.setLowestLastSynch(window.localStorage.getItem(tablename + ".mtime"));
			let txVirtual = {
					tablename : listname,
					record : [ "modified_after;" + modified_after ],
					resultMessage : window.localStorage.getItem(tablename + ".data"),
					mtime : window.localStorage.getItem(tablename + ".mtime")
			}
			this.#onListResponse(txVirtual);
			return;
		}

		let listargs = { setname : "dilboWeb", modified_after : modified_after };
		if (tablename == "logbook") {
			listargs.logbookname = userPreferences.logbook;
			listargs.sports_year_start = $_sports_year_start();
		}
		else if (tablename == "workbook") {
			listargs.workbookname = "clubwork";
			listargs.sports_year_start = $_sports_year_start();
		}
		dilboTxQueue.addNewTxToPending("list", listname, listargs, 0, null, null);
	}
	
	/**
	 * Activities on recieving a list response, or a cache load
	 */
	#onListResponse(tx) {
		let modifiedAfter = 0;
		for (let param of tx.record)
			if (param.split(";")[0] == 'modified_after')
				modifiedAfter = parseInt(param.split(";")[1]);
		// List definitions loading
		if (tx.tablename == "@definitions") {
			let defsArray = TfyhToolbox.csvToObject(tx.resultMessage);
			let firstTableToDownload = defsArray[0].from;
			for (let definition of defsArray) {
				// cache the definition, indexed by name
				this.#listDefs[definition.name] = definition;
				// (re)build the related table
				let tablename = definition.from;
				if (! this.tables[tablename]) {
					this.tables[tablename] = this.#createTable();
					this.tables[tablename].columns = definition.select.split(/\,/g);
					this.tables[tablename].config = dilboConfig.get_by_path(".appTables." + tablename);
					// NB: the JavaScript tables are a subset of columns and
					// rows of the server tables, but they share the metadata.
				}
				this.tables[tablename].refreshing = true;
			}
			// merge cached data first
			this.#loadingIndex = defsArray.length - 1;
			this.#getList(Object.keys(this.#listDefs)[this.#loadingIndex], modifiedAfter);
		}
		// any other list. Put the result into the respective table.
		else {
			let tablename = this.#listDefs[tx.tablename].from; 
			// use cache timestamp, if provided. API transaction do not have an
			// mtime property
			this.tables[tablename].mtime = (tx.mtime) ? tx.mtime : Math.floor(Date.now() / 1000);
			this.#mergeCsv(tablename, tx.resultMessage, (this.#cached.indexOf(tablename) >= 0));
			this.#refreshIndices(tablename);
			this.notifyListeners(tablename);
			this.tables[tablename].refreshing = false;
			if (this.#loadingIndex > 0) {
				// if this list response is part of the initial loading
				// trigger the next list or the next module to initialize.
				this.#loadingIndex --;
				if (this.#loadingIndex >= 0) 
					this.#getList(Object.keys(this.#listDefs)[this.#loadingIndex], modifiedAfter);
			}
			// at first load continue with module initialization.
			else if (this.#initializing) {
				this.#initializing = false;
				initModules();
				// while the initialization continues with the logbook display,
				// trigger full synchronisation.
				dilboTxQueue.setLowestLastSynch(0);
			} else
				this.notifyListeners(tablename);
		}
	}
	
	/**
	 * Handle the response of a successful record modification.
	 */
	#onModifyResponse(tx, mode) {
		// load the list from the server. This will merge all server side
		// changes into the record, in particular the numbering and trigger
		// refresh.
		this.tables[tx.tablename].refreshing = true;
		this.#getTable(tx.tablename, (Date.now() / 1000) - 300);
	}

	/**
	 * insert the element into the array, if not contained or remove, if
	 * contained
	 */
	#setArrayElement(array, element, include) {
		if (typeof array == 'undefined')
			return;
		let indexOf = array.indexOf(element); 
		if (indexOf < 0) {
			if (include) {
				array.push(element);
				array.sort();
			}
		} else {
			if (!include)
				array.splice(indexOf, 1);
		}
	}

	/**
	 * update the index entries for a single uid, upon insertion, change or
	 * delete of a record. Set isDelete to true for the latter. Updates both the
	 * table specific indices and the global indices.
	 */
	#updateIndicesSingle(tablename, uid, isDelete) {
		let forTable = this.tables[tablename];
		let dateNow = new Date();
		let secondsNow = Date.now() / 1000;
		let row = forTable.rows[uid]; 

		// uid index
		this.iUids[uid] = tablename;
		// uuid index
		if (forTable.columns.indexOf("uuid") >= 0) {
			let uuid = row.uuid;
			if (isDelete) {
				delete forTable.iUuids[uuid];
				delete this.iUuids[uuid];
			} else {
				forTable.iUuids[uuid] = row;
				this.iUuids[uuid] = tablename;
			}
		} 
		// name index
		if ((forTable.columns.indexOf("name") >= 0)
				|| (forTable.columns.indexOf("first_name") >= 0)) {
			let uuid = row.uuid;
			let name = row.name;
			if (tablename == 'persons')
				name = row.first_name + " " + row.last_name;
			if (uuid)
				if (isDelete)
					delete forTable.iUuids[uuid];
				else
					forTable.iUuids[uuid] = row;
			if (name)
				if (isDelete)
					delete forTable.iNames[uuid];
				else
					forTable.iNames[name] = row;
			if (! this.iNames[name])
				this.iNames[name] = [];
			if (this.iNames[name].indexOf(tablename) < 0)
				this.#setArrayElement(this.iNames[name], tablename, !isDelete);
		}

		// name lists
		let includeVersionized = (!row.invalid_from || (row.invalid_from > 8e15)) && !isDelete;
		if (this.validNames[tablename]) {
			if (tablename == 'persons') {
				let personName = row.first_name + " " + row.last_name;  
				this.#setArrayElement(this.validNames.persons, personName, includeVersionized);
			} else {
				this.#setArrayElement(this.validNames[tablename], row.name, includeVersionized);
			}
		}

		// special lists
		// open trips (used only for boat status display, not for logbook
		// display.
		else if (tablename == 'logbook') 
			this.#setArrayElement(this.openTrips, uid, !row.end && !isDelete);
		// open damages and open severe damages
		else if (tablename == 'damages') {
			this.#setArrayElement(this.openDamages, uid, !row.fixed && !isDelete);
			if (row.severity == 'NOTUSEABLE') 
				this.#setArrayElement(this.openSevereDamages, uid, !row.fixed && !isDelete);
		}
		// open reservations
		else if (tablename == 'reservations') {
			this.#setArrayElement(this.openReservations, uid, (row.end > dateNow) && !isDelete);
			this.#setArrayElement(this.currentReservations, uid, (row.start < dateNow) 
					&& (row.end > dateNow) && !isDelete);
		} 

	}
	/**
	 * refresh the indices of the respective table: names, uuid and the global
	 * indices on uids, uuids and names.
	 */
	#refreshIndices(tablename) {
		let forTable = this.tables[tablename];
		forTable.iUuids = {};
		forTable.iNames = {};
		for (let uid in forTable.rows)
			this.#updateIndicesSingle(tablename, uid, false);
	}

	/**
	 * Merge a record into a table. Set updateIndices to false to skip this step
	 */
	mergeRecord(row, cfg, targetTable, updateIndices) {
		if (!cfg)
			return;
		if (!row.uid)
			console.log("Record without uid.");
		else {
			// create new or overwrite existing
			let tableRow = (targetTable.rows[row.uid]) ? targetTable.rows[row.uid] : {};
			for (let field in row) {
				if (cfg.has_child(field)) {
					// Parse the value and then assign it to a field
					let type = cfg.get_child(field).get_type();
					let strVal = row[field];
					let val = TfyhData.parse(strVal, type);
					tableRow[field] = val;
				} 
			}
			// add the row to the table, or replace it
			targetTable.rows[row.uid] = tableRow;
			if (updateIndices)
				this.#updateIndicesSingle(cfg.get_name(), row.uid, false);
			this.notifyListeners(cfg.get_name());
		}
	}
	/**
	 * Read from csv-string into memory. Merges into existing rows and refreshes
	 * the indices. Set storeUpdate to store the result, if a chache shall be
	 * used.
	 */
	#mergeCsv(targetTablename, csvData, storeUpdate) {
		// Split csv data into string arrays
		let rows = TfyhToolbox.csvToObject(csvData);
		// if the table was not initialized (e.g. server form lookup lists)
		// provide the column names.
		if ((this.tables[targetTablename].columns.length == 0) && (rows.length > 0))
			this.tables[targetTablename].columns = Object.keys(rows[0]);
		let targetTable = this.tables[targetTablename];
		let cfg = targetTable.config;
		// parse rows and insert into table
		for (let row of rows) 
			this. mergeRecord(row, cfg, targetTable, false);
		// Store the result into the cache.
		if (storeUpdate) {
			// there is no new data and the existing table is not empty
			if (! rows[0] && window.localStorage.getItem(targetTablename))
				return;
			let csv = this.#writeCsv(targetTablename); 
			window.localStorage.setItem(targetTablename + ".data", csv);
			window.localStorage.setItem(targetTablename + ".mtime", Math.floor(Date.now() / 1000));
		}
		// refresh the indices
		this.#refreshIndices(targetTablename);
	}
	
	/**
	 * write from memory to csv-string
	 */
	#writeCsv (tablename) {
		let table = this.tables[tablename];
		let cfg = table.config;
		let csv = "";
		for (let column of table.columns) 
			csv += column + ";";
		csv = csv.substring(0, csv.length - 1) + "\n";
		for (let uid in table.rows) {
			let row = table.rows[uid];
			for (let column of table.columns) { 
				// get the data type
				let type = cfg.get_child(column).get_type();
				// format the value
				let strVal = TfyhData.format(row[column], type, "csv");
				// encode into csv
				csv += TfyhToolbox.encodeCsvEntry(strVal) + ";";
			}
			csv = csv.substring(0, csv.length - 1) + "\n";
		}
		return csv;
	}
	
	/* ----------------------------------------------------------------- */
    /* ------ HTML DISPLAY --------------------------------------------- */
    /* ----------------------------------------------------------------- */
	
	// get the HTML code for a single item
	recordToTableHTML (tablename, record) {
		let recordCfg = dilboConfig.get_by_path(".appTables." + tablename);
		let html = "<h3>" + recordCfg.get_descriptor_field("text_local_name") + "</h3>";
		html += "<table>";
		for (let childname in recordCfg.get_children()) {
			if (record[childname]) {
				if ((childname != "uid") && (childname != "uuid")) {
					let childCfg = recordCfg.get_child(childname);
					let column_reference = childCfg.get_descriptor_field("column_reference");
					let toDisplay = this.getValueToDisplay(record[childname], childCfg.get_type(), column_reference);
					html += "<tr><td>" + childCfg.get_descriptor_field("text_local_name")
							+ "</td><td>" 
							+ TfyhToolbox.escapeHtml(toDisplay, userPreferences.languageCode) 
							+ "</td></tr>";
				}
			}
		}
		return html + "</table>";
	}
	
	

}