/**  
 *
 *       dilbo - digital logbook for Rowing and Canoeing
 *       -----------------------------------------------
 *       https://www.dilbo.org
 *
 * Copyright  2023-2024  Martin Glade
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

/**
 * Queue of transaction requests for communication with dilboServer. Compile and
 * send messages, asynchronously collect of responses or communication errors,
 * parse responses and hand-over the response or error to cTxHandler.
 */

class DilboTxQueue {
	
	
	// timing prameters
	#apiPostURI = "../api/posttx.php";
	#apiPendingMillis = 200;
	#api_user_id = false;
	#api_session_id = false;
	
	// container size
	#txsPerContainer = 10;

	// Synchronisation defaults, must never be 0.
	// other write access check period (seconds)
	#synch_check_period = 90;
	#last_synch_check = Date.now() / 1000;
	// download synchronisation period (seconds)
	#synch_period = 900;
	#last_synch = Date.now() / 1000;
	// session inactivity and lifetime timeouts in seconds
	#keep_alive_period = 600;  // 10 minutes
	#session_lifetime = 10800;  // 3 hours
	// Construction of the queue will result in a session start request
	#last_session_activity = Date.now() / 1000; 
	#last_session_regenerate = Date.now() / 1000;

	// Server result codes. Shall be identical to the dilbo_api.php
	// $result_messages definition
	#api_resultMessages = {};
	
	// handling: for permanent errors transactions will be dropped
	// for temporary errors transactions will be resent until they are resolved.
	permanentContainerErrors = [ 400, 401, 402, 403, 500 ];
	temporaryContainerErrors = [ 405, 406, 408, 409 ];
	permanentTransactionErrors = [ 501, 502, 503 ];
	writeTransactions = [ "insert", "update", "delete" ];

	// the #MESSAGE_SEPARATOR_STRING must not contain any regex special
	// character
	// because it is used in "split()" calls as "regex" argument.
	#MESSAGE_SEPARATOR_STRING = "\n-#|#-\n";
	#MS_REPLACEMENT_STRING = "\n_#|#_\n";
	
	#queues = {
		txPending : [],
		txDone : [],
		txFailed : []
	}

	#apiVersion = 4;   
	// the first request always goes with lowest version, then the client shall
	// max out the version based on the server response
	#txID = 42;
	#txcID = 42;
	// the number of transactions sent is memorized to remove them on failures.
	#busy = 0;
	// the operation can be paused upon overload detection or timeout
	#paused = 0; // wait time in milliseconds
	
	#timer;
	static that;

	/**
	 * The transaction prototype to use for communication with the server
	 */
	#emptyTransaction = {
		// transaction values
		ID : -1,
		type : "",
		tablename : "",
		uid : "00000000", // only for local storage, not used at API
		record : [],
		// result for container
		cresultCode : -1,
		cresultMessage : "",
		// result for transaction
		resultCode : -1,
		resultMessage : "",
		// workflow
		callback : null,
		onError : null,
		// status control
		sentAt : 0,
		resultAt : 0,
		closedAt : 0,
		// cache to know where to put changed Ids to
		listRowPos : 0
	}

	/**
	 * the transaction container prototype to use for communication with the
	 * server
	 */
	#emptyContainer = {
		version : 0,
		cID : 0,
		cresultCode : 0,
		cresultMessage : "",
		txms : []
	}
	
	/**
	 * Constructor
	 */
	constructor(api_user_id, api_session_id) {
		// set credentials
		this.#api_user_id = api_user_id; 
		this.#api_session_id = api_session_id;
		// set result codes and messages of the api
		let resultCodes = dilboConfig.get_by_path(".framework.api.result_codes").get_value();
		for (let resultCode of resultCodes)
			this.#api_resultMessages[resultCode.substring(0, 3)] = resultCode.substring(4); 
		// start the queue with a session start
		this.addNewTxToPending ("session", "start", null, null, null);
		// add all stored pending write transactions to the queue
		let storageItemNames = Object.keys(window.localStorage);
		for (let storageItemName of storageItemNames) {
			if (storageItemName.startsWith("txPending")) {
				let plain = window.localStorage.getItem(storageItemName);
				let parts = TfyhToolbox.splitCsvRow(plain);
				let uid = storageItemName.substring("txPending-".length); 
				let type = parts[1];
				// add, if it is a write transaction
				if (this.writeTransactions.indexOf(type) >= 0) {
					let tablename = parts[2]; 
					let record = {};
					let i = 3;
					while ((i + 1) < parts.length) {
						record[parts[i]] = TfyhToolbox.encodeCsvEntry(parts[i + 1]);
						i++;
					}
				} else
					// remove, if not a write transaction
					window.localStorage.removeItem(storageItemName);
			}
		}
		// start the timer
		this.#timer = setInterval(this.#onTimerEvent, this.#apiPendingMillis);
		DilboTxQueue.that = this; // needed to be backreference within timer
								// events
	}
	
	/**
	 * return a queue status String for debugging
	 */
	getStatus () {
		let statusStr = "";
		if (this.#queues.txPending.length > 0) {
			statusStr += _("K3Mp55|Pending:_");
			this.#queues.txPending.forEach(function(tx) {
				statusStr += "#" + tx.ID + "-" + tx.type + " ";
			});
		} else if (this.#queues.txDone.length > 0) {
			statusStr += _("WPilW7|Done:_") + this.#queues.txDone.length;
		} else if (this.#queues.txFailed.length > 0) {
			statusStr += _("BqT8Qu|Failed:_") + this.#queues.txFailed.length;
		}
		return statusStr;
	}
	
	/**
	 * Get the time in seconds for the tables synchronisation. Use four times
	 * the synchronisation period for redundancy.
	 */
	getSynchLimit() {
		return this.#last_synch - 3 * this.#synch_period; 
	}
	
	/**
	 * Set the last synchronisation time (seconds since unix epoch. If the
	 * current value of this.#last_synch is lower than lastSynchCandidate,
	 * nothing will happen.
	 */
	setLowestLastSynch(lastSynchCandidate) {
		if (lastSynchCandidate < this.#last_synch)
			this.#last_synch = lastSynchCandidate; 
	}
	
	/* ------------------------------------------------------ */
	/* ------- FORMATTING AND STORAGE ----------------------- */
	/* ------------------------------------------------------ */

	/**
	 * provide a human readable String for a transaction (logging at close
	 * action)
	 */
	#txToString (tx) {
		let txStr = "#txID" + tx.ID + "; ";
		let recordStr = "";
		tx.record.forEach(function(nvp) {
			recordStr += nvp + ";"
		});
		if (recordStr.length > 50)
			recordStr = recordStr.substring(0, 47) + "..."
		txStr += tx.type + " " + tx.tablename + "(" + recordStr + "); ";
		let sentDate = new Date(tx.sentAt);
		let sentStr = (tx.sentAt == 0) ? _("Ti6aAG|not_yet_") : TfyhData.format(sentDate, "datetime");
		txStr += _("RL7YhL|sent:_") + sentStr + " ";
		let closedDate = new Date(tx.closedAt);
		let messageStr = tx.resultMessage;
		if (messageStr.length > 100)
			messageStr = messageStr.substring(0, 97) + "... (" + messageStr.length + " chars)"; 
		txStr += _("elDfgL|result:_") + tx.resultCode + " " + messageStr + " ";
		let closedStr = (tx.closedAt == 0) ? _("fiB2y8|not_yet_") : TfyhData.format(closedDate, "datetime");
		txStr += _("3f1zwM|closed:_") + closedStr + " ";
		return txStr;
	}

	/**
	 * store a transaction in the local storage.
	 */
	#toLocalStorage(tx) {
		let plain = tx.ID + ";" + tx.type + ";" + tx.tablename + ";";
		if (tx.record)
			for (let nvp of tx.record) {
				while (nvp.indexOf(this.#MESSAGE_SEPARATOR_STRING) >= 0)
					nvp = nvp.replace(this.#MESSAGE_SEPARATOR_STRING, this.#MS_REPLACEMENT_STRING);
				plain += nvp + ";";
			}
		plain = plain.substring(0, plain.length - 1);
		window.localStorage.setItem("txPending-" + tx.uid, plain);
	}

	/**
	 * Shift transactions from one queue to another, or update them.
	 */
	#updatePendingQueue (count, destination) {
		let i = 0;
		let nowMillis = Date.now();
		if (destination) {
			// move transaction from pending to failed or done.
			while ((this.#queues.txPending.length > 0) && (i < count)) {
				// remove item from source
				let txShifted = this.#queues.txPending.splice(0, 1)[0];
				window.localStorage.removeItem("txPending-" + txShifted.uid);
				// add timestamps
				txShifted.resultAt = nowMillis + i;
				txShifted.closedAt = nowMillis + i;
				// put to destination (memory only)
				this.#queues[destination].push(txShifted);
				i++;
			}
			console.log("txPending (" + this.#queues.txPending.length + ") => " + destination
					+ " (" + this.#queues[destination].length + ", +" + i + ")");
		} else {
			// update pending with the sentAt timestamp
			while ((i < this.#queues.txPending.length) && (i < count)) {
				let txUpdated = this.#queues.txPending[i];
				txUpdated.sentAt = nowMillis + i;
				this.#toLocalStorage(txUpdated);
				i++;
			}
		}
		return i;
	}

	/* ------------------------------------------------------ */
	/* ------- SENDING DATA --------------------------------- */
	/* ------------------------------------------------------ */
	
	/**
	 * Add a transaction based on the tx type and parameters to the pending
	 * queue. Set atFront to true, to add at front, else the new transaction is
	 * appended.
	 * 
	 * @params String type transaction type.
	 * @params String tablename the table name or the list name
	 * @params String record as object of strings
	 * @params function callback: a function which cann be executed after the
	 *         response was received, e.g. a data refresh
	 * @params function onError: a function which cann be executed if the
	 *         response indicates an error. A log is always posted to the
	 *         cobnsole, but if the user shall be notified, put the respectve
	 *         function here. It must take the transaction itself as function
	 *         argument.
	 * @return added transaction.
	 */
	addNewTxToPending (type, tablename, record, callback, onError) {
		// prepare the transaction cache. create a clone of the empty
		// transaction, set values and push it to the queue.
		var tx = Object.assign({}, this.#emptyTransaction);
		this.#txID++;
		tx.ID = this.#txID;
		tx.type = type;
		tx.tablename = tablename;
		tx.uid = TfyhToolbox.generateUid(6);
		let encodedRecord = [];
		for (key in record) 
			encodedRecord.push(TfyhToolbox.encodeCsvEntry(key) 
						+ ";" + TfyhToolbox.encodeCsvEntry("" + record[key]));
		tx.record = encodedRecord;
		tx.callback = callback;
		tx.onError = onError;
		this.#queues.txPending.push(tx);
		this.#toLocalStorage(tx);
		console.log("Pending queue ++: " + this.#queues.txPending.length);
		return tx;
	}

	/**
	 * Build a new container based on the provided ransaction. This Javascript
	 * web app omly uses one transaction per container.
	 */
	#buildContainer (txCnt) {
		this.#txcID++;
		// use the globally provided user and session id, from page load.
		let plain = this.#apiVersion + ";" + this.#txcID + ";" + this.#api_user_id + ";" 
			+ this.#api_session_id + ";";
		for (let i = 0; i < txCnt; i++) {
			let tx = this.#queues.txPending[i];
			plain += tx.ID + ";" + tx.type + ";" + tx.tablename + ";";
			if (tx.record)
				for (let nvp of tx.record) {
					while (nvp.indexOf(this.#MESSAGE_SEPARATOR_STRING) >= 0)
						nvp = nvp.replace(this.#MESSAGE_SEPARATOR_STRING, this.#MS_REPLACEMENT_STRING);
					plain += nvp + ";";
				}
			plain = plain.substring(0, plain.length - 1) + this.#MESSAGE_SEPARATOR_STRING;
		}
		plain = plain.substring(0, plain.length - this.#MESSAGE_SEPARATOR_STRING.length)
		let base64 = window.btoa(unescape(encodeURIComponent(plain)));
		let base64api = base64.replace(/\//g, '-').replace(/\+/g, '*').replace(
				/\=/g, '_');
		console.log("Container built: " + txCnt + " transactions in " + base64api.length + " base64 characters");
		return base64api;
	}
	
	/**
	 * Check for last other write access and start reload of tables, if true.
	 */
	#synchCheck() {
		let postRequest = new XMLHttpRequest();
		let that = this;
		// provide the callbacks
		postRequest.onload = function() {
			if (postRequest.status == 200) {
				var lowa = parseInt(postRequest.response.split(";")[0]);
				if (lowa > that.#last_synch) {
					// get updates. Use four times the synchronisation period
					// for redundancy
					dilboTables.load(false);
					that.#last_synch = (Date.now() / 1000);
				}
			}
		};
		// provide the callback for any error.
		postRequest.onerror = function() { /* do nothing */
		};
		postRequest.ontimeout = function() { /* do nothing */
		};
		// send the post request
		postRequest.open('POST', that.#apiPostURI, true);
		postRequest.setRequestHeader('Content-type',
				'application/x-www-form-urlencoded; charset=UTF-8');
		postRequest.send("lowa=" + that.#api_session_id);
		that.#last_synch_check = (Date.now() / 1000);
	}

	/**
	 * Check the queue, when the pending timer fires. If no transaction is busy,
	 * send the first transaction in the queue.
	 */
	#onTimerEvent () {
		// start transaction, if not busy nor enabled and a transaction is
		// pending
		let that = DilboTxQueue.that;
		let now = Date.now(); 
		if (now > that.#paused) {
			
			// send transactions, if idle
			if (that.#busy == 0) {
				if (that.#queues.txPending.length > 0) {
					that.#busy = that.#sendPendingTxs();
					that.#updatePendingQueue(that.#busy);
				}	
			} else {
				// log queue status, if waiting
				if (that.#queues.txPending.length > 0) 
					console.log("" + that.#queues.txPending.length + " pending.");
			}

			// regenerate session
			let regenerateDueAt = that.#last_session_regenerate + (that.#session_lifetime * 0.9);
			if ((that.#last_session_regenerate > 0) && (now / 1000) > regenerateDueAt) {
				// the first session start sets that.#last_session_regenerate.
				// Wait for this
				// to happen before starting session regeneration
				that.#last_session_regenerate = (now / 1000); 
				that.addNewTxToPending ("session", "regenerate", null, null, null);
				that.addNewTxToPending ("config", "appSettings", null, null, null);
				that.addNewTxToPending ("config", "clubSettings", null, null, null);
			}
			// revitalize session
			if ((now / 1000) > (that.#last_session_activity + (that.#keep_alive_period / 2)) ) {
				that.#last_session_activity = (now / 1000); 
				that.addNewTxToPending ("nop", "", { sleep : 0 }, null, null);
			}
			// schedule full synchronization
			if ((now / 1000) > (that.#last_synch + that.#synch_period) ) {
				// get updates. Use four times the synchronisation period for
				// redundancy
				dilboTables.load(false);
				that.#last_synch = (now / 1000); 
			}
			// schedule synchronization check
			else if ((now / 1000) > (that.#last_synch_check + that.#synch_check_period) )
				that.#synchCheck();
		}
	}

	/**
	 * wrapper for the send procedure. Creates the post request an assigns the
	 * callback procedure for container success (this.#handleResponse) and
	 * failure (this.#handleContainerError). Then sends the request.
	 */
	#sendPendingTxs () {
		let postRequest = new XMLHttpRequest();
		let that = this;
		// provide the callback for a response received
		postRequest.onload = function() {
			let txrc;
			try {
				if (postRequest.status == 500) {
					let txrc = {
							cresultCode : 500,
							cresultMessage : _("kZ6xuS|Server returned status c...")
						};
					that.#handleContainer(txrc);
				} else {
					txrc = that.#parseResponse(postRequest.response);
					that.#handleContainer(txrc);
				}
			} catch (e) {
				if (dilboModal)
					dilboModal.showException(e);
				else
					alert(e.stack);
			}
		};
		// provide the callback for timeout
		postRequest.ontimeout = function() {
			let txrc = {
					cresultCode : 408,
					cresultMessage : _("7vNdV1|Post request time out")
				};
			that.#handleContainer(txrc);
		};
		// provide the callback for any error.
		postRequest.onerror = function() {
			let txrc = {
					cresultCode : 405,
					cresultMessage : _("oiGoJ3|Communication error at c...")
				};
			that.#handleContainer(txrc);
		};
		// send the post request
		postRequest.open('POST', this.#apiPostURI, true);
		postRequest.setRequestHeader('Content-type', 'application/x-www-form-urlencoded; charset=UTF-8');
		let now = Date.now();
		let txCnt = Math.min(this.#queues.txPending.length, this.#txsPerContainer);
		let txc = this.#buildContainer(txCnt);
		postRequest.send("txc=" + txc);
		return txCnt; 
	}

	/* ------------------------------------------------------ */
	/* ------- RECEIVING DATA ------------------------------- */
	/* ------------------------------------------------------ */

	/**
	 * parse a server response into a new transaction container.
	 * 
	 * @params String response to be parsed.
	 * @return the transaction container.
	 */
	#parseResponse (response) {

		let plain = TfyhToolbox.base64apiToUtf8(response);  
		// javascript native atob doesn't work with UTF-8 Strings.
		
		let txc = Object.assign({}, this.#emptyContainer);
		let plainParts = plain.split(";", 4);
		let cApiVersion = parseInt(plainParts[0]);
		if ((plainParts.length < 3) || (isNaN(cApiVersion))) {
			txc.cresultCode = 404;
			txc.cresultMessage = "Invalid transaction response: " + response.substring(0, 100);;
			txc.txms = [];
			txc.txrs = [];
			return txc;
		}
		
		txc.cApiVersion = parseInt(plainParts[0]);
		txc.cID = parseInt(plainParts[1]);
		txc.cresultCode = parseInt(plainParts[2]);
		txc.cresultMessage = plainParts[3];
		// syntax error check.
		if (plainParts.length < 4) {
			txc.txms = [];
			txc.txrs = [];
			return txc;
		}

		// javascript substring is different from java.substring. The last
		// element is not the reaminder, but just the last element of the split
		// operation.
		let txcHeaderLength = plainParts[0].length + plainParts[1].length + plainParts[2].length + plainParts[3].length + 4;
		let txmsStr = plain.substring(txcHeaderLength);
		txc.txms = [];
		while (txmsStr.length > 0) {
			let indexOfEnd = txmsStr.indexOf(this.#MESSAGE_SEPARATOR_STRING);
			if (indexOfEnd >= 0) {
				txc.txms.push(txmsStr.substring(0, indexOfEnd));
				txmsStr = txmsStr.substring(indexOfEnd + this.#MESSAGE_SEPARATOR_STRING.length);
				indexOfEnd = txmsStr.indexOf(this.#MESSAGE_SEPARATOR_STRING);
			} else {
				txc.txms.push(txmsStr);
				txmsStr = "";
			}
		}

		// parse transaction responses within container.
		txc.txrs = [];
		for (let txm of txc.txms) {
			let txr = {};
			txr.ID = (txm.indexOf(";") >= 0) ? parseInt(txm.substring(0, txm.indexOf(";"))) : 0;
			txm = txm.substring(txm.indexOf(";") + 1)
			txr.resultCode = (txm.indexOf(";") >= 0) ? parseInt(txm.substring(0, txm.indexOf(";"))) : 500;
			txr.resultMessage = (txm.indexOf(";") >= 0) ? txm.substring(txm.indexOf(";") + 1) : "";
			txc.txrs.push(txr);
			for (let tx of this.#queues.txPending)
				if (tx.ID == txr.ID) {
					tx.resultCode = txr.resultCode;
					tx.resultMessage = txr.resultMessage;
				}
		}
		return txc;
	}

	/* ------------------------------------------------------ */
	/* ------- HANDLE RESPONSE ------------------------------ */
	/* ------------------------------------------------------ */

	/**
	 * Handle the error for a transaction container. The transactions go to the
	 * failed queue and for login failures an error page is displayed.
	 */
	#handleContainerError (txrc) {
		let failed = this.#busy;
		for (let i = 0; i < failed; i++) {
			// get the front message from the queue
			let tx = this.#queues.txPending[0];
			// set result
			tx.cresultCode = txrc.cresultCode;
			tx.cresultMessage = txrc.cresultMessage;
			// move it to the failed queue
			this.#updatePendingQueue(1, "txFailed");
		}
		this.#busy = 0;
		// provide an appropriate error, if the login failed.
		if ((txrc.cresultCode == 402) || (txrc.cresultCode == 403)) {
			
		}
			location.href = "../pages/error.php";
		// else: provide an alert to th user
		alert("dilboTxQueue.560: " + _("LFeLuc|Server communication fai...", 
				failed, "" + txrc.cresultCode + ":" + txrc.cresultMessage));
	}

	/**
	 * After parsing the response, handle the result in the tx response
	 * container. May be multiple transactions
	 */
	#handleContainer (txrc) {
		
		if (this.permanentContainerErrors.indexOf(txrc.cresultCode) >= 0) {
			this.#handleContainerError(txrc);
			return;
		}

		// errors which will go away
		if (this.temporaryContainerErrors.indexOf(txrc.cresultCode) >= 0) {
			console.log(_("RhWHn4|Temporary server error d..." + txrc.cresultCode 
					+ " - " + txrc.cresultMessage + ", continue trying after 5 minutes."));
			alert("Temporary error for transaction container " + txrc.cresultCode);
			// pause for 5 minutes
			this.#paused = Date.now() + 300000;
			// transactions stay in the pending queue. Retry after pause.
			this.#busy = 0;  
			return;
		}

		// container response with no error, but transaction may still contain
		// errors
		let now = Date.now();
		for (let txr of txrc.txrs) {
			let tx = this.#queues.txPending[0];
			if (tx) {
				tx.resultCode = txr.resultCode;
				tx.resultMessage = txr.resultMessage;
				if (this.permanentTransactionErrors.indexOf(txr.resultCode) >= 0) {
					// any permanent transaction error.
					this.#updatePendingQueue(1, "txFailed");
					if (!tx.resultMessage)
						tx.resultMessage = _(this.#api_resultMessages[tx.resultCode]);
					if (tx.onError)
						tx.onError(tx);
				} else {
					try {
						// successful transaction. The elegant way to address
						// functions by their name doesn't work inside classes.
						if (tx.type == 'nop')
							this.#nop_resp(tx);
						else if (tx.type == 'session')
							this.#session_resp(tx);
						else if (tx.type == 'config')
							this.#config_resp(tx);
						else if ([ "insert", "update", "delete", "list"].indexOf(tx.type) >= 0)
							dilboTables.onTxResponse(tx);
						else if (tx.callback && (typeof tx.callback == 'function'))
							tx.callback(tx);
					} catch (e) {
						// catch any exception to avoid a blocking head in the
						// pending queue
						alert(_("UNFZk2|transaction to server fa...") + e.stack);
						if (dilboModal)
							dilboModal.showException(e);
					}
					this.#updatePendingQueue(1, "txDone");
				}
				console.log(this.#txToString(tx));
			}
		}
		this.#busy = 0;
	}
	
	/**
	 * Response on nop request. Nothing to do but log the response.
	 */
	#nop_resp (tx) {
		console.log("NOP-Response received.\n");
	}
	
	/**
	 * Response on a session. Parse and refresh the session parameters.
	 */
	#session_resp (tx) {
		console.log("SESSION-Response received.\n");
		// Session close requested. This usually ends the application.
		if (tx.tablename == 'close') {
			if (typeof tx.callback == 'function')
				tx.callback();
			return;
		} 
		let responseElements = tx.resultMessage.split(/;/g);
		let apiParams = {};
		for (let responseElement of responseElements)
			if (responseElement.indexOf("=") > 0)
				apiParams[responseElement.split("=")[0]] = responseElement.split("=")[1];

		// Synchronisation periods.
		this.#synch_check_period = (parseInt(apiParams["synch_check_period"])) ? parseInt(apiParams["synch_check_period"]) : this.#synch_check_period;
		this.#synch_period = (parseInt(apiParams["synch_period"])) ? parseInt(apiParams["synch_period"]) : this.#synch_period;

		// Session management.
		this.#keep_alive_period = (parseInt(apiParams["max_session_keepalive"])) ? parseInt(apiParams["max_session_keepalive"]) : this.#keep_alive_period;
		this.#session_lifetime = (parseInt(apiParams["max_session_duration"])) ? parseInt(apiParams["max_session_duration"]) : this.#session_lifetime;
		this.#api_session_id = (apiParams["api_session_id"]) ? apiParams["api_session_id"] : "";
		let nowSeconds = Date.now() / 1000;
		this.#last_session_activity = nowSeconds;
		if ((tx.tablename == 'regenerate') || ((tx.tablename == 'start') && (this.#last_session_regenerate == 0)))
			this.#last_session_regenerate = nowSeconds;

		// log to the console
		let welcomeMessage = apiParams.server_welcome_message;
		console.log("SESSION-Response received:\n'" + ((welcomeMessage) ? welcomeMessage.replace(/\<br\>/g, "'\n") : "- none -"));
		$("#client-state").html(apiParams.server_welcome_message);
	}

	/**
	 * Response on config request. Nothing to do but log the response.
	 */
	#config_resp (tx) {
		console.log("CONFIG-Response received.\n");
    	window.localStorage.setItem(tx.tablename, tx.resultMessage);
    	window.localStorage.setItem(tx.tablename + ".mtime", Math.floor(Date.now() / 1000));
    	dilboConfig.includeSettings (tx.tablename);
	}
	
}
