/**
 * 
 * dilbo - digital logbook for Rowing and Canoeing
 * ----------------------------------------------- https://www.dilbo.org
 * 
 * Copyright 2023-2024 Martin Glade
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

class DilboUIprovider {
	
	#idParent;  // the parent in the DOM model to put the povider contents to.
	#type; // the provider type
	#table; // the table all provider contents refers to. Null for some like
			// menu or htnlView
	#horizontal; // the container orientation, used for the menu only which
					// can be used horizontally or vertically.
	#providerCfg; // the provider configuration
	#boatIconType; // the type of boat icon to be used, canoe or rowing
	
	#menuItems; // the items of the menu for appropriate display at menu and
				// tabs
	#statistics; // a cache for statistics tables retrieved fron the server.
	
	#calledMenuIds = []; // The menu ids called. Used for go back
	
	/**
	 * virtualFields are used by provider display templates as shortcuts for
	 * some preformatting
	 */
	static #virtualFields = {
		logbook : {
			startDate : {
				realField : "start", 
				getter : DilboUIprovider.#getDate
			},
			startTime : {
				realField : "start", 
				getter : DilboUIprovider.#getTime
			},
			endDate : {
				realField : "end", 
				getter : DilboUIprovider.#getDate
			},
			endTime : {
				realField : "end", 
				getter : DilboUIprovider.#getTime
			},
			tripPeriod : {
				realField : null, 
				getter : DilboUIprovider.#tripPeriod
			}
		},
		assets : {
			boatDescription : {
				realField : null, 
				getter : DilboUIprovider.#boatDescription
			},
			damageDescription : {
				realField : null, 
				getter : DilboUIprovider.#damageDescription
			}
		},
		reservations : {
			startDate : {
				realField : "start", 
				getter : DilboUIprovider.#getDate
			},
			endDate : {
				realField : "end", 
				getter : DilboUIprovider.#getDate
			}
		},
	}
	
	#statusDiv = "<div style='width:2rem;float:left'>";
	#textDiv = "<div style='float:left'>";
	#statusIcon = "<img src='../resources/icons/boats/{boatIcon}_{status}.png' style='width:1.75rem;margin-right:20%;'>";
	#reservationIcon = "<img src='../resources/icons/boats/{boatIcon}_booked.png' style='width:1.75rem;margin-right:20%;'>";
	#damageIcon = "<img src='../resources/icons/boats/{boatIcon}_damaged.png' style='width:1.75rem;margin-right:20%;'>";

	// STATIC FUNCTIONS FOR TEXT FORMATTING
	// ====================================
	/**
	 * Split the itemId into its parts
	 */
	static splitId(id) {
		let parts = id.split(/\_/g);
		// reassemble all remainder into third part
		while (parts.length > 3) 
			parts[2] = parts[2] + "_" + parts.splice(3, 1);
		return parts;
	}

	/**
	 * Split the proiders options string into an object
	 */
	static #parseOptions(providerCfg) {
		let options = providerCfg.get_child("options").get_value().split(/\;/g);
		let optionsObj = {};
		for (let option of options)
			optionsObj[option.split("=")[0]] = option.split("=")[1];
		return optionsObj;
	}
	
	/**
	 * Get the time of a datetime
	 */
	static #getTime(record, fieldname) {
		return TfyhData.format(
				TfyhData.parse(TfyhData.format(record[fieldname], "datetime", "csv").substring(11), "time"), 
				"time", userPreferences.languageCode).substring(0, 5);
	}
	
	/**
	 * Get the date of a datetime
	 */
	static #getDate(record, fieldname) {
		return TfyhData.format(
				TfyhData.parse(TfyhData.format(record[fieldname], "datetime", "csv").substring(0, 10), "date"), 
				"date", userPreferences.languageCode);
	}
	
	/**
	 * get a string representation of the trip period
	 */
	static #tripPeriod(record) {
		if (!record.end || (record.end < new Date("2000-01-01")))
			return DilboUIprovider.#getDate(record, "start") + " " + DilboUIprovider.#getTime(record, "start")
					+ " -> ";
		if (DilboUIprovider.#getDate(record, "start") == DilboUIprovider.#getDate(record, "end"))
			return DilboUIprovider.#getDate(record, "start") + ", " + DilboUIprovider.#getTime(record, "start") 
					+ "-" + DilboUIprovider.#getTime(record, "end");
		return DilboUIprovider.#getDate(record, "start") + " " + DilboUIprovider.#getTime(record, "start") 
					+ "-" + DilboUIprovider.#getDate(record, "end") + " " + DilboUIprovider.#getTime(record, "end");
	}
	
	/**
	 * Get es short descriptiuon of a boat
	 */
	static #boatDescription(record) {
		let variants = record.variant_options.split(/\|/g);
		let description = "";
		for (let variantname of variants) {
			let variant = dilboConfig.get_by_path(".appSettings.boat_variants." + variantname);
			if (variant)
				description += variant.get_descriptor_field("text_local_name") + "; ";
		}
		if (description.length > 1)
			description = description.substring(0, description.length - 2);
		else
			description = "-";
		return description;
	}
	
	/**
	 * Get es short descriptiuon of a damage
	 */
	static #damageDescription(record) {
		let uuid = record.uuid;
		let description = "";
		for (let damageUid in dilboTables.tables.damages.rows) 
			if (!dilboTables.tables.damages.rows[damageUid].fixed && 
					(dilboTables.tables.damages.rows[damageUid].asset_uuid == uuid))
				description += dilboTables.tables.damages.rows[damageUid].description + "; ";
		if (description.length > 1)
			description = description.substring(0, description.length - 2);
		return description;
	}
	
	/**
	 * Fill a template string with available values.
	 */
	static #fill(template, record, recordCfg) {
		let filled = template;

		// check for configuration path elements, whether a replacement is
		// requested
		let c1 = filled.indexOf("{.");
		let c2 = filled.indexOf("}", c1);
		while ((c1 >= 0) && (c2 >= 0)) {
			let configPath = filled.substring(c1 + 1, c2);
			let configItem = dilboConfig.get_by_path(configPath);
			if (configItem) {
				filled = filled.substring(0, c1) + configItem.get_value() + filled.substring(c2 + 1);
				c2 = 0;
			}
			c1 = filled.indexOf("{.", c2);
			c2 = filled.indexOf("}", c1);
		}
		if (!record)
			return filled;
		
		// check for virtual fields first, whether a replacement is requested
		let virtualFields = DilboUIprovider.#virtualFields[recordCfg.get_name()];
		if (virtualFields) {
			for (let virtualFieldName in virtualFields) {
				let search = "{" + virtualFieldName + "}";
				if (filled.indexOf(search) >= 0) {
					let vField = virtualFields[virtualFieldName];
					let vToDisplay = "???";
					if (record[vField.realField] || (vField.realField == null))
						vToDisplay = vField.getter(record, vField.realField);
					// String.prototype.replace does only replace the very first
					// instance of a String
					while (filled.indexOf(search) >= 0)
						filled = filled.replace(search, vToDisplay);
				}
			}
		}
		
		// check for record fields next, whether a replacement is requested
		for (let key in record) {
			let search = "{" + key + "}";
			if (filled.indexOf(search) >= 0) {
				let type = recordCfg.get_child(key).get_type();
				let column_reference = recordCfg.get_child(key).get_descriptor_field("column_reference");
				let toDisplay = dilboTables.getValueToDisplay(record[key], type, column_reference);
				// String.prototype.replace does only replace the very first
				// instance of a String
				while (filled.indexOf(search) >= 0)
					filled = filled.replace(search, toDisplay);
			}
		}
		return filled;
	}
	
	/**
	 * Parse the template on formatting: ~* = &lt;b&gt;; *~ = &lt;/b&gt;; ~_ =
	 * &lt;u&gt;; _~ = &lt;/u&gt;; ~/ = &lt;i&gt;; /~ = &lt;/i&gt;; ~\ =
	 * &lt;br&gt;;
	 */
	static #parseTemplateFormats(template) {
		let replacers = { 
				"~*" : "<b>", 
				"*~" : "</b>",
				"~_" : "<u>",
				"_~" : "</u>",
				"~/" : "<i>",
				"/~" : "</i>",
				"~[" : "<img class='boxed' src='", 
				"]~" : "' />",
				"~\\" : "<br>"
		}
		for (let replace in replacers)
			while (template.indexOf(replace) >= 0)
				template = template.replace(replace, replacers[replace]);
		return template;
	}
	

	
	// CONSTRUCTOR
	// ===========
	/**
	 * Append the provider content to the div with the given unique id, which
	 * consists of the container uid and a position integer, formatted as
	 * "tfyh-[uid].[pos]". the typ is specified and a settings object with the
	 * parameters to use by the provider.
	 */
	constructor(container, providerCfg) {
	
		// set the parameters
		this.#idParent = container.getId();
		this.#horizontal = container.getHorizontal();
		this.#providerCfg = providerCfg;
		this.#type = providerCfg.get_child("type").get_value();
		this.#table = providerCfg.get_child("table").get_value();
		let sportsType = this.#providerCfg.get_by_path(".clubSettings.habits.sports_type").get_value();
		this.#boatIconType = (sportsType == 'canoeing') ? "canoe" : "rowing_boat";
		
		// register the provider for notification on table changes, if it is
		// linked to a table..
		let ownName = this.#providerCfg.get_name();
		if (this.#table)
			dilboTables.addListener(this.#table, "uiProvider-" + ownName, this);
		
		if (this.#type.localeCompare("menu") == 0) {
			this.#displayMenuContainer();
			dilboMenu = this;
		}
		if (this.#type.localeCompare("eventList") == 0) 
			this.#displayEventListContainer();
		if (this.#type.localeCompare("boatList") == 0) 
			this.#displayBoatListContainer();
		if (this.#type.localeCompare("htmlView") == 0) 
			this.#displayHtmlContainer();
		if (this.#type.localeCompare("clock") == 0) 
			this.#displayClockContainer();
	}
	
	// PROVIDER EVENT INPUT
	// ====================
	/**
	 * There was a change in the relevant table for this UI provider.
	 */
	notify() {
		if (this.#type.localeCompare("eventList") == 0) 
			this.#displayEventListContainer();
		else if (this.#type.localeCompare("boatList") == 0) 
			this.#displayBoatListContainer();
	}
	
	/**
	 * handle an event forwarded to this provider
	 */
	handleEvent(itemId) {
		if (this.#type.localeCompare("menu") == 0) 
			this.#handleMenuEvent(itemId);
	}
	
	// MODAL TABS
	// ==========
	
	/**
	 * get a set of tabs, adapted to the onclick options of the triggering
	 * provider and the uid of the list item.
	 */
	#getModalTabs(callerUid, onclickOptions, calledOrUid) {
		let tabs = [];
		let record = {};
		let isUidCalled = (TfyhToolbox.isUid(calledOrUid));
		if (isUidCalled) {
			let tablename = dilboTables.iUids[calledOrUid];
			if (tablename)
				record = dilboTables.tables[tablename].rows[calledOrUid];
		}
			
		// compile the tabs
		for (let tab of onclickOptions) {
			tab = tab.trim();
			let hasField = ((tab.indexOf("{") > 0) && (tab.indexOf("}") > tab.indexOf("{")));
			if (hasField) {
				let fieldName = (hasField) ? tab.substring(tab.indexOf("{") + 1, tab.indexOf("}")) : false;
				let field = record[fieldName];
				if (fieldName == 'uid')
					tab = tab.trim().replace("{uid}", record.uid);
				else if (field && TfyhValidate.is_uuid(field) && dilboTables.iUuids[field]) {
					let targetRecord = dilboTables.tables[dilboTables.iUuids[field]].iUuids[field];
					tab = tab.trim().replace("{" + fieldName + "}", targetRecord.uid);
				} else 
					tab = "invalid_toBeDropped";
			} 
			let menuItemName = tab.trim().substring(0, tab.indexOf("_"));
			let calledName = tab.trim().substring(tab.indexOf("_") + 1);
			let calledCfg = dilboConfig.get_by_path(".uiSettings.providers." + calledName);
			let menuItem = dilboMenu.#menuItems[menuItemName];
			if (menuItem) {
				let text = (isUidCalled) ? menuItem.headline : calledCfg.get_descriptor_field("text_local_name");
				let tabWithId = {
						// localize the tab text
						text : text,
						id : callerUid + "_" + tab.trim()
				} 
				tabs.push(tabWithId);
			}
		}
		return tabs;
	}
	
	// EVENT LIST
	// ==========
	/**
	 * build the event list and display it within a container
	 */
	#displayEventListContainer() {
		let list = this.#buildEventList(this.#providerCfg);
		$('#' + this.#idParent).html(list);
		this.#bindListEvents();
	}

	/**
	 * Display an event list as configured in either the calling provider (in
	 * cases calledOrUid is a uid) or by the called provider (if calledOrUid is
	 * not a uid, i. e. a provider name). If the called provider needs a uid for
	 * filtering, set calledOrUid to the filter uid and called to the called
	 * provider.
	 */
	#displayEventListModal(callerUid, calledOrUid, called) {
		
		// get the configured tabs - if they exist
		let publicEventlists = dilboConfig.get_by_path(".clubSettings.habits.public_eventlists").get_value().split(/\;/g);
		let eventlistsTabOptions = [];
		for (let publicEventlist of publicEventlists)
			eventlistsTabOptions.push("displayEventList_" + publicEventlist.trim());
		let isUidCalled = (TfyhToolbox.isUid(calledOrUid));
		let callerCfg = dilboConfig.getItemForUid(callerUid);
		let tabOptions = (isUidCalled) ? callerCfg.get_child("onclick").get_value().split(/\;/g) : eventlistsTabOptions;
		// get the active tab.
		let providerCfg = callerCfg;
		let activeTab = 1;
		if (!isUidCalled) {
			if (!calledOrUid)
				calledOrUid = publicEventlists[0];
			providerCfg = dilboConfig.get_by_path(".uiSettings.providers." + calledOrUid);
			for (let eventlist of publicEventlists)
				if (calledOrUid != eventlist.trim())
					activeTab ++;
				else break;
		} else if (called)
			// special case: filtered provider for open damages
			providerCfg = dilboConfig.get_by_path(".uiSettings.providers." + called);

		// create the tabs
		let modalTabs = this.#getModalTabs(providerCfg.get_uid(), tabOptions, calledOrUid); 
		
		// create the list
		let headline = "<h3>" + providerCfg.get_descriptor_field("text_local_name") + "</h3>";
		let asset_uid = (TfyhToolbox.isUid(calledOrUid)) ? calledOrUid : false;
		let list = this.#buildEventList(providerCfg, asset_uid);
				
		// display and bind modal buttons and modal tabs events
		dilboModal.setTabs(modalTabs, activeTab);
		dilboModal.showHtml(headline + list, (this.#calledMenuIds.length > 1) ? this.#goBack : false);
		// bind list item events
		this.#bindListEvents();
	}
	
	/**
	 * Build the event list. The asset_uid can be set as filter, Omit it to get
	 * it for all assets.
	 */
	#buildEventList(providerCfg, asset_uid) {
		
		// resolve the context
		if (!providerCfg)
			return;
		let table = providerCfg.get_child("table").get_value();
		let template = providerCfg.get_child("template").get_value();
		template = DilboUIprovider.#parseTemplateFormats(template);
		let optionsObj = DilboUIprovider.#parseOptions(providerCfg);
		let onclickMenuItem = providerCfg.get_child("onclick").get_value();
		// onclickMenuItem refers always to the very first onclick option
		onclickMenuItem = (onclickMenuItem) ? onclickMenuItem.substring(0, onclickMenuItem.indexOf("_")) : "displayDetails";
		
		// build the list and apply filters
		let now = Date.now();	
		let listHtml = "";
		if (dilboTables) {
			let rows = dilboTables.tables[table].rows;
			let recordCfg = this.#providerCfg.get_by_path(".appTables." + table);
			let maxRows = (optionsObj.maxRows) ? parseInt(optionsObj.maxRows) : 50;
			let sort = (optionsObj.sort) ? optionsObj.sort : false;
			// table specific options
			let onlyOpen = ((table == 'logbook') && (optionsObj.open == "true")) || ((table == 'damages') && (optionsObj.fixed == "false"));
			let onlyFuture = (table == 'reservations') && (optionsObj.period == "future");
			let listElements = [];
			let r = 0;
			// sort rows. Must be done before formatting, because new rows are
			// always appended and if the count of rows is limited they will not
			// show up.
			let datetime = { logbook : "start", damages : "reported_on", reservations : "start" };
			let sortedRows = [];
			for (let uid in rows)
				sortedRows.push({ datetime : rows[uid][datetime[table]], row : rows[uid] } )
			sortedRows.sort(function(a, b) { 
				return (b.datetime == a.datetime) ? b.number - a.number : b.datetime - a.datetime;
			});
			// compile html
			for (let sortedRow of sortedRows) {
				let row = sortedRow.row;
				let useRow = (r < maxRows);
				if (table == 'logbook')
					useRow = useRow && (!(onlyOpen && row.end));
				if (table == 'reservations')
					useRow = useRow && (!onlyFuture || (row.end > now));
				if (table == 'damages')
					useRow = useRow && (!onlyOpen || (row.fixed == 0));
				if (asset_uid) {
					let asset = dilboTables.tables.assets.rows[asset_uid];
					useRow = useRow && (row.asset_uuid == asset.uuid);
				}
				if (useRow) {
					let icon = "";
					if (table == 'logbook')
						icon = this.#statusIcon.replace("{boatIcon}", this.#boatIconType).replace("{status}", "onthewater");
					if (table == 'reservations')
						icon = this.#reservationIcon.replace("{boatIcon}", this.#boatIconType);
					if (table == 'damages')
						icon = this.#damageIcon.replace("{boatIcon}", this.#boatIconType);
					// ids are: item provider (for the tabs) - menu action
					// (access control) - item (object to work wwith).
					let id = providerCfg.get_uid() + "_" + onclickMenuItem + "_" + row.uid;
					let listElement = "<div class='w3-row listitem' id='" + id + "'>"
										+ this.#statusDiv + icon + "</div>" 
										+ DilboUIprovider.#fill(template, row, recordCfg) + "</div>\n";
					listHtml += listElement;
					r++;
				}
			}
		}
		return listHtml;
	}

	// BOAT LIST
	// =========
	
	#displayBoatListContainer() {
		
		let template = this.#providerCfg.get_child("template").get_value();
		template = DilboUIprovider.#parseTemplateFormats(template);
		let optionsObj = DilboUIprovider.#parseOptions(this.#providerCfg);
		let onclickMenuItem = this.#providerCfg.get_child("onclick").get_value();
		// onclickMenuItem refers always to the very first onclick option
		onclickMenuItem = (onclickMenuItem) ? onclickMenuItem.substring(0, onclickMenuItem.indexOf("_")) : "displayDetails";

		let nowSecs = Date.now() / 1000;
		let list = "";
		let boat_variants = dilboConfig.get_by_path(".appSettings.boat_variants");
		let groupBy = (optionsObj.groupBy) ? optionsObj.groupBy : false;
		if (dilboTables) {
			let boats = dilboTables.tables.assets.rows;
			let boatCfg = this.#providerCfg.get_by_path(".appTables.assets");
			let statusFilter = (optionsObj.status) ? optionsObj.status.toUpperCase() : "AVAILABLE";
			let sorting = (optionsObj.sort) ? optionsObj.sort.split(".") : false;
			let listElements = [];
			for (let uid in boats) {
				let boat = boats[uid];
				let places = 1; 
				let firstVariantName = boat.variant_options.split(/\|/g)[0];
				if (boat_variants.has_child(firstVariantName))
					places = boat_variants.get_child(firstVariantName).get_child("places").get_value();
				let status = dilboTables.getAssetStatus(uid);
				if ((boat.invalid_from > nowSecs) && (boat.base_status != 'HIDE') && (boat.asset_type == 'BOAT')
						&& (status == statusFilter)) {
					let groupValue = (groupBy == 'places') ? places : "";  
					let group = (groupBy) ? " " + groupBy + "-" + groupValue : ""; 
					let icon = this.#statusIcon.replace("{boatIcon}", this.#boatIconType).replace("{status}", status.toLowerCase());
					let id = this.#providerCfg.get_uid() + "_" + onclickMenuItem + "_" + boat.uid;
					let listElement = "<div class='w3-row w3-show listitem" + group + "' id='" + id + "'>" 
										+ this.#statusDiv + icon + "</div>" + this.#textDiv
										+ DilboUIprovider.#fill(template, boat, boatCfg) + "</div></div>\n";
					listElements.push({
						name : boat.name,
						places : places,
						html : listElement
					});
				}
			}
			if (sorting)
				listElements.sort(function(a, b) {
					let diff = {
							name : a.name.localeCompare(b.name),
							places : a.places - b.places
					};
					return (diff[sorting[0]] * 1000) + ((sorting.length > 1) ? diff[sorting[1]] : 0);
			});
			let currentGroup = "";
			for (let listElement of listElements) {
				if (groupBy) {
					let groupOfElement = listElement[groupBy];
					if (groupOfElement != currentGroup) {
						list += "<div class='w3-row w3-show group-heading' id='" + groupBy + "-" + groupOfElement + "'>" 
								+ "<span class='caret'>▼ </span>"
								+ groupOfElement + "</div>\n";
						currentGroup = groupOfElement; 
					}
				}
				list += listElement.html + "\n";
			}
		}
		$('#' + this.#idParent).html(list);
		this.#bindListEvents(groupBy !== false);
	}
	
	// LIST EVENTS
	// ===========
	
	/**
	 * show details on click of listitem
	 */
	#bindListEvents(groupEvents) {
		// if this function is called more than once, events pile up.
		$(".listitem").unbind();
		let that = this;
		$(".listitem").click(function() {
			let itemId = $(this).attr("id");
			that.#onListItemClick(itemId);
		});
		if (groupEvents)
			$(".group-heading").click(function() {
				// get the group id
				let groupBy = $(this).attr("id");
				// toggle visibility of all items in the group
				$("." + groupBy).toggleClass("w3-hide");
				$("." + groupBy).toggleClass("w3-show");
				// switch the caret for the group heading
				$("#" + groupBy + " .caret").each(function() {
					$(this).html((this.innerText.startsWith("▶")) ? "▼ " : "▶ ");
				});
			});
	}
	
	/**
	 * Trigger the appropriate actions for a list item click, based on the list
	 * items type and status.
	 */
	#onListItemClick(itemId) {
		
		let splitId = DilboUIprovider.splitId(itemId);
		let poviderUid = splitId[0];
		let providerCfg = dilboConfig.getItemForUid(poviderUid);
		let itemUid = splitId[2];
		
		// get the context forms etc
		let onclickOptions = providerCfg.get_child("onclick").get_value().split(/\;/g);
		let formTabs = this.#getModalTabs(poviderUid, onclickOptions, itemUid);
		// some actions set their tabs dynmaically like displayEventLists.
		// but displayDetails or FormHandler do not mannage tabs, so set
		// the tabs before handling the menu event to make sure they fit.
		dilboModal.setTabs(formTabs, 1);
		dilboMenu.handleEvent(formTabs[0].id);
	}
	
	// STATISTICS
	// ==========
	
	/**
	 * get the statistics from the server for a uiProvider
	 */
	#getRankingFromServer(ranking) {
		let publicRankings = dilboConfig.get_by_path(".clubSettings.habits.public_rankings").get_value().split(/\;/g);
		if (!ranking) 
			ranking = publicRankings[0];
		let providerCfg = dilboConfig.get_by_path(".uiSettings.providers." + ranking);
		let headline = "<h3>" + providerCfg.get_descriptor_field("text_local_name") + "</h3>";
		let optionsObj = DilboUIprovider.#parseOptions(providerCfg);
		dilboTxQueue.addNewTxToPending ("list", "statistics", {
				type : optionsObj.type,
				field : optionsObj.field,
				providerName : ranking,
			}, this, null);
	}

	/**
	 * read the returned statistics result into the statistics cache
	 */
	buildStatistics(providerName, resultMessage) {
		let providerCfg = dilboConfig.get_by_path(".uiSettings.providers." + providerName);
		let optionsObj = DilboUIprovider.#parseOptions(providerCfg);
		if (!this.#statistics)
			this.#statistics = {};
		if (!this.#statistics[optionsObj.type])
			this.#statistics[optionsObj.type] = {};
		this.#statistics[optionsObj.type][optionsObj.field] = TfyhToolbox.csvToObject(resultMessage);
		if (optionsObj.type == 'total_distance') {
			this.#statistics[optionsObj.type][optionsObj.field].sort(function(a,b) {
				return parseFloat(b.total_distance) - parseFloat(a.total_distance);
			})
			this.#displayStatisticsModal(providerCfg, optionsObj.type, optionsObj.field);
		}
	}
	
	/**
	 * Display the statistics in the modal
	 */
	#displayStatisticsModal(providerCfg, type, field) {
		if (type == 'total_distance') {
			let publicRankings = dilboConfig.get_by_path(".clubSettings.habits.public_rankings").get_value();
			let rankingsTabOptions = [];
			for (let publicRanking of publicRankings.split(/\;/g))
				rankingsTabOptions.push("displayRanking_" + publicRanking.trim());
			let trimmed = [];
			for (let publicRanking of publicRankings.split(/\;/g))
				trimmed.push(publicRanking.trim());
			let activeTab = trimmed.indexOf(providerCfg.get_name()) + 1; 
			let headline = "<h3>" + providerCfg.get_descriptor_field("text_local_name") + "</h3>";
			let maximum = this.#statistics[type][field][0].total_distance;
			let list = "<table id='statistics'><tbody><tr><th style='width:40%'>"
				+ _("QtMS00|Name") + "</th>" + "<th style='width:40%'>" + _("cAOo7x|Distance")
				+ "</th><th style='width:20%'>" + _("cpSFae|km")
				+ "</th><th style='width:20%'>" + _("2law1v|Fessions")
				+ "</th></tr>";
			for (let row of this.#statistics[type][field]) 
				list += "<tr><td>" + row.name + "</td>" 
						+ "<td><div class='bar-container'><div class='bar' style='width:"
							+ Math.floor(row.total_distance / maximum * 100) + "%'>&nbsp;</div></div></td>" 
							+ "<td>" + row.total_distance + "</td>" 
							+ "<td>" + row.sessions_count + "</td></tr>";
			list += "</table>";
			
			// create the tabs
			let modalTabs = this.#getModalTabs(providerCfg.get_uid(), rankingsTabOptions, this.#providerCfg.get_name());
			dilboModal.setTabs(modalTabs, activeTab);
			dilboModal.showHtml(headline + list, (this.#calledMenuIds.length > 1) ? this.#goBack : false);
		}
	}
	
	// HTML View
	// =========
	#displayHtmlContainer() {
		let resource = this.#providerCfg.get_child("resources").get_value();
		let template = this.#providerCfg.get_child("template").get_value();
		if (template) {
			template = DilboUIprovider.#parseTemplateFormats(template);
			template = DilboUIprovider.#fill(template);
			$('#' + this.#idParent).html(template);
		} else if (resource) {
			$('#' + this.#idParent).load(resource);
		}
	}
	
	
	// DETAILS
	// =======
	#displayDetailsModal(uid) {
		let table = dilboTables.iUids[uid];
		let goBack = (this.#calledMenuIds.length > 1) ? this.#goBack : false; 
		if (!table) 
			dilboModal.showHtml(_("Pc6MzY|uid °%1° could not be re...", uid), goBack);
		else {
			let record = dilboTables.tables[table].rows[uid];
			if (!record) 
				dilboModal.showHtml(_("uuagNx|uid °%1° could not be fo...", uid, table), goBack);
			else
				dilboModal.showHtml(dilboTables.recordToTableHTML(table, record), goBack);
		}
	}
	
	// CLOCK
	// =====
    #displayClockContainer() {
    	setInterval(() => {
    	    let d = new Date(); // object of date()
    	    let hr = d.getHours();
    	    let min = d.getMinutes();
    	    let sec = d.getSeconds();
    	    let hr_rotation = 30 * hr + min / 2; // converting current time
    	    let min_rotation = 6 * min;
    	    let sec_rotation = 6 * sec;
    	    try {
        	    document.getElementById("clockHour").style.transform = `rotate(${hr_rotation}deg)`;
        	    document.getElementById("clockMinute").style.transform = `rotate(${min_rotation}deg)`;
        	    document.getElementById("clockSecond").style.transform = `rotate(${sec_rotation}deg)`;
    	    } catch (ignored) {}
    	}, 2000);
    	let w = $('#' + this.#idParent).innerWidth();
    	let h = $('#' + this.#idParent).innerHeight();
    	let size = (w > h) ? h : w;
    	const clockHtml = '<div id="clockSquareBox" style="width:' + size + 'px; height:' + size + 'px;overflow:hidden;"><div id="clockContainer">' 
    		+ '<div id="clockHour"></div><div id="clockMinute"></div><div id="clockSecond"></div></div></div>';
    	$('#' + this.#idParent).html(clockHtml);
    	$('#' + this.#idParent).removeClass("tfyh-tile").addClass("tfyh-tile-transparent").css("overflow", "hidden");
    }

	
	// MENU
	// ====
	// This is a sprecific provider, as it takes all on click events to manage
	// permissions.
	
	/**
	 * Create the menu based on the provided set.
	 */
	#displayMenuContainer() {
		let menu;
		let sportstype = this.#providerCfg.get_by_path(".clubSettings.habits.sports_type").get_value();
		let tripx = (sportstype == 'canoeing') ? "c" : "r";
		try {
			menu = TfyhToolbox.csvToObject (php_menu_csv);
		} catch (e) {
			alert("Menu definition not fonud. Error: " + e + ". Aborting");
		}
		let buttonTypes = {};
		let buttonHints = {};
		let buttonLinks = {};
		let buttonHide = {};
		let i = 1;
		this.#menuItems = {};
		for (let item of menu) {
			// localize the items headline.
			item.headline = _(item.headline);
			buttonTypes[i] = item.id
							.replace("startSession", "trip" + tripx + "Start")
							.replace("endSession", "trip" + tripx + "End");
			buttonHints[i] = item.headline;
			buttonLinks[i] = item.link;
			buttonHide[i] = item.permission.startsWith(".");
			this.#menuItems[item.link] = item;
			i++;
		}
		for (let index in buttonTypes){
			// add the button
			let html = $('#' + this.#idParent).html();
			if (!buttonHide[index])
				html += '<div id="menuItem_' + buttonLinks[index] + '" class="menuitem dilbo-menubutton" style="' 
					+ ((this.#horizontal) ? "height" : "width") + ': 100%; background: url(\'../resources/icons/menu/' 
					+ buttonTypes[index] + '.png\') no-repeat; background-size: contain;"'
					+ ' title="' + buttonHints[index] + '"></div>';
			$('#' + this.#idParent).html(html);
		}
		this.#bindMenuEvents();
	}
	
	// will bind all events by selecting all .menuitem elements belonging to
	// this
	#bindMenuEvents() {
		let menuitems = $('.menuitem'); // for debugging: do not inline
										// statement.
		let that = this;
		menuitems.click(function() {
			let thisElement = $(this); // for debugging: do not inline
										// statement.
			let id = thisElement.attr("id");
			if (!id)
				return;
			window.handleEvent(id);
		});
	}
	
	/**
	 * All events go through the menu. To go back it is sufficient to call the
	 * menu item that was used before this one.
	 */
	#goBack(clearHistory) {
		if (clearHistory)
			dilboMenu.#calledMenuIds = [];
		else {
			let lastItemId = dilboMenu.#calledMenuIds.pop();
			let previousItemId = dilboMenu.#calledMenuIds.pop();
			if (previousItemId)
				dilboMenu.#handleMenuEvent(previousItemId);
		}
	}

	/**
	 * Asynchronous call in ordert to close the API session first
	 */
	#gotoAdminUI() {
		window.location.href = '../pages/admin_home.php';
	}
	
	/**
	 * Handle all events coming in
	 */
	#handleMenuEvent(itemId) {
		
		this.#calledMenuIds.push(itemId);
		let action = DilboUIprovider.splitId(itemId);
		if (!this.#menuItems[action[1]]) {
			alert(_("mQvmev|the menu action %1 is no...", action[1]));
			return;
		}
		let menuItem = dilboMenu.#menuItems[action[1]];
		let tabWithoutId = {
					text : menuItem.headline,
					id : itemId
			} 

		if (action[1] == 'adminUI') 
			dilboTxQueue.addNewTxToPending ("session", "close", null, this.#gotoAdminUI, null);
		// forms
		else if ((action[1] == 'startSession') || (action[1] == 'editSession'))  
			dilboFormHandler.editSession_do(action[2], false);
		else if (action[1] == 'closeSession') 
			dilboFormHandler.editSession_do(action[2], true);
		else if (action[1] == 'editPreferences') 
			dilboFormHandler.editPreferences_do(action[2]);
		else if (action[1] == 'reportDamage') 
			dilboFormHandler.reportDamage_do(action[2]);
		else if (action[1] == 'requestReservation') 
			dilboFormHandler.requestReservation_do(action[2]);
		else if (action[1] == 'sendMessage') 
			dilboFormHandler.sendMessage_do();
		else if (action[1] == 'requestReservation') 
			dilboFormHandler.requestReservation_do(action[2]);
		// lists
		else if (action[1] == 'displayDetails') 
			this.#displayDetailsModal(action[2]);
		else if (action[1] == 'displayEventList') 
			this.#displayEventListModal(action[0], action[2]);
		else if (action[1] == 'displayDamages') 
			this.#displayEventListModal(action[0], action[2]);
		else if (action[1] == 'displayBoatDamages') 
			this.#displayEventListModal(action[0], action[2], "open_damages_per_boat");
		else if (action[1] == 'displayRanking') 
			this.#getRankingFromServer(action[2]);

		else
			alert(_("aJJzrZ|No such menu item %1.", action[1]));
	}
	
}
